Package: g6R
Title: Graph Visualisation Engine Widget for R and 'shiny' Apps
Version: 0.6.5
Authors@R: 
    c(
    person("David", "Granjon", , "dgranjon@ymail.com", role = c("aut", "cre")),
    person("David", "Schoch", , "david@schochastics.net", role = c("aut")),
    person(family = "cynkra GmbH", role = "fnd"),
    person(family = "Bristol Myers Squibb", role = "fnd")
    )
Description: Create stunning network experiences powered by the 'G6' graph visualisation engine
    'JavaScript' library <https://g6.antv.antgroup.com/en>. In 'shiny' mode, modify your
    graph directly from the server function to dynamically interact with nodes and edges.
    Select your favorite layout among 20 choices. 15 behaviors are available such as
    interactive edge creation, collapse-expand and brush select.
    17 plugins designed to improve the user experience such as a mini-map,
    toolbars and grid lines. Customise the look and feel of your graph with comprehensive
    options for nodes, edges and more.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://github.com/cynkra/g6R, https://cynkra.github.io/g6R/
BugReports: https://github.com/cynkra/g6R/issues
Imports: htmlwidgets, shiny
Suggests: knitr, rmarkdown, quarto, igraph, roxy.shinylive, testthat
        (>= 3.0.0), withr, stringr, htmltools, bslib
VignetteBuilder: quarto, knitr
Depends: R (>= 4.1.0)
LazyData: true
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-07-17 11:56:44 UTC; davidgranjon
Author: David Granjon [aut, cre],
  David Schoch [aut],
  cynkra GmbH [fnd],
  Bristol Myers Squibb [fnd]
Maintainer: David Granjon <dgranjon@ymail.com>
Repository: CRAN
Date/Publication: 2026-07-17 12:50:14 UTC
Built: R 4.6.1; ; 2026-07-17 17:10:51 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as_g6_combo,g6_combo)
S3method(as_g6_combo,list)
S3method(as_g6_combos,data.frame)
S3method(as_g6_combos,g6_combos)
S3method(as_g6_combos,list)
S3method(as_g6_data,g6_data)
S3method(as_g6_data,list)
S3method(as_g6_edge,g6_edge)
S3method(as_g6_edge,list)
S3method(as_g6_edges,data.frame)
S3method(as_g6_edges,g6_edges)
S3method(as_g6_edges,list)
S3method(as_g6_node,g6_node)
S3method(as_g6_node,list)
S3method(as_g6_nodes,data.frame)
S3method(as_g6_nodes,g6_nodes)
S3method(as_g6_nodes,list)
S3method(as_g6_port,g6_port)
S3method(as_g6_port,list)
S3method(as_g6_ports,g6_ports)
S3method(as_g6_ports,list)
S3method(validate_element,g6_combo)
S3method(validate_element,g6_edge)
S3method(validate_element,g6_element)
S3method(validate_element,g6_node)
S3method(validate_elements,g6_combos)
S3method(validate_elements,g6_edges)
S3method(validate_elements,g6_nodes)
S3method(validate_port,g6_port)
S3method(validate_ports,g6_ports)
S3method(validate_ports,list)
export(JS)
export(animation_config)
export(antv_dagre_layout)
export(as_g6_combo)
export(as_g6_combos)
export(as_g6_data)
export(as_g6_edge)
export(as_g6_edges)
export(as_g6_node)
export(as_g6_nodes)
export(as_g6_port)
export(as_g6_ports)
export(auto_adapt_label)
export(auto_fit_config)
export(background)
export(brush_select)
export(bubble_sets)
export(canvas_config)
export(circular_layout)
export(click_select)
export(collapse_expand)
export(combo_combined_layout)
export(combo_options)
export(compact_box_layout)
export(concentric_layout)
export(context_menu)
export(create_edge)
export(d3_force_layout)
export(dagre_layout)
export(dendrogram_layout)
export(drag_canvas)
export(drag_element)
export(drag_element_force)
export(edge_bundling)
export(edge_filter_lens)
export(edge_options)
export(edge_style_options)
export(fish_eye)
export(fix_element_size)
export(focus_element)
export(force_atlas2_layout)
export(fruchterman_layout)
export(fullscreen)
export(g6)
export(g6Output)
export(g6_add_combos)
export(g6_add_data)
export(g6_add_edges)
export(g6_add_nodes)
export(g6_add_plugin)
export(g6_behaviors)
export(g6_canvas_resize)
export(g6_collapse_combo)
export(g6_collapse_options)
export(g6_combo)
export(g6_combos)
export(g6_data)
export(g6_edge)
export(g6_edges)
export(g6_expand_combo)
export(g6_fit_center)
export(g6_focus_combos)
export(g6_focus_edges)
export(g6_focus_nodes)
export(g6_get_combos)
export(g6_get_edges)
export(g6_get_input_ports)
export(g6_get_nodes)
export(g6_get_output_ports)
export(g6_get_ports)
export(g6_hide_combos)
export(g6_hide_edges)
export(g6_hide_elements)
export(g6_hide_nodes)
export(g6_igraph)
export(g6_input_port)
export(g6_layout)
export(g6_node)
export(g6_nodes)
export(g6_options)
export(g6_output)
export(g6_output_port)
export(g6_plugins)
export(g6_port)
export(g6_ports)
export(g6_proxy)
export(g6_remove_combos)
export(g6_remove_edges)
export(g6_remove_nodes)
export(g6_set_combos)
export(g6_set_data)
export(g6_set_edges)
export(g6_set_nodes)
export(g6_set_options)
export(g6_set_theme)
export(g6_show_combos)
export(g6_show_edges)
export(g6_show_elements)
export(g6_show_nodes)
export(g6_update_behavior)
export(g6_update_combos)
export(g6_update_edges)
export(g6_update_layout)
export(g6_update_nodes)
export(g6_update_plugin)
export(g6_update_ports)
export(grid_line)
export(history)
export(hover_activate)
export(hull)
export(is_g6_collapse_options)
export(is_g6_combo)
export(is_g6_combos)
export(is_g6_data)
export(is_g6_edge)
export(is_g6_edges)
export(is_g6_node)
export(is_g6_nodes)
export(is_g6_port)
export(is_g6_ports)
export(lasso_select)
export(legend)
export(minimap)
export(node_options)
export(node_style_options)
export(optimize_viewport_transform)
export(radial_layout)
export(renderG6)
export(render_g6)
export(scroll_canvas)
export(set_g6_max_collapse_depth)
export(snapline)
export(timebar)
export(toolbar)
export(tooltips)
export(validate_edges_ports)
export(validate_element)
export(validate_elements)
export(validate_port)
export(validate_ports)
export(watermark)
export(zoom_canvas)
import(htmlwidgets)
importFrom(stats,setNames)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# g6R

<!-- badges: start -->

[![R-CMD-check](https://github.com/cynkra/g6R/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/cynkra/g6R/actions/workflows/R-CMD-check.yaml)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/g6R)](https://CRAN.R-project.org/package=g6R)
[![Codecov test
coverage](https://codecov.io/gh/cynkra/g6R/graph/badge.svg)](https://app.codecov.io/gh/cynkra/g6R)
<!-- badges: end -->

<p style="text-align: center;">

<img src="./man/figures/hex.png" style="width:50.0%"
alt="g6R hex logo" />
</p>

`{g6R}` provides R bindings to the G6 graph
[library](https://g6.antv.antgroup.com/en). It allows to create
interactive network in R, with stunning set of features, including:

- Support for various **layouts**, such as force-directed, radial,
  circular, and hierarchical layouts.
- Interactive **behaviors** like zooming, dragging, and selecting
  elements.
- **Plugins** for additional functionality, such as minimaps and
  tooltips, context menus, and node grouping features like bubble sets,
  hulls and legends.
- Various **data sources** including data frames, lists and remote JSON
  urls.
- Support for **combos** allowing for **nested** nodes.
- High **performance** rendering (\>20000 nodes).

<p style="text-align: center;">

<img src="./man/figures/g6-layers.png" style="width:50.0%"
alt="g6R layers example" />
</p>

## Installation

You can install the development version of `{g6R}` from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("cynkra/g6R")
```

## Example

To create a `{g6R}` graph:

``` r
library(g6R)
nodes <- data.frame(id = 1:10)

# Generate random edges
edges <- data.frame(
  source = c(2, 6, 7),
  target = c(1, 3, 9)
)

g6(nodes, edges) |>
  g6_options(
    node = list(
      style = list(
        labelBackground = TRUE,
        labelBackgroundFill = '#FFB6C1',
        labelBackgroundRadius = 4,
        labelFontFamily = 'Arial',
        labelPadding = c(0, 4),
        labelText = JS(
          "(d) => {
              return d.id
            }"
        )
      )
    )
  ) |>
  g6_layout(d3_force_layout()) |>
  g6_behaviors(
    "zoom-canvas",
    drag_element_force(fixed = TRUE),
    click_select(
      multiple = TRUE,
      onClick = JS(
        "(e) => {
            console.log(e);
          }"
      )
    ),
    brush_select(),
    create_edge()
  ) |>
  g6_plugins(
    "minimap",
    "tooltip",
    context_menu()
  )
```

<!-- README.md is generated from README.Rmd. Please edit that file -->

# g6R

<!-- badges: start -->

[![R-CMD-check](https://github.com/cynkra/g6R/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/cynkra/g6R/actions/workflows/R-CMD-check.yaml)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![CRAN
status](https://www.r-pkg.org/badges/version/g6R)](https://CRAN.R-project.org/package=g6R)
[![Codecov test
coverage](https://codecov.io/gh/cynkra/g6R/graph/badge.svg)](https://app.codecov.io/gh/cynkra/g6R)
<!-- badges: end -->

<p style="text-align: center;">
<img src="./man/figures/hex.png" style="width:50.0%"
alt="g6R hex logo" />
</p>

`{g6R}` provides R bindings to the G6 graph
[library](https://g6.antv.antgroup.com/en). It allows to create
interactive network in R, with stunning set of features, including:

- Support for various **layouts**, such as force-directed, radial,
  circular, and hierarchical layouts.
- Interactive **behaviors** like zooming, dragging, and selecting
  elements.
- **Plugins** for additional functionality, such as minimaps and
  tooltips, context menus, and node grouping features like bubble sets,
  hulls and legends.
- Various **data sources** including data frames, lists and remote JSON
  urls.
- Support for **combos** allowing for **nested** nodes.
- High **performance** rendering (\>20000 nodes).

<p style="text-align: center;">
<!-- Served from GitHub (build-ignored) to keep the package tarball under CRAN's 5 MB limit. -->
<img
src="https://raw.githubusercontent.com/cynkra/g6R/main/man/figures/g6R-deconstructed.gif"
style="width:92.0%"
alt="A g6 graph deconstructed into its data, layout, options, behaviors and plugins layers" />
</p>

## Installation

You can install the development version of `{g6R}` from
[GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("cynkra/g6R")
```

## Example

To create a `{g6R}` graph:

``` r
library(g6R)
nodes <- data.frame(id = 1:10)

# Generate random edges
edges <- data.frame(
  source = c(2, 6, 7),
  target = c(1, 3, 9)
)

g6(nodes, edges) |>
  g6_options(
    node = list(
      style = list(
        labelBackground = TRUE,
        labelBackgroundFill = '#FFB6C1',
        labelBackgroundRadius = 4,
        labelFontFamily = 'Arial',
        labelPadding = c(0, 4),
        labelText = JS(
          "(d) => {
              return d.id
            }"
        )
      )
    )
  ) |>
  g6_layout(d3_force_layout()) |>
  g6_behaviors(
    "zoom-canvas",
    drag_element_force(fixed = TRUE),
    click_select(
      multiple = TRUE,
      onClick = JS(
        "(e) => {
            console.log(e);
          }"
      )
    ),
    brush_select(),
    create_edge()
  ) |>
  g6_plugins(
    "minimap",
    "tooltip",
    context_menu()
  )
```

library(shiny)
library(bslib)
library(g6R)

themes <- list(
  "light" = list(
    theme = 'light',
    node = list(
      style = list(
        size = 4
      ),
      palette = list(
        type = 'group',
        field = 'cluster'
      )
    ),
    plugins = list(
      list(
        type = 'background',
        background = '#fff'
      )
    )
  ),
  "dark" = list(
    theme = 'dark',
    node = list(
      style = list(
        size = 4
      ),
      palette = list(
        type = 'group',
        field = 'cluster'
      )
    ),
    plugins = list(
      list(
        type = 'background',
        background = '#000'
      )
    )
  ),
  "blue" = list(
    theme = 'light',
    node = list(
      style = list(
        size = 4
      ),
      palette = list(
        type = 'group',
        field = 'cluster',
        color = "blues",
        invert = TRUE
      )
    ),
    plugins = list(
      list(
        type = 'background',
        background = '#f3faff'
      )
    )
  ),
  "yellow" = list(
    theme = 'light',
    node = list(
      style = list(
        size = 4
      ),
      palette = list(
        type = 'group',
        field = 'cluster',
        color = c(
          '#ffe7ba',
          '#ffd591',
          '#ffc069',
          '#ffa940',
          '#fa8c16',
          '#d46b08',
          '#ad4e00',
          '#873800',
          '#612500'
        )
      )
    ),
    plugins = list(
      list(
        type = 'background',
        background = '#fcf9f1'
      )
    )
  )
)

ui <- page_fluid(
  selectInput(
    "theme",
    "Theme",
    choices = c("light", "dark", "blue", "yellow"),
    selected = "default"
  ),
  g6Output("graph", height = "100vh")
)

server <- function(input, output, session) {
  output$graph <- renderG6({
    g6(
      jsonUrl = "https://assets.antv.antgroup.com/g6/20000.json"
    ) |>
      g6_options(
        animation = FALSE,
        autoFit = "view",
        padding = 20,
        node = list(
          style = list(size = 4),
          palette = list(
            type = "group",
            field = "cluster"
          )
        )
      ) |>
      g6_behaviors(
        "zoom-canvas",
        "drag-canvas",
        "optimize-viewport-transform"
      )
  })

  observeEvent(input$theme, {
    g6_proxy("graph") |>
      g6_set_theme(themes[[input$theme]])
  })

  observeEvent(input[["graph-initialized"]], {
    showNotification(
      "Graph initialized",
      type = "message"
    )
  })

  observeEvent(input[["graph-state"]], {
    showNotification(
      "Graph state updated",
      type = "message"
    )
  })
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(g6R)

default_node_style <- list(
  fill = "#1783FF",
  labelBackgroundFill = "#a0cafa",
  badges = list()
)

nodes <- list(
  list(
    id = "1",
    combo = "combo1",
    style = default_node_style
  ),
  list(
    id = "2",
    combo = "combo1",
    style = list(
      fill = "#ee705c",
      labelBackgroundFill = "#FFB6C1",
      badges = list(
        list(
          text = sprintf("Eval errors: '%s'", 1),
          placement = "right-bottom",
          backgroundFill = "#85847e"
        ),
        list(
          text = sprintf("Data errors: '%s'", 1),
          placement = "right",
          backgroundFill = "#edb528"
        ),
        list(
          text = sprintf("State errors: '%s'", 1),
          placement = "right-top",
          backgroundFill = "#ee705c"
        )
      )
    )
  ),
  list(
    id = "3",
    combo = "combo1",
    style = default_node_style
  ),
  list(
    id = "4",
    combo = "combo2",
    style = default_node_style
  ),
  list(
    id = "5",
    combo = "combo2",
    style = default_node_style
  )
)

edges <- list(
  list(
    type = "fly-marker-cubic",
    source = "1",
    target = "2"
  ),
  list(
    source = "2",
    target = "3",
    type = "cubic",
    style = list(
      stroke = "#ee705c",
      badgeText = "ERROR",
      badgeBackgroundFill = "#ee705c",
      badgeBackground = TRUE
    ),
    states = list("inactive")
  ),
  list(
    type = "fly-marker-cubic",
    source = "4",
    target = "5"
  )
)

combos <- list(
  list(
    id = "combo1",
    type = "rect",
    data = list(label = "Combo A"),
    style = list(
      stroke = "orange",
      fill = "orange",
      fillOpacity = 0.2,
      shadowColor = "orange",
      collapsedFill = "orange",
      collapsedStroke = "orange",
      iconFill = "orange",
      labelPlacement = "top"
    )
  ),
  list(
    id = "combo2",
    data = list(label = "Combo B"),
    style = list(
      stroke = "lightblue",
      fill = "lightblue",
      fillOpacity = 0.2,
      shadowColor = "lightblue",
      collapsedFill = "lightblue",
      collapsedStroke = "lightblue",
      iconFill = "lightblue",
      labelPlacement = "top"
    )
  )
)

ui <- page_fluid(
  g6Output("graph", height = "100vh"),
)

server <- function(input, output, session) {
  output$graph <- renderG6({
    g6(
      nodes,
      edges,
      combos
    ) |>
      g6_options(
        animation = FALSE,
        node = list(
          style = list(
            labelBackground = TRUE,
            labelBackgroundRadius = 4,
            labelFontFamily = "Arial",
            labelPadding = c(0, 4),
            labelText = JS(
              "(d) => {
                return d.id
              }"
            )
          )
        ),
        combo = list(
          animation = FALSE,
          type = "circle-combo-with-extra-button",
          style = list(
            labelText = JS(
              "(d) => {
                return d.data.label
              }"
            )
          )
        ),
        edge = list(
          style = list(
            endArrow = TRUE,
            lineDash = c(5, 5),
            labelText = JS(
              "(d) => {
                return d.label
              }"
            )
          )
        )
      ) |>
      g6_layout(d3_force_layout()) |>
      g6_behaviors(
        "zoom-canvas",
        "drag-canvas",
        drag_element(dropEffect = "link"),
        click_select(multiple = TRUE),
        "brush-select"
      ) |>
      g6_plugins(
        g6R::toolbar(
          style = list(
            backgroundColor = "#f5f5f5",
            padding = "8px",
            boxShadow = "0 2px 8px rgba(0, 0, 0, 0.15)",
            borderRadius = "8px",
            border = "1px solid #e8e8e8",
            opacity = "0.9",
            marginTop = "12px",
            marginLeft = "12px"
          ),
          position = "left"
        ),
        context_menu()
      )
  })
}

shinyApp(ui, server)

library(shiny)
library(g6R)

options(
  "g6R.mode" = "dev",
  "g6R.max_collapse_depth" = Inf,
  # In theory this is automatically set whenever a g6_node has some children ...
  "g6R.directed_graph" = TRUE
)

ui <- fluidPage(
  g6_output("dag", height = "1000px"),
  verbatimTextOutput("clicked_port"),
  verbatimTextOutput("removed_node")
)

server <- function(input, output, session) {
  output$dag <- render_g6(
    g6(
      nodes = g6_nodes(
        # a = dataset_block("iris") — outside combos
        g6_node(
          id = "a",
          type = "custom-rect-node",
          style = list(
            labelText = "dataset: iris"
          ),
          ports = g6_ports(
            g6_output_port(
              key = "output-a",
              placement = "bottom",
              label = "data"
            )
          ),
          children = c("b"),
          collapse = g6_collapse_options(
            collapsed = TRUE,
            visibility = "hover",
            placement = "right-top"
          )
        ),
        # b = head_block(n = 10) — combo1 (stack1)
        g6_node(
          id = "b",
          type = "custom-rect-node",
          style = list(
            labelText = "head(n=10)"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-b",
              placement = "top",
              label = "data"
            ),
            g6_output_port(
              key = "output-b",
              placement = "bottom",
              label = "data"
            )
          ),
          combo = "combo1",
          children = c("c", "f"),
          collapse = g6_collapse_options(
            placement = "right-top"
          )
        ),
        # c = subset_block() — combo1 (stack1)
        g6_node(
          id = "c",
          type = "custom-rect-node",
          style = list(
            labelText = "subset"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-c",
              placement = "top",
              label = "data"
            ),
            g6_output_port(
              key = "output-c",
              placement = "bottom",
              label = "data"
            )
          ),
          combo = "combo1",
          children = "d"
        ),
        # d = head_block(n = 5) — combo1 (stack1)
        g6_node(
          id = "d",
          type = "custom-rect-node",
          style = list(
            labelText = "head(n=5)"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-d",
              placement = "top",
              label = "data"
            ),
            g6_output_port(
              key = "output-d",
              placement = "bottom",
              label = "data"
            )
          ),
          combo = "combo1",
          children = c("e"),
          collapse = g6_collapse_options(
            collapsed = FALSE,
            placement = "right-top"
          )
        ),
        # e = rbind_block() — outside combos
        g6_node(
          id = "e",
          type = "custom-rect-node",
          style = list(
            labelText = "rbind"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-e-1",
              placement = "top",
              label = "1",
              arity = Inf
            ),
            g6_input_port(
              key = "input-e-2",
              placement = "top",
              label = "2",
              arity = Inf
            ),
            g6_output_port(
              key = "output-e",
              placement = "bottom",
              label = "data",
              arity = Inf
            )
          ),
          children = c("g"),
          collapse = g6_collapse_options(
            collapsed = FALSE,
            placement = "right-top"
          )
        ),
        # f = subset_block() — combo2 (stack2)
        g6_node(
          id = "f",
          type = "custom-rect-node",
          style = list(
            labelText = "subset"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-f",
              placement = "top",
              label = "data"
            ),
            g6_output_port(
              key = "output-f",
              placement = "bottom",
              label = "data"
            )
          ),
          combo = "combo2",
          children = c("e", "g"),
          collapse = g6_collapse_options(
            collapsed = FALSE,
            placement = "right-top"
          )
        ),
        # g = rbind_block() — combo2 (stack2)
        g6_node(
          id = "g",
          type = "custom-rect-node",
          style = list(
            labelText = "rbind"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-g-1",
              placement = "top",
              label = "1",
              arity = Inf
            ),
            g6_input_port(
              key = "input-g-2",
              placement = "top",
              label = "2",
              arity = Inf
            ),
            g6_output_port(
              key = "output-g",
              placement = "bottom",
              label = "data",
              arity = Inf
            )
          ),
          combo = "combo2",
          children = c("h"),
          collapse = g6_collapse_options(
            collapsed = FALSE,
            placement = "right-top"
          )
        ),
        # h = head_block(n = 3) — outside combos
        g6_node(
          id = "h",
          type = "custom-rect-node",
          style = list(
            labelText = "head(n=3)"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-h",
              placement = "top",
              label = "data"
            ),
            g6_output_port(
              key = "output-h",
              placement = "bottom",
              label = "data"
            )
          ),
          children = c("i"),
          collapse = g6_collapse_options(
            collapsed = FALSE,
            placement = "right-top"
          )
        ),
        # i = scatter_block(x = "Sepal.Length", y = "Sepal.Width") — outside combos
        g6_node(
          id = "i",
          type = "custom-rect-node",
          style = list(
            labelText = "scatter plot"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-i",
              placement = "top",
              label = "data"
            ),
            g6_output_port(
              key = "output-i",
              placement = "bottom",
              label = "data"
            )
          ),
          children = c("j"),
          collapse = g6_collapse_options(
            placement = "right-top"
          )
        ),
        # --- Bubble set nodes (descendants of i, no combo) ---
        g6_node(
          id = "j",
          type = "custom-rect-node",
          style = list(labelText = "filter"),
          ports = g6_ports(
            g6_input_port(key = "input-j", placement = "top", label = "data"),
            g6_output_port(
              key = "output-j",
              placement = "bottom",
              label = "data"
            )
          ),
          children = c("k", "l"),
          collapse = g6_collapse_options(placement = "right-top")
        ),
        g6_node(
          id = "k",
          type = "custom-rect-node",
          style = list(labelText = "summarise"),
          ports = g6_ports(
            g6_input_port(key = "input-k", placement = "top", label = "data"),
            g6_output_port(
              key = "output-k",
              placement = "bottom",
              label = "data"
            )
          ),
          children = c("m"),
          collapse = g6_collapse_options(placement = "right-top")
        ),
        # --- Hull nodes (descendants of i, no combo) ---
        g6_node(
          id = "l",
          type = "custom-rect-node",
          style = list(labelText = "mutate"),
          ports = g6_ports(
            g6_input_port(key = "input-l", placement = "top", label = "data"),
            g6_output_port(
              key = "output-l",
              placement = "bottom",
              label = "data"
            )
          ),
          children = c("n"),
          collapse = g6_collapse_options(placement = "right-top")
        ),
        g6_node(
          id = "m",
          type = "custom-rect-node",
          style = list(labelText = "arrange"),
          ports = g6_ports(
            g6_input_port(key = "input-m", placement = "top", label = "data"),
            g6_output_port(
              key = "output-m",
              placement = "bottom",
              label = "data"
            )
          )
        ),
        g6_node(
          id = "n",
          type = "custom-rect-node",
          style = list(labelText = "select"),
          ports = g6_ports(
            g6_input_port(key = "input-n", placement = "top", label = "data")
          )
        )
      ),
      edges = g6_edges(
        # a -> b (data)
        g6_edge(
          source = "a",
          target = "b",
          style = list(
            sourcePort = "output-a",
            targetPort = "input-b",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # b -> c (data)
        g6_edge(
          source = "b",
          target = "c",
          style = list(
            sourcePort = "output-b",
            targetPort = "input-c",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # c -> d (data)
        g6_edge(
          source = "c",
          target = "d",
          style = list(
            sourcePort = "output-c",
            targetPort = "input-d",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # b -> f (data) — cross-combo: combo1 -> combo2
        g6_edge(
          source = "b",
          target = "f",
          style = list(
            sourcePort = "output-b",
            targetPort = "input-f",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # d -> e (input 1) — combo1 -> outside
        g6_edge(
          source = "d",
          target = "e",
          style = list(
            sourcePort = "output-d",
            targetPort = "input-e-1",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # f -> e (input 2) — combo2 -> outside
        g6_edge(
          source = "f",
          target = "e",
          style = list(
            sourcePort = "output-f",
            targetPort = "input-e-2",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # e -> g (input 1) — outside -> combo2
        g6_edge(
          source = "e",
          target = "g",
          style = list(
            sourcePort = "output-e",
            targetPort = "input-g-1",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # f -> g (input 2) — inside combo2
        g6_edge(
          source = "f",
          target = "g",
          style = list(
            sourcePort = "output-f",
            targetPort = "input-g-2",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # g -> h (data) — combo2 -> outside
        g6_edge(
          source = "g",
          target = "h",
          style = list(
            sourcePort = "output-g",
            targetPort = "input-h",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # h -> i (data) — outside -> outside
        g6_edge(
          source = "h",
          target = "i",
          style = list(
            sourcePort = "output-h",
            targetPort = "input-i",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # i -> j
        g6_edge(
          source = "i",
          target = "j",
          style = list(
            sourcePort = "output-i",
            targetPort = "input-j",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # j -> k (bubble set member)
        g6_edge(
          source = "j",
          target = "k",
          style = list(
            sourcePort = "output-j",
            targetPort = "input-k",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # j -> l (bubble set member)
        g6_edge(
          source = "j",
          target = "l",
          style = list(
            sourcePort = "output-j",
            targetPort = "input-l",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # k -> m (hull member)
        g6_edge(
          source = "k",
          target = "m",
          style = list(
            sourcePort = "output-k",
            targetPort = "input-m",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # l -> n (hull member)
        g6_edge(
          source = "l",
          target = "n",
          style = list(
            sourcePort = "output-l",
            targetPort = "input-n",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        )
      ),
      combos = g6_combos(
        g6_combo(
          "combo1",
          collapse = g6_collapse_options(visibility = "hover")
        ),
        g6_combo("combo2", collapse = g6_collapse_options())
      )
    ) |>
      g6_layout(antv_dagre_layout(sortByCombo = TRUE)) |>
      g6_options(
        animation = FALSE,
        autoFit = TRUE,
        node = list(
          style = list(
            #fill = "#CED4D9",
            #fillOpacity = 0
          )
        ),
        combo = list(),
        renderer = JS("() => new SVGRenderer()"),
        edge = list(style = list(endArrow = TRUE))
      ) |>
      g6_behaviors(
        collapse_expand(
          enable = JS(
            "(e) => {
              return e.targetType === 'combo'
            }"
          )
        ),
        click_select(multiple = TRUE),
        drag_element(),
        drag_canvas(
          enable = JS(
            "(e) => {
          return e.targetType === 'canvas' && !e.shiftKey && !e.altKey;
        }"
          )
        ),
        zoom_canvas(),
        create_edge(
          enable = JS(
            "(e) => {
            return e.targetType === 'node' && e.targetType !== 'combo';
          }"
          ),
          onFinish = NULL
        )
      ) |>
      # Allow to dynamically remove an edge or node
      g6_plugins(
        bubble_sets(
          key = "bubble-set-1",
          members = c("j", "k"),
          label = TRUE,
          labelText = "bubble set",
          fill = "#F08F56",
          stroke = "#F08F56",
          labelBackground = TRUE,
          labelPlacement = "top",
          labelFill = "#fff",
          labelPadding = 2,
          labelBackgroundFill = "#F08F56",
          labelBackgroundRadius = 5
        ),
        hull(
          key = "hull-1",
          members = c("l", "m", "n"),
          labelText = "Super hull",
          labelAutoRotate = FALSE,
          labelCloseToPath = FALSE,
          fill = "#7B68EE",
          stroke = "#7B68EE",
          labelBackground = TRUE,
          labelPlacement = "top",
          labelFill = "#fff",
          labelPadding = 2,
          labelBackgroundFill = "#7B68EE",
          labelBackgroundRadius = 5
        ),
        context_menu(
          enable = JS("(e) => true"),
          getItems = JS(
            "(e) => {
              if (e.targetType === 'edge') {
                return [{ name: 'Remove edge', value: 'remove_edge' }];
              } else if (e.targetType === 'node') {
                return [{ name: 'Remove node', value: 'remove_node' }];
              }
              return [];
            }"
          ),
          onClick = JS(
            "(value, target, current) => {
              const graph = HTMLWidgets
                .find(`#${target.closest('.g6').id}`)
                .getWidget();
              if (current.id === undefined) return;
              if (value === 'remove_edge') {
              Shiny.setInputValue(target.closest('.g6').id + '-removed_edge',
                  {id: current.id},
                  {priority: 'event'});
              } else if (value === 'remove_node') {
                // Send node ID to Shiny before removing
                Shiny.setInputValue(target.closest('.g6').id + '-removed_node',
                  {id: current.id},
                  {priority: 'event'});
              }
            }
          "
          )
        )
      )
  )

  output$clicked_port <- renderPrint({
    input[["dag-selected_port"]]
  })

  output$removed_node <- renderPrint({
    input[["dag-removed_node"]]
  })

  proxy <- g6_proxy("dag")

  # Add a new node when a port is clicked from the guide
  # at the mouse position (close to the guide)
  observeEvent(input[["dag-selected_port"]], {
    new_id <- as.character(round(as.numeric(Sys.time())))
    pos <- input[["dag-mouse_position"]]
    parent_id <- as.character(input[["dag-selected_port"]][["node"]])

    proxy |>
      g6_add_nodes(
        g6_node(
          id = new_id,
          type = "custom-circle-node",
          style = list(
            x = pos$x + 50, # avoids overlapping with the guide.
            y = pos$y,
            labelText = paste("Node", new_id)
          ),
          ports = g6_ports(
            g6_input_port(
              key = sprintf("input-%s", new_id),
              placement = "left",
              arity = Inf
            ),
            g6_output_port(
              key = sprintf("output-%s", new_id),
              placement = "right",
              arity = Inf
            )
          )
        )
      ) |>
      g6_add_edges(
        g6_edge(
          source = parent_id,
          target = new_id,
          style = list(
            sourcePort = input[["dag-selected_port"]][["port"]],
            targetPort = sprintf("input-%s", new_id),
            endArrow = TRUE
          )
        )
      )
  })

  observeEvent(input[["dag-removed_edge"]], {
    proxy |>
      g6_remove_edges(input[["dag-removed_edge"]]$id)
  })

  # Handle node removal - update parent's children list
  observeEvent(input[["dag-removed_node"]], {
    proxy |>
      g6_remove_nodes(input[["dag-removed_node"]]$id)
  })
}

shinyApp(ui, server)

library(shiny)
library(g6R)

options(
  "g6R.mode" = "dev",
  "g6R.layout_on_data_change" = TRUE
)

ui <- fluidPage(
  g6_output("dag"),
  verbatimTextOutput("clicked_port")
)

server <- function(input, output, session) {
  output$dag <- render_g6(
    g6(
      nodes = g6_nodes(
        g6_node(
          id = 1,
          type = "custom-image-node",
          style = list(
            src = "https://gw.alipayobjects.com/mdn/rms_6ae20b/afts/img/A*N4ZMS7gHsUIAAAAAAAAAAABkARQnAQ",
            labelText = "Node 1"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-1",
              placement = c(0, 1),
              label = "port 1 (visible)"
              # visibility = "visible" is default
            ),
            g6_output_port(
              key = "output-1",
              placement = "right",
              label = "port 2 (hover)",
              visibility = "hover"
            ),
            g6_input_port(
              key = "input-12",
              placement = "top",
              label = "port 3 (hidden)",
              visibility = "hidden"
            )
          )
        ),
        g6_node(
          id = 2,
          type = "custom-circle-node",
          style = list(
            labelText = "Node 2"
          ),
          ports = g6_ports(
            g6_input_port(
              key = "input-2",
              placement = "left"
            ),
            g6_output_port(
              key = "output-2",
              placement = "right",
              arity = Inf
            )
          )
        ),
        # Nodes 3 and 4 reproduce the blockr.dag styling: a bottom label with a
        # background, and an output port placed at the label surface via
        # placement = "label-bottom". The input port stays at the node top.
        g6_node(
          id = 3,
          type = "custom-image-node",
          style = list(
            src = "https://gw.alipayobjects.com/mdn/rms_6ae20b/afts/img/A*N4ZMS7gHsUIAAAAAAAAAAABkARQnAQ",
            labelText = "Dataset",
            labelFill = "#6b7280",
            labelBackground = TRUE,
            labelBackgroundFill = "#f3f4f6",
            labelBackgroundStroke = "#e5e7eb",
            labelBackgroundRadius = 4,
            labelPlacement = "bottom",
            labelOffsetY = 8,
            labelBackgroundLineWidth = 1,
            labelBackgroundOpacity = 1,
            labelPadding = c(1, 6, 1, 6),
            labelFontSize = 11,
            labelFontFamily = "Open Sans, system-ui, sans-serif"
          ),
          ports = g6_ports(
            g6_input_port(key = "input-3", placement = "top"),
            g6_output_port(key = "output-3", placement = "label-bottom")
          )
        ),
        g6_node(
          id = 4,
          type = "custom-image-node",
          style = list(
            src = "https://gw.alipayobjects.com/mdn/rms_6ae20b/afts/img/A*N4ZMS7gHsUIAAAAAAAAAAABkARQnAQ",
            labelText = "Scatter",
            labelFill = "#6b7280",
            labelBackground = TRUE,
            labelBackgroundFill = "#f3f4f6",
            labelBackgroundStroke = "#e5e7eb",
            labelBackgroundRadius = 4,
            labelPlacement = "bottom",
            labelOffsetY = 8,
            labelBackgroundLineWidth = 1,
            labelBackgroundOpacity = 1,
            labelPadding = c(1, 6, 1, 6),
            labelFontSize = 11,
            labelFontFamily = "Open Sans, system-ui, sans-serif"
          ),
          ports = g6_ports(
            g6_input_port(key = "input-4", placement = "top"),
            g6_output_port(key = "output-4", placement = "label-bottom")
          )
        )
      ),
      edges = g6_edges(
        g6_edge(
          source = 1,
          target = 2,
          style = list(
            sourcePort = "output-1",
            targetPort = "input-2",
            endArrow = TRUE,
            startArrow = FALSE,
            endArrowType = "vee"
          )
        ),
        # Connects the label-bottom output of node 3 to the top input of node 4.
        g6_edge(
          source = 3,
          target = 4,
          style = list(
            sourcePort = "output-3",
            targetPort = "input-4",
            endArrow = TRUE,
            endArrowType = "vee"
          )
        )
      )
    ) |>
      g6_layout() |>
      g6_options(
        animation = FALSE,
        edge = list(style = list(endArrow = TRUE))
      ) |>
      g6_behaviors(
        click_select(multiple = TRUE),
        drag_element(),
        drag_canvas(
          enable = JS(
            "(e) => {
          return e.targetType === 'canvas' && !e.shiftKey && !e.altKey;
        }"
          )
        ),
        zoom_canvas(),
        create_edge(
          enable = JS("(e) => true"),
          onFinish = NULL
        )
      ) |>
      # Allow to dynamically remove an edge
      g6_plugins(
        context_menu(
          enable = JS("(e) => e.targetType === 'edge'"),
          getItems = JS(
            "() => {
              return [
                { name: 'Remove edge', value: 'remove_edge' }
              ];
            }"
          ),
          onClick = JS(
            "(value, target, current) => {
              const graph = HTMLWidgets
                .find(`#${target.closest('.g6').id}`)
                .getWidget();
              console.log(current.id);
              if (current.id === undefined) return;
              if (value === 'remove_edge') {
                graph.removeEdgeData([current.id]);
                graph.draw();
              }
            }
          "
          )
        )
      )
  )

  output$clicked_port <- renderPrint({
    input[["dag-selected_port"]]
  })

  proxy <- g6_proxy("dag")

  # Add a new node when a port is clicked from the guide
  # at the mouse position (close to the guide)
  observeEvent(input[["dag-selected_port"]], {
    new_id <- round(as.numeric(Sys.time()))
    pos <- input[["dag-mouse_position"]]

    proxy |>
      g6_add_nodes(
        g6_node(
          id = new_id,
          type = "custom-circle-node",
          style = list(
            x = pos$x + 50, # avoids overlapping with the guide.
            y = pos$y,
            labelText = paste("Node", new_id)
          ),
          ports = g6_ports(
            g6_input_port(
              key = sprintf("input-%s", new_id),
              placement = "left"
            ),
            g6_output_port(
              key = sprintf("output-%s", new_id),
              placement = "right",
              arity = Inf
            )
          )
        )
      ) |>
      g6_add_edges(
        g6_edge(
          source = input[["dag-selected_port"]][["node"]],
          target = new_id,
          style = list(
            sourcePort = input[["dag-selected_port"]][["port"]],
            targetPort = sprintf("input-%s", new_id),
            endArrow = TRUE
          )
        )
      )
  })
}

shinyApp(ui, server)

library(shiny)
library(g6R)

ui <- fluidPage(
  g6_output("dag", height = "800px")
)

server <- function(input, output, session) {
  output$dag <- render_g6(
    g6(
      nodes = g6_nodes(
        # Webhook (left)
        g6_node(
          id = "webhook",
          type = "custom-rect-node",
          style = list(
            labelText = "Webhook",
            x = 100,
            y = 300,
            size = c(50, 50),
            iconSrc = "data:image/svg+xml;base64,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",
            iconWidth = 20,
            iconHeight = 20,
            radius = 8
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "right")
          )
        ),
        # AI Agent (center)
        g6_node(
          id = "ai_agent",
          type = "custom-rect-node",
          style = list(
            labelText = "AI Agent\nTools Agent\n",
            labelPlacement = "top",
            size = c(150, 100),
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDM2IDM2IiB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIGFyaWEtaGlkZGVuPSJ0cnVlIiBjbGFzcz0iaWNvbmlmeSBpY29uaWZ5LS10d2Vtb2ppIj48ZWxsaXBzZSBmaWxsPSIjZjQ5MDBjIiBjeD0iMzMuNSIgY3k9IjE0LjUiIHJ4PSIyLjUiIHJ5PSIzLjUiLz48ZWxsaXBzZSBmaWxsPSIjZjQ5MDBjIiBjeD0iMi41IiBjeT0iMTQuNSIgcng9IjIuNSIgcnk9IjMuNSIvPjxwYXRoIGZpbGw9IiNmZmFjMzMiIGQ9Ik0zNCAxOWExIDEgMCAwIDEtMSAxaC0zYTEgMSAwIDAgMS0xLTF2LTlhMSAxIDAgMCAxIDEtMWgzYTEgMSAwIDAgMSAxIDF6TTcgMTlhMSAxIDAgMCAxLTEgMUgzYTEgMSAwIDAgMS0xLTF2LTlhMSAxIDAgMCAxIDEtMWgzYTEgMSAwIDAgMSAxIDF6Ii8+PHBhdGggZmlsbD0iI2ZmY2M0ZCIgZD0iTTI4IDVjMCAyLjc2MS00LjQ3OCA0LTEwIDRTOCA3Ljc2MSA4IDVzNC40NzctNSAxMC01IDEwIDIuMjM5IDEwIDUiLz48cGF0aCBmaWxsPSIjZjQ5MDBjIiBkPSJNMjUgNC4wODNDMjUgNS42OTQgMjEuODY1IDcgMTggN3MtNy0xLjMwNi03LTIuOTE3IDMuMTM0LTIuOTE3IDctMi45MTcgNyAxLjMwNiA3IDIuOTE3Ii8+PHBhdGggZmlsbD0iIzI2OSIgZD0iTTMwIDUuNUMzMCA2Ljg4MSAyOC44ODEgNyAyNy41IDdoLTE5QzcuMTE5IDcgNiA2Ljg4MSA2IDUuNVM3LjExOSAzIDguNSAzaDE5QTIuNSAyLjUgMCAwIDEgMzAgNS41Ii8+PHBhdGggZmlsbD0iIzU1YWNlZSIgZD0iTTMwIDZINmEyIDIgMCAwIDAtMiAydjI2aDI4VjhhMiAyIDAgMCAwLTItMiIvPjxwYXRoIGZpbGw9IiMzYjg4YzMiIGQ9Ik0zNSAzM3YtMWEyIDIgMCAwIDAtMi0ySDIyLjA3MWwtMy4zNjQgMy4zNjRhMSAxIDAgMCAxLTEuNDE0IDBMMTMuOTI5IDMwSDNhMiAyIDAgMCAwLTIgMnYxYzAgMS4xMDQtLjEwNCAyIDEgMmgzMmMxLjEwNCAwIDEtLjg5NiAxLTIiLz48Y2lyY2xlIGZpbGw9IiNmZmYiIGN4PSIyNC41IiBjeT0iMTQuNSIgcj0iNC41Ii8+PGNpcmNsZSBmaWxsPSIjZGQyZTQ0IiBjeD0iMjQuNSIgY3k9IjE0LjUiIHI9IjIuNzIxIi8+PGNpcmNsZSBmaWxsPSIjZmZmIiBjeD0iMTEuNSIgY3k9IjE0LjUiIHI9IjQuNSIvPjxwYXRoIGZpbGw9IiNmNWY4ZmEiIGQ9Ik0yOSAyNS41YTIuNSAyLjUgMCAwIDEtMi41IDIuNWgtMTdhMi41IDIuNSAwIDEgMSAwLTVoMTdhMi41IDIuNSAwIDAgMSAyLjUgMi41Ii8+PHBhdGggZmlsbD0iI2NjZDZkZCIgZD0iTTE3IDIzaDJ2NWgtMnptLTUgMGgydjVoLTJ6bTEwIDBoMnY1aC0yek03IDI1LjVhMi41IDIuNSAwIDAgMCAyIDIuNDV2LTQuOWEyLjUgMi41IDAgMCAwLTIgMi40NW0yMC0yLjQ1djQuODk5YTIuNSAyLjUgMCAwIDAgMC00Ljg5OSIvPjxjaXJjbGUgZmlsbD0iI2RkMmU0NCIgY3g9IjExLjUiIGN5PSIxNC41IiByPSIyLjcyMSIvPjwvc3ZnPg==",
            iconWidth = 40,
            iconHeight = 40,
            x = 320,
            y = 300,
            radius = 8
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "out", placement = "right"),
            g6_input_port(
              key = "down1",
              label = "chat model",
              placement = c(0.15, 1)
            ),
            g6_input_port(
              key = "down2",
              label = "mode memory",
              placement = c(0.32, 1)
            ),
            g6_input_port(
              key = "down3",
              label = "tool",
              arity = 3,
              placement = c(0.49, 1)
            ),
            g6_input_port(
              key = "down4",
              label = "output parser",
              placement = c(0.66, 1)
            )
          )
        ),
        # Switch (right of AI Agent)
        g6_node(
          id = "switch",
          type = "custom-rect-node",
          style = list(
            labelText = "Switch \n mode: rules",
            size = c(50, 50),
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCA2MCA2MCIgeG1sOnNwYWNlPSJwcmVzZXJ2ZSI+PHBhdGggc3R5bGU9ImZpbGw6bm9uZTtzdHJva2U6IzM4NDU0ZjtzdHJva2Utd2lkdGg6MjtzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMCIgZD0iTTI4LjY2NCA0VjFtMCA1OFY0MG0wLTIwdjQiLz48cGF0aCBzdHlsZT0iZmlsbDojMjNhMjRkIiBkPSJNNDguNTUzIDIwSDEwLjA2MWEyLjM5NyAyLjM5NyAwIDAgMS0yLjM5Ny0yLjM5N1Y2LjM5N0EyLjM5NyAyLjM5NyAwIDAgMSAxMC4wNjEgNGgzOC40OTJsOC43MTEgNy44NGEuMjE0LjIxNCAwIDAgMSAwIC4zMTl6Ii8+PHBhdGggc3R5bGU9ImZpbGw6I2ViYmExNiIgZD0iTTExLjQ0NyA0MGgzOC44MjFhMi4zOTcgMi4zOTcgMCAwIDAgMi4zOTctMi4zOTdWMjYuMzk3QTIuMzk3IDIuMzk3IDAgMCAwIDUwLjI2OCAyNEgxMS40NDdsLTguNzExIDcuODRhLjIxNC4yMTQgMCAwIDAgMCAuMzE5eiIvPjwvc3ZnPg==",
            iconWidth = 30,
            iconHeight = 30,
            x = 520,
            y = 300,
            radius = 8
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "get", placement = c(1, 0.2)),
            g6_output_port(key = "post", placement = c(1, 0.5)),
            g6_output_port(key = "delete", placement = c(1, 0.8))
          )
        ),
        # Right-side rectangular nodes (top to bottom)
        g6_node(
          id = "get_properties",
          type = "custom-rect-node",
          style = list(
            labelText = "Get properties",
            x = 700,
            y = 180,
            size = c(50, 50),
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDEwMCAxMDAiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PHJhZGlhbEdyYWRpZW50IGlkPSJhIiBjeD0iNTAlIiBjeT0iNTAlIiBmeD0iNTAlIiBmeT0iNTAlIiByPSI1MCUiPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiM2NzliY2I7c3RvcC1vcGFjaXR5Oi43NSIgb2Zmc2V0PSIwJSIvPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiMxMjMxNDE7c3RvcC1vcGFjaXR5OjEiIG9mZnNldD0iMTAwJSIvPjwvcmFkaWFsR3JhZGllbnQ+PC9kZWZzPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDYiIHJ5PSI0NCIgc3R5bGU9InN0cm9rZS13aWR0aDo0O3N0cm9rZTojZGRkO2ZpbGw6bm9uZSIvPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDciIHJ5PSI0NCIgc3R5bGU9ImZpbGw6dXJsKCNhKSIvPjxnIHN0eWxlPSJmaWxsOm5vbmU7c3Ryb2tlOiNhYWE7c3Ryb2tlLXdpZHRoOjEuNXB4O3N0cm9rZS1saW5lY2FwOmJ1dHQiPjxwYXRoIGQ9Ik01MCA5NGMtMTAtNC0yOC0xOC0yOC00NCAwLTI5IDE4LTM4IDI4LTQ0djg4czI5LTEwIDI4LTQ1UzUwIDYgNTAgNk00IDUwaDkyIi8+PHBhdGggZD0iTTE3IDIwczEyIDEyIDMzIDEyYzIyIDAgMzQtMTIgMzQtMTJNMTQgNzhzMTMtMTEgMzYtMTFjMjUgMCAzNiAxMSAzNiAxMSIvPjwvZz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20,
            radius = 8
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "out", placement = "right")
          )
        ),
        g6_node(
          id = "structure_response",
          type = "custom-rect-node",
          style = list(
            labelText = "Structure Response",
            x = 900,
            y = 180,
            size = c(50, 50),
            radius = 8,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCA1MTIgNTEyIiB4bWw6c3BhY2U9InByZXNlcnZlIj48cGF0aCBzdHlsZT0iZmlsbDojNmRjODJhIiBkPSJNNDk1LjMwNCA1MTJIMTYuNjk2QzcuNDc5IDUxMiAwIDUwNC41MjcgMCA0OTUuMzA0VjE2LjY5NkMwIDcuNDczIDcuNDc5IDAgMTYuNjk2IDBoNDc4LjYwOUM1MDQuNTIxIDAgNTEyIDcuNDczIDUxMiAxNi42OTZ2NDc4LjYwOWMwIDkuMjIyLTcuNDc5IDE2LjY5NS0xNi42OTYgMTYuNjk1Ii8+PHBhdGggc3R5bGU9ImZpbGw6IzYxYjMyNSIgZD0iTTQ5NS4zMDQgMEgyNTZ2NTEyaDIzOS4zMDRjOS4yMTcgMCAxNi42OTYtNy40NzMgMTYuNjk2LTE2LjY5NlYxNi42OTZDNTEyIDcuNDczIDUwNC41MjEgMCA0OTUuMzA0IDAiLz48cGF0aCBzdHlsZT0iZmlsbDojZmZmIiBkPSJNMTgzLjY1MiA0NDUuMjE3SDExNi44N2MtOS4yMTcgMC0xNi42OTYtNy40NzMtMTYuNjk2LTE2LjY5NlY4My40NzhjMC05LjIyMyA3LjQ3OS0xNi42OTYgMTYuNjk2LTE2LjY5Nmg2Ni43ODNjOS4yMTcgMCAxNi42OTYgNy40NzMgMTYuNjk2IDE2LjY5NnMtNy40NzkgMTYuNjk2LTE2LjY5NiAxNi42OTZoLTUwLjA4N3YzMTEuNjUyaDUwLjA4N2M5LjIxNyAwIDE2LjY5NiA3LjQ3MyAxNi42OTYgMTYuNjk2cy03LjQ4IDE2LjY5NS0xNi42OTcgMTYuNjk1Ii8+PHBhdGggc3R5bGU9ImZpbGw6I2ZmZTZiMyIgZD0iTTM5NS4xMyA0NDUuMjE3aC02Ni43ODNjLTkuMjE3IDAtMTYuNjk2LTcuNDczLTE2LjY5Ni0xNi42OTZzNy40NzktMTYuNjk2IDE2LjY5Ni0xNi42OTZoNTAuMDg3VjEwMC4xNzRoLTUwLjA4N2MtOS4yMTcgMC0xNi42OTYtNy40NzMtMTYuNjk2LTE2LjY5NnM3LjQ3OS0xNi42OTYgMTYuNjk2LTE2LjY5Nmg2Ni43ODNjOS4yMTcgMCAxNi42OTYgNy40NzMgMTYuNjk2IDE2LjY5NnYzNDUuMDQzYzAgOS4yMjMtNy40NzggMTYuNjk2LTE2LjY5NiAxNi42OTYiLz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "out", placement = "right")
          )
        ),
        g6_node(
          id = "post_url",
          type = "custom-rect-node",
          style = list(
            labelText = "Post URL",
            x = 700,
            y = 300,
            size = c(50, 50),
            radius = 8,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDEwMCAxMDAiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PHJhZGlhbEdyYWRpZW50IGlkPSJhIiBjeD0iNTAlIiBjeT0iNTAlIiBmeD0iNTAlIiBmeT0iNTAlIiByPSI1MCUiPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiM2NzliY2I7c3RvcC1vcGFjaXR5Oi43NSIgb2Zmc2V0PSIwJSIvPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiMxMjMxNDE7c3RvcC1vcGFjaXR5OjEiIG9mZnNldD0iMTAwJSIvPjwvcmFkaWFsR3JhZGllbnQ+PC9kZWZzPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDYiIHJ5PSI0NCIgc3R5bGU9InN0cm9rZS13aWR0aDo0O3N0cm9rZTojZGRkO2ZpbGw6bm9uZSIvPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDciIHJ5PSI0NCIgc3R5bGU9ImZpbGw6dXJsKCNhKSIvPjxnIHN0eWxlPSJmaWxsOm5vbmU7c3Ryb2tlOiNhYWE7c3Ryb2tlLXdpZHRoOjEuNXB4O3N0cm9rZS1saW5lY2FwOmJ1dHQiPjxwYXRoIGQ9Ik01MCA5NGMtMTAtNC0yOC0xOC0yOC00NCAwLTI5IDE4LTM4IDI4LTQ0djg4czI5LTEwIDI4LTQ1UzUwIDYgNTAgNk00IDUwaDkyIi8+PHBhdGggZD0iTTE3IDIwczEyIDEyIDMzIDEyYzIyIDAgMzQtMTIgMzQtMTJNMTQgNzhzMTMtMTEgMzYtMTFjMjUgMCAzNiAxMSAzNiAxMSIvPjwvZz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "out", placement = "right")
          )
        ),
        g6_node(
          id = "delete_return",
          type = "custom-rect-node",
          style = list(
            labelText = "Delete/Return",
            x = 700,
            y = 420,
            size = c(50, 50),
            radius = 8,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCA2MCA2MCIgeG1sOnNwYWNlPSJwcmVzZXJ2ZSI+PHBhdGggc3R5bGU9ImZpbGw6bm9uZTtzdHJva2U6IzM4NDU0ZjtzdHJva2Utd2lkdGg6MjtzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJva2UtbWl0ZXJsaW1pdDoxMCIgZD0iTTI4LjY2NCA0VjFtMCA1OFY0MG0wLTIwdjQiLz48cGF0aCBzdHlsZT0iZmlsbDojMjNhMjRkIiBkPSJNNDguNTUzIDIwSDEwLjA2MWEyLjM5NyAyLjM5NyAwIDAgMS0yLjM5Ny0yLjM5N1Y2LjM5N0EyLjM5NyAyLjM5NyAwIDAgMSAxMC4wNjEgNGgzOC40OTJsOC43MTEgNy44NGEuMjE0LjIxNCAwIDAgMSAwIC4zMTl6Ii8+PHBhdGggc3R5bGU9ImZpbGw6I2ViYmExNiIgZD0iTTExLjQ0NyA0MGgzOC44MjFhMi4zOTcgMi4zOTcgMCAwIDAgMi4zOTctMi4zOTdWMjYuMzk3QTIuMzk3IDIuMzk3IDAgMCAwIDUwLjI2OCAyNEgxMS40NDdsLTguNzExIDcuODRhLjIxNC4yMTQgMCAwIDAgMCAuMzE5eiIvPjwvc3ZnPg==",
            iconWidth = 30,
            iconHeight = 30
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "out1", placement = c(1, 0.3)),
            g6_output_port(key = "out2", placement = c(1, 0.7))
          )
        ),
        g6_node(
          id = "delete_url",
          type = "custom-rect-node",
          style = list(
            labelText = "Delete URL",
            x = 900,
            y = 380,
            size = c(50, 50),
            radius = 8,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDEwMCAxMDAiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PHJhZGlhbEdyYWRpZW50IGlkPSJhIiBjeD0iNTAlIiBjeT0iNTAlIiBmeD0iNTAlIiBmeT0iNTAlIiByPSI1MCUiPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiM2NzliY2I7c3RvcC1vcGFjaXR5Oi43NSIgb2Zmc2V0PSIwJSIvPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiMxMjMxNDE7c3RvcC1vcGFjaXR5OjEiIG9mZnNldD0iMTAwJSIvPjwvcmFkaWFsR3JhZGllbnQ+PC9kZWZzPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDYiIHJ5PSI0NCIgc3R5bGU9InN0cm9rZS13aWR0aDo0O3N0cm9rZTojZGRkO2ZpbGw6bm9uZSIvPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDciIHJ5PSI0NCIgc3R5bGU9ImZpbGw6dXJsKCNhKSIvPjxnIHN0eWxlPSJmaWxsOm5vbmU7c3Ryb2tlOiNhYWE7c3Ryb2tlLXdpZHRoOjEuNXB4O3N0cm9rZS1saW5lY2FwOmJ1dHQiPjxwYXRoIGQ9Ik01MCA5NGMtMTAtNC0yOC0xOC0yOC00NCAwLTI5IDE4LTM4IDI4LTQ0djg4czI5LTEwIDI4LTQ1UzUwIDYgNTAgNk00IDUwaDkyIi8+PHBhdGggZD0iTTE3IDIwczEyIDEyIDMzIDEyYzIyIDAgMzQtMTIgMzQtMTJNMTQgNzhzMTMtMTEgMzYtMTFjMjUgMCAzNiAxMSAzNiAxMSIvPjwvZz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "out", placement = "right")
          )
        ),
        g6_node(
          id = "return_output",
          type = "custom-rect-node",
          style = list(
            labelText = "Return output",
            x = 900,
            y = 480,
            size = c(50, 50),
            radius = 8,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDEwMCAxMDAiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PHJhZGlhbEdyYWRpZW50IGlkPSJhIiBjeD0iNTAlIiBjeT0iNTAlIiBmeD0iNTAlIiBmeT0iNTAlIiByPSI1MCUiPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiM2NzliY2I7c3RvcC1vcGFjaXR5Oi43NSIgb2Zmc2V0PSIwJSIvPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiMxMjMxNDE7c3RvcC1vcGFjaXR5OjEiIG9mZnNldD0iMTAwJSIvPjwvcmFkaWFsR3JhZGllbnQ+PC9kZWZzPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDYiIHJ5PSI0NCIgc3R5bGU9InN0cm9rZS13aWR0aDo0O3N0cm9rZTojZGRkO2ZpbGw6bm9uZSIvPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDciIHJ5PSI0NCIgc3R5bGU9ImZpbGw6dXJsKCNhKSIvPjxnIHN0eWxlPSJmaWxsOm5vbmU7c3Ryb2tlOiNhYWE7c3Ryb2tlLXdpZHRoOjEuNXB4O3N0cm9rZS1saW5lY2FwOmJ1dHQiPjxwYXRoIGQ9Ik01MCA5NGMtMTAtNC0yOC0xOC0yOC00NCAwLTI5IDE4LTM4IDI4LTQ0djg4czI5LTEwIDI4LTQ1UzUwIDYgNTAgNk00IDUwaDkyIi8+PHBhdGggZD0iTTE3IDIwczEyIDEyIDMzIDEyYzIyIDAgMzQtMTIgMzQtMTJNMTQgNzhzMTMtMTEgMzYtMTFjMjUgMCAzNiAxMSAzNiAxMSIvPjwvZz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_input_port(key = "in", placement = "left"),
            g6_output_port(key = "out", placement = "right")
          )
        ),
        # Bottom circle nodes (left to right, spaced out)
        g6_node(
          id = "google_gemini",
          type = "custom-circle-node",
          style = list(
            labelText = "Google Gemini\nChat Model",
            x = 160,
            y = 520,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCAzMiAzMiIgeG1sOnNwYWNlPSJwcmVzZXJ2ZSI+PHBhdGggZD0iTTguOSAxNmMwIC42LjEgMS4yLjIgMS44TDExIDE2bC0xLjgtMS44cS0uMy45LS4zIDEuOCIgc3R5bGU9ImZpbGw6bm9uZSIvPjxwYXRoIGQ9Ik0xNiAyMy4xYy0zLjMgMC02LTIuMi02LjgtNS4ybC02LjcgNi43QzUuMyAyOSAxMC4zIDMyIDE2IDMyYzMuMSAwIDYtLjkgOC41LTIuNWwtNi43LTYuN3EtLjkuMy0xLjguMyIgc3R5bGU9ImZpbGw6IzM0YTg1MyIvPjxwYXRoIGQ9Ik0zMiAxMy44Yy0uMS0uNS0uNS0uOC0xLS44SDE2Yy0uNiAwLTEgLjQtMSAxdjVjMCAuNi40IDEgMSAxaDUuM2MtLjkgMS40LTIuMiAyLjMtMy41IDIuOGw2LjcgNi43QzI5IDI2LjcgMzIgMjEuNyAzMiAxNnYtLjdxLjE1LS42IDAtMS41IiBzdHlsZT0iZmlsbDojNDI4NWY0Ii8+PHBhdGggZD0iTTguOSAxNmMwLS42LjEtMS4yLjItMS44TDIuNSA3LjVDLjkgMTAgMCAxMi45IDAgMTZzLjkgNiAyLjUgOC41bDYuNy02LjdxLS4zLS45LS4zLTEuOCIgc3R5bGU9ImZpbGw6I2ZiYmMwNSIvPjxwYXRoIGQ9Ik0yOC41IDZjLTEuMS0xLjQtMi41LTIuNi00LTMuNkMyMiAuOSAxOS4xIDAgMTYgMCAxMC4zIDAgNS4zIDMgMi41IDcuNWw2LjcgNi43QzEwIDExLjIgMTIuOCA5IDE2IDlxLjkgMCAxLjguM2MuOS4zIDEuNy44IDIuNiAxLjUuMy4zLjcuMyAxLjEuMWw2LjctMy4zYy4zLS4xLjUtLjQuNS0uNy4xLS4zIDAtLjYtLjItLjkiIHN0eWxlPSJmaWxsOiNlYTQzMzUiLz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "top")
          )
        ),
        g6_node(
          id = "proxmox_api_doc",
          type = "custom-circle-node",
          style = list(
            labelText = "Proxmox API\nDocumentation",
            x = 250,
            y = 600,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDEwMCAxMDAiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PHJhZGlhbEdyYWRpZW50IGlkPSJhIiBjeD0iNTAlIiBjeT0iNTAlIiBmeD0iNTAlIiBmeT0iNTAlIiByPSI1MCUiPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiM2NzliY2I7c3RvcC1vcGFjaXR5Oi43NSIgb2Zmc2V0PSIwJSIvPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiMxMjMxNDE7c3RvcC1vcGFjaXR5OjEiIG9mZnNldD0iMTAwJSIvPjwvcmFkaWFsR3JhZGllbnQ+PC9kZWZzPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDYiIHJ5PSI0NCIgc3R5bGU9InN0cm9rZS13aWR0aDo0O3N0cm9rZTojZGRkO2ZpbGw6bm9uZSIvPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDciIHJ5PSI0NCIgc3R5bGU9ImZpbGw6dXJsKCNhKSIvPjxnIHN0eWxlPSJmaWxsOm5vbmU7c3Ryb2tlOiNhYWE7c3Ryb2tlLXdpZHRoOjEuNXB4O3N0cm9rZS1saW5lY2FwOmJ1dHQiPjxwYXRoIGQ9Ik01MCA5NGMtMTAtNC0yOC0xOC0yOC00NCAwLTI5IDE4LTM4IDI4LTQ0djg4czI5LTEwIDI4LTQ1UzUwIDYgNTAgNk00IDUwaDkyIi8+PHBhdGggZD0iTTE3IDIwczEyIDEyIDMzIDEyYzIyIDAgMzQtMTIgMzQtMTJNMTQgNzhzMTMtMTEgMzYtMTFjMjUgMCAzNiAxMSAzNiAxMSIvPjwvZz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "top")
          )
        ),
        g6_node(
          id = "proxmox_api_wiki",
          type = "custom-circle-node",
          style = list(
            labelText = "Proxmox API\nWiki",
            x = 340,
            y = 600,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDEwMCAxMDAiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PHJhZGlhbEdyYWRpZW50IGlkPSJhIiBjeD0iNTAlIiBjeT0iNTAlIiBmeD0iNTAlIiBmeT0iNTAlIiByPSI1MCUiPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiM2NzliY2I7c3RvcC1vcGFjaXR5Oi43NSIgb2Zmc2V0PSIwJSIvPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiMxMjMxNDE7c3RvcC1vcGFjaXR5OjEiIG9mZnNldD0iMTAwJSIvPjwvcmFkaWFsR3JhZGllbnQ+PC9kZWZzPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDYiIHJ5PSI0NCIgc3R5bGU9InN0cm9rZS13aWR0aDo0O3N0cm9rZTojZGRkO2ZpbGw6bm9uZSIvPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDciIHJ5PSI0NCIgc3R5bGU9ImZpbGw6dXJsKCNhKSIvPjxnIHN0eWxlPSJmaWxsOm5vbmU7c3Ryb2tlOiNhYWE7c3Ryb2tlLXdpZHRoOjEuNXB4O3N0cm9rZS1saW5lY2FwOmJ1dHQiPjxwYXRoIGQ9Ik01MCA5NGMtMTAtNC0yOC0xOC0yOC00NCAwLTI5IDE4LTM4IDI4LTQ0djg4czI5LTEwIDI4LTQ1UzUwIDYgNTAgNk00IDUwaDkyIi8+PHBhdGggZD0iTTE3IDIwczEyIDEyIDMzIDEyYzIyIDAgMzQtMTIgMzQtMTJNMTQgNzhzMTMtMTEgMzYtMTFjMjUgMCAzNiAxMSAzNiAxMSIvPjwvZz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "top")
          )
        ),
        g6_node(
          id = "proxmox",
          type = "custom-circle-node",
          style = list(
            labelText = "Proxmox",
            x = 430,
            y = 520,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDEwMCAxMDAiIHhtbG5zPSJodHRwOi8vd3d3LnczLm9yZy8yMDAwL3N2ZyI+PGRlZnM+PHJhZGlhbEdyYWRpZW50IGlkPSJhIiBjeD0iNTAlIiBjeT0iNTAlIiBmeD0iNTAlIiBmeT0iNTAlIiByPSI1MCUiPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiM2NzliY2I7c3RvcC1vcGFjaXR5Oi43NSIgb2Zmc2V0PSIwJSIvPjxzdG9wIHN0eWxlPSJzdG9wLWNvbG9yOiMxMjMxNDE7c3RvcC1vcGFjaXR5OjEiIG9mZnNldD0iMTAwJSIvPjwvcmFkaWFsR3JhZGllbnQ+PC9kZWZzPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDYiIHJ5PSI0NCIgc3R5bGU9InN0cm9rZS13aWR0aDo0O3N0cm9rZTojZGRkO2ZpbGw6bm9uZSIvPjxlbGxpcHNlIGN4PSI1MCIgY3k9IjUwIiByeD0iNDciIHJ5PSI0NCIgc3R5bGU9ImZpbGw6dXJsKCNhKSIvPjxnIHN0eWxlPSJmaWxsOm5vbmU7c3Ryb2tlOiNhYWE7c3Ryb2tlLXdpZHRoOjEuNXB4O3N0cm9rZS1saW5lY2FwOmJ1dHQiPjxwYXRoIGQ9Ik01MCA5NGMtMTAtNC0yOC0xOC0yOC00NCAwLTI5IDE4LTM4IDI4LTQ0djg4czI5LTEwIDI4LTQ1UzUwIDYgNTAgNk00IDUwaDkyIi8+PHBhdGggZD0iTTE3IDIwczEyIDEyIDMzIDEyYzIyIDAgMzQtMTIgMzQtMTJNMTQgNzhzMTMtMTEgMzYtMTFjMjUgMCAzNiAxMSAzNiAxMSIvPjwvZz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "top")
          )
        ),
        g6_node(
          id = "auto_fixing_parser",
          type = "custom-rect-node",
          style = list(
            labelText = "Auto-fixing\nOutput Parser",
            labelPlacement = "right",
            size = c(100, 50),
            x = 520,
            y = 600,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHZpZXdCb3g9IjAgMCA2NCA2NCIgeG1sOnNwYWNlPSJwcmVzZXJ2ZSI+PHN0eWxlPi5zdDF7b3BhY2l0eTouMn0uc3Qye2ZpbGw6IzIzMWYyMH0uc3Qze2ZpbGw6IzRmNWQ3M30uc3Q1e2ZpbGw6I2ZmZn08L3N0eWxlPjxnIGlkPSJMYXllcl8xIj48Y2lyY2xlIGN4PSIzMiIgY3k9IjMyIiByPSIzMiIgc3R5bGU9ImZpbGw6Izc2YzJhZiIvPjxnIGNsYXNzPSJzdDEiPjxwYXRoIGNsYXNzPSJzdDIiIGQ9Ik00OS45IDQ2LjIgMzUuOCAzMi4xYy0xLTEtMi41LTEuNC0zLjgtMWwtOC4xLTgtMy01LjktNS45LTMtMi45IDIuOCAzIDUuOSA1LjkgMyA4LjEgOGMtLjQgMS4zIDAgMi44IDEgMy44bDE0LjEgMTQuMWMxLjYgMS42IDQuMSAxLjYgNS43IDAgMS42LTEuNSAxLjYtNCAwLTUuNiIvPjwvZz48cGF0aCBjbGFzcz0ic3QzIiBkPSJNMzQuNSAzNy4zIDE4LjQgMjEuNGwyLjktMi44IDE2LjEgMTUuOXoiLz48cGF0aCBkPSJNNDkuOSA0OS45Yy0xLjYgMS42LTQuMSAxLjYtNS43IDBMMzAuMSAzNS44Yy0xLjYtMS42LTEuNi00LjEgMC01LjdzNC4xLTEuNiA1LjcgMGwxNC4xIDE0LjFjMS42IDEuNiAxLjYgNC4xIDAgNS43IiBzdHlsZT0iZmlsbDojZjVjZjg3Ii8+PGcgY2xhc3M9InN0MSI+PHBhdGggY2xhc3M9InN0MiIgZD0ibTUzLjMgMTguNC02LjUgNi41Yy0xLjYgMS42LTQuMSAxLjYtNS43IDBzLTEuNi00LjEgMC01LjdsNi41LTYuNWMtMS4xLS40LTIuMy0uNy0zLjYtLjctNS41IDAtMTAgNC41LTEwIDEwIDAgMS4zLjMgMi41LjcgMy42bC0xMSAxMWMtMS4yLS4zLTIuNC0uNi0zLjctLjYtNS41IDAtMTAgNC41LTEwIDEwIDAgMS4zLjMgMi41LjcgMy42bDYuNS02LjVjMS42LTEuNiA0LjEtMS42IDUuNyAwczEuNiA0LjEgMCA1LjdsLTYuNSA2LjVjMS4xLjQgMi40LjcgMy42LjcgNS41IDAgMTAtNC41IDEwLTEwIDAtMS4zLS4zLTIuNS0uNy0zLjZsMTEtMTFjMS4xLjQgMi40LjcgMy42LjcgNS41IDAgMTAtNC41IDEwLTEwIC4xLTEuNC0uMi0yLjYtLjYtMy43Ii8+PC9nPjxwYXRoIGNsYXNzPSJzdDUiIGQ9Ik0yMy4wOSAzNS4yNTMgMzUuMTgyIDIzLjE2bDUuNjU3IDUuNjU3TDI4Ljc0NyA0MC45MXoiLz48cGF0aCBjbGFzcz0ic3QzIiBkPSJtMjEgMjMuOS01LjktMy0zLTUuOSAyLjgtMi45IDUuOSAzLjEgMy4xIDUuOXoiLz48cGF0aCBjbGFzcz0ic3Q1IiBkPSJNNDYuOCAyMi44Yy0xLjYgMS42LTQuMSAxLjYtNS43IDBzLTEuNi00LjEgMC01LjdsNi41LTYuNWMtMS4xLS4zLTIuMy0uNi0zLjYtLjYtNS41IDAtMTAgNC41LTEwIDEwczQuNSAxMCAxMCAxMCAxMC00LjUgMTAtMTBjMC0xLjMtLjMtMi41LS43LTMuNnpNMTcuMiA0MS4yYzEuNi0xLjYgNC4xLTEuNiA1LjcgMHMxLjYgNC4xIDAgNS43bC02LjUgNi41YzEuMS40IDIuNC43IDMuNi43IDUuNSAwIDEwLTQuNSAxMC0xMHMtNC41LTEwLTEwLTEwLTEwIDQuNS0xMCAxMGMwIDEuMy4zIDIuNS43IDMuNnoiLz48L2c+PGcgaWQ9IkxheWVyXzIiLz48L3N2Zz4=",
            iconWidth = 30,
            iconHeight = 30,
            radius = 20
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "top"),
            g6_input_port(key = "in1", placement = c(0.7, 1)),
            g6_input_port(key = "in2", placement = c(0.4, 1))
          )
        ),
        g6_node(
          id = "groq_chat_model",
          type = "custom-circle-node",
          style = list(
            labelText = "Groq Chat\nModel",
            x = 610,
            y = 750,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB3aWR0aD0iODAwIiBoZWlnaHQ9IjgwMCIgdmlld0JveD0iMCAwIDY0IDY0IiB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciPjxkZWZzPjxzdHlsZT4uY2xzLTR7ZmlsbDojNTQ1OTZlfS5jbHMtNXtmaWxsOiM2MWM1YTh9PC9zdHlsZT48L2RlZnM+PHRpdGxlPmNoYXQ8L3RpdGxlPjxnIGlkPSJjaGF0Ij48cGF0aCBkPSJNMzUgOEMyMC4wOSA4IDggMTcuNCA4IDI5YzAgNi4zOSAzLjY4IDEyLjExIDkuNDcgMTZMMTUgNTNsOC4yNi01LjA4QTMzLjUgMzMuNSAwIDAgMCAzNSA1MGMxNC45MSAwIDI3LTkuNCAyNy0yMVM0OS45MSA4IDM1IDgiIHN0eWxlPSJmaWxsOiNmOTk7c3Ryb2tlOiM1NDU5NmU7c3Ryb2tlLWxpbmVjYXA6cm91bmQ7c3Ryb2tlLWxpbmVqb2luOnJvdW5kO3N0cm9rZS13aWR0aDoycHgiLz48cGF0aCBkPSJNMjkgMTJDMTQuMDkgMTIgMiAyMS40IDIgMzNjMCA2LjM5IDMuNjggMTIuMTEgOS40NyAxNkw5IDU3bDguMjYtNS4wOEEzMy41IDMzLjUgMCAwIDAgMjkgNTRjMTQuOTEgMCAyNy05LjQgMjctMjFTNDMuOTEgMTIgMjkgMTIiIHN0eWxlPSJzdHJva2U6IzU0NTk2ZTtzdHJva2UtbGluZWNhcDpyb3VuZDtzdHJva2UtbGluZWpvaW46cm91bmQ7c3Ryb2tlLXdpZHRoOjJweDtmaWxsOiM5YWY0M2IiLz48cGF0aCBjbGFzcz0iY2xzLTQiIGQ9Ik0xMC44MSA1MS4xMSA5IDU3bDcuMjgtNC40OGEzMyAzMyAwIDAgMS01LjQ3LTEuNDFNMzAuMzQgMTJDNDEuMTcgMTQuNzYgNDkgMjIuNjcgNDkgMzJjMCAxMS42LTEyLjA5IDIxLTI3IDIxaC0xLjM0QTM0LjMgMzQuMyAwIDAgMCAyOSA1NGMxNC45MSAwIDI3LTkuNCAyNy0yMSAwLTExLjI1LTExLjM3LTIwLjQzLTI1LjY2LTIxIiBzdHlsZT0ib3BhY2l0eTouMSIvPjxjaXJjbGUgY2xhc3M9ImNscy00IiBjeD0iMzEiIGN5PSIyNyIgcj0iMiIvPjxjaXJjbGUgY2xhc3M9ImNscy00IiBjeD0iMTQiIGN5PSIyNyIgcj0iMiIvPjxjaXJjbGUgY2xhc3M9ImNscy01IiBjeD0iMzEiIGN5PSIzMiIgcj0iMiIvPjxjaXJjbGUgY2xhc3M9ImNscy01IiBjeD0iMTQiIGN5PSIzMiIgcj0iMiIvPjxlbGxpcHNlIGNsYXNzPSJjbHMtNCIgY3g9IjIyIiBjeT0iMzYiIHJ4PSI0IiByeT0iNSIvPjxlbGxpcHNlIGN4PSIyMiIgY3k9IjM5IiByeD0iMiIgcnk9IjEiIHN0eWxlPSJmaWxsOiNmZjkyN2QiLz48L2c+PC9zdmc+",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "top")
          )
        ),
        g6_node(
          id = "structured_output_parser",
          type = "custom-circle-node",
          style = list(
            labelText = "Structured\nOutput Parser",
            x = 450,
            y = 750,
            iconSrc = "data:image/svg+xml;base64,PHN2ZyB4bWxucz0iaHR0cDovL3d3dy53My5vcmcvMjAwMC9zdmciIHdpZHRoPSI4MDAiIGhlaWdodD0iODAwIiB2aWV3Qm94PSIwIDAgMzIgMzIiIHhtbDpzcGFjZT0icHJlc2VydmUiPjxwYXRoIGQ9Ik0yOSAzMkgzYTMgMyAwIDAgMS0zLTNWM2EzIDMgMCAwIDEgMy0zaDI2YTMgMyAwIDAgMSAzIDN2MjZhMyAzIDAgMCAxLTMgMyIgc3R5bGU9ImZpbGw6I2YyYzk5ZSIvPjxwYXRoIGQ9Ik0yNyAzMkgzYTMgMyAwIDAgMS0zLTNWM2EzIDMgMCAwIDEgMy0zaDI0YTMgMyAwIDAgMSAzIDN2MjZhMyAzIDAgMCAxLTMgMyIgc3R5bGU9ImZpbGw6I2Y5ZTBiZCIvPjxwYXRoIGQ9Ik0yOCAzdjE3YTEgMSAwIDAgMS0xIDFIM2ExIDEgMCAwIDEtMS0xVjNhMSAxIDAgMCAxIDEtMWgyNGExIDEgMCAwIDEgMSAxbS0xIDIxSDE2YTEgMSAwIDAgMCAwIDJoMTFhMSAxIDAgMCAwIDAtMm0tMTQgMGExIDEgMCAxIDAgMCAyIDEgMSAwIDAgMCAwLTIiIHN0eWxlPSJmaWxsOiM2NzYyNWQiLz48cGF0aCBkPSJNOS41IDhoLTNhLjUuNSAwIDAgMSAwLTFoM2EuNS41IDAgMCAxIDAgMW0uNSAxLjVhLjUuNSAwIDAgMC0uNS0uNWgtM2EuNS41IDAgMCAwIDAgMWgzYS41LjUgMCAwIDAgLjUtLjVtMCAyYS41LjUgMCAwIDAtLjUtLjVoLTNhLjUuNSAwIDAgMCAwIDFoM2EuNS41IDAgMCAwIC41LS41bTAgMmEuNS41IDAgMCAwLS41LS41aC0zYS41LjUgMCAwIDAgMCAxaDNhLjUuNSAwIDAgMCAuNS0uNSIgc3R5bGU9ImZpbGw6I2U2OWQ4YSIvPjxwYXRoIGQ9Ik0yNC41IDE2aC0xN2EuNS41IDAgMCAxIDAtMWgxN2EuNS41IDAgMCAxIDAgMU0yMyAxMy41YS41LjUgMCAwIDAtLjUtLjVoLTExYS41LjUgMCAwIDAgMCAxaDExYS41LjUgMCAwIDAgLjUtLjUiIHN0eWxlPSJmaWxsOiM5OGQzYmMiLz48cGF0aCBkPSJNNi41IDZoLTFhLjUuNSAwIDAgMSAwLTFoMWEuNS41IDAgMCAxIDAgMU0yNSA3LjVhLjUuNSAwIDAgMC0uNS0uNWgtMTNhLjUuNSAwIDAgMCAwIDFoMTNhLjUuNSAwIDAgMCAuNS0uNW0tMiAyYS41LjUgMCAwIDAtLjUtLjVoLTExYS41LjUgMCAwIDAgMCAxaDExYS41LjUgMCAwIDAgLjUtLjVtMiAyYS41LjUgMCAwIDAtLjUtLjVoLTEzYS41LjUgMCAwIDAgMCAxaDEzYS41LjUgMCAwIDAgLjUtLjVtLTE4IDZhLjUuNSAwIDAgMC0uNS0uNWgtMWEuNS41IDAgMCAwIDAgMWgxYS41LjUgMCAwIDAgLjUtLjUiIHN0eWxlPSJmaWxsOiNmZmYyZGYiLz48L3N2Zz4=",
            iconWidth = 20,
            iconHeight = 20
          ),
          ports = g6_ports(
            g6_output_port(key = "out", placement = "top")
          )
        )
      ),
      edges = g6_edges(
        # Main flow
        g6_edge(
          source = "webhook",
          target = "ai_agent",
          style = list(
            sourcePort = "out",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect",
            labelText = "GET",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          source = "ai_agent",
          target = "switch",
          style = list(
            sourcePort = "out",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect"
          )
        ),
        # Switch to right-side nodes
        g6_edge(
          type = "cubic",
          source = "switch",
          target = "get_properties",
          style = list(
            sourcePort = "get",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect",
            labelText = "GET",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          source = "get_properties",
          target = "structure_response",
          style = list(
            sourcePort = "out",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect"
          )
        ),
        g6_edge(
          source = "switch",
          target = "post_url",
          style = list(
            sourcePort = "post",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect",
            labelText = "POST",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "switch",
          target = "delete_return",
          style = list(
            sourcePort = "delete",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect",
            labelText = "DELETE",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "delete_return",
          target = "delete_url",
          style = list(
            sourcePort = "out1",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect",
            labelText = "true",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "delete_return",
          target = "return_output",
          style = list(
            sourcePort = "out2",
            targetPort = "in",
            endArrow = TRUE,
            stroke = "#c1c2cb",
            lineWidth = 1,
            endArrowType = "triangleRect",
            labelText = "false",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        # AI Agent downward edges (cubic/curved)
        g6_edge(
          source = "google_gemini",
          target = "ai_agent",
          style = list(
            type = "cubic",
            sourcePort = "out",
            targetPort = "down1",
            stroke = "#c1c2cb",
            endArrow = TRUE,
            lineDash = c(6, 4),
            labelText = "chat model",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "proxmox_api_doc",
          target = "ai_agent",
          style = list(
            sourcePort = "out",
            targetPort = "down3",
            stroke = "#c1c2cb",
            endArrow = TRUE,
            lineDash = c(6, 4),
            labelText = "tool",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "proxmox_api_wiki",
          target = "ai_agent",
          style = list(
            sourcePort = "out",
            targetPort = "down3",
            stroke = "#c1c2cb",
            endArrow = TRUE,
            lineDash = c(6, 4),
            labelText = "tool",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "proxmox",
          target = "ai_agent",
          style = list(
            sourcePort = "out",
            targetPort = "down3",
            stroke = "#c1c2cb",
            endArrow = TRUE,
            lineDash = c(6, 4),
            labelText = "tool",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "auto_fixing_parser",
          target = "ai_agent",
          style = list(
            sourcePort = "out",
            targetPort = "down4",
            stroke = "#c1c2cb",
            endArrow = TRUE,
            lineDash = c(6, 4),
            labelText = "output parser",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        # Output parser edges
        g6_edge(
          type = "cubic",
          source = "groq_chat_model",
          target = "auto_fixing_parser",
          style = list(
            sourcePort = "out",
            targetPort = "in1",
            stroke = "#c1c2cb",
            endArrow = TRUE,
            lineDash = c(6, 4),
            labelText = "model",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        ),
        g6_edge(
          type = "cubic",
          source = "structured_output_parser",
          target = "auto_fixing_parser",
          style = list(
            sourcePort = "out",
            targetPort = "in2",
            stroke = "#c1c2cb",
            endArrow = TRUE,
            lineDash = c(6, 4),
            labelText = "output parser",
            labelPlacement = "start",
            labelBackground = TRUE,
            labelBackgroundFill = "#545353ff"
          )
        )
      )
    ) |>
      #g6_layout(antv_dagre_layout(rankdir = "LR")) |>
      g6_options(
        animation = FALSE,
        background = "#1b1b1b",
        theme = "dark",
        node = list(
          style = list(
            fill = "#282828",
            stroke = "#c1c2cb",
            #labelFill = "#cdcfd3",
            lineWidth = 2
          ),
          state = list(
            selected = list(
              fill = "#282828",
              stroke = "#c1c2cb"
            )
          )
        )
      ) |>
      g6_behaviors(
        click_select(multiple = TRUE),
        drag_element(
          enable = JS(
            "(e) => {
          return !e.shiftKey && !e.altKey;
        }"
          )
        ),
        drag_canvas(
          enable = JS(
            "(e) => {
          return e.targetType === 'canvas' && !e.shiftKey && !e.altKey;
        }"
          )
        ),
        zoom_canvas(),
        create_edge(
          enable = JS(
            "(e) => {
        return e.shiftKey}"
          )
        )
      )
  )
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(g6R)

nodes <- list(
  list(
    id = "node1",
    combo = "combo1",
    style = list(
      halo = FALSE, # Whether to display the node halo
      port = TRUE,
      ports = list(
        list(
          key = "top",
          placement = c(0.5, 1),
          fill = "#7E92B5"
        ),
        list(
          key = "right",
          placement = c(1, 0.5),
          fill = "#F4664A"
        ),
        list(
          key = "bottom",
          placement = c(0.5, 0),
          fill = "#FFBE3A"
        ),
        list(
          key = "left",
          placement = c(0, 0.5),
          fill = "#D580FF"
        )
      ),
      portR = 3,
      portLineWidth = 1,
      portStroke = "#fff"
    )
  ),
  list(
    id = "node2",
    style = list(
      x = 150,
      y = 50,
      badge = TRUE, # Whether to display the badge
      badges = list(
        list(
          text = "A",
          placement = "right-top"
        ),
        list(
          text = "Important",
          placement = "right"
        ),
        list(
          text = "Notice",
          placement = "right-bottom"
        )
      ),
      badgePalette = c("#7E92B5", "#F4664A", "#FFBE3A"), # Badge background color palette
      badgeFontSize = 7 # Badge font size
    ),
    combo = "combo1"
  )
)

edges <- list(
  list(type = "fly-marker-cubic", source = "node1", target = "node2")
)
combos <- list(
  list(
    id = "combo1",
    type = "rect",
    data = list(label = "Combo A")
  )
)

ui <- page_fluid(
  g6Output("graph"),
  verbatimTextOutput("selected_elements"),
  actionButton("set_data", "Set data")
)

server <- function(input, output, session) {
  output$graph <- renderG6({
    g6() |>
      g6_options(
        edge = list(
          type = "fly-marker-cubic",
          style = list(
            endArrow = TRUE
          )
        )
      ) |>
      g6_layout(d3_force_layout()) |>
      g6_behaviors(
        zoom_canvas(),
        drag_element(dropEffect = "link"),
        click_select(
          multiple = TRUE,
          onClick = JS(
            "(e) => {
              console.log(e);  
            }"
          )
        ),
        brush_select(immediately = TRUE),
        collapse_expand(),
        drag_canvas(
          trigger = list(
            up = c("ArrowUp"),
            down = c("ArrowDown"),
            left = c("ArrowLeft"),
            right = c("ArrowRight")
          ),
          animation = list(duration = 100)
        )
      ) |>
      g6_plugins(
        minimap(),
        fullscreen(),
        tooltips(
          title = "Tooltip",
          enable = JS(
            "(e) => { return e.targetType === 'node';}"
          ),
          #getContent = JS(
          #  "(e, items) => {
          #      let result = `<h4>Custom Content</h4>`;
          #      items.forEach((item) => {
          #        console.log(item.id);
          #        result += `<p>Type: ${item.id}</p>`;
          #      });
          #      return result;
          #    }"
          #),
          onOpenChange = JS(
            "( open ) => {   
            console.log(open);
          }"
          ),
          trigger = "click"
        ),
        g6R::toolbar(
          style = list(
            backgroundColor = "#f5f5f5",
            padding = "8px",
            boxShadow = "0 2px 8px rgba(0, 0, 0, 0.15)",
            borderRadius = "8px",
            border = "1px solid #e8e8e8",
            opacity = "0.9",
            marginTop = "12px",
            marginLeft = "12px"
          ),
          getItems = JS(
            "( ) => [   
                { id : 'zoom-in' , value : 'zoom-in' } ,  
                { id : 'zoom-out' , value : 'zoom-out' } ,   
                { id : 'auto-fit' , value : 'auto-fit' } ,
                { id: 'delete', value: 'delete' }, 
                { id: 'request-fullscreen', value: 'request-fullscreen' },
                { id: 'exit-fullscreen', value: 'exit-fullscreen' }
              ]"
          ),
          key = "toolbar",
          className = NULL,
          position = c(
            "top-left",
            "top",
            "top-right",
            "right",
            "bottom-right",
            "bottom",
            "bottom-left",
            "left"
          ),
          onClick = JS(
            "( value, target, current ) => {   
                // Handle button click events
              const graph = HTMLWidgets.find(`#${target.closest('.g6').id}`).getWidget();
              const fullScreen = graph.getPluginInstance('fullscreen');
              const zoomLevel = graph.getZoom();
                if ( value === 'zoom-in' ) {   
                  graph.zoomTo (graph.getZoom() + 0.1);
                } else if ( value === 'zoom-out' ) {     
                  graph.zoomTo (graph.getZoom() - 0.1);
                } else if ( value === 'auto-fit' ) {     
                  graph.fitView ( ) ;
                } else if (value === 'delete') {
                  const selectedNodes = graph.getElementDataByState('node', 'selected').map((node) => {
                    return node.id
                  });
                  graph.removeNodeData(selectedNodes);
                  graph.draw();
                } else if (value === 'request-fullscreen') {
                  if (fullScreen !== undefined) {
                    fullScreen.request();
                  }
                } else if (value === 'exit-fullscreen') {
                  if (fullScreen !== undefined) {
                    fullScreen.exit();
                  }
                }
              }"
          )
        ),
        context_menu(
          enable = JS(
            "(e) => {
                  return e.targetType === 'node'
                }"
          ),
          getItems = JS(
            "() => {
                  return [
                    { name: 'Create edge', value: 'create_edge' },
                    { name: 'Remove node', value: 'remove_node' }
                  ];
                }"
          ),
          onClick = JS(
            "(value, target, current) => {
                  const graph = HTMLWidgets.find(`#${target.closest('.g6').id}`).getWidget();
                  if (value === 'create_edge') {
                    graph.updateBehavior({
                      key: 'create-edge', // Specify the behavior to update
                      enable: true,
                    });
                    // Select node
                    graph.setElementState(current.id, 'selected');
                    // Disable drag node as it is incompatible with edge creation
                    graph.updateBehavior({ key: 'drag-element', enable: false });
                  } else if (value === 'remove_node') {
                    graph.removeNodeData([current.id]);
                    graph.draw();
                  }
                }"
          )
        )
      )
  })

  observeEvent(input$set_data, {
    g6_proxy("graph") |>
      g6_set_data(
        list(
          nodes = nodes,
          edges = edges,
          combos = combos
        )
      )
  })

  output$selected_elements <- renderPrint({
    list(
      node = input[["graph-selected_node"]],
      edge = input[["graph-selected_edge"]],
      combo = input[["graph-selected_combo"]],
      node2_state = input[["graph-node2-state"]]$combo
    )
  })
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(g6R)

options("g6R.mode" = "dev")

nodes <- list(
  list(
    id = "node1",
    type = "custom-circle-node",
    combo = "combo1",
    style = list(
      ports = list(
        list(
          key = "output1",
          type = "output",
          placement = c(1, 0.5)
        )
      )
    )
  ),
  list(
    id = "node2",
    type = "custom-circle-node",
    style = list(
      ports = list(
        list(
          key = "input2",
          type = "input",
          placement = c(0.5, 1)
        )
      ),
      badge = TRUE, # Whether to display the badge
      badges = list(
        list(
          text = "A",
          placement = "right-top"
        ),
        list(
          text = "Important",
          placement = "right"
        ),
        list(
          text = "Notice",
          placement = "right-bottom"
        )
      ),
      badgePalette = c("#7E92B5", "#F4664A", "#FFBE3A"), # Badge background color palette
      badgeFontSize = 7 # Badge font size
    ),
    combo = "combo1"
  )
)

edges <- list(
  list(
    source = "node1",
    target = "node2",
    style = list(sourcePort = "output1", targetPort = "input2")
  )
)
combos <- list(
  list(
    id = "combo1",
    type = "rect",
    data = list(label = "Combo A")
  )
)

ui <- page_fluid(
  g6Output("graph"),
  verbatimTextOutput("selected_elements"),
  div(
    class = "d-flex align-items-center",
    actionButton("remove_node", "Remove node 1"),
    actionButton("remove_edge", "Remove edge"),
    actionButton("remove_combo", "Remove combo 1")
  ),
  div(
    class = "d-flex align-items-center",
    actionButton("add_node", "Add node and connect"),
    actionButton("update_node", "Update node 1 and node 2"),
    actionButton("update_edge", "Update edge 1"),
    actionButton("update_combo", "Update combo 1")
  ),
  div(
    class = "d-flex align-items-center",
    actionButton("canvas_resize", "Resize canvas"),
    actionButton("focus", "Focus node 1"),
    actionButton("show", "Show node 1"),
    actionButton("hide", "Hide node 1")
  ),
  div(
    class = "d-flex align-items-center",
    actionButton("expand", "Expand combo 1"),
    actionButton("collapse", "Collapse combo 1"),
    actionButton("set_options", "Set options")
  ),
  div(
    class = "d-flex align-items-center",
    actionButton("update_plugin", "Update minimap"),
    actionButton("update_behavior", "Disable zoom canvas")
  ),
  div(
    class = "d-flex align-items-center",
    actionButton("set_nodes", "Set nodes state"),
    actionButton("add_data", "Add new data"),
    actionButton("set_data", "Set data (overwrite")
  )
)

server <- function(input, output, session) {
  output$graph <- renderG6({
    g6(
      nodes,
      edges,
      combos
    ) |>
      g6_options(
        edge = edge_options(
          style = list(
            endArrow = TRUE
          )
        )
      ) |>
      g6_layout(d3_force_layout()) |>
      g6_behaviors(
        zoom_canvas(),
        drag_element(dropEffect = "link"),
        click_select(multiple = TRUE),
        brush_select(immediately = TRUE),
        collapse_expand(),
        drag_canvas(
          trigger = list(
            up = c("ArrowUp"),
            down = c("ArrowDown"),
            left = c("ArrowLeft"),
            right = c("ArrowRight")
          ),
          animation = list(duration = 100)
        )
      ) |>
      g6_plugins(
        minimap(),
        fullscreen(),
        tooltips(),
        g6R::toolbar(),
        context_menu()
      )
  })

  observeEvent(input$set_data, {
    g6_proxy("graph") |>
      g6_set_data(
        list(
          nodes = list(
            list(id = "add1")
          )
        )
      )
  })

  observeEvent(input$add_data, {
    g6_proxy("graph") |>
      g6_add_data(
        list(
          combos = list(
            list(id = "addcombo1")
          ),
          edges = list(
            list(source = "add1", target = "add2")
          ),
          nodes = list(
            list(id = "add1", combo = "addcombo1"),
            list(id = "add2")
          )
        )
      )
  })

  observeEvent(input$set_nodes, {
    g6_proxy("graph") |>
      g6_set_nodes(list(node1 = "selected", node2 = "disabled"))
  })

  observeEvent(input$remove_node, {
    g6_proxy("graph") |>
      g6_remove_nodes("node1")
  })

  observeEvent(input$remove_edge, {
    g6_proxy("graph") |>
      g6_remove_edges("node1-node2")
  })

  observeEvent(input$remove_combo, {
    g6_proxy("graph") |>
      g6_remove_combos("combo1")
  })

  observeEvent(input$add_node, {
    g6_proxy("graph") |>
      g6_add_combos(
        data.frame(
          id = "combo2"
        )
      ) |>
      g6_add_nodes(
        data.frame(
          id = c("node3", "node4"),
          combo = c("combo2", "combo2")
        )
      ) |>
      g6_add_edges(
        data.frame(
          source = "node3",
          target = "node4"
        )
      )
  })

  observeEvent(input$update_node, {
    g6_proxy("graph") |>
      g6_update_nodes(
        list(
          list(
            id = "node1",
            type = "star"
          ),
          list(id = "node2", states = list("selected"))
        )
      )
  })

  observeEvent(input$update_edge, {
    g6_proxy("graph") |>
      g6_update_edges(list(list(
        id = "node1-node2",
        style = list(stroke = "#1783F", lineWidth = 2)
      )))
  })

  observeEvent(input$update_combo, {
    g6_proxy("graph") |>
      g6_update_combos(list(list(id = "combo1", type = "circle")))
  })

  observeEvent(input$canvas_resize, {
    g6_proxy("graph") |>
      g6_canvas_resize(1000, 1000)
  })

  observeEvent(input$focus, {
    g6_proxy("graph") |>
      g6_focus_nodes("node1", animation = list(duration = 2000))
  })

  observeEvent(input$show, {
    g6_proxy("graph") |>
      g6_show_nodes("node1")
  })

  observeEvent(input$hide, {
    g6_proxy("graph") |>
      g6_hide_nodes("node1")
  })

  observeEvent(input$expand, {
    g6_proxy("graph") |>
      g6_expand_combo("combo1")
  })

  observeEvent(input$collapse, {
    g6_proxy("graph") |>
      g6_collapse_combo("combo1")
  })

  observeEvent(input$set_options, {
    g6_proxy("graph") |>
      g6_set_options(
        list(
          node = list(
            style = list(
              fill = '#91d5ff',
              stroke = '#40a9ff',
              lineWidth = 1,
              radius = 10
            )
          ),
          edge = list(
            style = list(
              stroke = '#91d5ff',
              lineWidth = 2,
              endArrow = TRUE
            )
          )
        )
      )
  })

  observeEvent(input$update_plugin, {
    g6_proxy("graph") |>
      g6_update_plugin(
        key = "minimap",
        position = "right-top"
      )
  })

  observeEvent(input$update_behavior, {
    g6_proxy("graph") |>
      g6_update_behavior(
        key = "zoom-canvas",
        enable = FALSE
      )
  })

  observeEvent(
    {
      req(input[["graph-initialized"]])
      input[["graph-state"]]
    },
    {
      g6_proxy("graph") |> g6_get_nodes(c("node1", "node2"))
    }
  )

  output$selected_elements <- renderPrint({
    list(
      node = input[["graph-selected_node"]],
      edge = input[["graph-selected_edge"]],
      combo = input[["graph-selected_combo"]],
      node2_combo = input[["graph-node2-state"]]$combo
    )
  })
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(g6R)

nodes <- data.frame(id = 1:100)

# Set a seed for reproducibility
set.seed(123)

# Define the number of edges to create (e.g., 200 random connections)
num_edges <- 50

# Generate random edges
edges <- data.frame(
  source = sample(nodes$id, num_edges, replace = TRUE),
  target = sample(nodes$id, num_edges, replace = TRUE)
)

edges$id <- paste0(edges$source, edges$target)
duplicated_id <- which(duplicated(edges$id) == TRUE)
if (length(duplicated_id)) {
  edges <- edges[-duplicated_id, ]
}

ui <- page_fluid(
  actionButton("add_hull", "Add hull"),
  verbatimTextOutput("state"),
  g6Output("graph", height = "100vh"),
)

server <- function(input, output, session) {
  output$graph <- renderG6({
    g6(nodes, edges) |>
      g6_options(
        autoFit = "view",
        node = list(
          style = list(
            labelBackground = TRUE,
            labelBackgroundFill = '#FFB6C1',
            labelBackgroundRadius = 4,
            labelFontFamily = 'Arial',
            labelPadding = c(0, 4),
            labelText = JS(
              "(d) => {
                return d.id
              }"
            )
          )
        )
      ) |>
      g6_layout() |>
      g6_behaviors(
        "zoom-canvas",
        drag_element_force(),
        click_select(multiple = TRUE),
        brush_select(),
        create_edge()
      ) |>
      g6_plugins(
        "minimap",
        context_menu()
      )
  })

  observeEvent(input$add_hull, {
    g6_proxy("graph") |>
      g6_add_plugin(
        hull(
          members = sample(nodes$id, 10),
          fill = "#F08F56",
          stroke = "#F08F56",
          labelText = "hull-a",
          labelPlacement = "top",
          labelBackground = TRUE,
          labelPadding = 5
        )
      )
  })

  observeEvent(req(input[["graph-initialized"]]), {
    print("Graph initialized")
  })

  observeEvent(req(input[["graph-state"]]), {
    print("Graph changed")
  })

  output$state <- renderPrint({
    list(
      node = input[["graph-selected_node"]],
      edge = input[["graph-selected_edge"]],
      combo = input[["graph-selected_combo"]]
    )
  })
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(g6R)

nodes <- data.frame(
  id = 1:5,
  combo = c("stack2", "stack2", "stack2", "stack1", "stack1")
)

# TBD add ports

# Set a seed for reproducibility
set.seed(123)

# Define the number of edges to create (e.g., 200 random connections)
num_edges <- 2

# Generate random edges
edges <- data.frame(
  source = sample(nodes$id, num_edges, replace = TRUE),
  target = sample(nodes$id, num_edges, replace = TRUE)
)

edges$id <- paste0(edges$source, edges$target)
duplicated_id <- which(duplicated(edges$id) == TRUE)
if (length(duplicated_id)) {
  edges <- edges[-duplicated_id, ]
}

combos <- data.frame(
  id = paste0("stack", c(1, 2))
)

ui <- page_fluid(
  input_dark_mode(),
  actionButton("add_hull", "Add hull"),
  g6Output("graph", height = "100vh")
)

server <- function(input, output, session) {
  output$graph <- renderG6({
    g6(nodes, edges, combos) |>
      g6_options(
        combo = list(
          animation = FALSE,
          type = "circle-combo-with-extra-button",
          style = list(
            labelText = JS(
              "(d) => {
              return d.id
            }"
            )
          )
        ),
        edge = list(
          type = "fly-marker-cubic",
          zIndex = 100,
          style = list(
            endArrow = TRUE,
            targetPort = "port-1",
            lineDash = c(5, 5)
          )
        ),
        node = list(
          type = "html",
          style = list(
            port = TRUE,
            ports = list(
              list(key = "port-1", placement = c(0, 0.5))
            ),
            labelBackground = TRUE,
            labelBackgroundFill = '#FFB6C1',
            labelBackgroundRadius = 4,
            labelFontFamily = 'Arial',
            labelPadding = c(0, 4),
            labelText = JS(
              "(d) => {
              return d.id
            }"
            ),
            innerHTML = JS(
              '(d) => {
          return `
                <div style="width: 13rem; user-select: none;" class="card bslib-card bslib-mb-spacing html-fill-item html-fill-container" data-bslib-card-init data-require-bs-caller="card()" data-require-bs-version="5">
  <div class="card-body bslib-gap-spacing html-fill-item html-fill-container" style="margin-top:auto;margin-bottom:auto;flex:1 1 auto; user-select: none; width = 13rem">
    <div class="form-group shiny-input-container" data-require-bs-version="5" data-require-bs-caller="input_switch()">
      <div class="bslib-input-switch form-switch form-check">
        <input id="${d.id}" class="form-check-input" type="checkbox" role="switch"/>
        <label class="form-check-label" for="${d.id}">
          <span>Add to dashboard</span>
        </label>
      </div>
    </div>
  </div>
  <script data-bslib-card-init>bslib.Card.initializeAllCards();</script>
</div>
          `
            }'
            )
          )
        )
      ) |>
      g6_layout(
        antv_dagre_layout(
          ranksep = 500,
          nodesep = 100,
          sortByCombo = TRUE,
          controlPoints = TRUE
        )
      ) |>
      g6_behaviors(
        "zoom-canvas",
        drag_element(),
        click_select(multiple = TRUE),
        brush_select(),
        "collapse-expand"
      ) |>
      g6_plugins(
        "minimap",
        "tooltip",
        "fullscreen"
      )
  })

  observeEvent(input$add_hull, {
    g6_proxy("graph") |>
      g6_add_plugin(
        hull(
          members = sample(nodes$id, 5),
          labelText = "hull-a",
          labelPlacement = "top",
          labelBackground = TRUE,
          labelPadding = 5
        )
      )
  })

  observe({
    print(input[["graph-selected_node"]])
    print(input[["graph-selected_edge"]])
  })
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(g6R)

nodes <- data.frame(id = as.character(1:10))

# Generate random edges
edges <- data.frame(
  source = c("2", "6", "7"),
  target = c("1", "3", "9")
)

ui <- page_fluid(
  actionButton("update", "Update plugin"),
  h6(
    class = "my-2",
    "Right click on a node (the default context_menu target). Then, click on
    update and right click on a node again."
  ),
  g6_output("graph")
)

server <- function(input, output, session) {
  output$graph <- render_g6({
    g6(nodes, edges) |>
      g6_options(animation = FALSE) |>
      g6_layout(d3_force_layout()) |>
      g6_plugins(
        context_menu()
      )
  })

  observeEvent(input$update, {
    g6_proxy("graph") |>
      g6_update_plugin(
        key = "contextmenu",
        getItems = JS(
          "() => {
        return [
          { name: 'plop', value: 'plop' }
        ];
      }"
        )
      )
  })
}

shinyApp(ui, server)

#' Create a proxy object to modify an existing g6 graph instance
#'
#' This function creates a proxy object that can be used to update an existing g6 graph
#' instance after it has been rendered in the UI. The proxy allows for server-side
#' modifications of the graph without completely re-rendering it.
#'
#' @param id Character string matching the ID of the g6 graph instance to be modified.
#' @param session The Shiny session object within which the graph exists.
#'   By default, this uses the current reactive domain.
#'
#' @return A proxy object of class "g6_proxy" that can be used with g6 proxy methods
#'   such as `g6_add_nodes()`, `g6_remove_nodes()`, etc.
#'
#' @export
g6_proxy <- function(id, session = shiny::getDefaultReactiveDomain()) {
  if (is.null(session)) {
    stop(
      "g6_proxy must be called from the server function of a Shiny app."
    )
  }

  structure(
    list(id = id, session = session),
    class = "g6_proxy"
  )
}

#' @keywords internal
g6_data_proxy <- function(graph, el, action, type) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_add_* with g6 object. Only within shiny and using g6_proxy."
    )
  }
  # No elements to process
  if (length(el) == 0) {
    return(graph)
  }

  if (action != "remove") {
    if (inherits(el, "data.frame")) {
      el <- lapply(seq_len(nrow(el)), \(i) {
        setNames(as.list(el[i, ]), colnames(el))
      })
    }
  } else {
    if (!is.null(el)) {
      if (length(el) == 1) {
        el <- list(el)
      }
    }
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-data", graph$id),
    list(
      el = el,
      action = action,
      type = type,
      layout = get_g6_layout_on_data_change()
    )
  )
  graph
}

#' @keywords internal
validate_g6_layout_on_data_change <- function(val) {
  if (!is.logical(val) || length(val) != 1) {
    stop(
      "`g6R.layout_on_data_change` option must be a single logical value (TRUE or FALSE)."
    )
  }
  invisible(val)
}

#' @keywords internal
get_g6_layout_on_data_change <- function() {
  val <- getOption("g6R.layout_on_data_change", FALSE)
  validate_g6_layout_on_data_change(val)
}

#' @keywords internal
g6_add <- function(graph, el, type) {
  g6_data_proxy(graph, el, action = "add", type = type)
}

#' @keywords internal
g6_remove <- function(graph, el, type) {
  g6_data_proxy(graph, el, action = "remove", type = type)
}

#' @keywords internal
g6_update <- function(graph, el, type) {
  g6_data_proxy(graph, el, action = "update", type = type)
}

#' @keywords internal
g6_set <- function(graph, el, type) {
  g6_data_proxy(graph, el, action = "set", type = type)
}

#' @keywords internal
g6_get <- function(graph, el, type) {
  g6_data_proxy(graph, el, action = "get", type = type)
}

#' Get the state of nodes/edges/combos in a g6 graph via proxy
#'
#' This function gets the state of one or more nodes/edges/combos to an existing g6 graph instance
#' using a proxy object.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param nodes,edges,combos A string or character vector with the IDs of the nodes/edges/combos.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' If a node with the same ID already exists, it will not be added again.
#' See \url{https://g6.antv.antgroup.com/en/api/data#graphgetnodedata} for more details.
#'
#' @seealso \code{\link{g6_proxy}}
#' @rdname g6-get
#' @export
#' @examples
#'  if (interactive()) {
#'    library(shiny)
#'    library(bslib)
#'
#'     ui <- page_fluid(
#'       verbatimTextOutput("res"),
#'       g6Output("graph")
#'     )
#'
#'     server <- function(input, output, session) {
#'       output$graph <- renderG6({
#'         g6(
#'           nodes = data.frame(id = c("node1", "node2"))
#'         ) |>
#'           g6_options(animation = FALSE) |>
#'           g6_layout() |>
#'           g6_behaviors(click_select())
#'       })
#'
#'       # Send query to JS
#'       observeEvent(req(input[["graph-initialized"]]), {
#'         g6_proxy("graph") |> g6_get_nodes(c("node1", "node2"))
#'       })
#'
#'       # Recover query result inside input[["<GRAPH_ID>-<ELEMENT_ID>-state"]]
#'       output$res <- renderPrint({
#'         list(
#'           node1_state = input[["graph-node1-state"]],
#'           node2_state = input[["graph-node2-state"]]
#'         )
#'       })
#'     }
#'     shinyApp(ui, server)
#'  }
g6_get_nodes <- function(graph, nodes) {
  g6_get(graph, nodes, type = "Node")
}

#' @rdname g6-get
#' @export
g6_get_edges <- function(graph, edges) {
  g6_get(graph, edges, type = "Edge")
}

#' @rdname g6-get
#' @export
g6_get_combos <- function(graph, combos) {
  g6_get(graph, combos, type = "Combo")
}

#' Set the state of nodes/edges/combos in a g6 graph via proxy
#'
#' This function sets the state of one or more nodes/edges/combos to an existing g6 graph instance
#' using a proxy object. This allows updating the graph without completely
#' re-rendering it. Valid states are "selected", "active", "inactive", "disabled", or "highlight".
#'
#' \link{g6_set_data} allows to set all graph data at once (nodes, edges and combos).
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param nodes A key value pair list with the node id and its state.
#' @param edges A key value pair list with the edge id and its state.
#' @param combos A key value pair list with the combo id and its state.
#' @param data A nested list containing all nodes, edges and combo data.
#' Alternatively you can use \code{\link{g6_data}} to generate compatible data.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' If a node with the same ID already exists, it will not be added again.
#' See \url{https://g6.antv.antgroup.com/en/api/element#graphsetelementstateid-state-options} for more details.
#'
#' @seealso \code{\link{g6_proxy}}
#' @rdname g6-set
#' @export
#' @examples
#' if (interactive()) {
#'   # Example setting node states
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   nodes <- data.frame(id = 1:2)
#'   edges <- data.frame(source = 1, target = 2)
#'
#'   ui <- page_fluid(
#'     g6_output("graph"),
#'     actionButton("set_state", "Set Node States")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(nodes = nodes, edges = edges) |> g6_layout()
#'     })
#'
#'     observeEvent(input$set_state, {
#'       g6_set_nodes(
#'         g6_proxy("graph"),
#'         list(`1` = "selected", `2` = "disabled")
#'       )
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#'
#'   # Replace data dynamically
#'   ui <- page_fluid(
#'     g6_output("graph"),
#'     actionButton("remove", "Remove All"),
#'     actionButton("reset", "Reset Graph"),
#'     verbatimTextOutput("state")
#'   )
#'
#'   server <- function(input, output, session) {
#'     # Store initial state after first render
#'     initial_state <- reactiveVal(NULL)
#'
#'     output$graph <- render_g6({
#'       g6(nodes = nodes, edges = edges) |>
#'         g6_layout() |>
#'         g6_options(animation = FALSE)
#'     })
#'
#'     # Save initial state once available
#'     observe({
#'       state <- input[["graph-state"]]
#'       if (!is.null(state) && is.null(initial_state())) {
#'         initial_state(state)
#'       }
#'     })
#'
#'     # Remove all nodes and edges
#'     observeEvent(input$remove, {
#'       g6_set_data(g6_proxy("graph"), list(nodes = list(), edges = list()))
#'     })
#'
#'     # Reset graph to initial state
#'     observeEvent(input$reset, {
#'       state <- initial_state()
#'       if (!is.null(state)) {
#'         g6_set_data(g6_proxy("graph"), state)
#'       }
#'     })
#'
#'     output$state <- renderPrint({
#'       input[["graph-state"]]
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
g6_set_nodes <- function(graph, nodes) {
  g6_set(graph, nodes, type = "Node")
}

#' @rdname g6-set
#' @export
g6_set_edges <- function(graph, edges) {
  g6_set(graph, edges, type = "Edge")
}

#' @rdname g6-set
#' @export
g6_set_combos <- function(graph, combos) {
  g6_set(graph, combos, type = "Combo")
}

#' @rdname g6-set
#' @export
g6_set_data <- function(graph, data) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_add_* with g6 object. Only within shiny and using g6_proxy."
    )
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-data", graph$id),
    list(data = as_g6_data(data), action = "set", type = "Data")
  )
  graph
}

#' Add nodes/edges/combos to a g6 graph via proxy
#'
#' This function adds one or more nodes/edges/combos to an existing g6 graph instance
#' using a proxy object. This allows updating the graph without completely
#' re-rendering it.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param ... Nodes or edges or combos. You can pass a list of nodes/edges/combos, a dataframe
#' or leverage the g6_nodes(), g6_edges() or g6_combos() helpers or pass individual elements like
#' g6_node(), g6_edge() or g6_combo(). Elements structure must be compliant with specifications listed at
#' \url{https://g6.antv.antgroup.com/manual/element/overview}.
#' @param data A nested list possibly containing nodes, edges and combo data. Can also be
#' created with the \code{\link{g6_data}} helper.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' If a node with the same ID already exists, it will not be added again.
#' See \url{https://g6.antv.antgroup.com/en/api/data#graphaddnodedata},
#' \url{https://g6.antv.antgroup.com/en/api/data#graphaddedgedata} and
#' \url{https://g6.antv.antgroup.com/en/api/data#graphaddcombodata} for more details.
#'
#' @seealso \code{\link{g6_proxy}}, \code{\link{g6_remove_nodes}}
#' @rdname g6-add
#' @export
#' @examples
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   # Static data defined globally
#'   nodes <- data.frame(id = 1:3)
#'   edges <- data.frame(source = c(1, 2), target = c(2, 3))
#'
#'   ui <- page_fluid(
#'     title = "Add Nodes Dynamically",
#'     g6_output("graph"),
#'     actionButton("add_node", "Add Node")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(nodes = nodes, edges = edges) |> g6_layout()
#'     })
#'
#'     next_id <- reactiveVal(max(nodes$id) + 1)
#'
#'     observeEvent(input$add_node, {
#'       g6_add_nodes(g6_proxy("graph"), data.frame(id = next_id()))
#'       next_id(next_id() + 1)
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
g6_add_nodes <- function(graph, ...) {
  nodes <- as_g6_elements(..., coerc_func = as_g6_nodes)
  g6_add(graph, nodes, type = "Node")
}

#' @rdname g6-add
#' @export
g6_add_edges <- function(graph, ...) {
  edges <- as_g6_elements(..., coerc_func = as_g6_edges)
  g6_add(graph, edges, type = "Edge")
}

#' @rdname g6-add
#' @export
g6_add_combos <- function(graph, ...) {
  combos <- as_g6_elements(..., coerc_func = as_g6_combos)
  g6_add(graph, combos, type = "Combo")
}

#' @rdname g6-add
#' @export
g6_add_data <- function(graph, data) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_add_data with g6 object. Only within shiny and using g6_proxy."
    )
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-data", graph$id),
    list(data = as_g6_data(data), action = "add", type = "Data")
  )
  graph
}

#' Remove nodes/edge/combos from a g6 graph via proxy
#'
#' This function removes one or more nodes/edges/combos from an existing g6 graph instance
#' using a proxy object. This allows updating the graph without completely
#' re-rendering it.
#'
#' @note When a node is removed, its connected edges are also removed.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param ids Character vector or list containing the IDs of the nodes/edges/combos to be removed.
#' If a single ID is provided, it will be converted to a list internally. You can't mix
#' nodes, edges and combos ids, elements have to be of the same type.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' See \url{https://g6.antv.antgroup.com/en/api/data#graphremovenodedata},
#' \url{https://g6.antv.antgroup.com/en/api/data#graphremoveedgedata} and
#' \url{https://g6.antv.antgroup.com/en/api/data#graphremovecombodata} for more details.
#'
#' @seealso \code{\link{g6_proxy}}
#' @rdname g6-remove
#' @export
#' @examples
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   # Static data defined globally
#'   nodes <- data.frame(id = 1:3)
#'   edges <- data.frame(source = c(1, 2), target = c(2, 3))
#'
#'   ui <- page_fluid(
#'     title = "Remove Nodes Dynamically",
#'     g6_output("graph"),
#'     actionButton("remove_node", "Remove Last Node")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(nodes = nodes, edges = edges) |> g6_layout()
#'     })
#'
#'     # Track the next node id and current node ids
#'     current_ids <- reactiveVal(nodes$id)
#'
#'     observeEvent(input$remove_node, {
#'       ids <- current_ids()
#'       if (length(ids) > 0) {
#'         remove_id <- tail(ids, 1)
#'         g6_remove_nodes(g6_proxy("graph"), remove_id)
#'         current_ids(ids[-length(ids)])
#'       }
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#'
#' }
g6_remove_nodes <- function(graph, ids) {
  g6_remove(graph, ids, type = "Node")
}

#' @rdname g6-remove
#' @export
g6_remove_edges <- function(graph, ids) {
  g6_remove(graph, ids, type = "Edge")
}

#' @rdname g6-remove
#' @export
g6_remove_combos <- function(graph, ids) {
  g6_remove(graph, ids, type = "Combo")
}

#' Update nodes/edges/combos to a g6 graph via proxy
#'
#' This function updates one or more nodes/edges/combos to an existing g6 graph instance
#' using a proxy object. This allows updating the graph without completely
#' re-rendering it.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param ... Nodes or edges or combos. You can pass a list of nodes/edges/combos, a dataframe
#' or leverage the g6_nodes(), g6_edges() or g6_combos() helpers or pass individual elements like
#' g6_node(), g6_edge() or g6_combo(). Elements structure must be compliant with specifications listed at
#' \url{https://g6.antv.antgroup.com/manual/element/overview}.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' See \url{https://g6.antv.antgroup.com/en/api/data#graphupdatenodedata},
#' \url{https://g6.antv.antgroup.com/en/api/data#graphupdateedgedata} and
#' \url{https://g6.antv.antgroup.com/en/api/data#graphupdatecombodata} for more details.
#'
#' @seealso \code{\link{g6_proxy}}, \code{\link{g6_remove_nodes}}
#' @rdname g6-update
#' @export
#' @examples
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   # Static data defined globally
#'   nodes <- data.frame(id = 1:3)
#'   edges <- data.frame(source = c(1, 2), target = c(2, 3))
#'
#'   ui <- page_fluid(
#'     title = "Update Nodes Dynamically",
#'     g6_output("graph"),
#'     actionButton("update_node", "Update Node 1 Label")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(nodes = nodes, edges = edges) |> g6_layout()
#'     })
#'
#'     observeEvent(input$update_node, {
#'       # Update label for node 1
#'       g6_update_nodes(
#'         g6_proxy("graph"),
#'         g6_node(id = 1, style = list(labelText = "Node label updated"))
#'       )
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
#'
g6_update_nodes <- function(graph, ...) {
  nodes <- as_g6_elements(..., coerc_func = as_g6_nodes)
  g6_update(graph, nodes, type = "Node")
}

#' @rdname g6-update
#' @export
g6_update_edges <- function(graph, ...) {
  edges <- as_g6_elements(..., coerc_func = as_g6_edges)
  g6_update(graph, edges, type = "Edge")
}

#' @rdname g6-update
#' @export
g6_update_combos <- function(graph, ...) {
  combos <- as_g6_elements(..., coerc_func = as_g6_combos)
  g6_update(graph, combos, type = "Combo")
}

#' Resize the canvas of a g6 graph via proxy
#'
#' This function changes the size of the canvas of an existing g6 graph instance
#' using a proxy object. This allows updating the graph dimensions without completely
#' re-rendering it.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param width Numeric value specifying the new width of the canvas in pixels.
#' @param height Numeric value specifying the new height of the canvas in pixels.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' See \url{https://g6.antv.antgroup.com/en/api/canvas#graphsetsizewidth-height} for more details.
#'
#' @seealso \code{\link{g6_proxy}}
#' @export
g6_canvas_resize <- function(graph, width, height) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_canvas_resize with g6 object. Only within shiny and using g6_proxy."
    )
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-canvas-resize", graph$id),
    list(width = width, height = height)
  )
  graph
}

#' Center graph
#'
#' This function pans the graph to the center of the viewport
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param animation Optional list containing animation configuration parameters for the focus action.
#'   Common parameters include:
#'   \itemize{
#'     \item \code{duration}: Duration of the animation in milliseconds.
#'     \item \code{easing}: Animation easing function name (e.g., "ease-in", "ease-out").
#'   }
#'   If NULL, no animation will be applied.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' See \url{https://g6.antv.antgroup.com/en/api/viewport#graphfitcenteranimation} for more details.
#'
#' @seealso \code{\link{g6_proxy}}
#' @export
g6_fit_center <- function(graph, animation = NULL) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_fit_center with g6 object. Only within shiny and using g6_proxy."
    )
  }

  if (!is.null(animation)) {
    stopifnot(is.list(animation))
  }

  if (is.null(animation)) {
    animation <- list()
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-fit-center", graph$id),
    animation
  )
  graph
}

#' @keywords internal
g6_element_action <- function(graph, ids, animation = NULL, action) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_*_element with g6 object. Only within shiny and using g6_proxy."
    )
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-element-action", graph$id),
    list(ids = ids, animation = animation, action = action)
  )
  graph
}

#' Focus on specific elements in a g6 graph via proxy
#'
#' This function focuses on one or more elements (nodes/edges) in an existing g6 graph instance
#' using a proxy object. It highlights the specified elements and can optionally
#' animate the view to focus on them.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param ids Character vector containing the IDs of the elements (nodes/edges/combos) to focus on.
#' @param animation Optional list containing animation configuration parameters for the focus action.
#'   Common parameters include:
#'   \itemize{
#'     \item \code{duration}: Duration of the animation in milliseconds.
#'     \item \code{easing}: Animation easing function name (e.g., "ease-in", "ease-out").
#'   }
#'   If NULL, no animation will be applied.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' See \url{https://g6.antv.antgroup.com/en/api/element#graphfocuselementid-animation} for more details.
#'
#' @seealso \code{\link{g6_proxy}}
#' @rdname g6-focus
#' @keywords internal
g6_focus_elements <- function(graph, ids, animation = NULL) {
  g6_element_action(graph, ids, animation, action = "focus")
}

#' Focus nodes in a g6 graph via proxy
#' @export
#' @rdname g6-focus
g6_focus_nodes <- function(graph, ids, animation = NULL) {
  g6_focus_elements(graph, ids, animation)
}

#' Focus edges in a g6 graph via proxy
#' @export
#' @rdname g6-focus
g6_focus_edges <- function(graph, ids, animation = NULL) {
  g6_focus_elements(graph, ids, animation)
}

#' Focus combos in a g6 graph via proxy
#' @export
#' @rdname g6-focus
g6_focus_combos <- function(graph, ids, animation = NULL) {
  g6_focus_elements(graph, ids, animation)
}

#' Hide/show elements in a g6 graph
#'
#' This function hides/shows specified elements (nodes, edges, or combos) in a g6 graph.
#' Hidden elements are removed from view but remain in the graph data structure.
#'
#' @param graph A g6 graph object or a g6_proxy object for Shiny applications.
#' @param ids Character vector specifying the IDs of elements to hide/show.
#'   Can include node IDs, edge IDs, or combo IDs.
#' @param animation Boolean to toggle animation.
#'
#' @return The modified g6 graph or g6_proxy object, allowing for method chaining.
#'
#' @details
#' When elements are hidden, they are removed from the visual display but still exist
#' in the underlying data structure. This means they can be shown again later using
#' \code{\link{g6_show_elements}} without having to recreate them.
#'
#' Hidden elements will not participate in layout calculations, which may cause other
#' elements to reposition. When elements are shown again, the graph
#' may recalculate layout positions, which can cause other elements to reposition.
#'
#' @seealso
#' \code{\link{g6_show_elements}}
#' @rdname g6-element-toggle
#' @export
g6_hide_elements <- function(graph, ids, animation = NULL) {
  g6_element_action(graph, ids, animation, action = "hide")
}

#' Hide nodes in a g6 graph via proxy
#' @export
#' @rdname g6-element-toggle
g6_hide_nodes <- function(graph, ids, animation = NULL) {
  g6_hide_elements(graph, ids, animation)
}

#' Hide edges in a g6 graph via proxy
#' @export
#' @rdname g6-element-toggle
g6_hide_edges <- function(graph, ids, animation = NULL) {
  g6_hide_elements(graph, ids, animation)
}

#' Hide combos in a g6 graph via proxy
#' @export
#' @rdname g6-element-toggle
g6_hide_combos <- function(graph, ids, animation = NULL) {
  g6_hide_elements(graph, ids, animation)
}

#' @rdname g6-element-toggle
#' @export
g6_show_elements <- function(graph, ids, animation = NULL) {
  g6_element_action(graph, ids, animation, action = "show")
}

#' Show nodes in a g6 graph via proxy
#' @export
#' @rdname g6-element-toggle
g6_show_nodes <- function(graph, ids, animation = NULL) {
  g6_show_elements(graph, ids, animation)
}

#' Show edges in a g6 graph via proxy
#' @export
#' @rdname g6-element-toggle
g6_show_edges <- function(graph, ids, animation = NULL) {
  g6_show_elements(graph, ids, animation)
}

#' Show combos in a g6 graph via proxy
#' @export
#' @rdname g6-element-toggle
g6_show_combos <- function(graph, ids, animation = NULL) {
  g6_show_elements(graph, ids, animation)
}

#' @keywords internal
g6_combo_action <- function(graph, id, options = NULL, action) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_*_combo with g6 object. Use only within shiny and using g6_proxy."
    )
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-combo-action", graph$id),
    list(id = id, options = options, action = action)
  )
  graph
}

#' Collapse or expand a combo element in a g6 graph
#'
#' This function collapses/expands a specified combo (a group of nodes) in a g6 graph,
#' hiding its member nodes and edges while maintaining the combo itself visible.
#' This is useful for simplifying complex graphs with multiple hierarchical groups.
#'
#' @param graph A g6 graph object or a g6_proxy object for Shiny applications.
#' @param id Character string specifying the ID of the combo to collapse/expand.
#' @param options List containing optional configuration parameters for the collapse/expand action:
#'   \itemize{
#'     \item \code{animate}: Logical value indicating whether to animate the collapsing process. Default is \code{TRUE}.
#'     \item \code{align}: Logical value to ensure the position of expanded/collapsed nodes remains unchanged.
#'   }
#'
#' @return The modified g6 graph or g6_proxy object, allowing for method chaining.
#'
#' @details
#' When a combo is collapsed, its member nodes and edges are hidden from view
#' while the combo itself remains visible, typically shown as a single node.
#' This helps to reduce visual complexity in large graphs with hierarchical groupings.
#'
#' @rdname g6-combo-action
#' @references \url{https://g6.antv.antgroup.com/en/api/element#graphcollapseelementid-options},
#' \url{https://g6.antv.antgroup.com/en/api/element#graphexpandelementid-options}
#' @export
g6_collapse_combo <- function(graph, id, options = NULL) {
  g6_combo_action(graph, id, options, action = "collapse")
}

#' @rdname g6-combo-action
#' @export
g6_expand_combo <- function(graph, id, options = NULL) {
  g6_combo_action(graph, id, options, action = "expand")
}

#' Set options for a g6 graph via proxy
#'
#' This function allows updating various configuration options of an existing g6 graph
#' instance using a proxy object within a Shiny application.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param ... Named arguments representing the options to update and their new values.
#'   These can include any valid g6 graph options such as fitView, animate, modes, etc.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' The function allows updating various graph options dynamically without having to
#' re-render the entire graph. This is useful for changing behavior, appearance,
#' or interaction modes in response to user input.
#'
#' @seealso \code{\link{g6_proxy}}
#' @export
g6_set_options <- function(graph, ...) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_*_combo with g6 object. Use only within shiny and using g6_proxy."
    )
  }
  # TBD: support JS wrapped options
  graph$session$sendCustomMessage(
    sprintf("%s_g6-set-options", graph$id),
    list(...)
  )
  graph
}

#' Execute layout for a g6 graph via proxy
#'
#' This function order the execution of the layout of the current
#' graph. It can also update layout options before running it.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param ... Any option to pass to the layout. If so, the layout is updated
#' before running it.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' @seealso \code{\link{g6_proxy}}
#' @export
g6_update_layout <- function(graph, ...) {
  graph$session$sendCustomMessage(
    sprintf("%s_g6-update-layout", graph$id),
    list(...)
  )
  graph
}

#' Update a plugin in a g6 graph via proxy
#'
#' This function allows updating the configuration of an existing plugin in a g6 graph
#' instance using a proxy object within a Shiny application.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param key Character string identifying the plugin to update.
#' @param ... Named arguments representing the plugin configuration options to update
#'   and their new values.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' The function allows dynamically updating the configuration of an existing plugin
#' without having to reinitialize it. This is useful for changing plugin behavior
#' or appearance in response to user interactions.
#'
#' @seealso \code{\link{g6_proxy}}, \code{\link{g6_add_plugin}}
#' @export
#' @examples
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   color_to_hex <- function(col) {
#'     rgb <- col2rgb(col)
#'     sprintf("#%02X%02X%02X", rgb[1], rgb[2], rgb[3])
#'   }
#'
#'   nodes <- data.frame(id = c("node1", "node2", "node3"))
#'   edges <- data.frame(source = "node1", target = "node2")
#'
#'   ui <- page_fluid(
#'     g6_output("graph"),
#'     actionButton("add_hull", "Add Hull Plugin"),
#'     selectInput(
#'       "hull_color",
#'       "Hull Color",
#'       choices = c("red", "blue", "green", "orange", "purple"),
#'       selected = "red"
#'     )
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(nodes = nodes, edges = edges) |>
#'         g6_layout() |>
#'         g6_options(animation = FALSE)
#'     })
#'
#'     observeEvent(input$add_hull, {
#'       g6_add_plugin(
#'         g6_proxy("graph"),
#'         hull(
#'           members = c("node1", "node2", "node3"),
#'           fill = color_to_hex(input$hull_color),
#'           fillOpacity = 0.2
#'         )
#'       )
#'     })
#'
#'     observeEvent(input$hull_color, {
#'       # Only update if hull plugin exists
#'       g6_update_plugin(
#'         g6_proxy("graph"),
#'         key = "hull",
#'         fill = color_to_hex(input$hull_color),
#'         stroke = color_to_hex(input$hull_color)
#'       )
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
g6_update_plugin <- function(graph, key, ...) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_update_plugin with g6 object. Use only within shiny and using g6_proxy."
    )
  }
  opts <- list(key = key, ...)
  graph$session$sendCustomMessage(
    sprintf("%s_g6-update-plugin", graph$id),
    # Mark any element of options wrapped by htmlwidgets::JS
    # so it can be evaluated on the JS side.
    list(opts = opts, evals = JSEvals(opts))
  )
  graph
}

#' Add a plugin to a g6 graph via proxy
#'
#' This function adds one or more plugins to an existing g6 graph instance
#' using a proxy object within a Shiny application.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param ... Named arguments where each name is a plugin type and each value is a
#'   list of configuration options for that plugin.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' G6 plugins extend the functionality of the graph visualization with features like
#' minimaps, toolbar controls, contextual menus, and more. This function allows adding
#' these plugins dynamically after the graph has been initialized.
#'
#' @seealso \code{\link{g6_proxy}}, \code{\link{g6_update_plugin}}
#' @export
g6_add_plugin <- function(graph, ...) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_add_plugin with g6 object. Use only within shiny and using g6_proxy."
    )
  }
  # TBD: support JS wrapped options
  graph$session$sendCustomMessage(
    sprintf("%s_g6-add-plugin", graph$id),
    list(...)
  )
  graph
}

#' Update a behavior in a g6 graph via proxy
#'
#' This function allows updating the configuration of an existing behavior in a g6 graph
#' instance using a proxy object within a Shiny application.
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param key Character string identifying the behavior to update.
#' @param ... Named arguments representing the behavior configuration options to update
#'   and their new values.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#'
#' @details
#' This function can only be used with a g6_proxy object within a Shiny application.
#' It will not work with regular g6 objects outside of Shiny.
#'
#' Behaviors in G6 define how the graph responds to user interactions like dragging,
#' zooming, clicking, etc. This function allows dynamically updating the configuration
#' of these behaviors without having to reinitialize the graph.
#'
#' @seealso \code{\link{g6_proxy}}
#' @export
#' @examples
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   nodes <- data.frame(id = c("node1", "node2", "node3"))
#'   edges <- data.frame(source = "node1", target = "node2")
#'
#'   ui <- page_fluid(
#'     g6_output("graph"),
#'     checkboxInput("enable_click_select", "Enable Click Select", value = TRUE)
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(nodes = nodes, edges = edges) |>
#'         g6_layout() |>
#'         g6_options(animation = FALSE) |>
#'         g6_behaviors(
#'           click_select()
#'         )
#'     })
#'
#'     observeEvent(input$enable_click_select, {
#'       g6_update_behavior(
#'         g6_proxy("graph"),
#'         key = "click-select",
#'         enable = input$enable_click_select
#'       )
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
#'
g6_update_behavior <- function(graph, key, ...) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_update_behavior with g6 object. Use only within shiny and using g6_proxy."
    )
  }
  opts <- list(key = key, ...)
  graph$session$sendCustomMessage(
    sprintf("%s_g6-update-behavior", graph$id),
    # Mark any element of options wrapped by htmlwidgets::JS
    # so it can be evaluated on the JS side.
    list(opts = opts, evals = JSEvals(opts))
  )
  graph
}

#' Set the theme for a g6 graph via proxy
#'
#' This function sets the theme for an existing g6 graph instance
#'
#' @param graph A g6_proxy object created with \code{\link{g6_proxy}}.
#' @param theme A character string representing the theme to apply to the graph.
#' There are 2 internal predefined themes: `light` and `dark`.
#' Alternatively, you can pass a custom theme object that conforms to the G6 theme specifications,
#' according to the documentation at \url{https://g6.antv.antgroup.com/en/manual/theme/custom-theme}.
#'
#' @return The g6_proxy object (invisibly), allowing for method chaining.
#' @seealso [g6_proxy()]
#' @export
#' @examples
#' if (interactive()) {
#'   library(shiny)
#'   library(bslib)
#'   library(g6R)
#'
#'   themes <- list(
#'     "light" = list(
#'       theme = 'light',
#'       node = list(
#'         style = list(
#'           size = 4
#'         ),
#'         palette = list(
#'           type = 'group',
#'           field = 'cluster'
#'         )
#'       ),
#'       plugins = list(
#'         list(
#'           type = 'background',
#'           background = '#fff'
#'         )
#'       )
#'     ),
#'     "dark" = list(
#'       theme = 'dark',
#'       node = list(
#'         style = list(
#'           size = 4
#'         ),
#'         palette = list(
#'           type = 'group',
#'           field = 'cluster'
#'         )
#'       ),
#'       plugins = list(
#'         list(
#'           type = 'background',
#'           background = '#000'
#'         )
#'       )
#'     ),
#'     "blue" = list(
#'       theme = 'light',
#'       node = list(
#'         style = list(
#'           size = 4
#'         ),
#'         palette = list(
#'           type = 'group',
#'           field = 'cluster',
#'           color = "blues",
#'           invert = TRUE
#'         )
#'       ),
#'       plugins = list(
#'         list(
#'           type = 'background',
#'           background = '#f3faff'
#'         )
#'       )
#'     ),
#'     "yellow" = list(
#'       theme = 'light',
#'       node = list(
#'         style = list(
#'           size = 4
#'         ),
#'         palette = list(
#'           type = 'group',
#'           field = 'cluster',
#'           color = c(
#'             '#ffe7ba',
#'             '#ffd591',
#'             '#ffc069',
#'             '#ffa940',
#'             '#fa8c16',
#'             '#d46b08',
#'             '#ad4e00',
#'             '#873800',
#'             '#612500'
#'           )
#'         )
#'       ),
#'       plugins = list(
#'         list(
#'           type = 'background',
#'           background = '#fcf9f1'
#'         )
#'       )
#'     )
#'   )
#'
#'   ui <- page_fluid(
#'     selectInput(
#'       "theme",
#'       "Theme",
#'       choices = c("light", "dark", "blue", "yellow"),
#'       selected = "default"
#'     ),
#'     g6Output("graph", height = "100vh")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- renderG6({
#'       g6(
#'         jsonUrl = "https://assets.antv.antgroup.com/g6/20000.json"
#'       ) |>
#'         g6_options(
#'           animation = FALSE,
#'           autoFit = "view",
#'           padding = 20,
#'           node = list(
#'             style = list(size = 4),
#'             palette = list(
#'               type = "group",
#'               field = "cluster"
#'             )
#'           )
#'         ) |>
#'         g6_behaviors(
#'           "zoom-canvas",
#'           "drag-canvas",
#'           "optimize-viewport-transform"
#'         )
#'     })
#'
#'     observeEvent(input$theme, {
#'       g6_proxy("graph") |>
#'         g6_set_theme(themes[[input$theme]])
#'     })
#'
#'     observeEvent(input[["graph-initialized"]], {
#'       showNotification(
#'         "Graph initialized",
#'         type = "message"
#'       )
#'     })
#'
#'     observeEvent(input[["graph-state"]], {
#'       showNotification(
#'         "Graph state updated",
#'         type = "message"
#'       )
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#'
#' }
g6_set_theme <- function(graph, theme) {
  if (!any(class(graph) %in% "g6_proxy")) {
    stop(
      "Can't use g6_set_theme with g6 object. Use only within shiny and using g6_proxy."
    )
  }
  graph$session$sendCustomMessage(
    sprintf("%s_g6-set-theme", graph$id),
    list(theme = theme)
  )
  graph
}

#' @keywords internal
prefix_port_key <- function(key, nid) {
  prefix <- paste0(nid, "-")
  ifelse(startsWith(key, prefix), key, paste0(prefix, key))
}

#' Update ports for one or more nodes in a g6 graph via proxy
#'
#' This function sends a message to the client
#' to update (add, remove, or modify) ports
#' for one or more nodes. The actual update is handled
#' on the JS side.
#'
#' Removing a port that is currently used by an edge
#' removes the edge as well. Conversely, removing an edge
#' does not remove the ports it was using.
#'
#' @param graph A g6_proxy object.
#' @param ops A named list of operations, one entry per node (named by node ID).
#' Each entry can contain:
#'   \itemize{
#'     \item \code{add}: a list of port objects to add.
#'     \item \code{remove}: a character vector of port keys to remove
#'     \item \code{update}: a list of port objects (with key) to update
#'   }
#' @return The g6_proxy object.
#' @export
#' @examples
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'
#'   ui <- fluidPage(
#'     g6_output("graph", height = "500px"),
#'     actionButton("update_ports", "Update Ports")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(
#'         nodes = g6_nodes(
#'           g6_node(
#'             id = "A",
#'             type = "custom-rect-node",
#'             ports = g6_ports(
#'               g6_input_port(key = "in1", label = "in1", placement = "left"),
#'               g6_output_port(key = "out1", label = "out1", placement = "right"),
#'               g6_output_port(key = "out2", label = "out2", placement = c(1, 1))
#'             ),
#'             style = list(x = 100, y = 200, labelText = "Node A")
#'           ),
#'           g6_node(
#'             id = "B",
#'             type = "custom-rect-node",
#'             ports = g6_ports(
#'               g6_input_port(key = "in2", label = "in2", placement = "left"),
#'               g6_output_port(key = "out3", label = "out3", placement = "right"),
#'               g6_output_port(key = "out4", label = "out4", placement = c(1, 0))
#'             ),
#'             style = list(x = 300, y = 200, labelText = "Node B")
#'           )
#'         ),
#'         edges = g6_edges(
#'           g6_edge(source = "A", target = "B", style = list(sourcePort = "out1", targetPort = "in2"))
#'         )
#'       ) |> g6_behaviors(click_select(), drag_element(), drag_canvas())
#'     })
#'
#'     observeEvent(input$update_ports, {
#'       g6_update_ports(
#'         g6_proxy("graph"),
#'         list(
#'           A = list(remove = c("out1", "out2")),
#'           B = list(
#'             add = list(g6_port(key = "new", label = "new", placement = "top")),
#'             update = list(g6_port(key = "in2", label = "Updated label"))
#'           )
#'         )
#'       )
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
g6_update_ports <- function(graph, ops) {
  if (!is.list(ops) || is.null(names(ops))) {
    stop("'ops' must be a named list with node IDs as names.")
  }

  ids <- names(ops)

  if (anyDuplicated(ids)) {
    stop(
      "Duplicate node IDs in 'ops': ",
      paste(unique(ids[duplicated(ids)]), collapse = ", ")
    )
  }

  for (nid in ids) {
    node_ops <- ops[[nid]]

    # Coerce, validate, and prefix add
    if (length(node_ops$add)) {
      ports <- as_g6_ports(node_ops$add)
      for (i in seq_along(ports)) {
        ports[[i]]$key <- prefix_port_key(ports[[i]]$key, nid)
      }
      ops[[nid]]$add <- ports
    }

    # Coerce, validate, and prefix update
    if (length(node_ops$update)) {
      ports <- as_g6_ports(node_ops$update)
      for (i in seq_along(ports)) {
        ports[[i]]$key <- prefix_port_key(ports[[i]]$key, nid)
      }
      ops[[nid]]$update <- ports
    }

    # Validate and prefix remove
    if (length(node_ops$remove)) {
      if (!is.character(node_ops$remove)) {
        stop(sprintf(
          "For node '%s', 'remove' must be a character vector of port keys",
          nid
        ))
      }
      ops[[nid]]$remove <- prefix_port_key(node_ops$remove, nid)
    }
  }

  graph$session$sendCustomMessage(
    sprintf("%s_g6-update-ports", graph$id),
    list(ids = ids, ops = ops)
  )
  graph
}

#' Extract ports from a g6 graph state via proxy
#'
#' These helpers extract ports from the graph state stored in Shiny input,
#' accessed via a g6_proxy object. Ports are grouped by node and can be filtered by type.
#'
#' @param graph A g6_proxy object.
#' @return A named list of ports for each node, optionally filtered by type.
#'
#' @rdname get-ports
#' @export
g6_get_ports <- function(graph) {
  stopifnot(inherits(graph, "g6_proxy"))
  ns_prefix <- graph$session$ns("")
  id <- graph$id
  if (!is.null(ns_prefix) && nzchar(ns_prefix)) {
    # Remove namespace prefix if present
    if (startsWith(id, ns_prefix)) {
      id <- sub(paste0("^", ns_prefix), "", id)
    }
  }
  state <- graph$session$input[[sprintf("%s-state", id)]]
  nodes <- state$nodes
  if (is.null(nodes)) {
    return(list())
  }
  ports <- lapply(nodes, function(node) node$style$ports)
  ports[!vapply(ports, is.null, logical(1))]
}

g6_get_type_ports <- function(graph, type = c("input", "output")) {
  type <- match.arg(type)
  all_ports <- g6_get_ports(graph)
  lapply(all_ports, function(port_list) {
    Filter(function(port) identical(port$type, type), port_list)
  })
}

#' @rdname get-ports
#' @export
g6_get_input_ports <- function(graph) {
  g6_get_type_ports(graph, "input")
}

#' @rdname get-ports
#' @export
g6_get_output_ports <- function(graph) {
  g6_get_type_ports(graph, "output")
}

library(testthat)
library(shiny)

test_that("g6_proxy creation works correctly", {
  session <- list()
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_s3_class(proxy, "g6_proxy")
  expect_equal(proxy$id, "test_id")
  expect_equal(proxy$session, session)
  expect_error(
    g6_proxy("test_id", NULL),
    "g6_proxy must be called from the server function of a Shiny app"
  )
})

test_that("g6_data_proxy handles inputs correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  test_df <- data.frame(id = c("1", "2"), label = c("Node 1", "Node 2"))
  expect_error(g6_data_proxy(proxy, test_df, "add", "Node"), NA)
  expect_snapshot(error = TRUE, {
    g6_data_proxy(list(), test_df, "add", "Node")
  })
})

test_that("g6_add_nodes handles various input forms and flattening", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  nodes_df <- data.frame(id = c("1", "2"))
  expect_error(g6_add_nodes(proxy, nodes_df), NA)
  nodes_list <- list(list(id = "3"), list(id = "4"))
  expect_error(g6_add_nodes(proxy, nodes_list), NA)
  expect_error(g6_add_nodes(proxy, list(id = "5"), list(id = "6")), NA)
  expect_error(
    g6_add_nodes(proxy, list(list(list(id = "7")))),
    "Input is too deeply nested"
  )
  expect_error(
    g6_add_nodes(proxy, g6_nodes(g6_node(id = "8"), g6_node(id = "9"))),
    NA
  )
  expect_error(g6_add_nodes(proxy, g6_node(id = "10"), g6_node(id = "11")), NA)
})

test_that("g6_remove_nodes works correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(g6_remove_nodes(proxy, c("node1", "node2")), NA)
})

test_that("g6_update_nodes works correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  update_df <- data.frame(id = "1")
  expect_error(g6_update_nodes(proxy, update_df), NA)
})

test_that("g6_canvas_resize works correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(g6_canvas_resize(proxy, 800, 600), NA)
  expect_error(
    g6_canvas_resize(list(), 800, 600),
    "Can't use g6_canvas_resize with g6 object"
  )
})

test_that("g6_fit_center works correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(g6_fit_center(proxy), NA)
  animation <- list(duration = 300, easing = "ease-out")
  expect_error(g6_fit_center(proxy, animation), NA)
  expect_snapshot(error = TRUE, {
    g6_fit_center(proxy, "invalid")
  })
})

test_that("g6_focus_elements works correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(g6_focus_elements(proxy, "node1"), NA)
  expect_error(g6_focus_elements(proxy, c("node1", "node2")), NA)
  animation <- list(duration = 300, easing = "ease-out")
  expect_error(g6_focus_elements(proxy, "node1", animation), NA)
})

test_that("g6_hide_elements and g6_show_elements work correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(g6_hide_elements(proxy, c("node1", "node2")), NA)
  expect_error(g6_show_elements(proxy, c("node1", "node2")), NA)
})

test_that("g6_collapse_combo and g6_expand_combo work correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(g6_collapse_combo(proxy, "combo1"), NA)
  expect_error(g6_expand_combo(proxy, "combo1"), NA)
  options <- list(animate = TRUE, align = TRUE)
  expect_error(g6_collapse_combo(proxy, "combo1", options), NA)
  expect_error(g6_expand_combo(proxy, "combo1", options), NA)
})

test_that("g6_set_options works correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(
    g6_set_options(
      proxy,
      fitView = TRUE,
      animate = TRUE,
      modes = list(default = list(type = "drag-canvas"))
    ),
    NA
  )
})

test_that("g6_add_plugin and g6_update_plugin work correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(
    g6_add_plugin(
      proxy,
      minimap = list(size = list(width = 100, height = 100))
    ),
    NA
  )
  expect_error(
    g6_update_plugin(
      proxy,
      "minimap",
      size = list(width = 150, height = 150)
    ),
    NA
  )
})

test_that("g6_update_behavior works correctly", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(
    g6_update_behavior(
      proxy,
      "drag-canvas",
      enableOptimize = TRUE
    ),
    NA
  )
})

test_that("g6_add_edges enforces source/target for creation", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)
  expect_error(g6_add_edges(proxy, list(source = "1", target = "2")), NA)
  edges_df <- data.frame(source = "1", target = "2")
  expect_error(g6_add_edges(proxy, edges_df), NA)
})

test_that("as_g6_elements flattens any level of nesting and errors on excessive nesting", {
  expect_equal(
    flatten_g6_elements(list(list(source = 1, target = 2))),
    list(list(source = 1, target = 2))
  )
  expect_equal(
    flatten_g6_elements(list(list(list(source = 1, target = 2)))),
    list(list(source = 1, target = 2))
  )
  expect_error(
    as_g6_elements(list(list(list(list(id = "7")))), coerc_func = as_g6_nodes),
    "Input is too deeply nested"
  )
})

test_that("g6_update_layout proxy call works and errors on invalid proxy", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("test_id", session)

  # should not error for a valid proxy
  expect_error(g6_update_layout(proxy, type = "grid"), NA)

  # invalid proxy (not a g6_proxy) should error
  expect_error(g6_update_layout(list(), type = "grid"))
})

test_that("g6_update_ports validates and sends correct structure", {
  session <- list(
    sendCustomMessage = function(type, message) {
      list(type = type, message = message)
    }
  )
  class(session) <- "ShinySession"
  proxy <- g6_proxy("graph1", session)

  # Valid: add and remove
  expect_error(
    g6_update_ports(
      proxy,
      ops = list(
        A = list(add = g6_ports(g6_port("out1")), remove = c("in1")),
        B = list(
          update = g6_ports(g6_port("in2", label = "lbl")),
          remove = character()
        )
      )
    ),
    NA
  )

  # Error: ops not a named list
  expect_error(
    g6_update_ports(
      proxy,
      ops = list(list(remove = "x"))
    ),
    "'ops' must be a named list"
  )

  # Error: duplicate node IDs
  expect_error(
    g6_update_ports(
      proxy,
      ops = list(A = list(remove = "x"), A = list(remove = "y"))
    ),
    "Duplicate node IDs"
  )

  # Error: remove not character
  expect_error(
    g6_update_ports(
      proxy,
      ops = list(A = list(remove = 123))
    ),
    "remove' must be a character vector"
  )

  # Error: add not valid ports
  expect_error(
    g6_update_ports(
      proxy,
      ops = list(A = list(add = list(list(type = "input")))) # missing key
    ),
    "key"
  )
})

test_that("g6_get_ports returns all ports grouped by node", {
  proxy <- structure(
    list(
      id = "graph1",
      session = list(
        ns = function(x = "") "",
        input = list(
          "graph1-state" = list(
            nodes = list(
              nodeA = list(
                style = list(
                  ports = list(
                    port1 = list(type = "input"),
                    port2 = list(type = "output")
                  )
                )
              ),
              nodeB = list(
                style = list(
                  ports = list(
                    port3 = list(type = "input")
                  )
                )
              )
            )
          )
        )
      )
    ),
    class = "g6_proxy"
  )
  ports <- g6_get_ports(proxy)
  expect_type(ports, "list")
  expect_named(ports, c("nodeA", "nodeB"))
  expect_true(all(c("port1", "port2") %in% names(ports$nodeA)))
  expect_true("port3" %in% names(ports$nodeB))
})

test_that("g6_get_type_ports filters ports by type", {
  proxy <- structure(
    list(
      id = "graph1",
      session = list(
        ns = function(x = "") "",
        input = list(
          "graph1-state" = list(
            nodes = list(
              nodeA = list(
                style = list(
                  ports = list(
                    port1 = list(type = "input"),
                    port2 = list(type = "output")
                  )
                )
              ),
              nodeB = list(
                style = list(
                  ports = list(
                    port3 = list(type = "input")
                  )
                )
              )
            )
          )
        )
      )
    ),
    class = "g6_proxy"
  )
  input_ports <- g6_get_input_ports(proxy)
  expect_true("port1" %in% names(input_ports$nodeA))
  expect_false("port2" %in% names(input_ports$nodeA))
  expect_true("port3" %in% names(input_ports$nodeB))

  output_ports <- g6_get_output_ports(proxy)
  expect_true("port2" %in% names(output_ports$nodeA))
  expect_false("port1" %in% names(output_ports$nodeA))
  expect_length(output_ports$nodeB, 0)
})

test_that("g6_get_ports returns empty list if no nodes", {
  proxy <- structure(
    list(
      id = "graph1",
      session = list(
        ns = function(x = "") "",
        input = list(
          "graph1-state" = list(nodes = NULL)
        )
      )
    ),
    class = "g6_proxy"
  )
  ports <- g6_get_ports(proxy)
  expect_equal(ports, list())
})

test_that("g6_get_type_ports errors on invalid type", {
  proxy <- structure(
    list(
      id = "graph1",
      session = list(
        ns = function(x = "") "",
        input = list(
          "graph1-state" = list(
            nodes = list(
              nodeA = list(
                style = list(
                  ports = list(
                    port1 = list(type = "input")
                  )
                )
              )
            )
          )
        )
      )
    ),
    class = "g6_proxy"
  )
  expect_error(g6_get_type_ports(proxy, "foo"))
})

#' Configure Global Options for G6 Graph
#'
#' Sets up the global configuration options for a G6 graph including node, edge and
#' combo styles, layout, canvas, animation, and interactive behavior settings.
#'
#' @details
#' The `g6_options` function provides a comprehensive configuration interface for G6 graphs.
#' It allows you to control all aspects of graph rendering and behavior, from styling of
#' individual elements to global visualization settings.
#'
#' @param graph g6 graph instance.
#' @param node Node configuration. Controls the default appearance and behavior of nodes.
#'   Created with \code{node_options()}. Default: NULL.
#'
#' @param edge Edge configuration. Controls the default appearance and behavior of edges.
#'   Created with \code{edge_options()}. Default: NULL.
#'
#' @param combo Combo configuration. Controls the default appearance and behavior of combo nodes.
#'   Created with \code{combo_options()}. Default: NULL.
#'
#' @param autoFit Automatically fit the graph content to the canvas.
#'   Created with \code{auto_fit_config()}. Default: NULL.
#'
#' @param canvas Canvas configuration for the graph rendering surface.
#'   Created with \code{canvas_config()}. Default: NULL.
#'
#' @param animation Global animation configuration for graph transitions.
#'   Created with \code{animation_config()}. Default: TRUE.
#'
#' @param autoResize Whether the graph should automatically resize when the window size changes.
#'   Default: FALSE.
#'
#' @param background Background color of the graph. If not specified, the background will be transparent.
#'   Default: NULL.
#'
#' @param cursor Default mouse cursor style when hovering over the graph.
#'   Options include: "default", "pointer", "move", etc.
#'   Default: "default".
#'
#' @param devicePixelRatio Device pixel ratio for rendering on high-DPI displays. If NULL,
#'   the browser's default device pixel ratio will be used.
#'   Default: NULL.
#'
#' @param renderer Rendering engine to use. A JS function. To render
#' as svg, you can pass `() => new SVGRenderer()`.
#'   Default: NULL (G6 will choose the appropriate renderer).
#'
#' @param padding Padding around the graph content in pixels. Can be a single number for equal padding
#'   on all sides or a vector of four numbers \code{[top, right, bottom, left]}.
#'   Default: NULL.
#'
#' @param rotation Rotation angle of the entire graph in degrees.
#'   Default: 0.
#'
#' @param x X-coordinate of the graph's center relative to the container.
#'   Default: NULL (will use container center).
#'
#' @param y Y-coordinate of the graph's center relative to the container.
#'   Default: NULL (will use container center).
#'
#' @param zoom Initial zoom level of the graph. 1 represents 100% (original size).
#'   Default: 1.
#'
#' @param zoomRange Minimum and maximum allowed zoom levels, specified as a vector
#'   with two elements: c(min_zoom, max_zoom).
#'   Default: c(0.01, 10).
#'
#' @param theme Color theme for the graph. Either `light` or `dark` or a list
#' representing a custom theme:
#' see \url{https://g6.antv.antgroup.com/en/manual/theme/custom-theme}.
#' @param ... Other configuration parameters.
#'
#' @return A list containing all specified G6 graph configuration options.
#'
#' @examples
#' # Basic usage with defaults
#' opts <- g6_options(g6())
#'
#' # Customize node and edge styles
#' opts <- g6_options(
#'   g6(),
#'   node = node_options(
#'     type = "circle",
#'     style = node_style_options(
#'       fill = "#1783FF",
#'       stroke = "#0066CC"
#'     )
#'   ),
#'   edge = edge_options(
#'     type = "cubic",
#'     style = edge_style_options(
#'       stroke = "#999999",
#'       lineWidth = 1.5
#'     )
#'   )
#' )
#'
#' # Configure graph with dark theme, auto-resize, and custom background
#' opts <- g6_options(
#'   g6(),
#'   theme = "dark",
#'   autoResize = TRUE,
#'   background = "#222222",
#'   padding = 20,
#'   zoom = 0.8,
#'   zoomRange = c(0.5, 2)
#' )
#'
#' # Configure with custom animations
#' opts <- g6_options(
#'   g6(),
#'   animation = animation_config(
#'     duration = 500,
#'     easing = "easeCubic"
#'   ),
#'   autoFit = auto_fit_config(duration = 300, easing = "ease-out")
#' )
#' @export
g6_options <- function(
  graph,
  node = NULL,
  edge = NULL,
  combo = NULL,
  autoFit = NULL,
  canvas = NULL,
  animation = TRUE,
  autoResize = FALSE,
  background = NULL,
  cursor = valid_cursors,
  devicePixelRatio = NULL,
  renderer = NULL,
  padding = NULL,
  rotation = 0,
  x = NULL,
  y = NULL,
  zoom = 1,
  zoomRange = c(0.01, 10),
  theme = "light",
  ...
) {
  if (!inherits(graph, "g6")) {
    stop("g6_options must be called on a g6 instance")
  }

  if (!is.null(renderer) && !is_js(renderer)) {
    stop("'renderer' must be a JavaScript function wrapped with JS()")
  }

  if ((isTRUE(animation) || is.list(animation)) && get_g6_preserve_position()) {
    warning(
      "'g6R.preserve_elements_position' 
      only works when animation is FALSE. It will be ignored."
    )
  }

  cursor <- match.arg(cursor)

  arg_names <- names(formals())
  arg_names <- arg_names[!(arg_names %in% c("...", "graph"))]
  # Get values of only the named parameters
  config <- mget(arg_names)

  # Drop NULL elements
  graph$x <- c(graph$x, dropNulls(c(config, list(...))))
  graph
}

#' Create Auto-Fit Configuration for G6 Graphs
#'
#' Configures the auto-fit behavior for a G6 graph. Auto-fit automatically adjusts
#' the view to fit all elements or centers them within the canvas.
#'
#' @details
#' The auto-fit feature helps ensure that graph elements remain visible within the canvas.
#' It can be configured to either fit all elements to view or center them, and can be
#' triggered under different conditions.
#'
#' @param type The auto-fit mode to use. Options:
#'   \itemize{
#'     \item \code{"view"}: Scale and translate the graph to fit all elements within the view (default)
#'     \item \code{"center"}: Only translate the graph to center elements without scaling
#'   }
#'
#' @param when When the auto-fit should be triggered. Options:
#'   \itemize{
#'     \item \code{"overflow"}: Trigger auto-fit only when elements overflow the canvas (default)
#'     \item \code{"always"}: Always perform auto-fit when the graph data changes
#'   }
#'
#' @param direction The direction for auto-fit adjustment. Options:
#'   \itemize{
#'     \item \code{"x"}: Adjust only along the x-axis
#'     \item \code{"y"}: Adjust only along the y-axis
#'     \item \code{"both"}: Adjust in both x and y directions (default)
#'   }
#'
#' @param duration The duration of the auto-fit animation in milliseconds (default: 1000)
#'
#' @param easing The animation easing function to use. Options:
#'   \itemize{
#'     \item \code{"ease-in-out"}: Slow at the beginning and end of the animation (default)
#'     \item \code{"ease"}: Standard easing
#'     \item \code{"ease-in"}: Slow at the beginning
#'     \item \code{"ease-out"}: Slow at the end
#'     \item \code{"linear"}: Constant speed throughout
#'     \item \code{"cubic-bezier"}: Custom cubic-bezier curve
#'     \item \code{"step-start"}: Jump immediately to the end state
#'     \item \code{"step-end"}: Jump at the end to the end state
#'   }
#'
#' @return A list containing the auto-fit configuration that can be passed to [g6_options()].
#'
#' @export
#'
#' @examples
#' # Basic auto-fit configuration with default settings
#' config <- auto_fit_config()
#'
#' # Auto-fit with only centering (no scaling)
#' config <- auto_fit_config(type = "center")
#'
#' # Auto-fit that always triggers when graph data changes
#' config <- auto_fit_config(when = "always")
#'
#' # Auto-fit only in the x direction
#' config <- auto_fit_config(direction = "x")
#'
#' # Auto-fit with a fast animation
#' config <- auto_fit_config(duration = 300, easing = "ease-out")
auto_fit_config <- function(
  type = c("view", "center"),
  when = c("overflow", "always"),
  direction = c("x", "y", "both"),
  duration = 1000,
  easing = c(
    "ease-in-out",
    "ease",
    "ease-in",
    "ease-out",
    "linear",
    "cubic-bezier",
    "step-start",
    "step-end"
  )
) {
  # Match arguments against their possible values
  type <- match.arg(type)
  when <- match.arg(when)
  direction <- match.arg(direction)
  easing <- match.arg(easing)

  # Validate numeric parameters
  if (!is.numeric(duration) || duration < 0) {
    stop("'duration' must be a non-negative numeric value")
  }

  # Create the main structure
  list(
    type = type,
    options = list(
      when = when,
      direction = direction
    ),
    animation = list(
      duration = duration,
      easing = easing
    )
  )
}

#' Create Canvas Configuration for G6 Graphs
#'
#' Configures the canvas settings for a G6 graph. The canvas is the rendering
#' surface where the graph is drawn.
#'
#' @details
#' Canvas configuration controls how the graph is rendered, including its size,
#' scaling, background, and rendering layer settings. This function provides a
#' structured way to configure all canvas-related options.
#'
#' Note that many of these settings (container, width, height, devicePixelRatio,
#' background, cursor) can also be set directly in the main graph configuration,
#' which will be automatically converted to canvas configuration items.
#'
#' @param container The container element for the canvas. Can be a CSS selector string
#'   or an HTML element reference.
#'
#' @param devicePixelRatio The device pixel ratio to use for rendering. Higher values
#'   provide sharper rendering on high-DPI displays but may impact performance.
#'   If not specified, the device's pixel ratio will be used.
#'
#' @param width The width of the canvas in pixels.
#'
#' @param height The height of the canvas in pixels.
#'
#' @param cursor The CSS cursor style to use when hovering over the canvas.
#'   Common values include "default", "pointer", "move", etc.
#'
#' @param background The background color of the canvas. Can be any valid CSS color
#'   value (hex, rgb, rgba, named colors).
#'
#' @param renderer A function that returns a renderer for different layers.
#'   The function takes a layer parameter which can be 'background', 'main',
#'   'label', or 'transient'.
#'
#' @param enableMultiLayer Whether to enable multi-layer rendering. This is a
#'   non-dynamic parameter and is only effective during initialization.
#'   Multi-layer rendering can improve performance for complex graphs by
#'   separating elements into different rendering layers.
#'
#' @return A list containing the canvas configuration that can be passed to [g6_options()].
#'
#' @export
#'
#' @examples
#' # Basic canvas configuration
#' config <- canvas_config(
#'   container = "#graph-container",
#'   width = 800,
#'   height = 600
#' )
#'
#' # Canvas with multi-layer rendering enabled
#' config <- canvas_config(
#'   container = "#graph-container",
#'   width = 1000,
#'   height = 700,
#'   enableMultiLayer = TRUE,
#'   cursor = "grab"
#' )
canvas_config <- function(
  container = NULL,
  devicePixelRatio = NULL,
  width = NULL,
  height = NULL,
  cursor = NULL,
  background = NULL,
  renderer = NULL,
  enableMultiLayer = NULL
) {
  # Validate container
  if (!is.null(container) && !is.character(container) && !is_js(container)) {
    stop(
      "'container' must be a CSS selector string or an HTML element reference"
    )
  }

  # Validate devicePixelRatio
  if (!is.null(devicePixelRatio)) {
    if (!is.numeric(devicePixelRatio) || devicePixelRatio <= 0) {
      stop("'devicePixelRatio' must be a positive number")
    }
  }

  # Validate width
  if (!is.null(width)) {
    if (!is.numeric(width) || width <= 0) {
      stop("'width' must be a positive number")
    }
  }

  # Validate height
  if (!is.null(height)) {
    if (!is.numeric(height) || height <= 0) {
      stop("'height' must be a positive number")
    }
  }

  # Validate cursor
  if (!is.null(cursor) && !is.character(cursor)) {
    stop("'cursor' must be a string (CSS cursor value)")
  }

  # Validate background
  if (!is.null(background) && !is.character(background)) {
    stop("'background' must be a string (CSS color value)")
  }

  # Validate renderer
  if (!is.null(renderer) && !is_js(renderer)) {
    stop("'renderer' must be a JavaScript function")
  }

  # Validate enableMultiLayer
  if (!is.null(enableMultiLayer) && !is.logical(enableMultiLayer)) {
    stop("'enableMultiLayer' must be a boolean value")
  }

  # Get all argument names
  arg_names <- names(formals())
  # Create list of argument values
  dropNulls(mget(arg_names))
}

#' Create Animation Configuration for G6 Graphs
#'
#' Configures animation settings for G6 graph elements. These settings control
#' how graph elements animate when changes occur.
#'
#' @details
#' Animation configuration allows fine-tuning the timing and behavior of animations
#' in G6 graphs. This includes controlling the duration, delay, easing function,
#' direction, and other aspects of how graph elements animate.
#'
#' @param delay Animation delay time in milliseconds. The time to wait before the
#'   animation begins. Must be a non-negative numeric value.
#'
#' @param direction Animation playback direction. Options:
#'   \itemize{
#'     \item \code{"forward"}: Plays normally (default)
#'     \item \code{"alternate"}: Plays forward, then in reverse
#'     \item \code{"alternate-reverse"}: Plays in reverse, then forward
#'     \item \code{"normal"}: Same as forward
#'     \item \code{"reverse"}: Plays in reverse direction
#'   }
#'
#' @param duration Animation duration in milliseconds. The length of time the
#'   animation will take to complete one cycle. Must be a non-negative numeric value.
#'
#' @param easing Animation easing function. Controls the rate of change during
#'   the animation. Common values include "linear", "ease", "ease-in", "ease-out",
#'   "ease-in-out", or cubic-bezier values.
#'
#' @param fill Fill mode after animation ends. Options:
#'   \itemize{
#'     \item \code{"none"}: Element returns to its initial state when animation ends (default)
#'     \item \code{"auto"}: Follows the rules of the animation effect
#'     \item \code{"backwards"}: Element retains first keyframe values during delay period
#'     \item \code{"both"}: Combines forwards and backwards behavior
#'     \item \code{"forwards"}: Element retains final keyframe values after animation ends
#'   }
#'
#' @param iterations Number of times the animation should repeat. A value of
#'   \code{Inf} will cause the animation to repeat indefinitely. Must be a
#'   non-negative numeric value.
#'
#' @return A list containing animation configuration that can be passed to [g6_options()].
#'
#' @export
#'
#' @examples
#' # Basic animation with duration
#' config <- animation_config(
#'   duration = 500
#' )
#'
#' # Complex animation configuration
#' config <- animation_config(
#'   delay = 100,
#'   duration = 800,
#'   easing = "ease-in-out",
#'   direction = "alternate",
#'   fill = "forwards",
#'   iterations = 2
#' )
#'
#' # Infinite animation
#' config <- animation_config(
#'   duration = 1000,
#'   easing = "linear",
#'   iterations = Inf
#' )
animation_config <- function(
  delay = NULL,
  direction = c(
    "forward",
    "alternate",
    "alternate-reverse",
    "normal",
    "reverse"
  ),
  duration = NULL,
  easing = NULL,
  fill = c("none", "auto", "backwards", "both", "forwards"),
  iterations = NULL
) {
  direction <- match.arg(direction)
  fill <- match.arg(fill)

  # Validate numeric parameters
  if (!is.null(delay) && (!is.numeric(delay) || delay < 0)) {
    stop("'delay' must be a non-negative numeric value")
  }

  if (!is.null(duration) && (!is.numeric(duration) || duration < 0)) {
    stop("'duration' must be a non-negative numeric value")
  }

  if (!is.null(iterations)) {
    if (!is.numeric(iterations) || iterations < 0) {
      stop("'iterations' must be a non-negative numeric value")
    }

    # Convert Inf to proper JavaScript representation if needed
    if (is.infinite(iterations)) {
      iterations <- "Infinity"
    }
  }

  # Validate easing parameter
  if (!is.null(easing) && !is.character(easing)) {
    stop("'easing' must be a string value (e.g., 'linear', 'ease-in-out')")
  }

  # Get all argument names
  arg_names <- names(formals())
  # Create list of argument values
  config <- mget(arg_names)
  # Drop NULL elements
  dropNulls(config)
}

#' Create Node Options Configuration for G6 Graphs
#'
#' Configures the general options for nodes in a G6 graph. These settings control
#' the type, style, state, palette, and animation of nodes.
#'
#' @details
#' Node options allow defining how nodes appear and behave in a G6 graph. This includes
#' selecting node types, setting styles, configuring state-based appearances, defining
#' color palettes, and specifying animation effects.
#'
#' @param type Node type. Can be a built-in node type name or a custom node name.
#'   Built-in types include "circle", "rect", "ellipse", "diamond", "triangle", etc.
#'   Default: "circle".
#'
#' @param style Node style configuration. Controls the appearance of nodes including color,
#'   size, border, etc. Can be created with \code{node_style_options()}.
#'   Default: NULL.
#'
#' @param state Defines the style of the node in different states, such as hover, selected,
#'   disabled, etc. Should be a list mapping state names to style configurations.
#'   Default: NULL.
#'
#' @param palette Defines the color palette of the node, used to map colors based on different data.
#'   Default: NULL.
#'
#' @param animation Defines the animation effect of the node. Can be created with
#'   \code{animation_config()}.
#'   Default: NULL.
#'
#' @return A list containing node options configuration that can be passed to [g6_options()].
#'
#' @export
#'
#' @examples
#' # Basic node options with default circle type
#' options <- node_options()
#'
#' # Rectangle node with custom style
#' options <- node_options(
#'   type = "rect",
#'   style = node_style_options(
#'     fill = "#E8F7FF",
#'     stroke = "#1783FF",
#'     lineWidth = 2
#'   )
#' )
node_options <- function(
  type = c(
    "circle",
    "rect",
    "ellipse",
    "diamond",
    "triangle",
    "star",
    "image",
    "modelRect"
  ),
  style = node_style_options(),
  state = NULL,
  palette = NULL,
  animation = NULL
) {
  # Match type argument against possible values
  type <- match.arg(type)

  # Get all argument names
  arg_names <- names(formals())
  # Create list of argument values
  config <- mget(arg_names)
  # Drop NULL elements
  dropNulls(config)
}

#' Create Node Style Options for G6 Graphs
#'
#' Configures the styling options for nodes in a G6 graph. These settings control
#' the appearance and interaction behavior of nodes. Used in \link{node_options}.
#'
#' @details
#' Node style options allow fine-grained control over how nodes are rendered and
#' behave in a G6 graph. This includes colors, sizes, borders, shadows, visibility,
#' positioning, and interaction states.
#'
#' @param collapsed Whether the current node/group is collapsed. Default: FALSE.
#'
#' @param cursor Node mouse hover cursor style. Common values include "default",
#'   "pointer", "move", etc. Default: "default".
#'
#' @param fill Node fill color. Can be any valid CSS color value. Default: "#1783FF".
#'
#' @param fillOpacity Node fill color opacity. Value between 0 and 1. Default: 1.
#'
#' @param increasedLineWidthForHitTesting When lineWidth is small, this value increases
#'   the interactive area to make "thin lines" easier to interact with. Default: 0.
#'
#' @param lineCap Node stroke end style. Options: "round", "square", "butt". Default: "butt".
#'
#' @param lineDash Node stroke dash style. Vector of numbers specifying dash pattern.
#'
#' @param lineDashOffset Node stroke dash offset. Default: NULL.
#'
#' @param lineJoin Node stroke join style. Options: "round", "bevel", "miter". Default: "miter".
#'
#' @param lineWidth Node stroke width. Default: 1.
#'
#' @param opacity Node overall opacity. Value between 0 and 1. Default: 1.
#'
#' @param shadowBlur Node shadow blur amount. Default: NULL.
#'
#' @param shadowColor Node shadow color. Default: NULL.
#'
#' @param shadowOffsetX Node shadow offset in the x-axis direction. Default: NULL.
#'
#' @param shadowOffsetY Node shadow offset in the y-axis direction. Default: NULL.
#'
#' @param shadowType Node shadow type. Options: "inner", "outer". Default: "outer".
#'
#' @param size Node size. Can be a single number for equal width/height or a vector
#'   of two numbers \code{[width, height]}. Default: 32.
#'
#' @param stroke Node stroke (border) color. Default: "#000".
#'
#' @param strokeOpacity Node stroke color opacity. Value between 0 and 1. Default: 1.
#'
#' @param transform CSS transform attribute to rotate, scale, skew, or translate the node.
#'   Default: NULL.
#'
#' @param transformOrigin Rotation and scaling center point. Default: NULL.
#'
#' @param visibility Whether the node is visible. Options: "visible", "hidden". Default: "visible".
#'
#' @param x Node x coordinate. Default: 0.
#'
#' @param y Node y coordinate. Default: 0.
#'
#' @param z Node z coordinate (for 3D). Default: 0.
#'
#' @param zIndex Node rendering level (for layering). Default: 0.
#' @param ... Other parameters.
#'
#' @return A list containing node style options that can be passed to [node_options()].
#'
#' @export
#'
#' @examples
#' # Basic node style with blue fill and red border
#' styles <- node_style_options(
#'   fill = "#1783FF",
#'   stroke = "#FF0000",
#'   lineWidth = 2
#' )
#'
#' # Create a node with shadow effects
#' styles <- node_style_options(
#'   fill = "#FFFFFF",
#'   stroke = "#333333",
#'   lineWidth = 1,
#'   shadowBlur = 10,
#'   shadowColor = "rgba(0,0,0,0.3)",
#'   shadowOffsetX = 5,
#'   shadowOffsetY = 5
#' )
#'
#' # Custom sized node with dashed border
#' styles <- node_style_options(
#'   size = c(100, 50),
#'   fill = "#E8F7FF",
#'   stroke = "#1783FF",
#'   lineDash = c(5, 5),
#'   opacity = 0.8
#' )
node_style_options <- function(
  collapsed = FALSE,
  cursor = "default",
  fill = "#1783FF",
  fillOpacity = 1,
  increasedLineWidthForHitTesting = 0,
  lineCap = c("butt", "round", "square"),
  lineDash = NULL,
  lineDashOffset = NULL,
  lineJoin = c("miter", "round", "bevel"),
  lineWidth = 1,
  opacity = 1,
  shadowBlur = NULL,
  shadowColor = NULL,
  shadowOffsetX = NULL,
  shadowOffsetY = NULL,
  shadowType = c("outer", "inner"),
  size = 32,
  stroke = "#000",
  strokeOpacity = 1,
  transform = NULL,
  transformOrigin = NULL,
  visibility = c("visible", "hidden"),
  x = NULL,
  y = NULL,
  z = NULL,
  zIndex = NULL,
  ...
) {
  # Validate boolean parameters
  if (!is.null(collapsed) && !is.logical(collapsed)) {
    stop("'collapsed' must be a boolean value (TRUE or FALSE)")
  }

  # Validate string parameters
  string_params <- c(
    "cursor",
    "fill",
    "lineCap",
    "lineJoin",
    "shadowColor",
    "shadowType",
    "stroke",
    "transform",
    "transformOrigin",
    "visibility"
  )

  for (param in string_params) {
    value <- get(param)
    if (!is.null(value) && !is.character(value)) {
      stop(sprintf("'%s' must be a string value", param))
    }
  }

  # Validate specific enum values
  lineCap <- match.arg(lineCap)
  lineJoin <- match.arg(lineJoin)
  shadowType <- match.arg(shadowType)
  visibility <- match.arg(visibility)

  # Validate numeric parameters
  numeric_params <- c(
    "fillOpacity",
    "increasedLineWidthForHitTesting",
    "lineDashOffset",
    "lineWidth",
    "opacity",
    "shadowBlur",
    "shadowOffsetX",
    "shadowOffsetY",
    "strokeOpacity",
    "x",
    "y",
    "z",
    "zIndex"
  )

  for (param in numeric_params) {
    value <- get(param)
    if (!is.null(value) && (!is.numeric(value) || length(value) != 1)) {
      stop(sprintf("'%s' must be a single numeric value", param))
    }
  }

  # Validate opacity values
  opacity_params <- c("fillOpacity", "opacity", "strokeOpacity")

  for (param in opacity_params) {
    value <- get(param)
    if (!is.null(value) && (value < 0 || value > 1)) {
      stop(sprintf("'%s' must be a number between 0 and 1", param))
    }
  }

  # Validate lineDash
  if (!is.null(lineDash) && (!is.numeric(lineDash) || length(lineDash) < 1)) {
    stop("'lineDash' must be a numeric vector")
  }

  # Validate size
  if (!is.null(size)) {
    if (!is.numeric(size)) {
      stop("'size' must be a numeric value or vector of length 2")
    }
    if (length(size) > 2) {
      stop("'size' can only have 1 or 2 elements")
    }
  }

  # Get all argument names and values
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Create Edge Options Configuration for G6 Graphs
#'
#' Configures the general options for edges in a G6 graph. These settings control
#' the type, style, state, palette, and animation of edges.
#'
#' @details
#' Edge options allow defining how edges appear and behave in a G6 graph. This includes
#' selecting edge types, setting styles, configuring state-based appearances, defining
#' color palettes, and specifying animation effects.
#'
#' @param type Edge type. Can be a built-in edge type name or a custom edge name.
#'   Built-in types include "line", "polyline", "arc", "quadratic", "cubic", "cubic-vertical",
#'   "cubic-horizontal", "loop", etc. Default: "line".
#'
#' @param style Edge style configuration. Controls the appearance of edges including color,
#'   width, dash patterns, etc. Can be created with \code{edge_style_options()}.
#'   Default: NULL.
#'
#' @param state Defines the style of the edge in different states, such as hover, selected,
#'   disabled, etc. Should be a list mapping state names to style configurations.
#'   Default: NULL.
#'
#' @param palette Defines the color palette of the edge, used to map colors based on different data.
#'   Default: NULL.
#'
#' @param animation Defines the animation effect of the edge. Can be created with
#'   \code{animation_config()}.
#'   Default: NULL.
#'
#' @return A list containing edge options configuration that can be passed to [g6_options()].
#'
#' @export
#'
#' @examples
#' # Basic edge options with default line type
#' options <- edge_options()
#'
#' # Curved edge with custom style
#' options <- edge_options(
#'   type = "cubic",
#'   style = edge_style_options(
#'     stroke = "#1783FF",
#'     lineWidth = 2,
#'     endArrow = TRUE
#'   )
#' )
edge_options <- function(
  type = c(
    "line",
    "polyline",
    "arc",
    "quadratic",
    "cubic",
    "cubic-vertical",
    "cubic-horizontal",
    "loop"
  ),
  style = edge_style_options(),
  state = NULL,
  palette = NULL,
  animation = NULL
) {
  # Match type argument against possible values
  type <- match.arg(type)

  # Get all argument names
  arg_names <- names(formals())
  # Create list of argument values
  config <- mget(arg_names)
  # Drop NULL elements
  dropNulls(config)
}

#' @keywords internal
valid_pointer_events <- c(
  "visible",
  "visiblepainted",
  "visiblestroke",
  "non-transparent-pixel",
  "visiblefill",
  "painted",
  "fill",
  "stroke",
  "all",
  "none",
  "auto",
  "inherit",
  "initial",
  "unset"
)

#' @keywords internal
valid_cursors <- c(
  'default',
  'auto',
  'none',
  'context-menu',
  'help',
  'pointer',
  'progress',
  'wait',
  'cell',
  'crosshair',
  'text',
  'vertical-text',
  'alias',
  'copy',
  'move',
  'no-drop',
  'not-allowed',
  'grab',
  'grabbing',
  'all-scroll',
  'col-resize',
  'row-resize',
  'n-resize',
  'e-resize',
  's-resize',
  'w-resize',
  'ne-resize',
  'nw-resize',
  'se-resize',
  'sw-resize',
  'ew-resize',
  'ns-resize',
  'nesw-resize',
  'nwse-resize',
  'zoom-in',
  'zoom-out'
)

#' Create Edge Style Options for G6 Graphs
#'
#' Configures the styling options for edges in a G6 graph. These settings control
#' the appearance and interaction behavior of edges.
#'
#' @details
#' Edge style options allow fine-grained control over how edges are rendered and
#' behave in a G6 graph. This includes colors, widths, line styles, shadows, visibility,
#' and interaction properties.
#'
#' @param class Edge class name for custom styling with CSS. Default: NULL.
#'
#' @param cursor Edge mouse hover cursor style. Common values include "default",
#'   "pointer", "move", etc. Default: "default".
#'
#' @param fill Edge area fill color (for edges with area, like loops). Default: NULL.
#'
#' @param fillRule Edge internal fill rule. Options: "nonzero", "evenodd". Default: NULL.
#'
#' @param filter Edge shadow filter effect. Default: NULL.
#'
#' @param increasedLineWidthForHitTesting When the edge width is too small, this value increases
#'   the interaction area to make edges easier to interact with. Default: NULL.
#'
#' @param isBillboard Effective in 3D scenes, always facing the screen so the line width
#'   is not affected by perspective projection. Default: TRUE.
#'
#' @param lineDash Edge dashed line style. Numeric vector specifying dash pattern. Default: 0.
#'
#' @param lineDashOffset Edge dashed line offset. Default: 0.
#'
#' @param lineWidth Edge width in pixels. Default: 1.
#'
#' @param opacity Overall opacity of the edge. Value between 0 and 1. Default: 1.
#'
#' @param pointerEvents Whether the edge responds to pointer events. Default: NULL.
#'
#' @param shadowBlur Edge shadow blur effect amount. Default: NULL.
#'
#' @param shadowColor Edge shadow color. Default: NULL.
#'
#' @param shadowOffsetX Edge shadow X-axis offset. Default: NULL.
#'
#' @param shadowOffsetY Edge shadow Y-axis offset. Default: NULL.
#'
#' @param shadowType Edge shadow type. Options: "inner", "outer", "both". Default: NULL.
#'
#' @param sourcePort Source port of the edge connection. Default: NULL.
#'
#' @param stroke Edge color. Default: "#000".
#'
#' @param strokeOpacity Edge color opacity. Value between 0 and 1. Default: 1.
#'
#' @param targetPort Target port of the edge connection. Default: NULL.
#'
#' @param transform CSS transform attribute to rotate, scale, skew, or translate the edge.
#'   Default: NULL.
#'
#' @param transformOrigin Rotation and scaling center point. Default: NULL.
#'
#' @param visibility Whether the edge is visible. Options: "visible", "hidden". Default: "visible".
#'
#' @param zIndex Edge rendering level (for layering). Default: 1.
#' @param ... Extra parameters.
#'
#' @return A list containing edge style options that can be passed to [edge_options()].
#'
#' @export
edge_style_options <- function(
  class = NULL,
  cursor = valid_cursors,
  fill = NULL,
  fillRule = c("nonzero", "evenodd"),
  filter = NULL,
  increasedLineWidthForHitTesting = NULL,
  isBillboard = TRUE,
  lineDash = 0,
  lineDashOffset = 0,
  lineWidth = 1,
  opacity = 1,
  pointerEvents = NULL,
  shadowBlur = NULL,
  shadowColor = NULL,
  shadowOffsetX = NULL,
  shadowOffsetY = NULL,
  shadowType = NULL,
  sourcePort = NULL,
  stroke = "#000",
  strokeOpacity = 1,
  targetPort = NULL,
  transform = NULL,
  transformOrigin = NULL,
  visibility = c("visible", "hidden"),
  zIndex = -1e4,
  ...
) {
  # Validate parameters
  cursor <- match.arg(cursor)
  fillRule <- match.arg(fillRule)

  if (!is.null(shadowType) && !(shadowType %in% c("inner", "outer", "both"))) {
    stop("'shadowType' must be one of: 'inner', 'outer', 'both'")
  }

  visibility <- match.arg(visibility)

  if (!is.null(pointerEvents) && !(pointerEvents %in% valid_pointer_events)) {
    stop(
      sprintf("'pointerEvents' must be one of: %s"),
      valid_pointer_events
    )
  }

  if (!is.logical(isBillboard)) {
    stop("'isBillboard' must be a boolean value (TRUE or FALSE)")
  }

  # Get all argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Create Combo Options Configuration for G6 Graphs
#'
#' Configures the general options for combos in a G6 graph. These settings control
#' the type, style, state, palette, and animation of combos.
#'
#' @details
#' Combo options allow defining how combos (node groupings) appear and behave in a G6 graph.
#' This includes selecting combo types, setting styles, configuring state-based appearances,
#' defining color palettes, and specifying animation effects.
#'
#' @param type Combo type. Can be a built-in combo type name or a custom combo name.
#'   Built-in types include "circle", "rect", "polygon", etc. Default: "circle".
#'
#' @param style Combo style configuration. Controls the appearance of combos including color,
#'   size, border, etc.
#'   Default: NULL.
#'
#' @param state Defines the style of the combo in different states, such as hover, selected,
#'   disabled, etc. Should be a list mapping state names to style configurations.
#'   Default: NULL.
#'
#' @param palette Defines the color palette of the combo, used to map colors based on different data.
#'   Default: NULL.
#'
#' @param animation Defines the animation effect of the combo. Can be created with
#'   \code{animation_config()}.
#'   Default: NULL.
#'
#' @return A list containing combo options configuration that can be passed to [g6_options()].
#'
#' @export
#'
#' @examples
#' # Basic combo options with default circle type
#' options <- combo_options()
#'
#' # Rectangle combo with custom style
#' options <- combo_options(
#'   type = "rect",
#'   style = list(
#'     fill = "#F6F6F6",
#'     stroke = "#CCCCCC",
#'     lineWidth = 1
#'   )
#' )
combo_options <- function(
  type = "circle",
  style = NULL,
  state = NULL,
  palette = NULL,
  animation = NULL
) {
  # Validate type parameter
  if (!is.character(type) || length(type) != 1) {
    stop("'type' must be a single string value")
  }

  # Validate style parameter
  if (!is.null(style) && !is.list(style)) {
    stop(
      "'style' must be a list, preferably created with combo_style_options()"
    )
  }

  # Validate state parameter
  if (!is.null(state)) {
    if (!is.list(state)) {
      stop("'state' must be a list mapping state names to style configurations")
    }
    for (state_name in names(state)) {
      if (!is.list(state[[state_name]])) {
        stop(sprintf(
          "State '%s' must be a list of style properties",
          state_name
        ))
      }
    }
  }

  # Validate palette parameter
  if (!is.null(palette) && !is.list(palette)) {
    stop("'palette' must be a list defining color mappings")
  }

  # Validate animation parameter
  if (!is.null(animation) && !is.list(animation)) {
    stop(
      "'animation' must be a list, preferably created with animation_config()"
    )
  }

  # Get all argument names
  arg_names <- names(formals())

  # Create list of argument values
  config <- mget(arg_names)

  # Drop NULL elements
  dropNulls(config)
}

test_that("g6_options creates proper configuration", {
  opts <- g6() |>
    g6_options(
      autoFit = "view",
      fitView = TRUE,
      animate = TRUE
    )

  expect_true(all(c("autoFit", "fitView", "animate") %in% names(opts$x)))
})

test_that("node_options creates proper node configuration", {
  node_opts <- node_options(
    type = "circle",
    style = list(fill = "blue", stroke = "red")
  )

  expect_type(node_opts, "list")
  expect_equal(node_opts$type, "circle")
  expect_equal(node_opts$style$fill, "blue")
  expect_equal(node_opts$style$stroke, "red")
})

test_that("edge_options creates proper edge configuration", {
  edge_opts <- edge_options(
    type = "line",
    style = list(stroke = "green", lineWidth = 2)
  )

  expect_type(edge_opts, "list")
  expect_equal(edge_opts$type, "line")
  expect_equal(edge_opts$style$stroke, "green")
  expect_equal(edge_opts$style$lineWidth, 2)
})

test_that("combo_options creates proper combo configuration", {
  combo_opts <- combo_options(
    type = "rect",
    style = list(fill = "yellow", stroke = "orange")
  )

  expect_type(combo_opts, "list")
  expect_equal(combo_opts$type, "rect")
  expect_equal(combo_opts$style$fill, "yellow")
  expect_equal(combo_opts$style$stroke, "orange")
})

test_that("node_style_options creates proper style configuration", {
  style_opts <- node_style_options(
    fill = "purple",
    stroke = "black",
    lineWidth = 3,
    radius = 10
  )

  expect_type(style_opts, "list")
  expect_equal(style_opts$fill, "purple")
  expect_equal(style_opts$stroke, "black")
  expect_equal(style_opts$lineWidth, 3)
  expect_equal(style_opts$radius, 10)
})

test_that("edge_style_options creates proper style configuration", {
  style_opts <- edge_style_options(
    stroke = "cyan",
    lineWidth = 4,
    lineDash = c(5, 5)
  )

  expect_type(style_opts, "list")
  expect_equal(style_opts$stroke, "cyan")
  expect_equal(style_opts$lineWidth, 4)
  expect_equal(style_opts$lineDash, c(5, 5))
})

test_that("canvas_config creates proper canvas configuration", {
  canvas_opts <- canvas_config(
    background = "white",
    cursor = "pointer"
  )

  expect_type(canvas_opts, "list")
  expect_equal(canvas_opts$background, "white")
  expect_equal(canvas_opts$cursor, "pointer")
})

test_that("animation_config creates proper animation configuration", {
  anim_opts <- animation_config(
    duration = 1000,
    easing = "linear"
  )

  expect_type(anim_opts, "list")
  expect_equal(anim_opts$duration, 1000)
  expect_equal(anim_opts$easing, "linear")
})

test_that("auto_fit_config creates proper autoFit configuration", {
  autofit_opts <- auto_fit_config(
    type = "view",
    duration = 300,
    easing = "ease-out"
  )

  expect_type(autofit_opts, "list")
  expect_equal(autofit_opts$type, "view")
  expect_type(autofit_opts$animation, "list")
  expect_length(autofit_opts$animation, 2)
  expect_equal(autofit_opts$animation$duration, 300)
  expect_equal(autofit_opts$animation$easing, "ease-out")
})

#' Create a G6 Graph from an igraph Object
#'
#' @description
#' Converts an `igraph` graph object into the format required by the `g6()` function
#' and creates an interactive G6 graph visualization. This is a convenience wrapper around
#' [g6()] that allows you to work directly with `igraph` objects.
#'
#' @details
#' This function extracts the node and edge data from an `igraph` object, converts them
#' into the appropriate format for G6, and passes them to [g6()]. Node styling is derived
#' from vertex attributes, and edge styling from edge attributes.
#'
#' If the graph is directed, edges will automatically be rendered with arrows.
#'
#' @inheritParams g6
#' @param graph An object of class [`igraph::igraph`] containing the graph to visualize.
#'
#' @return An htmlwidget object that can be rendered in R Markdown, Shiny apps, or the R console.
#'
#' @seealso [g6()] for more information about node, edge, and combo structure.
#'
#' @examples
#' if (require(igraph)) {
#'   g <- igraph::make_ring(5)
#'   g6_igraph(g)
#' }
#'
#' @export
g6_igraph <- function(
  graph,
  combos = NULL,
  width = "100%",
  height = NULL,
  elementId = NULL
) {
  if (!requireNamespace("igraph", quietly = TRUE)) {
    stop(
      "The package 'igraph' is required for this function to work. Please install it."
    )
  }
  igraph::V(graph)$id <- as.character(seq_len(igraph::vcount(graph)))

  nodes_df <- igraph::as_data_frame(graph, what = "vertices")
  nodes_df <- shape2type(nodes_df)
  nodes <- style_from_df(nodes_df, node_style_options)

  # TODO: find better solution
  names <- igraph::V(graph)$name %||% igraph::V(graph)$id
  igraph::V(graph)$name <- igraph::V(graph)$id
  edges_df <- igraph::as_data_frame(graph, what = "edges")
  names(edges_df)[1:2] <- c("source", "target")
  if (igraph::is_directed(graph)) {
    edges_df$style$endArrow <- TRUE
  }
  edges <- style_from_df(edges_df, edge_style_options)
  igraph::V(graph)$name <- names

  g6(
    nodes = nodes,
    edges = edges,
    combos = combos,
    width = width,
    height = height,
    elementId = elementId
  )
}

#' @keywords internal
reserved_properties <- function() {
  c(
    "id",
    "type",
    "data",
    "style",
    "states",
    "combo",
    "source",
    "target",
    "children"
  )
}

#' @keywords internal
style_from_df <- function(df, style_fn) {
  formal_args <- names(formals(style_fn))
  has_dots <- "..." %in% formal_args
  style_args <- setdiff(formal_args, "...")

  items <- unname(split(df, seq(nrow(df))))
  items <- lapply(items, function(row) {
    item <- as.list(row)

    # TBD: David Schoch: implement mapping between g6R and igraph
    # for attributes names ...
    style_inputs <- item[names(item) %in% style_args]
    extra_inputs <- item[
      !names(item) %in% style_args & !names(item) %in% reserved_properties()
    ]

    if (has_dots) {
      style <- do.call(style_fn, c(style_inputs, extra_inputs))
    } else {
      style <- do.call(style_fn, style_inputs)
    }

    item$style <- style
    # Should not be at the top level
    item[names(item) %in% style_args] <- NULL
    item[names(item) %in% names(extra_inputs)] <- NULL
    return(item)
  })

  items
}

#' @keywords internal
#' @note In igraph type is reserved for bipartite graphs, whereas in g6 it defines the
#' shape of the node
shape2type <- function(nodes_df) {
  if ("shape" %in% names(nodes_df)) {
    if (!"type" %in% names(nodes_df)) {
      nodes_df$type <- nodes_df$shape
    } else {
      nodes_df$type_igraph <- nodes_df$type
      nodes_df$type <- nodes_df$shape
    }
  }
  nodes_df
}

test_that("g6_igraph works with igraph objects", {
  skip_if_not_installed("igraph")

  # Create a simple igraph object
  library(igraph)
  g_igraph <- make_graph(c("A", "B", "B", "C", "C", "A"))

  # Convert to g6
  g6_graph <- g6_igraph(g_igraph)

  expect_s3_class(g6_graph, "htmlwidget")
  expect_s3_class(g6_graph, "g6")
  expect_true("nodes" %in% names(g6_graph$x$data))
  expect_true("edges" %in% names(g6_graph$x$data))

  # Check that nodes have proper structure
  expect_length(g6_graph$x$data$nodes, 3) # A, B, C
  expect_length(g6_graph$x$data$edges, 3) # 3 edges

  # Check node IDs are characters
  node_ids <- sapply(g6_graph$x$data$nodes, function(x) x$id)
  expect_type(node_ids, "character")
})

test_that("g6_igraph handles vertex and edge attributes", {
  skip_if_not_installed("igraph")

  library(igraph)

  # Create igraph with attributes
  g_igraph <- make_graph(c("A", "B", "B", "C"))
  V(g_igraph)$name <- c("Node A", "Node B", "Node C")
  V(g_igraph)$color <- c("red", "blue", "green")
  E(g_igraph)$weight <- c(1.5, 2.0)
  E(g_igraph)$type <- c("solid", "dashed")

  g6_graph <- g6_igraph(g_igraph)

  # Check that attributes are preserved
  expect_true(any(sapply(g6_graph$x$data$nodes, function(x) {
    !is.null(x$style$name)
  })))
  expect_true(any(sapply(g6_graph$x$data$nodes, function(x) {
    !is.null(x$style$color)
  })))
  expect_true(any(sapply(g6_graph$x$data$edges, function(x) {
    !is.null(x$style$weight)
  })))
})

test_that("g6_igraph handles graphs with isolated vertices", {
  skip_if_not_installed("igraph")

  library(igraph)

  # Create graph with isolated vertices
  g_igraph <- make_graph(c("A", "B")) # Only one edge
  g_igraph <- add_vertices(g_igraph, 1, name = "C") # Add isolated vertex

  g6_graph <- g6_igraph(g_igraph)

  expect_length(g6_graph$x$data$nodes, 3) # A, B, C
  expect_length(g6_graph$x$data$edges, 1) # Only A-B edge
})

test_that("g6_igraph errors for non-igraph objects", {
  skip_if_not_installed("igraph")
  expect_snapshot(
    {
      g6_igraph("not an igraph")
      g6_igraph(list())
      g6_igraph(123)
    },
    error = TRUE
  )
})

#' Create a G6 Graph Visualization
#'
#' @description
#' Creates an interactive graph visualization using the G6 graph visualization library.
#' This function is the main entry point for creating G6 graph visualizations in R.
#'
#' @details
#' The `g6` function creates a G6 graph as an htmlwidget that can be used in R Markdown,
#' Shiny applications, or rendered to HTML. It takes graph data in the form of nodes, edges,
#' and optional combo groupings, along with various configuration options for customizing
#' the appearance and behavior of the graph.
#'
#' \subsection{Nodes}{
#' The `nodes` parameter should be a data frame or list of nodes with at least an `id` field
#' for each node. Additional fields can include:
#' \itemize{
#'   \item \code{id} (required): Unique identifier for the node. Must be a character.
#'   \item \code{type}: Node type (e.g., "circle", "rect", "diamond").
#'   \item \code{data}: Custom data associated with the node.
#'   \item \code{style}: List of style attributes (color, size, etc.).
#'   \item \code{states}: String. Initial states for the node, such as selected, active, hover, etc.
#'   \item \code{combo}: ID of the combo this node belongs to.
#'   \item \code{ports}: Node ports. Can also be passed inside the style field.
#' }}
#'
#' \subsection{Edges}{
#' The `edges` parameter should be a data frame or list of edges with at least `source` and
#' `target` fields identifying the connected nodes. Additional fields can include:
#' \itemize{
#'   \item \code{source} (required): ID of the source node. Must be a character.
#'   \item \code{target} (required): ID of the target node. Must be a character.
#'   \item \code{id}: Unique identifier for the edge.
#'   \item \code{type}: Edge type (e.g., "line", "cubic", "arc").
#'   \item \code{data}: Custom data associated with the edge.
#'   \item \code{style}: List of style attributes (color, width, etc.).
#'   \item \code{states}: String. Initial states for the edge.
#' }}
#'
#'
#' \subsection{Combos}{
#' The `combos` parameter is used for grouping nodes and can be a data frame or list with
#' combo definitions. Fields include:
#' \itemize{
#'   \item \code{id} (required): Unique identifier for the combo. Must be a character.
#'   \item \code{type}: String: Combo type. It can be the type of built-in Combo, or the custom Combo.
#'   \item \code{data}: Custom data associated with the combo.
#'   \item \code{style}: List of style attributes.
#'   \item \code{states}: String. Initial states for the combo.
#'   \item \code{combo}: String. Parent combo ID. If there is no parent combo, it is null.
#' }}
#'
#' Nodes are assigned to combos by setting their `combo` field to the ID of the combo.
#'
#' @param nodes A data frame or list of nodes in the graph. Each node should have at least
#'   an "id" field. See 'Data Structure' section for more details.
#'   Default: NULL.
#'
#' @param edges A data frame or list of edges in the graph. Each edge should have "source"
#'   and "target" fields identifying the connected nodes. See 'Data Structure' section
#'   for more details.
#'   Default: NULL.
#'
#' @param combos A data frame or list of combo groups in the graph. Each combo should have
#'   at least an "id" field. Nodes can be assigned to combos using their "combo" field.
#'   See 'Data Structure' section for more details.
#'   Default: NULL.
#'
#' @param jsonUrl An url pointing to a valid JSON containing the graph data in G6 format.
#' See \url{https://assets.antv.antgroup.com/g6/20000.json} for
#' an example. Can't be used at the same time as `nodes`, `edges`, and `combos`.
#' @param iconsUrl A URL pointing to a JavaScript file containing custom icons. Default provides
#' reasonable set of icons from
#' \url{https://at.alicdn.com/t/project/2678727/caef142c-804a-4a2f-a914-ae82666a31ee.html?spm=a313x.7781069.1998910419.35}.
#'
#' @param width Width of the graph container in pixels or as a valid CSS unit.
#'   Default: NULL (automatic sizing).
#'
#' @param height Height of the graph container in pixels or as a valid CSS unit.
#'   Default: NULL (automatic sizing).
#'
#' @param elementId A unique ID for the graph HTML element.
#'   Default: NULL (automatically generated).
#'
#' @return An htmlwidget object that can be printed, included in R Markdown documents,
#' or used in Shiny applications. This widget contains the graph data and configuration
#' necessary to render the G6 graph visualization.
#'
#' @examples
#' # Create a simple graph with two nodes and one edge
#' nodes <- data.frame(
#'   id = c("node1", "node2")
#' )
#'
#' edges <- data.frame(
#'   source = "node1",
#'   target = "node2"
#' )
#'
#' g6(nodes = nodes, edges = edges)
#' @export
g6 <- function(
  nodes = NULL,
  edges = NULL,
  combos = NULL,
  jsonUrl = NULL,
  iconsUrl = "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js",
  width = "100%",
  height = NULL,
  elementId = NULL
) {
  stopifnot(!is.null(iconsUrl))
  if (
    !is.null(jsonUrl) &&
      (!is.null(nodes) || !is.null(edges) || !is.null(combos))
  ) {
    stop(
      "Can't use jsonUrl` argument at the same time as `nodes`, `edges`, and `combos`."
    )
  }

  dat <- NULL
  if (is.null(jsonUrl)) {
    # Convert data frames to lists of records
    dat <- g6_data(nodes, edges, combos)
    validate_edges_ports(dat[["edges"]], dat[["nodes"]])
  }

  # Build properly named list of parameters to pass to widget
  x <- list(
    data = dat,
    jsonUrl = jsonUrl,
    iconsUrl = iconsUrl,
    mode = get_g6_mode(),
    preservePosition = get_g6_preserve_position(),
    directed = get_g6_directed_graph(),
    maxCollapseDepth = get_g6_max_collapse_depth()
  )

  # In case we need it ...
  hookFunc <- function(widget) {
    # TBD
  }

  # create widget
  htmlwidgets::createWidget(
    name = "g6",
    x,
    width = width,
    height = height,
    package = "g6R",
    elementId = elementId,
    preRenderHook = NULL
  )
}

#' @keywords internal
validate_g6_mode <- function(mode) {
  if (!(mode %in% c("dev", "prod"))) {
    stop(sprintf(
      "`g6R.mode` option must be one of 'dev' or 'prod'. Current value: '%s'",
      mode
    ))
  }
  invisible(mode)
}

#' @keywords internal
get_g6_mode <- function() {
  mode <- getOption("g6R.mode", "prod")
  validate_g6_mode(mode)
}

#' @keywords internal
validate_g6_preserve_position <- function(val) {
  if (!is.logical(val) || length(val) != 1) {
    stop(
      "`g6R.preserve_elements_position` option must be a single logical value (TRUE or FALSE)."
    )
  }
  invisible(val)
}

#' @keywords internal
get_g6_preserve_position <- function() {
  val <- getOption("g6R.preserve_elements_position", FALSE)
  validate_g6_preserve_position(val)
}

#' @keywords internal
validate_g6_directed_graph <- function(val) {
  if (!is.logical(val) || length(val) != 1) {
    stop(
      "`g6R.directed_graph` option must be a single logical value (TRUE or FALSE)."
    )
  }
  invisible(val)
}

#' @keywords internal
get_g6_directed_graph <- function() {
  val <- getOption("g6R.directed_graph", FALSE)
  validate_g6_directed_graph(val)
}

#' @keywords internal
set_g6_directed_graph <- function(val) {
  validate_g6_directed_graph(val)
  options("g6R.directed_graph" = val)
  invisible(val)
}

#' @keywords internal
validate_g6_max_collapse_depth <- function(val) {
  if (!is.numeric(val) || length(val) != 1 || val < -1) {
    stop(
      "`g6R.max_collapse_depth` option must be a single number >= -1."
    )
  }
  invisible(val)
}

#' @keywords internal
get_g6_max_collapse_depth <- function() {
  val <- getOption("g6R.max_collapse_depth", Inf)
  validate_g6_max_collapse_depth(val)
}

#' Set max collapse depth
#'
#' Controls which nodes display a collapse button based on their depth in
#' the graph. Only nodes at depth \code{<= maxCollapseDepth} will show
#' collapse buttons. Set to \code{Inf} (the default) to allow all nodes
#' with children to be collapsible. Set to \code{0} to only allow root
#' nodes to collapse. Set to \code{-1} to disable collapsing entirely.
#'
#' @param val A single number >= -1. Use \code{Inf} for no limit,
#' \code{-1} to disable all collapsing.
#'
#' @return Invisibly returns \code{val}.
#' @export
set_g6_max_collapse_depth <- function(val) {
  validate_g6_max_collapse_depth(val)
  options("g6R.max_collapse_depth" = val)
  invisible(val)
}

#' Shiny bindings for g6
#'
#' Output and render functions for using g6 within Shiny
#' applications and interactive Rmd documents.
#'
#' @param outputId output variable to read from
#' @param width,height Must be a valid CSS unit (like \code{'100\%'},
#'   \code{'400px'}, \code{'auto'}) or a number, which will be coerced to a
#'   string and have \code{'px'} appended.
#' @param expr An expression that generates a g6
#' @param env The environment in which to evaluate \code{expr}.
#' @param quoted Is \code{expr} a quoted expression (with \code{quote()})? This
#'   is useful if you want to save an expression in a variable.
#'
#' @name g6-shiny
#'
#' @return `g6Output` and `g6_output`
#' return a Shiny output function that can be used in the UI part of a Shiny app.
#' `renderG6` and `render_g6` return a
#' Shiny render function that can be used in the server part of a Shiny app to
#' render a `g6` element.
#'
#' @export
g6Output <- function(outputId, width = "100%", height = "400px") {
  #nocov start
  htmlwidgets::shinyWidgetOutput(
    outputId,
    "g6",
    width,
    height,
    package = "g6R"
  ) #nocov end
}

#' Alias to \link{g6Output}
#' @export
#' @rdname g6-shiny
g6_output <- g6Output

#' @rdname g6-shiny
#' @export
renderG6 <- function(expr, env = parent.frame(), quoted = FALSE) {
  #nocov start
  if (!quoted) {
    expr <- substitute(expr)
  } # force quoted
  htmlwidgets::shinyRenderWidget(expr, g6Output, env, quoted = TRUE)
} #nocov end

#' Alias to \link{renderG6}
#' @export
#' @rdname g6-shiny
render_g6 <- renderG6

test_that("g6 function creates proper htmlwidget", {
  # Test with minimal input
  g <- g6()
  expect_s3_class(g, "htmlwidget")
  expect_s3_class(g, "g6")
  expect_length(g$x$data, 0)

  expect_snapshot(error = TRUE, {
    # Test with invalid input
    g6(iconsUrl = NULL)
    g6(
      nodes = data.frame(),
      jsonUrl = "https://assets.antv.antgroup.com/g6/cluster.json"
    )
  })

  # Test with data frames
  nodes_df <- data.frame(
    id = c("node1", "node2", "node3"),
    stringsAsFactors = FALSE
  )

  edges_df <- data.frame(
    source = c("node1", "node2"),
    target = c("node2", "node3"),
    stringsAsFactors = FALSE
  )

  g <- g6(nodes = nodes_df, edges = edges_df)
  expect_s3_class(g, "htmlwidget")
  expect_length(g$x$data$nodes, 3)
  expect_length(g$x$data$edges, 2)

  # Check node structure
  expect_equal(g$x$data$nodes[[1]]$id, "node1")

  # Check edge structure
  expect_equal(g$x$data$edges[[1]]$source, "node1")
  expect_equal(g$x$data$edges[[1]]$target, "node2")
})

test_that("g6 function handles list input", {
  # Test with lists
  nodes_list <- list(
    list(id = "a"),
    list(id = "b")
  )

  edges_list <- list(
    list(source = "a", target = "b", type = "line")
  )

  g <- g6(nodes = nodes_list, edges = edges_list)
  expect_s3_class(g, "htmlwidget")
  expect_length(g$x$data$nodes, 2)
  expect_length(g$x$data$edges, 1)
  expect_equal(g$x$data$edges[[1]]$type, "line")
})

test_that("g6 function handles combos", {
  nodes_df <- data.frame(
    id = c("node1", "node2"),
    combo = c("combo1", "combo1"),
    stringsAsFactors = FALSE
  )

  combos_df <- data.frame(
    id = "combo1",
    stringsAsFactors = FALSE
  )

  g <- g6(nodes = nodes_df, combos = combos_df)
  expect_s3_class(g, "htmlwidget")
  expect_length(g$x$data$nodes, 2)
  expect_length(g$x$data$combos, 1)
  expect_equal(g$x$data$combos[[1]]$id, "combo1")
  expect_equal(g$x$data$nodes[[1]]$combo, "combo1")
})

test_that("g6 function handles dimensions", {
  g1 <- g6(width = "500px", height = "300px")
  expect_equal(g1$width, "500px")
  expect_equal(g1$height, "300px")

  g2 <- g6(width = 500, height = 300)
  expect_equal(g2$width, 500)
  expect_equal(g2$height, 300)
})

test_that("g6 function handles elementId", {
  g <- g6(elementId = "my-graph")
  expect_equal(g$elementId, "my-graph")
})

test_that("get_g6_max_collapse_depth returns Inf by default", {
  withr::local_options(g6R.max_collapse_depth = NULL)
  expect_equal(get_g6_max_collapse_depth(), Inf)
})

test_that("set_g6_max_collapse_depth sets and retrieves correctly", {
  withr::local_options(g6R.max_collapse_depth = NULL)
  set_g6_max_collapse_depth(0)
  expect_equal(get_g6_max_collapse_depth(), 0)
  set_g6_max_collapse_depth(3)
  expect_equal(get_g6_max_collapse_depth(), 3)
  set_g6_max_collapse_depth(Inf)
  expect_equal(get_g6_max_collapse_depth(), Inf)
})

test_that("set_g6_max_collapse_depth accepts -1 to disable collapsing", {
  withr::local_options(g6R.max_collapse_depth = NULL)
  set_g6_max_collapse_depth(-1)
  expect_equal(get_g6_max_collapse_depth(), -1)
})

test_that("set_g6_max_collapse_depth rejects invalid values", {
  expect_error(set_g6_max_collapse_depth(-2), ">= -1")
  expect_error(set_g6_max_collapse_depth("a"), ">= -1")
  expect_error(set_g6_max_collapse_depth(c(1, 2)), ">= -1")
})

test_that("g6 widget includes maxCollapseDepth in x", {
  withr::local_options(g6R.max_collapse_depth = NULL)
  g <- g6()
  expect_equal(g$x$maxCollapseDepth, Inf)

  withr::local_options(g6R.max_collapse_depth = 0)
  g2 <- g6()
  expect_equal(g2$x$maxCollapseDepth, 0)
})

#' Check if an object is a g6_port
#'
#' @param x An object to check.
#' @return Logical indicating if \code{x} is of class \code{g6_port}.
#' @examples
#' p <- g6_port("input-1", type = "input")
#' is_g6_port(p)
#' is_g6_port(list(key = "input-1", type = "input"))
#' is_g6_port(as_g6_port(list(key = "input-1", type = "input")))
#' @export
#' @rdname is_g6_port
is_g6_port <- function(x) {
  inherits(x, "g6_port")
}

#' @export
#' @rdname is_g6_port
#' @examples
#' is_g6_ports(g6_ports(
#'   g6_port("input-1", type = "input"),
#'   g6_port("output-1", type = "output")
#' ))
is_g6_ports <- function(x) {
  inherits(x, "g6_ports")
}

#' Create a G6 Port
#'
#' @param key Character. Unique identifier for the port (required).
#' @param label Character. Label text for the port (optional).
#' @param type Character. Either "input" or "output" (required).
#' @param arity Numeric. Maximum number of
#' connections this port can accept (default: 1).
#' Use 0, Inf, or any non-negative integer.
#' @param visibility Character. Controls port visibility behavior:
#' \itemize{
#'   \item \code{"visible"}: Ports are always shown (default).
#'   \item \code{"hover"}: Ports appear only when hovering over the node.
#'   \item \code{"hidden"}: Ports are never visible.
#' }
#' @param ... Additional port style parameters, e.g. \code{placement} and
#' \code{r} (radius, default 6). See
#' \url{https://g6.antv.antgroup.com/en/manual/element/node/base-node#portstyleprops}.
#' In addition to G6's native placements (\code{"left"}, \code{"right"},
#' \code{"top"}, \code{"bottom"}, or a \code{c(x, y)} pair), g6R adds
#' \code{placement = "label-bottom"}: the port snaps to the bottom-centre of the
#' node's label background when the node has a bottom label, falling back to a
#' normal bottom-of-node port otherwise. Pass \code{ripple = FALSE} to disable
#' the hover ripple for a port, or \code{haloFill} to override the colour of the
#' ring that makes the port look set into the node surface.
#' @return An S3 object of class 'g6_port'.
#' @examples
#' g6_port("input-1", label = "port 1", type = "input", arity = 2, placement = "left")
#' g6_port("output-1", label = "port 2", type = "output", placement = "right")
#' g6_port("input-2", type = "input", visibility = "hover")
#' @export
g6_port <- function(
  key,
  label = key,
  type = c("input", "output"),
  arity = 1,
  visibility = c("visible", "hover", "hidden"),
  ...
) {
  type <- match.arg(type)
  visibility <- match.arg(visibility)
  port <- structure(
    c(
      list(
        key = key,
        type = type,
        arity = arity,
        visibility = visibility
      ),
      if (!is.null(label)) list(label = label),
      list(...)
    ),
    class = "g6_port"
  )
  validate_port(port)
}

#' Create a single-connection input port
#' @export
#' @param fill Character. Color of the port
#' (default: "#52C41A" for input ports,
#' "#FF4D4F" for output ports).
#' @rdname g6_port
#' @note To create an (input/output) port with multiple connections,
#' simply set the arity to \code{Inf} or any positive integer.
g6_input_port <- function(
  key,
  label = key,
  arity = 1,
  visibility = c("visible", "hover", "hidden"),
  fill = "#52C41A",
  ...
) {
  visibility <- match.arg(visibility)
  g6_port(
    key = key,
    label = label,
    type = "input",
    arity = arity,
    visibility = visibility,
    fill = fill,
    ...
  )
}

#' Create an output port (single connection by default)
#' @export
#' @rdname g6_port
g6_output_port <- function(
  key,
  label = key,
  arity = 1,
  visibility = c("visible", "hover", "hidden"),
  fill = "#FF4D4F",
  ...
) {
  visibility <- match.arg(visibility)
  g6_port(
    key = key,
    label = label,
    type = "output",
    arity = arity,
    visibility = visibility,
    fill = fill,
    ...
  )
}

#' Validate a single G6 port
#'
#' @param x An object of class 'g6_port'.
#' @param ... Generic consistency.
#' @return The validated port (invisibly).
#' @examples
#' validate_port(g6_port("input-1", type = "input"))
#' @export
validate_port <- function(x, ...) {
  UseMethod("validate_port")
}

#' @rdname validate_port
#' @export
validate_port.g6_port <- function(x, ...) {
  if (
    !is.character(x$key) ||
      length(x$key) != 1 ||
      is.na(x$key) ||
      x$key == ""
  ) {
    stop("'key' must be a non-empty character string.")
  }

  # Avoid issues with JSON conversion of Inf
  if (is.infinite(x$arity) || identical(x$arity, "Infinity")) {
    x$arity <- "Infinity"
  } else {
    if (
      !is.numeric(x$arity) ||
        length(x$arity) != 1 ||
        is.na(x$arity) ||
        x$arity < 0
    ) {
      stop(
        "'arity' must be a single non-negative number (0, Inf, or positive integer)."
      )
    }
  }

  # Ensure ports are displayed: doc says
  # If set to undefined, the port is treated as a point,
  # not displayed on canvas. Default is set to 6.
  if (is.null(x[["r"]])) {
    x[["r"]] <- 6
  } else {
    if (
      !is.numeric(x[["r"]]) ||
        length(x[["r"]]) != 1 ||
        is.na(x[["r"]]) ||
        x[["r"]] <= 0
    ) {
      stop("'r' (radius) must be a single positive number.")
    }
  }

  # Ensure ports are on the node border (avoids 0.5/0.5)
  placement <- x$placement
  if (is.numeric(placement) && length(placement) == 2) {
    # Must be on an edge: at least one value is exactly 0 or 1
    if (!any(placement == 0 | placement == 1)) {
      stop(
        "Invalid port placement: when placement is a numeric vector, at least one value must be 0 or 1 (i.e., on the node edge). ",
        "You supplied: c(",
        paste(placement, collapse = ", "),
        ")"
      )
    }
  }
  x
}

#' Create a List of G6 Ports
#'
#' @param ... One or more g6_port objects.
#' @return An S3 object of class 'g6_ports'.
#' @examples
#' g6_ports(
#'   g6_port("input-1", type = "input", placement = "left"),
#'   g6_port("output-1", type = "output", placement = "right")
#' )
#' @export
g6_ports <- function(...) {
  ports <- list(...)
  validate_ports(ports)
  structure(ports, class = "g6_ports")
}

#' Validate a list of G6 ports
#'
#' @param x A list of g6_port objects.
#' @param ... Generic consistency.
#' @return The validated list (invisibly).
#' @examples
#' validate_ports(list(
#'   g6_port("input-1", type = "input"),
#'   g6_port("output-1", type = "output")
#' ))
#' @export
validate_ports <- function(x, ...) {
  UseMethod("validate_ports")
}

#' @rdname validate_ports
#' @export
validate_ports.list <- function(x, ...) {
  if (length(x) == 0) {
    stop("No ports provided.")
  }
  if (!all(lgl_ply(x, is_g6_port))) {
    stop("All elements must be of class 'g6_port'.")
  }
  keys <- vapply(x, function(p) p$key, character(1))
  if (anyDuplicated(keys)) {
    stop(
      "All port 'key' values must be unique. Duplicates: ",
      paste(unique(keys[duplicated(keys)]), collapse = ", ")
    )
  }
  invisible(x)
}

#' @rdname validate_ports
#' @export
validate_ports.g6_ports <- function(x, ...) {
  validate_ports.list(x, ...)
}

#' Coerce to a g6_port object
#'
#' @param x An object to coerce.
#' @param ... Additional arguments (unused).
#' @return An object of class \code{g6_port}.
#' @examples
#' as_g6_port(list(key = "input-1", type = "input", placement = "left"))
#' as_g6_port(g6_port("input-1", type = "input"))
#' @export
as_g6_port <- function(x, ...) {
  UseMethod("as_g6_port")
}

#' @rdname as_g6_port
#' @export
as_g6_port.g6_port <- function(x, ...) {
  x
}

#' @rdname as_g6_port
#' @export
as_g6_port.list <- function(x, ...) {
  port_ctor <- g6_port
  if (length(x[["type"]])) {
    port_ctor <- switch(
      x[["type"]],
      "input" = g6_input_port,
      "output" = g6_output_port
    )
    x[["type"]] <- NULL
  }
  do.call(port_ctor, x)
}

#' Coerce to a list of g6_port objects (g6_ports)
#'
#' @param x An object to coerce.
#' @param ... Additional arguments (unused).
#' @return An object of class \code{g6_ports}.
#' @examples
#' as_g6_ports(list(
#'   list(key = "input-1", type = "input", placement = "left"),
#'   list(key = "output-1", type = "output", placement = "right")
#' ))
#' as_g6_ports(g6_ports(
#'   g6_port("input-1", type = "input"),
#'   g6_port("output-1", type = "output")
#' ))
#' @export
as_g6_ports <- function(x, ...) {
  UseMethod("as_g6_ports")
}

#' @rdname as_g6_ports
#' @export
as_g6_ports.g6_ports <- function(x, ...) {
  x
}

#' @rdname as_g6_ports
#' @export
as_g6_ports.list <- function(x, ...) {
  ports <- lapply(x, as_g6_port)
  do.call(g6_ports, ports)
}

extract_port_field <- function(nodes, field) {
  unlist(
    lapply(nodes, function(node) {
      ports <- node$style$ports
      if (is.null(ports) || length(ports) == 0) {
        return(NULL)
      }
      vapply(ports, function(port) port[[field]], character(1))
    }),
    use.names = FALSE
  )
}

#' Validate that all edges connect input/output ports
#'
#' @param nodes List of g6_node objects.
#' @param edges List of g6_edge objects.
#' @return Invisibly TRUE if all edges are valid, otherwise stops with an error.
#' @export
validate_edges_ports <- function(edges, nodes) {
  keys <- extract_port_field(nodes, "key")
  types <- extract_port_field(nodes, "type")
  port_types <- setNames(types, keys)

  if (length(port_types) == 0) {
    return(invisible(TRUE))
  }

  source_ports <- vapply(
    edges,
    function(e) e$style$sourcePort,
    character(1)
  )
  target_ports <- vapply(
    edges,
    function(e) e$style$targetPort,
    character(1)
  )

  if (length(source_ports) == 0 || length(target_ports) == 0) {
    return(invisible(TRUE))
  }

  type1 <- port_types[source_ports]
  type2 <- port_types[target_ports]

  if (anyNA(type1) || anyNA(type2)) {
    stop("Edge refers to unknown port key.")
  }

  same_type <- type1 == type2
  if (any(same_type)) {
    idx <- which(same_type)[1]
    stop(
      sprintf(
        "Invalid edge: cannot connect two '%s' ports ('%s' <-> '%s').",
        type1[idx],
        source_ports[idx],
        target_ports[idx]
      )
    )
  }

  invisible(TRUE)
}

library(testthat)
library(g6R)

test_that("g6_port constructs and validates a port", {
  p <- g6_port("input-1", type = "input", placement = "left")
  expect_s3_class(p, "g6_port")
  expect_equal(p$key, "input-1")
  expect_equal(p$type, "input")
})

test_that("g6_ports constructs and validates a list of ports", {
  ports <- g6_ports(
    g6_port("input-1", type = "input"),
    g6_port("output-1", type = "output")
  )
  expect_s3_class(ports, "g6_ports")
  expect_length(ports, 2)
  expect_equal(ports[[2]]$key, "output-1")
})

test_that("g6_port validation fails for missing or invalid key/type/multiple", {
  expect_snapshot(error = TRUE, {
    g6_port(key = "", type = "input")
  })
  expect_snapshot(error = TRUE, {
    g6_port(key = "x", type = "foo")
  })
  expect_snapshot(error = TRUE, {
    g6_port(key = "x", type = "input", arity = -1)
  })
})

test_that("validate_port.g6_port handles 'r' (radius) correctly", {
  # Default: r is NULL, should be set to 6
  p <- structure(list(key = "x", type = "input", arity = 1), class = "g6_port")
  res <- validate_port.g6_port(p)
  expect_equal(res$r, 6)

  # Valid: r is a positive number
  p <- structure(
    list(key = "x", type = "input", arity = 1, r = 10),
    class = "g6_port"
  )
  res <- validate_port.g6_port(p)
  expect_equal(res$r, 10)

  # Invalid: r is negative
  p <- structure(
    list(key = "x", type = "input", arity = 1, r = -1),
    class = "g6_port"
  )
  expect_error(
    validate_port.g6_port(p),
    "'r' \\(radius\\) must be a single positive number."
  )

  # Invalid: r is NA
  p <- structure(
    list(key = "x", type = "input", arity = 1, r = NA_real_),
    class = "g6_port"
  )
  expect_error(
    validate_port.g6_port(p),
    "'r' \\(radius\\) must be a single positive number."
  )

  # Invalid: r is not numeric
  p <- structure(
    list(key = "x", type = "input", arity = 1, r = "foo"),
    class = "g6_port"
  )
  expect_error(
    validate_port.g6_port(p),
    "'r' \\(radius\\) must be a single positive number."
  )
})

test_that("g6_ports fails if keys are not unique or elements are not g6_port", {
  expect_snapshot(error = TRUE, {
    g6_ports(
      g6_port("a", type = "input"),
      g6_port("a", type = "output")
    )
  })
  expect_snapshot(error = TRUE, {
    g6_ports(
      g6_port("a", type = "input"),
      list(key = "b", type = "output")
    )
  })
})

test_that("is_g6_port and coercion functions work", {
  p <- g6_port("input-1", type = "input")
  expect_true(is_g6_port(p))
  expect_identical(as_g6_port(p), p)
  expect_s3_class(as_g6_port(list(key = "x", type = "input")), "g6_port")

  ports <- as_g6_ports(list(
    list(key = "a", type = "input"),
    list(key = "b", type = "output")
  ))
  expect_s3_class(ports, "g6_ports")
  expect_true(all(lgl_ply(ports, is_g6_port)))
})

test_that("is_g6_ports works", {
  ports <- g6_ports(
    g6_port("a", type = "input"),
    g6_port("b", type = "output")
  )
  expect_true(is_g6_ports(ports))
})

test_that("g6_input_port and g6_output_port construct ports", {
  inp <- g6_input_port("in1")
  out <- g6_output_port("out1")
  expect_s3_class(inp, "g6_port")
  expect_equal(inp$type, "input")
  expect_s3_class(out, "g6_port")
  expect_equal(out$type, "output")
})

test_that("validate_edges_ports allows only input-output connections", {
  nodes <- list(
    list(
      style = list(
        ports = list(
          list(key = "input-1", type = "input"),
          list(key = "output-1", type = "output")
        )
      )
    ),
    list(
      style = list(
        ports = list(
          list(key = "input-2", type = "input"),
          list(key = "output-2", type = "output")
        )
      )
    )
  )

  # Valid: input -> output
  edges_valid <- list(
    list(style = list(sourcePort = "output-1", targetPort = "input-2"))
  )
  expect_invisible(validate_edges_ports(edges_valid, nodes))

  # Invalid: input -> input
  edges_invalid <- list(
    list(style = list(sourcePort = "input-1", targetPort = "input-2"))
  )
  expect_error(
    validate_edges_ports(edges_invalid, nodes),
    "cannot connect two 'input' ports"
  )

  # Invalid: output -> output
  edges_invalid2 <- list(
    list(style = list(sourcePort = "output-1", targetPort = "output-2"))
  )
  expect_error(
    validate_edges_ports(edges_invalid2, nodes),
    "cannot connect two 'output' ports"
  )

  # Invalid: unknown port key
  edges_unknown <- list(
    list(style = list(sourcePort = "foo", targetPort = "input-2"))
  )
  expect_error(validate_edges_ports(edges_unknown, nodes), "unknown port key")
})

test_that("numeric placement must be on node edge", {
  expect_error(
    g6_input_port(key = "bad", placement = c(0.5, 0.5)),
    "at least one value must be 0 or 1"
  )
})

#' Create a List of G6 Plugins
#'
#' Combines multiple G6 plugins into a list that can be passed to a G6 graph configuration.
#' G6 plugins extend the functionality of the base graph visualization with additional features.
#'
#' @details
#' G6 plugins provide extended functionality beyond the core graph visualization capabilities.
#' Plugins are divided into several categories:
#'
#' \subsection{Visual Style Enhancement}{
#'   \itemize{
#'     \item \strong{Grid Line (grid-line):} Displays grid reference lines on the canvas
#'     \item \strong{Background (background):} Adds background images or colors to the canvas
#'     \item \strong{Watermark (watermark):} Adds a watermark to the canvas to protect copyright
#'     \item \strong{Hull (hull):} Creates an outline for a specified set of nodes
#'     \item \strong{Bubble Sets (bubble-sets):} Creates smooth bubble-like element outlines
#'     \item \strong{Snapline (snapline):} Displays alignment reference lines when dragging elements
#'   }
#' }
#'
#' \subsection{Navigation and Overview}{
#'   \itemize{
#'     \item \strong{Minimap (minimap):} Displays a thumbnail preview of the graph, supporting navigation
#'     \item \strong{Fullscreen (fullscreen):} Supports full-screen display and exit for charts
#'     \item \strong{Timebar (timebar):} Provides filtering and playback control for temporal data
#'   }
#' }
#'
#' \subsection{Interactive Controls}{
#'   \itemize{
#'     \item \strong{Toolbar (toolbar):} Provides a collection of common operation buttons
#'     \item \strong{Context Menu (contextmenu):} Displays a menu of selectable operations on right-click
#'     \item \strong{Tooltip (tooltip):} Displays detailed information about elements on hover
#'     \item \strong{Legend (legend):} Displays categories and corresponding style descriptions of chart data
#'   }
#' }
#'
#' \subsection{Data Exploration}{
#'   \itemize{
#'     \item \strong{Fisheye (fisheye):} Provides a focus + context exploration experience
#'     \item \strong{Edge Filter Lens (edge-filter-lens):} Filters and displays edges within a specified area
#'     \item \strong{Edge Bundling (edge-bundling):} Bundles edges with similar paths together to reduce visual clutter
#'   }
#' }
#'
#' \subsection{Advanced Features}{
#'   \itemize{
#'     \item \strong{History (history):} Supports undo/redo operations
#'     \item \strong{Camera Setting (camera-setting):} Configures camera parameters in a 3D scene
#'   }
#' }
#'
#' @param graph G6 graph instance.
#' @param ... G6 plugin configuration objects created with plugin-specific functions
#' @note You can also build your own plugins as described at
#' \url{https://g6.antv.antgroup.com/en/manual/plugin/custom-plugin}.
#'
#' @return A list of G6 plugin configurations that can be passed to a G6 graph.
#'
#' @export
#'
#' @examples
#' # Create a configuration with multiple plugins
#' plugins <- g6_plugins(
#'   g6(),
#'   minimap(),
#'   grid_line(),
#'   tooltips(
#'     getContent = JS("(e, items) => {
#'       return `<div>${items[0].id}</div>`;
#'     }")
#'   )
#' )
#'
#' # Add a context menu and toolbar
#' plugins <- g6_plugins(
#'   g6(),
#'   context_menu(
#'     key = "my-context-menu",
#'     className = "my-context-menu",
#'     trigger = "click",
#'     offset = c(10, 10),
#'     getItems = JS("(event) => {
#'       const type = event.itemType;
#'       const isNode = type === 'node';
#'       return [
#'         { key: 'delete', text: 'Delete' },
#'         { key: 'edit', text: 'Edit' },
#'         { key: 'details', text: 'View Details', disabled: !isNode }
#'       ];
#'     }"),
#'     onClick = JS("(value, target, current) => {
#'       if (value === 'delete') {
#'         // do stuff
#'     }")
#'   ),
#'   g6R::toolbar(
#'     position = "top-right",
#'     getItems = JS("() => [
#'       { id: 'zoom-in', value: 'zoom-in' },
#'       { id: 'zoom-out', value: 'zoom-out' },
#'       { id: 'fit', value: 'fit' }
#'     ]"),
#'     onClick = JS("(value) => {
#'       if (value === 'zoom-in') graph.zoomTo(1.1);
#'       else if (value === 'zoom-out') graph.zoomTo(0.9);
#'       else if (value === 'fit') graph.fitView();
#'     }")
#'   )
#' )
g6_plugins <- function(graph, ...) {
  plugins <- list(...)
  if (length(plugins)) {
    graph$x$plugins <- lapply(plugins, validate_plugin)
  } else {
    stop("You must provide at least one plugin configuration.")
  }
  graph
}

#' @keywords internal
validate_plugin <- function(x) {
  validate_component(x, "plugin")
}

#' Configure Background Plugin for G6
#'
#' Creates a configuration object for the background plugin in G6.
#' This plugin adds a customizable background to the graph canvas.
#'
#' @param key Unique identifier for updates (string, default: NULL).
#' @param width Background width (string, default: "100%").
#' @param height Background height (string, default: "100%").
#' @param backgroundColor Background color (string, default: NULL).
#' @param backgroundImage Background image URL (string, default: NULL).
#' @param backgroundSize Background size (string, default: "cover").
#' @param backgroundPosition Background position (string, default: NULL).
#' @param backgroundRepeat Background repeat (string, default: NULL).
#' @param opacity Background opacity (string, default: NULL).
#' @param transition Transition animation (string, default: "background 0.5s").
#' @param zIndex Stacking order (string, default: "-1").
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/background}.
#'
#' @return A list with the configuration settings for the background plugin.
#' @export
#'
#' @examples
#' # Basic background color
#' bg <- background(backgroundColor = "#f0f0f0")
#'
#' # Background with image
#' bg <- background(
#'   backgroundImage = "https://example.com/background.jpg",
#'   backgroundSize = "contain",
#'   backgroundRepeat = "no-repeat",
#'   backgroundPosition = "center"
#' )
#'
#' # Semi-transparent background with transition
#' bg <- background(
#'   backgroundColor = "#000000",
#'   opacity = "0.3",
#'   transition = "all 1s ease-in-out"
#' )
background <- function(
  key = NULL,
  width = "100%",
  height = "100%",
  backgroundColor = NULL,
  backgroundImage = NULL,
  backgroundSize = "cover",
  backgroundPosition = NULL,
  backgroundRepeat = NULL,
  opacity = NULL,
  transition = "background 0.5s",
  zIndex = "-1",
  ...
) {
  # Create the configuration list with required type
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "background"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Bubble Sets Plugin for G6
#'
#' Creates a configuration object for the bubble-sets plugin in G6.
#' This plugin creates bubble-like contours around groups of specified elements.
#'
#' @param members Member elements, including nodes and edges (character/numeric vector, required).
#' @param key Unique identifier for updates (string, default: NULL).
#' @param avoidMembers Elements to avoid when drawing contours (character/numeric vector, default: NULL).
#' @param label Whether to display labels (boolean, default: TRUE).
#' @param labelPlacement Label position (string, default: "bottom").
#' @param labelBackground Whether to display background (boolean, default: FALSE).
#' @param labelPadding Label padding (numeric or numeric vector, default: 0).
#' @param labelCloseToPath Whether the label is close to the contour (boolean, default: TRUE).
#' @param labelAutoRotate Whether the label rotates with the contour (boolean, default: TRUE).
#' @param labelOffsetX Label x-axis offset (numeric, default: 0).
#' @param labelOffsetY Label y-axis offset (numeric, default: 0).
#' @param labelMaxWidth Maximum width of the text (numeric, default: NULL).
#' @param maxRoutingIterations Maximum iterations for path calculation (numeric, default: 100).
#' @param maxMarchingIterations Maximum iterations for contour calculation (numeric, default: 20).
#' @param pixelGroup Number of pixels per potential area group (numeric, default: 4).
#' @param edgeR0 Edge radius parameter R0 (numeric, default: NULL).
#' @param edgeR1 Edge radius parameter R1 (numeric, default: NULL).
#' @param nodeR0 Node radius parameter R0 (numeric, default: NULL).
#' @param nodeR1 Node radius parameter R1 (numeric, default: NULL).
#' @param morphBuffer Morph buffer size (numeric, default: NULL).
#' @param threshold Threshold (numeric, default: NULL).
#' @param memberInfluenceFactor Member influence factor (numeric, default: NULL).
#' @param edgeInfluenceFactor Edge influence factor (numeric, default: NULL).
#' @param nonMemberInfluenceFactor Non-member influence factor (numeric, default: NULL).
#' @param virtualEdges Whether to use virtual edges (boolean, default: NULL).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/bubble-sets}.
#'
#' @return A list with the configuration settings for the bubble-sets plugin.
#' @export
#'
#' @examples
#' # Basic bubble set around specific nodes
#' bubble <- bubble_sets(
#'   members = c("node1", "node2", "node3"),
#'   label = TRUE
#' )
#'
#' # More customized bubble set
#' bubble <- bubble_sets(
#'   key = "team-a",
#'   members = c("node1", "node2", "node3", "edge1", "edge2"),
#'   avoidMembers = c("node4", "node5"),
#'   labelPlacement = "top",
#'   labelBackground = TRUE,
#'   labelPadding = c(4, 2),
#'   maxRoutingIterations = 150
#' )
#'
#' # Bubble set with advanced parameters
#' bubble <- bubble_sets(
#'   members = c("node1", "node2", "node3"),
#'   pixelGroup = 6,
#'   edgeR0 = 10,
#'   nodeR0 = 5,
#'   memberInfluenceFactor = 0.8,
#'   edgeInfluenceFactor = 0.5,
#'   nonMemberInfluenceFactor = 0.3,
#'   virtualEdges = TRUE
#' )
bubble_sets <- function(
  members,
  key = "bubble-sets",
  avoidMembers = NULL,
  label = TRUE,
  labelPlacement = c("bottom", "left", "right", "top", "center"),
  labelBackground = FALSE,
  labelPadding = 0,
  labelCloseToPath = TRUE,
  labelAutoRotate = TRUE,
  labelOffsetX = 0,
  labelOffsetY = 0,
  labelMaxWidth = NULL,
  maxRoutingIterations = 100,
  maxMarchingIterations = 20,
  pixelGroup = 4,
  edgeR0 = NULL,
  edgeR1 = NULL,
  nodeR0 = NULL,
  nodeR1 = NULL,
  morphBuffer = NULL,
  threshold = NULL,
  memberInfluenceFactor = NULL,
  edgeInfluenceFactor = NULL,
  nonMemberInfluenceFactor = NULL,
  virtualEdges = NULL,
  ...
) {
  if (label) {
    labelPlacement <- match.arg(labelPlacement)
  }
  # Check for required parameters
  if (!is.character(members) && !is.numeric(members) || length(members) == 0) {
    stop("'members' must be a non-empty character/numeric vector")
  }

  members <- as.character(members)

  if (
    !is.null(avoidMembers) &&
      (!is.character(avoidMembers) &&
        !is.numeric(avoidMembers) ||
        length(avoidMembers) == 0)
  ) {
    stop("'avoidMembers' must be a character/numeric vector or NULL")
  }

  if (!is.null(avoidMembers)) {
    avoidMembers <- as.character(avoidMembers)
  }

  # Get argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "bubble-sets"

  # Overlay plugins (bubble sets, hull) are visual-only. Their Contour
  # shape is appended to the canvas after nodes (on AFTER_RENDER), so in
  # SVG mode it sits on top in the DOM. We fix two issues:
  # - pointerEvents: "none" lets clicks/drags pass through to nodes
  # - zIndex: -1 renders the overlay behind nodes so it doesn't
  #   visually cover collapse buttons or other node UI
  dots <- list(...)
  if (is.null(dots$pointerEvents)) {
    dots$pointerEvents <- "none"
  }
  if (is.null(dots$zIndex)) {
    dots$zIndex <- -1L
  }

  # Drop NULL elements
  dropNulls(c(config, dots))
}

#' Configure Context Menu Behavior
#'
#' Creates a configuration object for the context-menu behavior in G6.
#' This allows users to display a context menu when right-clicking or clicking on graph elements.
#'
#' @param key Unique identifier for the behavior, used for
#' subsequent operations (string, default: "context-menu").
#' @param className Additional class name for the menu DOM (string, default: "g6-contextmenu").
#' @param trigger How to trigger the menu: "contextmenu" for right-click,
#' "click" for click (string, default: "contextmenu").
#' @param offset Offset of the menu display in X and Y directions (numeric vector, default: c(4, 4)).
#' @param onClick Callback method triggered after menu item is clicked (JS function). Our default allows
#' to create edge or either remove the current node.
#' @param getItems Returns the list of menu items, supports Promise (JS function, default: NULL).
#' @param getContent Returns the content of the menu, supports Promise (JS function, default: NULL).
#' @param loadingContent Menu content used when getContent returns a
#' Promise (string or HTML element, default: NULL).
#' @param enable Whether the context menu is available (boolean or JS function, default: TRUE).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/contextmenu}.
#'
#' @return A list with the configuration settings for the context menu plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- context_menu()
#'
#' # Custom configuration with JavaScript functions
#' config <- context_menu(
#'   key = "my-context-menu",
#'   className = "my-context-menu",
#'   trigger = "click",
#'   offset = c(10, 10),
#'   getItems = JS("(event) => {
#'     const type = event.itemType;
#'     const isNode = type === 'node';
#'     return [
#'       { key: 'delete', text: 'Delete' },
#'       { key: 'edit', text: 'Edit' },
#'       { key: 'details', text: 'View Details', disabled: !isNode }
#'     ];
#'   }"),
#'   onClick = JS("(value, target, current) => {
#'     if (value === 'delete') {
#'       // do stuff
#'   }")
#' )
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   nodes <- data.frame(id = c("node1", "node2"))
#'   edges <- data.frame(source = "node1", target = "node2")
#'
#'   ui <- page_fluid(
#'     g6_output("graph"),
#'     verbatimTextOutput("contextmenu_info")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(
#'         nodes = nodes,
#'         edges = edges
#'       ) |>
#'         g6_layout() |>
#'         g6_plugins(
#'           context_menu()
#'         )
#'     })
#'
#'     output$contextmenu_info <- renderPrint({
#'       input[["graph-contextmenu"]]
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
context_menu <- function(
  key = "contextmenu",
  className = "g6-contextmenu",
  trigger = "contextmenu",
  offset = c(4, 4),
  onClick = NULL,
  getItems = NULL,
  getContent = NULL,
  loadingContent = NULL,
  enable = JS("(e) => e.targetType === 'node'"),
  ...
) {
  # Validate inputs
  if (!is.character(className)) {
    stop("'className' should be a string")
  }

  valid_triggers <- c("contextmenu", "click")
  if (!trigger %in% valid_triggers) {
    stop("'trigger' should be one of 'contextmenu' or 'click'")
  }

  if (!is.numeric(offset) || length(offset) != 2) {
    stop("'offset' should be a numeric vector of length 2")
  }

  # Validate JS functions (no conversion)
  if (!is.null(onClick) && !is_js(onClick)) {
    stop(
      "'onClick' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(getItems) && !is_js(getItems)) {
    stop(
      "'getItems' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(getContent) && !is_js(getContent)) {
    stop(
      "'getContent' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' must be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  if (
    !is.null(loadingContent) &&
      !is.character(loadingContent) &&
      !is_js(loadingContent)
  ) {
    stop("'loadingContent' must be a string or JavaScript expression")
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "contextmenu"

  # Provide default for onSelect for Shiny context
  if (is.null(config$onClick)) {
    config$onClick <- JS(
      "(value, target, current) => {
        const graph = HTMLWidgets
          .find(`#${target.closest('.g6').id}`)
          .getWidget();
        if (current.id === undefined) return;
        if (value === 'create_edge') {
          graph.updateBehavior({
            key: 'create-edge', // Specify the behavior to update
            enable: true,
          });
          // Select node
          graph.setElementState(current.id, 'selected');
          // Disable drag node as it is incompatible with edge creation
          graph.updateBehavior({ key: 'drag-element', enable: false });
          graph.updateBehavior({ key: 'drag-element-force', enable: false });
        } else if (value === 'remove_node') {
          graph.removeNodeData([current.id]);
          graph.draw();
        }
      }
    "
    )
  }

  if (is.null(config$getItems)) {
    config$getItems <- JS(
      "() => {
        return [
          { name: 'Create edge', value: 'create_edge' },
          { name: 'Remove node', value: 'remove_node' }
        ];
      }"
    )
  }

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Edge Bundling Plugin
#'
#' Creates a configuration object for the edge-bundling plugin in G6.
#' This plugin automatically bundles similar edges together to reduce visual clutter.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param bundleThreshold Edge compatibility threshold, determines which edges
#' should be bundled together (number, default: 0.6).
#' @param cycles Number of simulation cycles (number, default: 6).
#' @param divisions Initial number of cut points (number, default: 1).
#' @param divRate Growth rate of cut points (number, default: 2).
#' @param iterations Number of iterations executed in the first cycle (number, default: 90).
#' @param iterRate Iteration decrement rate (number, default: 2/3).
#' @param K Edge strength, affects attraction and repulsion between edges (number, default: 0.1).
#' @param lambda Initial step size (number, default: 0.1).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/edge-bundling}.
#'
#' @return A list with the configuration settings for the edge-bundling plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- edge_bundling()
#'
#' # Custom configuration
#' config <- edge_bundling(
#'   key = "my-edge-bundling",
#'   bundleThreshold = 0.8,
#'   cycles = 8,
#'   K = 0.2
#' )
edge_bundling <- function(
  key = "edge-bundling",
  bundleThreshold = 0.6,
  cycles = 6,
  divisions = 1,
  divRate = 2,
  iterations = 90,
  iterRate = 2 / 3,
  K = 0.1,
  lambda = 0.1,
  ...
) {
  # Validate inputs
  if (
    !is.numeric(bundleThreshold) || bundleThreshold < 0 || bundleThreshold > 1
  ) {
    stop("'bundleThreshold' must be a number between 0 and 1")
  }

  if (!is.numeric(cycles) || cycles <= 0) {
    stop("'cycles' must be a positive number")
  }

  if (!is.numeric(divisions) || divisions <= 0) {
    stop("'divisions' must be a positive number")
  }

  if (!is.numeric(divRate) || divRate <= 0) {
    stop("'divRate' must be a positive number")
  }

  if (!is.numeric(iterations) || iterations <= 0) {
    stop("'iterations' must be a positive number")
  }

  if (!is.numeric(iterRate) || iterRate <= 0) {
    stop("'iterRate' must be a positive number")
  }

  if (!is.numeric(K) || K <= 0) {
    stop("'K' must be a positive number")
  }

  if (!is.numeric(lambda) || lambda <= 0) {
    stop("'lambda' must be a positive number")
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "edge-bundling"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Edge Filter Lens Plugin
#'
#' Creates a configuration object for the edge-filter-lens plugin in G6.
#' This plugin creates a lens that filters and displays edges within a specific area.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param trigger Method to move the lens: "pointermove", "click", or "drag" (string, default: "pointermove").
#' @param r Radius of the lens (number, default: 60).
#' @param maxR Maximum radius of the lens (number, default: NULL - half of the smaller canvas dimension).
#' @param minR Minimum radius of the lens (number, default: 0).
#' @param scaleRBy Method to scale the lens radius (string, default: "wheel").
#' @param nodeType Edge display condition: "both", "source", "target", or "either" (string, default: "both").
#' @param filter Filter out elements that are never displayed in the lens (JS function, default: NULL).
#' @param style Style of the lens (list, default: NULL).
#' @param nodeStyle Style of nodes in the lens (list or JS function, default: list(label = FALSE)).
#' @param edgeStyle Style of edges in the lens (list or JS function, default: list(label = TRUE)).
#' @param preventDefault Whether to prevent default events (boolean, default: TRUE).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/edge-filter-lens}.
#'
#' @return A list with the configuration settings for the edge-filter-lens plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- edge_filter_lens()
#'
#' # Custom configuration
#' config <- edge_filter_lens(
#'   key = "my-edge-lens",
#'   trigger = "drag",
#'   r = 100,
#'   nodeType = "either",
#'   style = list(
#'     fill = "rgba(200, 200, 200, 0.3)",
#'     stroke = "#999",
#'     lineWidth = 2
#'   ),
#'   filter = JS("(id, type) => {
#'     // Only display edges connected to specific nodes
#'     if (type === 'edge') {
#'       const edge = graph.getEdgeData(id);
#'       return edge.source === 'node1' || edge.target === 'node1';
#'     }
#'     return true;
#'   }")
#' )
edge_filter_lens <- function(
  key = "edge-filter-lens",
  trigger = c("pointermove", "click", "drag"),
  r = 60,
  maxR = NULL, # TBD check the default
  minR = 0,
  scaleRBy = "wheel",
  nodeType = c("both", "source", "target", "either"),
  filter = NULL,
  style = NULL,
  nodeStyle = list(label = FALSE),
  edgeStyle = list(label = TRUE),
  preventDefault = TRUE,
  ...
) {
  # Validate inputs
  trigger <- match.arg(trigger)

  if (!is.numeric(r) || r < 0) {
    stop("'r' must be a non-negative number")
  }

  if (!is.null(maxR) && (!is.numeric(maxR) || maxR < r)) {
    stop("'maxR' must be a number greater than or equal to 'r'")
  }

  if (!is.numeric(minR) || minR < 0 || minR > r) {
    stop("'minR' must be a non-negative number less than or equal to 'r'")
  }

  if (!is.character(scaleRBy) || scaleRBy != "wheel") {
    stop("'scaleRBy' must be 'wheel'")
  }

  nodeType <- match.arg(nodeType)

  if (!is.null(filter) && !is_js(filter)) {
    stop(
      "'filter' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(style) && !is.list(style)) {
    stop("'style' must be a list")
  }

  if (!is.null(nodeStyle) && !is.list(nodeStyle)) {
    stop(
      "'nodeStyle' must be a list"
    )
  }

  if (!is.null(edgeStyle) && !is.list(edgeStyle)) {
    stop(
      "'edgeStyle' must be a list"
    )
  }

  if (!is.logical(preventDefault)) {
    stop("'preventDefault' must be a boolean value")
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "edge-filter-lens"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Fish Eye Plugin
#'
#' Creates a configuration object for the fisheye plugin in G6.
#' This plugin creates a fisheye lens effect that magnifies elements within a specific area.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param trigger Method to move the fisheye: "pointermove", "click", or "drag" (string, default: "pointermove").
#' @param r Radius of the fisheye (number, default: 120).
#' @param maxR Maximum adjustable radius of the fisheye
#' (number, default: NULL - half of the smaller canvas dimension).
#' @param minR Minimum adjustable radius of the fisheye (number, default: 0).
#' @param d Distortion factor (number, default: 1.5).
#' @param maxD Maximum adjustable distortion factor (number, default: 5).
#' @param minD Minimum adjustable distortion factor (number, default: 0).
#' @param scaleRBy Method to adjust the fisheye radius: "wheel" or "drag" (string, default: NULL).
#' @param scaleDBy Method to adjust the fisheye distortion factor: "wheel" or "drag" (string, default: NULL).
#' @param showDPercent Whether to show the distortion factor value in the fisheye (boolean, default: TRUE).
#' @param style Style of the fisheye (list, default: NULL).
#' @param nodeStyle Style of nodes in the fisheye (list or JS function, default: list(label = TRUE)).
#' @param preventDefault Whether to prevent default events (boolean, default: TRUE).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/fisheye}.
#'
#' @return A list with the configuration settings for the fisheye plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- fish_eye()
#'
#' # Custom configuration
#' config <- fish_eye(
#'   key = "my-fisheye",
#'   trigger = "drag",
#'   r = 200,
#'   d = 2.5,
#'   scaleRBy = "wheel",
#'   scaleDBy = "drag",
#'   style = list(
#'     stroke = "#1890ff",
#'     fill = "rgba(24, 144, 255, 0.1)",
#'     lineWidth = 2
#'   ),
#'   nodeStyle = JS("(datum) => {
#'     return {
#'       label: true,
#'       labelCfg: {
#'         style: {
#'           fill: '#003a8c',
#'           fontSize: 14
#'         }
#'       }
#'     };
#'   }")
#' )
fish_eye <- function(
  key = "fish-eye",
  trigger = c("pointermove", "click", "drag"),
  r = 120,
  maxR = NULL,
  minR = 0,
  d = 1.5,
  maxD = 5,
  minD = 0,
  scaleRBy = NULL,
  scaleDBy = NULL,
  showDPercent = TRUE,
  style = NULL,
  nodeStyle = list(label = TRUE),
  preventDefault = TRUE,
  ...
) {
  # Validate inputs
  trigger <- match.arg(trigger)

  if (!is.numeric(r) || r < 0) {
    stop("'r' must be a non-negative number")
  }

  if (!is.null(maxR) && (!is.numeric(maxR) || maxR < r)) {
    stop("'maxR' must be a number greater than or equal to 'r'")
  }

  if (!is.numeric(minR) || minR < 0 || minR > r) {
    stop("'minR' must be a non-negative number less than or equal to 'r'")
  }

  if (!is.numeric(d) || d < 0) {
    stop("'d' must be a non-negative number")
  }

  if (!is.numeric(maxD) || maxD < d) {
    stop("'maxD' must be a number greater than or equal to 'd'")
  }

  if (!is.numeric(minD) || minD < 0 || minD > d) {
    stop("'minD' must be a non-negative number less than or equal to 'd'")
  }

  valid_scale_methods <- c("wheel", "drag")
  if (!is.null(scaleRBy) && !scaleRBy %in% valid_scale_methods) {
    stop("'scaleRBy' must be one of 'wheel' or 'drag'")
  }

  if (!is.null(scaleDBy) && !scaleDBy %in% valid_scale_methods) {
    stop("'scaleDBy' must be one of 'wheel' or 'drag'")
  }

  if (!is.logical(showDPercent)) {
    stop("'showDPercent' must be a boolean value")
  }

  if (!is.null(style) && !is.list(style)) {
    stop("'style' must be a list")
  }

  if (!is.null(nodeStyle) && !is.list(nodeStyle) && !is_js(nodeStyle)) {
    stop(
      "'nodeStyle' must be a list or a JS function returning style properties."
    )
  }

  if (!is.logical(preventDefault)) {
    stop("'preventDefault' must be a boolean value")
  }

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "fisheye"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Fullscreen Plugin
#'
#' Creates a configuration object for the fullscreen plugin in G6.
#' This plugin enables fullscreen mode for the graph.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param autoFit Whether to auto-fit the canvas size to the screen when in
#' fullscreen mode (boolean, default: TRUE).
#' @param trigger Methods to trigger fullscreen, e.g.,
#' list(request = "button", exit = "escape") (list, default: NULL).
#' @param onEnter Callback function after entering fullscreen mode (JS function, default: NULL).
#' @param onExit Callback function after exiting fullscreen mode (JS function, default: NULL).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/fullscreen}.
#'
#' @return A list with the configuration settings for the fullscreen plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- fullscreen()
#'
#' # Custom configuration
#' config <- fullscreen(
#'   key = "my-fullscreen",
#'   autoFit = TRUE,
#'   trigger = list(
#'     request = "F",
#'     exit = "Esc"
#'   ),
#'   onEnter = JS("() => {
#'     console.log('Entered fullscreen mode');
#'   }"),
#'   onExit = JS("() => {
#'     console.log('Exited fullscreen mode');
#'   }")
#' )
fullscreen <- function(
  key = "fullscreen",
  autoFit = TRUE,
  trigger = NULL,
  onEnter = NULL,
  onExit = NULL,
  ...
) {
  # Validate inputs
  if (!is.logical(autoFit)) {
    stop("'autoFit' must be a boolean value")
  }

  if (!is.null(trigger) && !is.list(trigger)) {
    stop("'trigger' must be a list with 'request' and/or 'exit' properties")
  }

  if (!is.null(trigger)) {
    if (!is.null(trigger$request) && !is.character(trigger$request)) {
      stop("'trigger$request' must be a string")
    }
    if (!is.null(trigger$exit) && !is.character(trigger$exit)) {
      stop("'trigger$exit' must be a string")
    }
  }

  if (!is.null(onEnter) && !is_js(onEnter)) {
    stop(
      "'onEnter' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(onExit) && !is_js(onExit)) {
    stop(
      "'onExit' must be a JavaScript function wrapped with JS()"
    )
  }

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "fullscreen"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Grid Line Plugin
#'
#' Creates a configuration object for the grid-line plugin in G6.
#' This plugin adds a background grid to the graph canvas.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param border Whether to display the border (boolean, default: TRUE).
#' @param borderLineWidth Border line width (number, default: 1).
#' @param borderStroke Border color (string, default: "#eee").
#' @param borderStyle Border style (string, default: "solid").
#' @param follow Whether the grid follows canvas movements (boolean or list, default: FALSE).
#' @param lineWidth Grid line width (number or string, default: 1).
#' @param size Grid unit size in pixels (number, default: 20).
#' @param stroke Grid line color (string, default: "#eee").
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/grid-line}.
#'
#' @return A list with the configuration settings for the grid-line plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- grid_line()
#'
#' # Custom configuration
#' config <- grid_line(
#'   key = "my-grid",
#'   border = TRUE,
#'   borderLineWidth = 2,
#'   borderStroke = "#ccc",
#'   borderStyle = "dashed",
#'   follow = list(
#'     translate = TRUE,
#'     zoom = FALSE
#'   ),
#'   lineWidth = 0.5,
#'   size = 30,
#'   stroke = "#e0e0e0"
#' )
grid_line <- function(
  key = "grid-line",
  border = TRUE,
  borderLineWidth = 1,
  borderStroke = "#eee",
  borderStyle = "solid",
  follow = FALSE,
  lineWidth = 1,
  size = 20,
  stroke = "#eee",
  ...
) {
  # Validate inputs
  if (!is.logical(border)) {
    stop("'border' must be a boolean value")
  }

  if (!is.numeric(borderLineWidth) || borderLineWidth < 0) {
    stop("'borderLineWidth' must be a non-negative number")
  }

  if (!is.character(borderStroke)) {
    stop("'borderStroke' must be a string representing a color")
  }

  valid_border_styles <- c(
    "solid",
    "dashed",
    "dotted",
    "double",
    "groove",
    "ridge",
    "inset",
    "outset",
    "none",
    "hidden"
  )
  if (!is.character(borderStyle) || !borderStyle %in% valid_border_styles) {
    stop("'borderStyle' must be a valid CSS border style")
  }

  if (!is.logical(follow) && !is.list(follow)) {
    stop(
      "'follow' must be a boolean or a list with 'translate' and/or 'zoom' properties"
    )
  }

  if (is.list(follow)) {
    if (!is.null(follow$translate) && !is.logical(follow$translate)) {
      stop("'follow$translate' must be a boolean")
    }
    if (!is.null(follow$zoom) && !is.logical(follow$zoom)) {
      stop("'follow$zoom' must be a boolean")
    }
  }

  if (!is.numeric(lineWidth) && !is.character(lineWidth)) {
    stop("'lineWidth' must be a number or string")
  }

  if (is.numeric(lineWidth) && lineWidth < 0) {
    stop("'lineWidth' as a number must be non-negative")
  }

  if (!is.numeric(size) || size <= 0) {
    stop("'size' must be a positive number")
  }

  if (!is.character(stroke)) {
    stop("'stroke' must be a string representing a color")
  }

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "grid-line"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure History Plugin
#'
#' Creates a configuration object for the history plugin in G6.
#' This plugin enables undo/redo functionality for graph operations.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param afterAddCommand Callback function called after a command is added to
#' the undo/redo queue (JS function, default: NULL).
#' @param beforeAddCommand Callback function called before a command is added to
#' the undo/redo queue (JS function, default: NULL).
#' @param executeCommand Callback function called when executing a command (JS function, default: NULL).
#' @param stackSize Maximum length of history records to be recorded, 0 means unlimited (number, default: 0).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/history}.
#'
#' @return A list with the configuration settings for the history plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- history()
#'
#' # Custom configuration
#' config <- history(
#'   key = "my-history",
#'   stackSize = 50,
#'   beforeAddCommand = JS("function(cmd, revert) {
#'     console.log('Before adding command:', cmd);
#'     // Only allow certain operations to be recorded
#'     return cmd.method !== 'update';
#'   }"),
#'   afterAddCommand = JS("function(cmd, revert) {
#'     console.log('Command added to ' + (revert ? 'undo' : 'redo') + ' stack');
#'   }"),
#'   executeCommand = JS("function(cmd) {
#'     console.log('Executing command:', cmd);
#'   }")
#' )
history <- function(
  key = "history",
  afterAddCommand = NULL,
  beforeAddCommand = NULL,
  executeCommand = NULL,
  stackSize = 0,
  ...
) {
  # Validate inputs
  if (!is.null(afterAddCommand) && !is_js(afterAddCommand)) {
    stop(
      "'afterAddCommand' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(beforeAddCommand) && !is_js(beforeAddCommand)) {
    stop(
      "'beforeAddCommand' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(executeCommand) && !is_js(executeCommand)) {
    stop(
      "'executeCommand' must be a JavaScript function wrapped with JS()"
    )
  }

  if (
    !is.numeric(stackSize) || stackSize < 0 || stackSize != round(stackSize)
  ) {
    stop("'stackSize' must be a non-negative integer")
  }

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "history"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Hull Plugin
#'
#' Creates a configuration object for the hull plugin in G6.
#' This plugin creates a hull (convex or concave) that surrounds specified graph elements.
#'
#' @param members Elements within the hull, including nodes and edges (character/numeric vector, required).
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param concavity Concavity parameter, larger values create less concave hulls (number, default: Infinity).
#' @param corner Corner type: "rounded", "smooth", or "sharp" (string, default: "rounded").
#' @param padding Padding around the elements (number, default: 10).
#' @param label Whether to display the label (boolean, default: TRUE).
#' @param labelText Label text content. Default to NULL.
#' @param labelPlacement Label position: "left", "right", "top", "bottom", or "center" (string, default: "bottom").
#' @param labelBackground Whether to display the background (boolean, default: FALSE).
#' @param labelPadding Label padding (number or numeric vector, default: 0).
#' @param labelCloseToPath Whether the label is close to the hull (boolean, default: TRUE).
#' @param labelAutoRotate Whether the label rotates with the hull, effective only when
#' closeToPath is true (boolean, default: TRUE).
#' @param labelOffsetX X-axis offset (number, default: 0).
#' @param labelOffsetY Y-axis offset (number, default: 0).
#' @param labelMaxWidth Maximum width of the text, exceeding will be ellipsized (number or NULL, default: NULL).
#' @param ... Other options.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/hull}.
#'
#' @return A list with the configuration settings for the hull plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- hull(members = c("node1", "node2", "node3"))
#'
#' # Custom configuration for a cluster
#' config <- hull(
#'   key = "cluster-hull",
#'   members = c("node1", "node2", "node3", "node4"),
#'   concavity = 0.8,
#'   corner = "smooth",
#'   padding = 15,
#'   label = TRUE,
#'   labelPlacement = "top",
#'   labelBackground = TRUE,
#'   labelPadding = c(4, 8),
#'   labelMaxWidth = 100
#' )
hull <- function(
  members,
  key = "hull",
  concavity = Inf,
  corner = c("rounded", "smooth", "sharp"),
  padding = 10,
  label = TRUE,
  labelText = NULL,
  labelPlacement = c("bottom", "left", "right", "top", "center"),
  labelBackground = FALSE,
  labelPadding = 0,
  labelCloseToPath = TRUE,
  labelAutoRotate = TRUE,
  labelOffsetX = 0,
  labelOffsetY = 0,
  labelMaxWidth = NULL,
  ...
) {
  # Validate inputs
  if (!is.character(members) && !is.numeric(members) || length(members) == 0) {
    stop("'members' must be a non-empty character/numeric vector")
  }
  members <- as.character(members)

  if (!is.numeric(concavity) || concavity <= 0) {
    stop("'concavity' must be a positive number or Infinity")
  }

  corner <- match.arg(corner)

  if (!is.numeric(padding) || padding < 0) {
    stop("'padding' must be a non-negative number")
  }

  if (!is.logical(label)) {
    stop("'label' must be a boolean value")
  }

  labelPlacement <- match.arg(labelPlacement)

  if (!is.logical(labelBackground)) {
    stop("'labelBackground' must be a boolean value")
  }

  # Check if labelPadding is a number or numeric vector
  if (!is.numeric(labelPadding)) {
    stop("'labelPadding' must be a number or numeric vector")
  }

  if (!is.logical(labelCloseToPath)) {
    stop("'labelCloseToPath' must be a boolean value")
  }

  if (!is.logical(labelAutoRotate)) {
    stop("'labelAutoRotate' must be a boolean value")
  }

  if (!is.numeric(labelOffsetX)) {
    stop("'labelOffsetX' must be a number")
  }

  if (!is.numeric(labelOffsetY)) {
    stop("'labelOffsetY' must be a number")
  }

  if (!is.null(labelMaxWidth) && (!is.numeric(labelMaxWidth) || labelMaxWidth < 0)) {
    stop("'labelMaxWidth' must be a non-negative number or NULL")
  }

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "hull"

  # Hull shapes (like bubble sets) are visual-only overlays appended to the
  # canvas after nodes, so in SVG mode they sit on top in the DOM.
  # - pointerEvents: "none" lets clicks/drags pass through to nodes
  # - zIndex: -1 renders the hull behind nodes so it doesn't visually cover
  #   collapse buttons or other node UI
  dots <- list(...)
  if (is.null(dots$pointerEvents)) {
    dots$pointerEvents <- "none"
  }
  if (is.null(dots$zIndex)) {
    dots$zIndex <- -1L
  }

  # Drop NULL elements
  dropNulls(c(config, dots))
}

#' Configure Legend Plugin
#'
#' Creates a configuration object for the legend plugin in G6.
#' This plugin adds a legend to the graph, allowing users to identify and interact with
#' different categories of elements.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param trigger How legend items trigger highlighting: "hover" or "click" (string, default: "hover").
#' @param position Relative position of the legend on the canvas (string, default: "bottom").
#' @param container Container to which the legend is mounted (HTML element or string, default: NULL).
#' @param className Legend canvas class name (string, default: NULL).
#' @param containerStyle Style of the legend container (list or JS object, default: NULL).
#' @param nodeField Node classification identifier (string or JS function, default: NULL).
#' @param edgeField Edge classification identifier (string or JS function, default: NULL).
#' @param comboField Combo classification identifier (string or JS function, default: NULL).
#' @param orientation Layout direction: "horizontal" or "vertical" (string, default: "horizontal").
#' @param layout Layout method: "flex" or "grid" (string, default: "flex").
#' @param showTitle Whether to display the title (boolean, default: FALSE).
#' @param titleText Title content (string, default: "").
#' @param x Relative horizontal position (number, default: NULL).
#' @param y Relative vertical position (number, default: NULL).
#' @param width Width of the legend (number, default: 240).
#' @param height Height of the legend (number, default: 160).
#' @param itemSpacing Spacing between text and marker (number, default: 4).
#' @param rowPadding Spacing between rows (number, default: 10).
#' @param colPadding Spacing between columns (number, default: 10).
#' @param itemMarkerSize Size of the legend item marker (number, default: 16).
#' @param itemLabelFontSize Font size of the legend item text (number, default: 16).
#' @param gridCol Maximum number of columns for grid layout (number, default: NULL).
#' @param gridRow Maximum number of rows for grid layout (number, default: NULL).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/legend}.
#'
#' @return A list with the configuration settings for the legend plugin.
#' @export
#'
#' @examples
#' # Basic configuration for node categories
#' config <- legend(
#'   nodeField = "category"
#' )
#'
#' # Advanced configuration
#' config <- legend(
#'   key = "my-legend",
#'   position = "top-right",
#'   nodeField = "type",
#'   edgeField = "relation",
#'   orientation = "vertical",
#'   layout = "grid",
#'   showTitle = TRUE,
#'   titleText = "Graph Elements",
#'   width = 300,
#'   height = 200,
#'   gridCol = 2,
#'   containerStyle = list(
#'     background = "#f9f9f9",
#'     border = "1px solid #ddd",
#'     borderRadius = "4px",
#'     padding = "8px"
#'   )
#' )
#'
#' # Using a function for classification
#' config <- legend(
#'   nodeField = JS("(item) => {
#'     return item.data.importance > 0.5 ? 'Important' : 'Regular';
#'   }")
#' )
legend <- function(
  key = "legend",
  trigger = c("hover", "click"),
  position = c(
    "bottom",
    "top",
    "left",
    "right",
    "top-left",
    "top-right",
    "bottom-left",
    "bottom-right"
  ),
  container = NULL,
  className = NULL,
  containerStyle = NULL,
  nodeField = NULL,
  edgeField = NULL,
  comboField = NULL,
  orientation = c("horizontal", "vertical"),
  layout = c("flex", "grid"),
  showTitle = FALSE,
  titleText = "",
  x = NULL,
  y = NULL,
  width = 240,
  height = 160,
  itemSpacing = 4,
  rowPadding = 10,
  colPadding = 10,
  itemMarkerSize = 16,
  itemLabelFontSize = 16,
  gridCol = NULL,
  gridRow = NULL,
  ...
) {
  # Validate inputs
  trigger <- match.arg(trigger)

  position <- match.arg(position)

  if (
    !is.null(container) &&
      !is.character(container) &&
      !inherits(container, "JS_EVAL")
  ) {
    stop(
      "'container' must be a string selector or HTMLElement reference wrapped with JS()"
    )
  }

  if (!is.null(className) && !is.character(className)) {
    stop("'className' must be a string")
  }

  if (
    !is.null(containerStyle) &&
      !is.list(containerStyle) &&
      !inherits(containerStyle, "JS_EVAL")
  ) {
    stop(
      "'containerStyle' must be a list or a JavaScript object wrapped with JS()"
    )
  }

  if (
    !is.null(nodeField) &&
      !is.character(nodeField) &&
      !is_js(nodeField)
  ) {
    stop(
      "'nodeField' must be a string or a JavaScript function wrapped with JS()"
    )
  }

  if (
    !is.null(edgeField) &&
      !is.character(edgeField) &&
      !is_js(edgeField)
  ) {
    stop(
      "'edgeField' must be a string or a JavaScript function wrapped with JS()"
    )
  }

  if (
    !is.null(comboField) &&
      !is.character(comboField) &&
      !is_js(comboField)
  ) {
    stop(
      "'comboField' must be a string or a JavaScript function wrapped with JS()"
    )
  }

  orientation <- match.arg(orientation)

  layout <- match.arg(layout)

  if (!is.logical(showTitle)) {
    stop("'showTitle' must be a boolean value")
  }

  if (!is.character(titleText)) {
    stop("'titleText' must be a string")
  }

  if (!is.null(x) && (!is.numeric(x) || length(x) != 1)) {
    stop("'x' must be a single numeric value")
  }

  if (!is.null(y) && (!is.numeric(y) || length(y) != 1)) {
    stop("'y' must be a single numeric value")
  }

  if (!is.numeric(width) || width <= 0) {
    stop("'width' must be a positive number")
  }

  if (!is.numeric(height) || height <= 0) {
    stop("'height' must be a positive number")
  }

  if (!is.numeric(itemSpacing) || itemSpacing < 0) {
    stop("'itemSpacing' must be a non-negative number")
  }

  if (!is.numeric(rowPadding) || rowPadding < 0) {
    stop("'rowPadding' must be a non-negative number")
  }

  if (!is.numeric(colPadding) || colPadding < 0) {
    stop("'colPadding' must be a non-negative number")
  }

  if (!is.numeric(itemMarkerSize) || itemMarkerSize <= 0) {
    stop("'itemMarkerSize' must be a positive number")
  }

  if (!is.numeric(itemLabelFontSize) || itemLabelFontSize <= 0) {
    stop("'itemLabelFontSize' must be a positive number")
  }

  if (
    !is.null(gridCol) &&
      (!is.numeric(gridCol) || gridCol <= 0 || gridCol != as.integer(gridCol))
  ) {
    stop("'gridCol' must be a positive integer")
  }

  if (
    !is.null(gridRow) &&
      (!is.numeric(gridRow) || gridRow <= 0 || gridRow != as.integer(gridRow))
  ) {
    stop("'gridRow' must be a positive integer")
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "legend"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Minimap Plugin
#'
#' Creates a configuration object for the minimap plugin in G6.
#' This plugin adds a minimap/thumbnail view of the entire graph.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param className Class name of the thumbnail canvas (string, default: NULL).
#' @param container Container to which the thumbnail is mounted (HTML element or string, default: NULL).
#' @param containerStyle Style of the thumbnail container (list or JS object, default: NULL).
#' @param delay Delay update time in milliseconds for performance optimization (number, default: 128).
#' @param filter Function to filter elements to display in minimap (JS function, default: NULL).
#' @param maskStyle Style of the mask (list or JS object, default: NULL).
#' @param padding Padding around the minimap (number or numeric vector, default: 10).
#' @param position Position of the thumbnail relative to the canvas
#' (string or numeric vector, default: "right-bottom").
#' @param renderer Custom renderer (JS object, default: NULL).
#' @param shape Method for generating element thumbnails (string or JS function, default: "key").
#' @param size Width and height of the minimap \code{[width, height]} (numeric vector, default: c(240, 160)).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/minimap}.
#'
#' @return A list with the configuration settings for the minimap plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- minimap()
#'
#' # Custom configuration
#' config <- minimap(
#'   key = "my-minimap",
#'   position = "left-top",
#'   size = c(200, 150),
#'   padding = 15,
#'   containerStyle = list(
#'     border = "1px solid #ddd",
#'     borderRadius = "4px",
#'     boxShadow = "0 0 8px rgba(0,0,0,0.1)"
#'   ),
#'   maskStyle = list(
#'     stroke = "#1890ff",
#'     strokeWidth = 2,
#'     fill = "rgba(24, 144, 255, 0.1)"
#'   )
#' )
#'
#' # With custom filtering function
#' config <- minimap(
#'   filter = JS("(id, elementType) => {
#'     // Only show nodes and important edges in the minimap
#'     if (elementType === 'node') return true;
#'     if (elementType === 'edge') {
#'       // Assuming edges have an 'important' attribute
#'       const edge = graph.findById(id);
#'       return edge.getModel().important === true;
#'     }
#'     return false;
#'   }")
#' )
minimap <- function(
  key = "minimap",
  className = NULL,
  container = NULL,
  containerStyle = NULL,
  delay = 128,
  filter = NULL,
  maskStyle = NULL,
  padding = 10,
  position = "right-bottom",
  renderer = NULL,
  shape = "key",
  size = c(240, 160),
  ...
) {
  # Validate inputs
  if (!is.null(className) && !is.character(className)) {
    stop("'className' must be a string")
  }

  if (
    !is.null(container) &&
      !is.character(container) &&
      !is_js(container)
  ) {
    stop(
      "'container' must be a string selector or HTMLElement reference wrapped with JS()"
    )
  }

  if (
    !is.null(containerStyle) &&
      !is.list(containerStyle) &&
      !is_js(containerStyle)
  ) {
    stop(
      "'containerStyle' must be a list or a JavaScript object wrapped with JS()"
    )
  }

  if (!is.numeric(delay) || delay < 0) {
    stop("'delay' must be a non-negative number")
  }

  if (!is.null(filter) && !is_js(filter)) {
    stop(
      "'filter' must be a JavaScript function wrapped with JS()"
    )
  }

  if (
    !is.null(maskStyle) &&
      !is.list(maskStyle) &&
      !is_js(maskStyle)
  ) {
    stop(
      "'maskStyle' must be a list or a JavaScript object wrapped with JS()"
    )
  }

  # Check if padding is a number or numeric vector
  if (!is.numeric(padding)) {
    stop("'padding' must be a number or numeric vector")
  }

  valid_positions <- c(
    "left",
    "right",
    "top",
    "bottom",
    "left-top",
    "left-bottom",
    "right-top",
    "right-bottom",
    "top-left",
    "top-right",
    "bottom-left",
    "bottom-right",
    "center"
  )
  if (!is.character(position) && !is.numeric(position)) {
    stop("'position' must be a string or numeric vector [x, y]")
  }

  if (is.character(position) && !position %in% valid_positions) {
    stop("'position' as string must be one of the valid position values")
  }

  if (is.numeric(position) && length(position) != 2) {
    stop("'position' as numeric vector must have exactly 2 elements [x, y]")
  }

  if (!is.null(renderer) && !is_js(renderer)) {
    stop(
      "'renderer' must be a JavaScript object wrapped with JS()"
    )
  }

  if (!is.character(shape) && !is_js(shape)) {
    stop(
      "'shape' must be a string or a JavaScript function wrapped with JS()"
    )
  }

  if (!is.numeric(size) || length(size) != 2 || any(size <= 0)) {
    stop("'size' must be a numeric vector of length 2 with positive values")
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "minimap"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Snapline Plugin
#'
#' Creates a configuration object for the snapline plugin in G6.
#' This plugin provides alignment guidelines when moving nodes.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param tolerance The alignment accuracy in pixels (number, default: 5).
#' @param offset The extension distance of the snapline (number, default: 20).
#' @param autoSnap Whether to enable automatic snapping (boolean, default: TRUE).
#' @param shape Specifies which shape to use as reference: "key" or a function
#' (string or JS function, default: "key").
#' @param verticalLineStyle Vertical snapline style (list or JS object, default: list(stroke = "#1783FF")).
#' @param horizontalLineStyle Horizontal snapline style (list or JS object, default: list(stroke = "#1783FF")).
#' @param filter Function to filter nodes that don't participate in alignment (JS function, default: NULL).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/snapline}.
#'
#' @return A list with the configuration settings for the snapline plugin.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- snapline()
#'
#' # Custom configuration
#' config <- snapline(
#'   key = "my-snapline",
#'   tolerance = 8,
#'   offset = 30,
#'   verticalLineStyle = list(
#'     stroke = "#f00",
#'     strokeWidth = 1.5,
#'     lineDash = c(5, 2)
#'   ),
#'   horizontalLineStyle = list(
#'     stroke = "#00f",
#'     strokeWidth = 1.5,
#'     lineDash = c(5, 2)
#'   )
#' )
#'
#' # With custom filter function
#' config <- snapline(
#'   filter = JS("(node) => {
#'     // Only allow regular nodes to participate in alignment
#'     // Exclude special nodes like 'start' or 'end'
#'     const model = node.getModel();
#'     return model.type !== 'start' && model.type !== 'end';
#'   }")
#' )
snapline <- function(
  key = "snapline",
  tolerance = 5,
  offset = 20,
  autoSnap = TRUE,
  shape = "key",
  verticalLineStyle = list(stroke = "#1783FF"),
  horizontalLineStyle = list(stroke = "#1783FF"),
  filter = NULL,
  ...
) {
  # Validate inputs
  if (!is.numeric(tolerance) || tolerance < 0) {
    stop("'tolerance' must be a non-negative number")
  }

  if (!is.numeric(offset) || offset < 0) {
    stop("'offset' must be a non-negative number")
  }

  if (!is.logical(autoSnap)) {
    stop("'autoSnap' must be a boolean value")
  }

  if (!is.character(shape) && !is_js(shape)) {
    stop(
      "'shape' must be a string or a JavaScript function wrapped with JS()"
    )
  }

  if (is.character(shape) && shape != "key") {
    stop("'shape' as string must be 'key'")
  }

  if (!is.null(verticalLineStyle) && !is.list(verticalLineStyle)) {
    stop(
      "'verticalLineStyle' must be a list"
    )
  }

  if (!is.null(horizontalLineStyle) && !is.list(horizontalLineStyle)) {
    stop(
      "'horizontalLineStyle' must be a list"
    )
  }

  if (!is.null(filter) && !is_js(filter)) {
    stop(
      "'filter' must be a JavaScript function wrapped with JS()"
    )
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "snapline"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Timebar Plugin
#'
#' Creates a configuration object for the timebar plugin in G6.
#' This plugin adds a timeline or chart-based control for time-related data visualization.
#'
#' @param data Time data, either a vector of timestamps or a list of objects with time and value (required).
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param className Additional class name for the timebar DOM (string, default: "g6-timebar").
#' @param x X position, will be ignored if position is set (number, default: NULL).
#' @param y Y position, will be ignored if position is set (number, default: NULL).
#' @param width Timebar width (number, default: 450).
#' @param height Timebar height (number, default: 60).
#' @param position Timebar position: "bottom" or "top" (string, default: "bottom").
#' @param padding Padding around the timebar (number or numeric vector, default: 10).
#' @param timebarType Display type: "time" or "chart" (string, default: "time").
#' @param elementTypes Filter element types: vector of "node", "edge", and/or "combo"
#' (character vector, default: c("node")).
#' @param mode Control element filtering method: "modify" or "visibility" (string, default: "modify").
#' @param values Current time value (number, vector of two numbers, Date, or vector of two Dates, default: NULL).
#' @param loop Whether to loop playback (boolean, default: FALSE).
#' @param getTime Method to get element time (JS function, default: NULL).
#' @param labelFormatter Custom time formatting in chart mode (JS function, default: NULL).
#' @param onChange Callback when time interval changes (JS function, default: NULL).
#' @param onReset Callback when reset (JS function, default: NULL).
#' @param onSpeedChange Callback when playback speed changes (JS function, default: NULL).
#' @param onPlay Callback when playback starts (JS function, default: NULL).
#' @param onPause Callback when paused (JS function, default: NULL).
#' @param onBackward Callback when moving backward (JS function, default: NULL).
#' @param onForward Callback when moving forward (JS function, default: NULL).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/timebar}.
#'
#' @return A list with the configuration settings for the timebar plugin.
#' @export
#'
#' @examples
#' # Basic timebar with array of timestamps
#' config <- timebar(
#'   data = c(1609459200000, 1609545600000, 1609632000000)  # Jan 1-3, 2021 in milliseconds
#' )
#'
#' # Chart-type timebar with time-value pairs
#' config <- timebar(
#'   data = list(
#'     list(time = 1609459200000, value = 10),
#'     list(time = 1609545600000, value = 25),
#'     list(time = 1609632000000, value = 15)
#'   ),
#'   timebarType = "chart",
#'   width = 600,
#'   height = 100,
#'   position = "top"
#' )
#'
#' # With custom callbacks
#' config <- timebar(
#'   data = c(1609459200000, 1609545600000, 1609632000000),
#'   onChange = JS("(values) => {
#'     console.log('Time changed:', values);
#'   }"),
#'   onPlay = JS("() => {
#'     console.log('Playback started');
#'   }")
#' )
#'
#' # With custom time getter function for elements
#' config <- timebar(
#'   data = c(1609459200000, 1609545600000, 1609632000000),
#'   getTime = JS("(datum) => {
#'     return datum.created_at; // Get time from created_at property
#'   }")
#' )
timebar <- function(
  data,
  key = "timebar",
  className = "g6-timebar",
  x = NULL,
  y = NULL,
  width = 450,
  height = 60,
  position = c("bottom", "top"),
  padding = 10,
  timebarType = c("time", "chart"),
  elementTypes = c("node", "edge", "combo"),
  mode = c("modify", "visibility"),
  values = NULL,
  loop = FALSE,
  getTime = NULL,
  labelFormatter = NULL,
  onChange = NULL,
  onReset = NULL,
  onSpeedChange = NULL,
  onPlay = NULL,
  onPause = NULL,
  onBackward = NULL,
  onForward = NULL,
  ...
) {
  # Validate inputs
  if (!is.null(x) && !is.numeric(x)) {
    stop("'x' must be a number")
  }

  if (!is.null(y) && !is.numeric(y)) {
    stop("'y' must be a number")
  }

  if (!is.numeric(width) || width <= 0) {
    stop("'width' must be a positive number")
  }

  if (!is.numeric(height) || height <= 0) {
    stop("'height' must be a positive number")
  }

  position <- match.arg(position)

  if (
    !is.numeric(padding) &&
      !is.null(padding) &&
      !is.vector(padding, mode = "numeric")
  ) {
    stop("'padding' must be a number or a numeric vector")
  }

  # Check if data is a numeric vector or a list of objects with time and value properties
  if (!is.numeric(data) && !is.list(data)) {
    stop(
      "'data' must be either a numeric vector or a list of objects with time and value properties"
    )
  }

  timebarType <- match.arg(timebarType)
  elementTypes <- match.arg(elementTypes)
  mode <- match.arg(mode)

  if (
    !is.null(values) &&
      !is.numeric(values) &&
      !inherits(values, "Date") &&
      !(is.vector(values) &&
        length(values) == 2 &&
        (all(sapply(values, is.numeric)) ||
          all(sapply(values, function(x) inherits(x, "Date")))))
  ) {
    stop("'values' must be a number, a Date, or a vector of two numbers/Dates")
  }

  if (!is.logical(loop)) {
    stop("'loop' must be a boolean value")
  }

  callback_params <- c(
    "getTime",
    "labelFormatter",
    "onChange",
    "onReset",
    "onSpeedChange",
    "onPlay",
    "onPause",
    "onBackward",
    "onForward"
  )

  for (param in callback_params) {
    param_value <- get(param)
    if (!is.null(param_value) && !is_js(param_value)) {
      stop(paste0(
        "'",
        param,
        "' must be a JavaScript function wrapped with JS()"
      ))
    }
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "timebar"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Toolbar Plugin
#'
#' Creates a configuration object for the toolbar plugin in G6.
#' This plugin adds a customizable toolbar with items for graph operations.
#'
#' @param getItems Function that returns the list of toolbar items (JS function, required).
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param className Additional CSS class name for the toolbar DOM element (string, default: NULL).
#' @param position Toolbar position relative to the canvas (string, default: "top-left").
#' @param style Custom style for the toolbar DOM element (list or JS object, default: NULL).
#' @param onClick Callback function after a toolbar item is clicked (JS function, default: NULL).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/toolbar}.
#'
#' @return A list with the configuration settings for the toolbar plugin.
#' @export
#'
#' @examples
#' # Basic toolbar with zoom controls
#' config <- g6R::toolbar(
#'   position = "top-right",
#'   getItems = JS("() => [
#'     { id: 'zoom-in', value: 'zoom-in' },
#'     { id: 'zoom-out', value: 'zoom-out' },
#'     { id: 'undo', value: 'undo' },
#'     { id: 'redo', value: 'redo' },
#'     { id: 'auto-fit', value: 'fit' }
#'   ]"),
#'   onClick = JS("(value) => {
#'     // redo, undo need to be used with the history plugin
#'     const history = graph.getPluginInstance('history');
#'     switch (value) {
#'       case 'zoom-in':
#'         graph.zoomTo(1.1);
#'         break;
#'       case 'zoom-out':
#'         graph.zoomTo(0.9);
#'         break;
#'       case 'undo':
#'         history?.undo();
#'         break;
#'       case 'redo':
#'         history?.redo();
#'         break;
#'       case 'fit':
#'         graph.fitView();
#'         break;
#'       default:
#'         break;
#'     }
#'   }")
#' )
toolbar <- function(
  getItems = NULL,
  key = "toolbar",
  className = NULL,
  position = c(
    "top-left",
    "top",
    "top-right",
    "right",
    "bottom-right",
    "bottom",
    "bottom-left",
    "left"
  ),
  style = NULL,
  onClick = NULL,
  ...
) {
  # Check if required parameter is provided
  if (!is.null(className) && !is.character(className)) {
    stop("'className' must be a string")
  }

  position <- match.arg(position)

  if (!is.null(style) && !is.list(style)) {
    stop("'style' must be a list")
  }

  if (!is.null(getItems) && !is_js(getItems)) {
    stop(
      "'getItems' must be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(onClick) && !is_js(onClick)) {
    stop(
      "'onClick' must be a JavaScript function wrapped with JS()"
    )
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "toolbar"

  # Default
  if (is.null(config$getItems)) {
    config$getItems <- JS(
      "( ) => [
        { id : 'zoom-in' , value : 'zoom-in' },
        { id : 'zoom-out' , value : 'zoom-out' },
        { id : 'auto-fit' , value : 'auto-fit' },
        { id: 'delete', value: 'delete' },
        { id: 'request-fullscreen', value: 'request-fullscreen' },
        { id: 'exit-fullscreen', value: 'exit-fullscreen' },
      ]"
    )
  }

  if (is.null(config$onClick)) {
    config$onClick <- JS(
      "( value, target, current ) => {   
        // Handle button click events
        const graph = HTMLWidgets
          .find(`#${target.closest('.g6').id}`)
          .getWidget();
        const fullScreen = graph.getPluginInstance('fullscreen');
        const zoomLevel = graph.getZoom();
        if ( value === 'zoom-in' ) {   
          graph.zoomTo (graph.getZoom() + 0.25);
        } else if ( value === 'zoom-out' ) {     
          graph.zoomTo (graph.getZoom() - 0.25);
        } else if ( value === 'auto-fit' ) {     
          graph.fitView ( ) ;
        } else if (value === 'delete') {
          const selectedNodes = graph
            .getElementDataByState('node', 'selected')
            .map((node) => {
              return node.id
            }
          );
          graph.removeNodeData(selectedNodes);
          graph.draw();
        } else if (value === 'request-fullscreen') {
          if (fullScreen !== undefined) {
            fullScreen.request();
          }
        } else if (value === 'exit-fullscreen') {
          if (fullScreen !== undefined) {
            fullScreen.exit();
          }
        }
      }
    "
    )
  }

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Tooltip Plugin
#'
#' Creates a configuration object for the tooltip plugin in G6.
#' This plugin displays tooltips when interacting with graph elements.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param position Tooltip position relative to the element (string, default: "top-right").
#' @param enable Whether the plugin is enabled (boolean or function, default: TRUE).
#' @param getContent Function to generate custom tooltip content (JS function, default: NULL).
#' @param onOpenChange Callback for tooltip show/hide events (JS function, default: NULL).
#' @param trigger Trigger behavior: "hover" or "click" (string, default: "hover").
#' @param container Custom rendering container for tooltip (string or HTML element, default: NULL).
#' @param offset Offset distance as a vector of two numbers \code{[x, y]} (numeric vector, default: c(10, 10)).
#' @param enterable Whether the pointer can enter the tooltip (boolean, default: FALSE).
#' @param title Title for the tooltip (string, default: NULL).
#' @param style Custom style for the tooltip (list or JS object, default: list(".tooltip" = list(visibility = "hidden"))).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/tooltip}.
#'
#' @return A list with the configuration settings for the tooltip plugin.
#' @export
#'
#' @examples
#' # Basic tooltip
#' config <- tooltips()
#'
#' # Tooltip with custom position and content
#' config <- tooltips(
#'   position = "bottom",
#'   getContent = JS("(event, items) => {
#'     let result = `<h4>Custom Content</h4>`;
#'     items.forEach((item) => {
#'       result += `<p>Type: ${item.data.description}</p>`;
#'     });
#'     return result;
#'   }")
#' )
#'
#' # Click-triggered tooltip with custom style
#' config <- tooltips(
#'   trigger = "click",
#'   position = "bottom-left",
#'   offset = c(15, 20),
#'   style = list(
#'     ".tooltip" = list(
#'       backgroundColor = "#fff",
#'       border = "1px solid #ccc",
#'       borderRadius = "4px",
#'       boxShadow = "0 2px 6px rgba(0,0,0,0.1)",
#'       padding = "10px",
#'       maxWidth = "300px"
#'     )
#'   )
#' )
#'
#' # Conditional tooltip based on node type
#' config <- tooltips(
#'   enable = JS("(event, items) => {
#'     // Only show tooltip for nodes with type 'important'
#'     const item = items[0];
#'     return item.type === 'important';
#'   }"),
#'   enterable = TRUE,
#'   onOpenChange = JS("(open) => {
#'     console.log('Tooltip visibility changed:', open);
#'   }")
#' )
tooltips <- function(
  key = "tooltip",
  position = c(
    "top-right",
    "top",
    "bottom",
    "left",
    "right",
    "top-left",
    "bottom-left",
    "bottom-right"
  ),
  enable = TRUE,
  getContent = NULL,
  onOpenChange = NULL,
  trigger = c("hover", "click"),
  container = NULL,
  offset = c(10, 10),
  enterable = FALSE,
  title = NULL,
  style = NULL,
  ...
) {
  position <- match.arg(position)

  if (!is.logical(enable) && !is_js(enable)) {
    stop("'enable' must be a boolean or a JavaScript function")
  }

  if (!is.null(getContent) && !is_js(getContent)) {
    stop("'getContent' must be a JavaScript function")
  }

  if (!is.null(onOpenChange) && !is_js(onOpenChange)) {
    stop("'onOpenChange' must be a JavaScript function")
  }

  trigger <- match.arg(trigger)

  if (!is.null(container) && !is.character(container)) {
    stop("'container' must be a string selector or an HTML element")
  }

  if (!is.numeric(offset) || length(offset) != 2) {
    stop("'offset' must be a numeric vector of length 2 [x, y]")
  }

  if (!is.logical(enterable)) {
    stop("'enterable' must be a boolean value")
  }

  if (!is.null(title) && !is.character(title)) {
    stop("'title' must be a string")
  }

  if (!is.null(style) && !is.list(style)) {
    stop("'style' must be a list")
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "tooltip"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Watermark Plugin
#'
#' Creates a configuration object for the watermark plugin in G6.
#' This plugin adds a watermark to the graph canvas.
#'
#' @param key Unique identifier for the plugin (string, default: NULL).
#' @param width Width of a single watermark (number, default: 200).
#' @param height Height of a single watermark (number, default: 100).
#' @param opacity Opacity of the watermark (number, default: 0.2).
#' @param rotate Rotation angle of the watermark in radians (number, default: pi/12).
#' @param imageURL Image watermark URL, higher priority than text watermark (string, default: NULL).
#' @param text Watermark text content (string, default: NULL).
#' @param textFill Color of the text watermark (string, default: "#000").
#' @param textFontSize Font size of the text watermark (number, default: 16).
#' @param textFontFamily Font of the text watermark (string, default: NULL).
#' @param textFontWeight Font weight of the text watermark (string, default: NULL).
#' @param textFontVariant Font variant of the text watermark (string, default: NULL).
#' @param textAlign Text alignment of the watermark (string, default: "center").
#' @param textBaseline Baseline alignment of the text watermark (string, default: "middle").
#' @param backgroundRepeat Repeat mode of the watermark (string, default: "repeat").
#' @param backgroundAttachment Background attachment behavior of the watermark (string, default: NULL).
#' @param backgroundBlendMode Background blend mode of the watermark (string, default: NULL).
#' @param backgroundClip Background clip of the watermark (string, default: NULL).
#' @param backgroundColor Background color of the watermark (string, default: NULL).
#' @param backgroundImage Background image of the watermark (string, default: NULL).
#' @param backgroundOrigin Background origin of the watermark (string, default: NULL).
#' @param backgroundPosition Background position of the watermark (string, default: NULL).
#' @param backgroundPositionX Horizontal position of the watermark background (string, default: NULL).
#' @param backgroundPositionY Vertical position of the watermark background (string, default: NULL).
#' @param backgroundSize Background size of the watermark (string, default: NULL).
#' @param ... Extra parameters.
#' See \url{https://g6.antv.antgroup.com/en/manual/plugin/watermark}.
#'
#' @return A list with the configuration settings for the watermark plugin.
#' @export
#'
#' @examples
#' # Basic text watermark
#' config <- watermark(
#'   text = "G6 Graph",
#'   opacity = 0.1
#' )
#'
#' # Image watermark
#' config <- watermark(
#'   imageURL = "https://gw.alipayobjects.com/os/s/prod/antv/assets/image/logo-with-text-73b8a.svg",
#'   width = 150,
#'   height = 75,
#'   opacity = 0.15,
#'   rotate = 0
#' )
#'
#' # Customized text watermark
#' config <- watermark(
#'   text = "CONFIDENTIAL",
#'   textFill = "#ff0000",
#'   textFontSize = 24,
#'   textFontWeight = "bold",
#'   opacity = 0.08,
#'   rotate = pi/6,
#'   backgroundRepeat = "repeat-x"
#' )
#'
#' # Watermark with background styling
#' config <- watermark(
#'   text = "Draft Document",
#'   textFill = "#333",
#'   backgroundColor = "#f9f9f9",
#'   backgroundClip = "content-box",
#'   backgroundSize = "cover"
#' )
watermark <- function(
  key = "watermark",
  width = 200,
  height = 100,
  opacity = 0.2,
  rotate = pi / 12,
  imageURL = NULL,
  text = NULL,
  textFill = "#000",
  textFontSize = 16,
  textFontFamily = NULL,
  textFontWeight = NULL,
  textFontVariant = NULL,
  textAlign = c("center", "end", "left", "right", "start"),
  textBaseline = c(
    "alphabetic",
    "bottom",
    "hanging",
    "ideographic",
    "middle",
    "top"
  ),
  backgroundRepeat = "repeat",
  backgroundAttachment = NULL,
  backgroundBlendMode = NULL,
  backgroundClip = NULL,
  backgroundColor = NULL,
  backgroundImage = NULL,
  backgroundOrigin = NULL,
  backgroundPosition = NULL,
  backgroundPositionX = NULL,
  backgroundPositionY = NULL,
  backgroundSize = NULL,
  ...
) {
  # Validate inputs
  if (!is.numeric(width) || width <= 0) {
    stop("'width' must be a positive number")
  }

  if (!is.numeric(height) || height <= 0) {
    stop("'height' must be a positive number")
  }

  if (!is.numeric(opacity) || opacity < 0 || opacity > 1) {
    stop("'opacity' must be a number between 0 and 1")
  }

  if (!is.numeric(rotate)) {
    stop("'rotate' must be a number (angle in radians)")
  }

  if (!is.null(imageURL) && !is.character(imageURL)) {
    stop("'imageURL' must be a string")
  }

  if (!is.null(text) && !is.character(text)) {
    stop("'text' must be a string")
  }

  if (is.null(imageURL) && is.null(text)) {
    warning(
      "Neither 'imageURL' nor 'text' is provided; watermark may not be visible"
    )
  }

  if (!is.character(textFill)) {
    stop("'textFill' must be a string (color value)")
  }

  if (!is.numeric(textFontSize) || textFontSize <= 0) {
    stop("'textFontSize' must be a positive number")
  }

  if (!is.null(textFontFamily) && !is.character(textFontFamily)) {
    stop("'textFontFamily' must be a string")
  }

  if (!is.null(textFontWeight) && !is.character(textFontWeight)) {
    stop("'textFontWeight' must be a string")
  }

  if (!is.null(textFontVariant) && !is.character(textFontVariant)) {
    stop("'textFontVariant' must be a string")
  }

  textAlign <- match.arg(textAlign)
  textBaseline <- match.arg(textBaseline)

  if (!is.character(backgroundRepeat)) {
    stop("'backgroundRepeat' must be a string")
  }

  # Check background properties are strings if provided
  bg_props <- c(
    "backgroundAttachment",
    "backgroundBlendMode",
    "backgroundClip",
    "backgroundColor",
    "backgroundImage",
    "backgroundOrigin",
    "backgroundPosition",
    "backgroundPositionX",
    "backgroundPositionY",
    "backgroundSize"
  )

  for (prop in bg_props) {
    value <- get(prop)
    if (!is.null(value) && !is.character(value)) {
      stop(paste0("'", prop, "' must be a string"))
    }
  }

  # Get all function argument names
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "watermark"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' @keywords internal
valid_plugins <- c(
  "background" = background,
  "bubble-sets" = bubble_sets,
  "contextmenu" = context_menu,
  "edge-bundling" = edge_bundling,
  "edge-filter-lens" = edge_filter_lens,
  "fisheye" = fish_eye,
  "fullscreen" = fullscreen,
  "grid-line" = grid_line,
  "history" = history,
  "hull" = hull,
  "legend" = legend,
  "minimap" = minimap,
  "snapline" = snapline,
  "timebar" = timebar,
  "toolbar" = toolbar,
  "tooltip" = tooltips,
  "watermark" = watermark
)

test_that("g6_plugins creates proper plugin configuration", {
  expect_snapshot(error = TRUE, {
    g6() |> g6_plugins()
  })

  g <- g6() |>
    g6_plugins(
      "minimap",
      grid_line()
    )

  expect_type(g$x$plugins, "list")
  expect_length(g$x$plugins, 2)
  keys <- vapply(g$x$plugins, function(x) x$key, character(1))
  expect_true("minimap" %in% keys)
  expect_true("grid-line" %in% keys)
})

test_that("validate plugins works", {
  expect_snapshot(error = TRUE, {
    g6() |>
      g6_plugins(
        "blabla"
      )

    g6() |> g6_plugins(list(test = 1))
  })
})

test_that("individual plugin functions work correctly", {
  lapply(names(valid_plugins), function(plugin) {
    print(plugin)
    expect_snapshot(
      {
        str(
          g6() |>
            g6_plugins(plugin)
        )
      },
      # Some plugins need default values
      error = if (plugin %in% c("bubble-sets", "hull", "timebar")) {
        TRUE
      } else {
        FALSE
      }
    )
  })
})

test_that("plugin validation works correctly", {
  # Test each plugin function can be called and returns correct type
  bg1 <- background()
  expect_type(bg1, "list")

  bg2 <- background(backgroundColor = "#f0f0f0", width = "50%")
  expect_type(bg2, "list")

  # Test bubble_sets validation
  expect_error(bubble_sets())
  bubble1 <- bubble_sets(members = c("node1", "node2"))
  expect_type(bubble1, "list")

  # Test context_menu validation
  expect_error(context_menu(className = 123))
  expect_error(context_menu(trigger = "invalid"))
  expect_error(context_menu(offset = c(1)))
  cm1 <- context_menu()
  expect_type(cm1, "list")

  # Test edge_bundling validation
  expect_error(edge_bundling(bundleThreshold = -0.1))
  expect_error(edge_bundling(bundleThreshold = 1.5))
  expect_error(edge_bundling(cycles = -1))
  expect_error(edge_bundling(K = -0.1))
  eb1 <- edge_bundling()
  expect_type(eb1, "list")

  # Test edge_filter_lens validation
  expect_error(edge_filter_lens(r = -5))
  expect_error(edge_filter_lens(minR = -1))
  expect_error(edge_filter_lens(maxR = 10, r = 20))
  expect_error(edge_filter_lens(scaleRBy = "invalid"))
  expect_error(edge_filter_lens(nodeStyle = "invalid"))
  efl1 <- edge_filter_lens()
  expect_type(efl1, "list")

  # Test fish_eye validation
  expect_error(fish_eye(r = -10))
  expect_error(fish_eye(d = -1))
  expect_error(fish_eye(maxD = 2, d = 5))
  expect_error(fish_eye(scaleRBy = "invalid"))
  fe1 <- fish_eye()
  expect_type(fe1, "list")

  # Test fullscreen validation
  expect_error(fullscreen(autoFit = "true"))
  expect_error(fullscreen(trigger = "invalid"))
  fs1 <- fullscreen()
  expect_type(fs1, "list")

  # Test grid_line validation
  expect_error(grid_line(border = "true"))
  expect_error(grid_line(borderLineWidth = -1))
  expect_error(grid_line(borderStyle = "invalid"))
  expect_error(grid_line(size = -10))
  gl1 <- grid_line()
  expect_type(gl1, "list")

  # Test history validation
  expect_error(history(stackSize = -1))
  expect_error(history(stackSize = 1.5))
  h1 <- history()
  expect_type(h1, "list")

  # Test hull validation - fix the error pattern
  expect_error(hull(members = c()))
  hull1 <- hull(members = c("node1", "node2"))
  expect_type(hull1, "list")

  # Test legend validation
  expect_error(legend(width = -100))
  expect_error(legend(height = -50))
  expect_error(legend(itemSpacing = -1))
  expect_error(legend(gridCol = -1))
  expect_error(legend(gridCol = 1.5))
  l1 <- legend()
  expect_type(l1, "list")

  # Test minimap validation
  expect_error(minimap(delay = -1))
  expect_error(minimap(size = c(100)))
  expect_error(minimap(size = c(-100, 50)))
  mm1 <- minimap()
  expect_type(mm1, "list")

  # Test snapline validation
  expect_error(snapline(tolerance = -1))
  expect_error(snapline(offset = -5))
  expect_error(snapline(autoSnap = "true"))
  expect_error(snapline(shape = 123))
  sl1 <- snapline()
  expect_type(sl1, "list")

  # Test timebar validation
  expect_error(timebar(data = "invalid"))
  expect_error(timebar(data = c(1, 2, 3), width = -100))
  expect_error(timebar(data = c(1, 2, 3), height = -50))
  expect_error(timebar(data = c(1, 2, 3), loop = "false"))
  tb1 <- timebar(data = c(1609459200000, 1609545600000))
  expect_type(tb1, "list")

  # Test toolbar validation
  expect_error(toolbar(style = "invalid"))
  tb2 <- toolbar()
  expect_type(tb2, "list")

  # Test tooltips validation
  expect_error(tooltips(enable = "true"))
  expect_error(tooltips(offset = c(1)))
  expect_error(tooltips(enterable = "false"))
  expect_error(tooltips(style = "invalid"))
  tt1 <- tooltips()
  expect_type(tt1, "list")

  # Test watermark validation
  text <- "Watermark Test"
  expect_error(watermark(width = -100))
  expect_error(watermark(height = -50))
  expect_error(watermark(opacity = 1.5))
  expect_error(watermark(opacity = -0.1))
  expect_error(watermark(textFontSize = -10, text = text))
  wm1 <- watermark(text = text)
  expect_type(wm1, "list")
})

test_that("plugin match.arg validation works", {
  # Test match.arg cases for various plugins
  bubble1 <- bubble_sets(members = c("n1"), labelPlacement = "top")
  expect_type(bubble1, "list")

  edge_lens1 <- edge_filter_lens(trigger = "click", nodeType = "source")
  expect_type(edge_lens1, "list")

  fisheye1 <- fish_eye(trigger = "drag")
  expect_type(fisheye1, "list")

  hull1 <- hull(members = c("n1"), corner = "sharp")
  expect_type(hull1, "list")

  legend1 <- legend(trigger = "click", position = "top-right")
  expect_type(legend1, "list")

  timebar1 <- timebar(
    data = c(1, 2, 3),
    position = "top",
    timebarType = "chart"
  )
  expect_type(timebar1, "list")

  toolbar1 <- toolbar(position = "bottom-right")
  expect_type(toolbar1, "list")

  tooltip1 <- tooltips(position = "bottom-left", trigger = "click")
  expect_type(tooltip1, "list")

  watermark1 <- watermark(
    text = "test",
    textAlign = "right",
    textBaseline = "hanging"
  )
  expect_type(watermark1, "list")
})

test_that("plugin complex configurations work", {
  # Test complex configurations to hit more code paths
  bubble_complex <- bubble_sets(
    members = c("n1", "n2", "n3"),
    avoidMembers = c("n4", "n5"),
    labelPlacement = "top",
    labelPadding = c(2, 4),
    maxRoutingIterations = 150,
    pixelGroup = 6
  )
  expect_type(bubble_complex, "list")

  # Test edge_filter_lens with complex configuration
  lens_complex <- edge_filter_lens(
    trigger = "drag",
    r = 100,
    maxR = 200,
    minR = 10,
    nodeType = "either"
  )
  expect_type(lens_complex, "list")

  # Test fish_eye with scaling options
  fisheye_complex <- fish_eye(
    trigger = "click",
    r = 150,
    d = 2.0,
    maxD = 4.0,
    minD = 0.5,
    scaleRBy = "wheel",
    scaleDBy = "drag"
  )
  expect_type(fisheye_complex, "list")

  # Test grid_line with follow configuration
  grid_complex <- grid_line(
    follow = list(translate = TRUE, zoom = FALSE),
    borderStyle = "dashed",
    size = 30
  )
  expect_type(grid_complex, "list")

  # Test hull with all label options
  hull_complex <- hull(
    members = c("n1", "n2"),
    corner = "smooth",
    labelPlacement = "center",
    labelBackground = TRUE,
    labelPadding = c(2, 4, 6),
    labelOffsetX = 5,
    labelOffsetY = 10
  )
  expect_type(hull_complex, "list")

  # Test legend with grid layout
  legend_complex <- legend(
    orientation = "vertical",
    layout = "grid",
    gridCol = 3,
    gridRow = 2,
    showTitle = TRUE,
    titleText = "Legend"
  )
  expect_type(legend_complex, "list")

  # Test minimap with numeric position
  minimap_complex <- minimap(
    position = c(10, 20),
    size = c(300, 200),
    padding = c(5, 10)
  )
  expect_type(minimap_complex, "list")

  # Test watermark with all text options
  watermark_complex <- watermark(
    text = "DRAFT",
    textAlign = "left",
    textBaseline = "top",
    textFontWeight = "bold",
    rotate = pi / 4,
    opacity = 0.1
  )
  expect_type(watermark_complex, "list")
})

#' Set the layout algorithm for a g6 graph
#'
#' This function configures the layout algorithm used to position nodes in a g6 graph.
#'
#' @param graph A g6 graph object created with \code{g6()}.
#' @param layout An existing layout function like \link{circular_layout} or a string like `circular-layout`.
#'   At minimum, this can be a list that should contain a \code{type} element specifying the layout algorithm.
#'   Additional parameters depend on the layout type chosen, for instance \code{list(type = "force")}.
#'
#' @return The modified g6 graph object with the specified layout, allowing for method chaining.
#'
#' @details
#' G6 provides several layout algorithms, each suitable for different graph structures:
#'
#' \itemize{
#'   \item \strong{force}: Force-directed layout using physical simulation of forces.
#'   \item \strong{random}: Random layout placing nodes randomly.
#'   \item \strong{circular}: Arranges nodes on a circle.
#'   \item \strong{radial}: Radial layout with nodes arranged outward from a central node.
#'   \item \strong{grid}: Arranges nodes in a grid pattern.
#'   \item \strong{concentric}: Concentric circles with important nodes in the center.
#'   \item \strong{dagre}: Hierarchical layout for directed acyclic graphs.
#'   \item \strong{fruchterman}: Force-directed layout based on the Fruchterman-Reingold algorithm.
#'   \item \strong{mds}: Multidimensional scaling layout.
#'   \item \strong{comboForce}: Force-directed layout specially designed for combo graphs.
#' }
#'
#' Each layout algorithm has specific configuration options. See the G6 documentation
#' for detailed information on each layout and its parameters:
#' \url{https://g6.antv.antgroup.com/en/manual/layout/overview}.
#'
#' @seealso [g6()]
#' @export
g6_layout <- function(graph, layout = d3_force_layout()) {
  if (is.null(layout)) {
    stop("layout must be specified.")
  }
  graph$x$layout <- validate_layout(layout)
  graph
}

#' @keywords internal
validate_layout <- function(x) {
  validate_component(x, "layout")
}

#' @keywords internal
build_layout <- function(type, ...) {
  # Which function called build_layout?
  caller_fun <- sys.function(sys.parent())
  caller_env <- parent.frame()

  #Get formal argument names from caller function
  caller_formals <- setdiff(names(formals(caller_fun)), "...")

  named_params <- mget(
    caller_formals,
    envir = caller_env,
    ifnotfound = list(NULL)
  )

  named_params <- dropNulls(named_params)
  extra_params <- list(...)
  c(list(type = type), named_params, extra_params)
}

#' Generate G6 AntV Dagre layout configuration
#'
#' This function creates a configuration list for G6 AntV Dagre layout
#' with all available options as parameters.
#'
#' @param rankdir Layout direction: "TB" (top to bottom), "BT" (bottom to top),
#'   "LR" (left to right), or "RL" (right to left).
#' @param align Node alignment: "UL" (upper left), "UR" (upper right),
#'   "DL" (down left), or "DR" (down right).
#' @param nodesep Node spacing (px). When rankdir is "TB" or "BT", it's the horizontal.
#'   spacing of nodes; when rankdir is "LR" or "RL", it's the vertical spacing of nodes.
#' @param nodesepFunc Function to customize node spacing for different nodes,
#'   in the form of function(node) that returns a number. Has higher priority than nodesep.
#' @param ranksep Layer spacing (px). When rankdir is "TB" or "BT", it's the vertical spacing
#'   between adjacent layers; when rankdir is "LR" or "RL", it's the horizontal spacing.
#' @param ranksepFunc Function to customize layer spacing, in the form of function(node)
#'   that returns a number. Has higher priority than ranksep.
#' @param ranker Algorithm for assigning ranks to nodes: "network-simplex", "tight-tree",
#'   or "longest-path".
#' @param nodeSize Node size for collision detection. Can be a single number (same width/height),
#'   an array \code{[width, height]}, or a function that returns either.
#' @param controlPoints Whether to retain edge control points.
#' @param begin Alignment position of the upper left corner of the layout.
#' Can be \code{[x, y]} or \code{[x, y, z]}.
#' @param sortByCombo Whether to sort nodes on the same layer by parentId to prevent combo overlap.
#' @param edgeLabelSpace Whether to leave space for edge labels.
#' @param nodeOrder Reference array of node order on the same layer, containing node id values.
#' @param radial Whether to perform a radial layout based on dagre.
#' @param focusNode Focused node (only used when radial=TRUE). Can be a node ID or node object.
#' @param preset Node positions to reference during layout calculation.
#' @param ... Additional parameters to pass to the layout.
#'
#' @return A list containing the configuration for G6 AntV Dagre layout.
#' @export
#'
#' @examples
#' # Basic dagre layout
#' dagre_config <- antv_dagre_layout()
#'
#' # Horizontal layout with custom spacing
#' dagre_config <- antv_dagre_layout(
#'   rankdir = "LR",
#'   align = "UL",
#'   nodesep = 80,
#'   ranksep = 150
#' )
#'
#' # Radial layout with focus node
#' dagre_config <- antv_dagre_layout(
#'   radial = TRUE,
#'   focusNode = "node1",
#'   ranker = "tight-tree"
#' )
antv_dagre_layout <- function(
  rankdir = c("TB", "BT", "LR", "RL"),
  align = c("UL", "UR", "DL", "DR"),
  nodesep = 50,
  nodesepFunc = NULL,
  ranksep = 100,
  ranksepFunc = NULL,
  ranker = c("network-simplex", "tight-tree", "longest-path"),
  nodeSize = NULL,
  controlPoints = FALSE,
  begin = NULL,
  sortByCombo = FALSE,
  edgeLabelSpace = TRUE,
  nodeOrder = NULL,
  radial = FALSE,
  focusNode = NULL,
  preset = NULL,
  ...
) {
  # Validate rankdir
  rankdir <- match.arg(rankdir)
  # Validate align
  align <- match.arg(align)

  # Validate nodesep
  if (!is.numeric(nodesep) || nodesep < 0) {
    stop("nodesep must be a non-negative number")
  }

  # Validate nodesepFunc
  if (!is.null(nodesepFunc) && !is_js(nodesepFunc)) {
    stop("nodesepFunc must be a JS function")
  }

  # Validate ranksep
  if (!is.numeric(ranksep) || ranksep < 0) {
    stop("ranksep must be a non-negative number")
  }

  # Validate ranksepFunc
  if (!is.null(ranksepFunc) && !is_js(ranksepFunc)) {
    stop("ranksepFunc must be a JS function")
  }

  # Validate ranker
  ranker <- match.arg(ranker)

  # Validate nodeSize
  if (!is.null(nodeSize)) {
    if (!is.numeric(nodeSize) && !is.function(nodeSize)) {
      if (is.numeric(nodeSize) && !is.vector(nodeSize)) {
        if (length(nodeSize) != 2) {
          stop(
            "nodeSize as a vector must have exactly 2 elements [width, height]"
          )
        }
      } else {
        stop(
          "nodeSize must be a number, a 2-element vector [width, height], or a function"
        )
      }
    }
  }

  # Validate controlPoints
  if (!is.logical(controlPoints)) {
    stop("controlPoints must be a logical value (TRUE or FALSE)")
  }

  # Validate begin
  if (!is.null(begin)) {
    if (!is.numeric(begin) || !(length(begin) == 2 || length(begin) == 3)) {
      stop("begin must be a numeric vector with 2 or 3 elements")
    }
  }

  # Validate sortByCombo
  if (!is.logical(sortByCombo)) {
    stop("sortByCombo must be a logical value (TRUE or FALSE)")
  }

  # Validate edgeLabelSpace
  if (!is.logical(edgeLabelSpace)) {
    stop("edgeLabelSpace must be a logical value (TRUE or FALSE)")
  }

  # Validate nodeOrder
  if (!is.null(nodeOrder) && !is.character(nodeOrder)) {
    stop("nodeOrder must be a character vector of node IDs")
  }

  # Validate radial
  if (!is.logical(radial)) {
    stop("radial must be a logical value (TRUE or FALSE)")
  }

  # Validate focusNode
  if (!is.null(focusNode) && radial == FALSE) {
    warning("focusNode is only effective when radial is TRUE")
  }
  build_layout("antv-dagre", ...)
}

#' Generate G6 AntV circular layout configuration
#'
#' This function creates a configuration list for G6 AntV circular layout
#' with all available options as parameters.
#'
#' @param angleRatio How many `2*PI` are there between the first node and the last node.
#' @param center Center of layout as vector `c(x, y)` or `c(x, y, z)`.
#' @param clockwise Is it arranged clockwise?
#' @param divisions Number of segments that nodes are placed on the ring.
#' @param nodeSize Node size (diameter) for collision detection.
#' @param nodeSpacing Minimum distance between rings.
#' @param ordering Basis for sorting nodes ("topology", "topology-directed", or "degree").
#' @param radius Radius of the circle (overrides startRadius and endRadius).
#' @param startAngle Starting angle of the layout.
#' @param endAngle End angle of the layout.
#' @param startRadius Starting radius of the spiral layout.
#' @param endRadius End radius of the spiral layout.
#' @param width Width of layout.
#' @param height Height of layout.
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/circular-layout}.
#'
#' @return A list containing the configuration for G6 AntV circular layout.
#' @export
#'
#' @examples
#' circular_config <- circular_layout(
#'   radius = 200,
#'   startAngle = 0,
#'   endAngle = pi,
#'   clockwise = FALSE
#' )
circular_layout <- function(
  angleRatio = 1,
  center = NULL,
  clockwise = TRUE,
  divisions = 1,
  nodeSize = 10,
  nodeSpacing = 10,
  ordering = NULL,
  radius = NULL,
  startAngle = 0,
  endAngle = 2 * pi,
  startRadius = NULL,
  endRadius = NULL,
  width = NULL,
  height = NULL,
  ...
) {
  # Validation checks
  if (!is.numeric(angleRatio) || angleRatio <= 0) {
    stop("angleRatio must be a positive number")
  }

  if (!is.null(center)) {
    if (!(is.numeric(center) && (length(center) == 2 || length(center) == 3))) {
      stop("center must be a numeric vector of length 2 or 3")
    }
  }

  if (!is.logical(clockwise)) {
    stop("clockwise must be a logical value (TRUE or FALSE)")
  }

  if (!is.numeric(divisions) || divisions < 1) {
    stop("divisions must be a positive integer")
  }

  if (!is.numeric(nodeSize) && !is_js(nodeSize) && !is.null(nodeSize)) {
    stop(
      "nodeSize must be a number, a JavaScript function wrapped by JS, or NULL"
    )
  }

  if (
    !is.numeric(nodeSpacing) &&
      !is_js(nodeSpacing) &&
      !is.null(nodeSpacing)
  ) {
    stop(
      "nodeSpacing must be a number, a JavaScript function wrapped by JS, or NULL"
    )
  }

  if (!is.null(ordering)) {
    valid_ordering <- c("topology", "topology-directed", "degree")
    if (!(ordering %in% valid_ordering)) {
      stop(paste(
        "ordering must be one of:",
        paste(valid_ordering, collapse = ", ")
      ))
    }
  }

  if (!is.null(radius) && !is.numeric(radius)) {
    stop("radius must be a numeric value")
  }

  if (!is.numeric(startAngle)) {
    stop("startAngle must be a numeric value")
  }

  if (!is.numeric(endAngle)) {
    stop("endAngle must be a numeric value")
  }

  if (!is.null(startRadius) && !is.numeric(startRadius)) {
    stop("startRadius must be a numeric value")
  }

  if (!is.null(endRadius) && !is.numeric(endRadius)) {
    stop("endRadius must be a numeric value")
  }

  if (!is.null(width) && (!is.numeric(width) || width <= 0)) {
    stop("width must be a positive numeric value")
  }

  if (!is.null(height) && (!is.numeric(height) || height <= 0)) {
    stop("height must be a positive numeric value")
  }

  build_layout("circular", ...)
}

#' Generate G6 AntV Compact Box layout configuration
#'
#' This function creates a configuration list for G6 AntV Compact Box layout
#' with all available options as parameters. The Compact Box layout is designed
#' for efficiently laying out trees and hierarchical structures.
#'
#' @param direction Layout direction: "LR" (left to right), "RL" (right to left),
#'   "TB" (top to bottom), "BT" (bottom to top), "H" (horizontal), or "V" (vertical)
#' @param getSide Function to set the nodes to be arranged on the left/right side
#'   of the root node. If not set, the algorithm automatically assigns the nodes
#'   to the left/right side. Note: This parameter is only effective when the layout
#'   direction is "H". Function format: \code{function(node) { return "left" or "right" }}
#' @param getId Callback function for generating node IDs.
#' Function format: \code{function(node) { return string }}
#' @param getWidth Function to calculate the width of each node.
#' Function format: \code{function(node) { return number }}
#' @param getHeight Function to calculate the height of each node.
#' Function format: \code{function(node) { return number }}
#' @param getHGap Function to calculate the horizontal gap for each node.
#' Function format: \code{function(node) { return number }}
#' @param getVGap Function to calculate the vertical gap for each node.
#' Function format: \code{function(node) { return number }}
#' @param radial Whether to enable radial layout
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/compact-box-layout}.
#'
#' @return A list containing the configuration for G6 AntV Compact Box layout.
#' @export
#'
#' @examples
#' # Basic compact box layout
#' box_config <- compact_box_layout()
#'
#' # Vertical compact box layout
#' box_config <- compact_box_layout(
#'   direction = "TB"
#' )
#'
#' # Radial layout
#' box_config <- compact_box_layout(
#'   radial = TRUE
#' )
compact_box_layout <- function(
  direction = c("LR", "RL", "TB", "BT", "H", "V"),
  getSide = NULL,
  getId = NULL,
  getWidth = NULL,
  getHeight = NULL,
  getHGap = NULL,
  getVGap = NULL,
  radial = FALSE,
  ...
) {
  # Validate direction
  direction <- match.arg(direction)

  # Validate getSide
  if (!is.null(getSide)) {
    if (!is.function(getSide)) {
      stop("getSide must be a function that returns 'left' or 'right'")
    }
    if (direction != "H") {
      warning("getSide is only effective when direction is 'H'")
    }
  }

  # Validate getId
  if (!is.null(getId) && !is_js(getId)) {
    stop("getId must be a JS function that returns a string")
  }

  # Validate getWidth
  if (!is.null(getWidth) && !is_js(getWidth)) {
    stop("getWidth must be a JS function that returns a number")
  }

  # Validate getHeight
  if (!is.null(getHeight) && !is_js(getHeight)) {
    stop("getHeight must be a JS function that returns a number")
  }

  # Validate getHGap
  if (!is.null(getHGap) && !is_js(getHGap)) {
    stop("getHGap must be a JS function that returns a number")
  }

  # Validate getVGap
  if (!is.null(getVGap) && !is_js(getVGap)) {
    stop("getVGap must be a JS function that returns a number")
  }

  # Validate radial
  if (!is.logical(radial)) {
    stop("radial must be a logical value (TRUE or FALSE)")
  }

  build_layout("compact-box", ...)
}

#' Generate G6 D3 Force layout configuration
#'
#' This function creates a configuration list for G6 D3 Force layout
#' with all available options as parameters.
#'
#' @param link A list specifying force parameters for links (edges), with components:
#'   \describe{
#'     \item{id}{Edge id generation function, format: \code{function(edge, index, edges) { return string }}. Default is \code{function(e) e.id}}
#'     \item{distance}{Ideal edge length that edges will tend toward. Can be a number or a function
#'       \code{function(edge, index, edges) { return number }}. Default is 30}
#'     \item{strength}{The strength of the force. Higher values make edge lengths closer to the ideal length.
#'       Can be a number or a function \code{function(edge, index, edges) { return number }}. Default is 1}
#'     \item{iterations}{Number of iterations of link force. Default is 1}
#'   }
#' @param collide A list specifying collision force parameters for nodes, with components:
#'   \describe{
#'     \item{radius}{Collision radius. Nodes closer than this distance will experience a repulsive force.
#'       Can be a number or a function \code{function(node, index, nodes) { return number }}. Default is 10}
#'     \item{strength}{The strength of the repulsive force. Higher values produce more obvious repulsion. Default is 1}
#'     \item{iterations}{The number of iterations for collision detection. Default is 1}
#'   }
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/d3-force}.
#'
#' @return A list containing the configuration for G6 AntV D3 Force layout.
#' @export
#'
#' @examples
#' # Basic D3 force layout
#' d3_force_config <- d3_force_layout()
#'
#' # Custom link distance and collision radius
#' d3_force_config <- d3_force_layout(
#'   link = list(
#'     distance = 150,
#'     strength = 0.5,
#'     iterations = 3
#'   ),
#'   collide = list(
#'     radius = 30,
#'     strength = 0.8
#'   )
#' )
d3_force_layout <- function(
  link = list(
    distance = 100,
    strength = 2
  ),
  collide = list(radius = 40),
  ...
) {
  # Validate link parameter
  if (!is.null(link)) {
    if (!is.list(link)) {
      stop("The 'link' parameter must be a list")
    }

    if (!is.null(link$id) && !is_js(link$id)) {
      stop("link$id must be a JS function that generates edge IDs")
    }

    if (!is.null(link$distance)) {
      if (!is.numeric(link$distance) && !is_js(link$distance)) {
        stop("link$distance must be a number or a JS function")
      }
      if (is.numeric(link$distance) && link$distance < 0) {
        stop("link$distance must be a positive number")
      }
    }

    if (!is.null(link$strength)) {
      if (!is.numeric(link$strength) && !is_js(link$strength)) {
        stop("link$strength must be a number or a JS function")
      }
      if (is.numeric(link$strength) && link$strength < 0) {
        stop("link$strength must be a positive number")
      }
    }

    if (!is.null(link$iterations)) {
      if (
        !is.numeric(link$iterations) ||
          !is.integer(as.integer(link$iterations)) ||
          link$iterations < 1
      ) {
        stop("link$iterations must be a positive integer")
      }
    }
  }

  # Validate collide parameter
  if (!is.null(collide)) {
    if (!is.list(collide)) {
      stop("The 'collide' parameter must be a list")
    }

    if (!is.null(collide$radius)) {
      if (!is.numeric(collide$radius) && !is_js(collide$radius)) {
        stop("collide$radius must be a number or a JS function")
      }
      if (is.numeric(collide$radius) && collide$radius < 0) {
        stop("collide$radius must be a positive number")
      }
    }

    if (!is.null(collide$strength)) {
      if (!is.numeric(collide$strength)) {
        stop("collide$strength must be a number")
      }
      if (collide$strength < 0 || collide$strength > 1) {
        stop("collide$strength should be between 0 and 1")
      }
    }

    if (!is.null(collide$iterations)) {
      if (
        !is.numeric(collide$iterations) ||
          !is.integer(as.integer(collide$iterations)) ||
          collide$iterations < 1
      ) {
        stop("collide$iterations must be a positive integer")
      }
    }
  }
  build_layout("d3-force", ...)
}

#' Generate G6 AntV Concentric layout configuration
#'
#' This function creates a configuration list for G6 AntV Concentric layout
#' with all available options as parameters. The Concentric layout places nodes in
#' concentric circles based on a centrality measure.
#'
#' @param center The center position of the circular layout \code{[x, y]} or \code{[x, y, z]}.
#'   By default, it's the center position of the current container.
#' @param clockwise Whether nodes are arranged in clockwise order.
#' @param equidistant Whether the distances between rings are equal.
#' @param width The width of the layout. By default, the container width is used.
#' @param height The height of the layout. By default, the container height is used.
#' @param sortBy Specify the sorting basis (node attribute name). The higher the value,
#'   the more central the node will be placed. If it is "degree", the degree of the node
#'   will be calculated. The higher the degree, the more central the node will be placed.
#' @param maxLevelDiff If the maximum attribute difference of nodes in the same layer
#'   is undefined, it will be set to maxValue / 4, where maxValue is the maximum
#'   attribute value of the sorting basis. For example, if sortBy is 'degree', then
#'   maxValue is the degree of the node with the largest degree among all nodes.
#' @param nodeSize Node size (diameter). Used to prevent collision detection when nodes overlap.
#'   Can be a number, a 2-element vector \code{[width, height]}, or a function that returns a number.
#' @param nodeSpacing Minimum distance between rings, used to adjust the radius.
#'   Can be a number, a vector of numbers, or a function that returns a number.
#' @param preventOverlap Whether to prevent node overlap. Must be coordinated with the nodeSize
#'   attribute or the data.size attribute in the node data. Only when data.size is set in the
#'   data or the nodeSize value that is the same as the current graph node size is configured
#'   in the layout, can node overlap collision detection be performed.
#' @param startAngle The arc at which to start layout of nodes (in radians)
#' @param sweep If the radian difference between the first and last nodes in the same layer
#'   is undefined, it will be set to \code{2 * Math.PI * (1 - 1 / |level.nodes|)}, where level.nodes
#'   is the number of nodes in each layer calculated by the algorithm, and |level.nodes| is
#'   the number of nodes in the layer.
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/concentric-layout}.
#'
#' @return A list containing the configuration for G6 AntV Concentric layout
#' @export
#'
#' @examples
#' # Basic concentric layout
#' concentric_config <- concentric_layout()
#'
#' # Custom concentric layout with degree-based sorting and overlap prevention
#' concentric_config <- concentric_layout(
#'   clockwise = TRUE,
#'   sortBy = "degree",
#'   preventOverlap = TRUE,
#'   nodeSize = 30,
#'   nodeSpacing = 20
#' )
#'
#' # Custom concentric layout with specific center and dimensions
#' concentric_config <- concentric_layout(
#'   center = c(300, 300),
#'   width = 600,
#'   height = 600,
#'   equidistant = TRUE,
#'   startAngle = pi
#' )
concentric_layout <- function(
  center = NULL,
  clockwise = FALSE,
  equidistant = FALSE,
  width = NULL,
  height = NULL,
  sortBy = "degree",
  maxLevelDiff = NULL,
  nodeSize = 30,
  nodeSpacing = 10,
  preventOverlap = FALSE,
  startAngle = 3 / 2 * pi,
  sweep = NULL,
  ...
) {
  # Validate center
  if (!is.null(center)) {
    if (!is.numeric(center) || !(length(center) == 2 || length(center) == 3)) {
      stop("center must be a numeric vector with 2 or 3 elements")
    }
  }

  # Validate clockwise
  if (!is.logical(clockwise)) {
    stop("clockwise must be a logical value (TRUE or FALSE)")
  }

  # Validate equidistant
  if (!is.logical(equidistant)) {
    stop("equidistant must be a logical value (TRUE or FALSE)")
  }

  # Validate width
  if (!is.null(width)) {
    if (!is.numeric(width) || width <= 0) {
      stop("width must be a positive number")
    }
  }

  # Validate height
  if (!is.null(height)) {
    if (!is.numeric(height) || height <= 0) {
      stop("height must be a positive number")
    }
  }

  # Validate sortBy
  if (!is.character(sortBy)) {
    stop("sortBy must be a character string")
  }

  # Validate maxLevelDiff
  if (!is.null(maxLevelDiff)) {
    if (!is.numeric(maxLevelDiff) || maxLevelDiff <= 0) {
      stop("maxLevelDiff must be a positive number")
    }
  }

  # Validate nodeSize
  if (!is.null(nodeSize)) {
    if (!is.numeric(nodeSize) && !is_js(nodeSize)) {
      if (is.numeric(nodeSize) && is.vector(nodeSize)) {
        if (length(nodeSize) != 2) {
          stop(
            "nodeSize as a vector must have exactly 2 elements [width, height]"
          )
        }
      } else {
        stop(
          "nodeSize must be a number, a 2-element vector [width, height], or a JS function"
        )
      }
    }
    if (is.numeric(nodeSize) && !is.vector(nodeSize) && nodeSize <= 0) {
      stop("nodeSize as a single number must be positive")
    }
  }

  # Validate nodeSpacing
  if (!is.null(nodeSpacing)) {
    if (!is.numeric(nodeSpacing) && !is_js(nodeSpacing)) {
      if (is.numeric(nodeSpacing) && is.vector(nodeSpacing)) {
        # It's a vector of numbers, which is allowed
      } else {
        stop("nodeSpacing must be a number, a numeric vector, or a JS function")
      }
    }
    if (is.numeric(nodeSpacing) && !is.vector(nodeSpacing) && nodeSpacing < 0) {
      stop("nodeSpacing as a single number must be non-negative")
    }
  }

  # Validate preventOverlap
  if (!is.logical(preventOverlap)) {
    stop("preventOverlap must be a logical value (TRUE or FALSE)")
  }

  # Validate startAngle
  if (!is.numeric(startAngle)) {
    stop("startAngle must be a number")
  }

  # Validate sweep
  if (!is.null(sweep)) {
    if (!is.numeric(sweep)) {
      stop("sweep must be a number")
    }
  }
  build_layout("concentric", ...)
}

#' Generate G6 AntV Dagre layout configuration
#'
#' This function creates a configuration list for G6 AntV Dagre layout
#' with all available options as parameters. The Dagre layout is designed
#' for directed graphs, creating hierarchical layouts with nodes arranged in layers.
#'
#' @param rankdir Layout direction: "TB" (top to bottom), "BT" (bottom to top),
#'   "LR" (left to right), or "RL" (right to left).
#' @param align Node alignment: "UL" (upper left), "UR" (upper right),
#'   "DL" (down left), or "DR" (down right).
#' @param nodesep Node spacing (px). When rankdir is "TB" or "BT", it's the horizontal
#'   spacing of nodes; when rankdir is "LR" or "RL", it's the vertical spacing of nodes.
#' @param ranksep Interlayer spacing (px). When rankdir is "TB" or "BT", it's the
#'   spacing between adjacent layers in the vertical direction; when rankdir is "LR" or "RL",
#'   it represents the spacing between adjacent layers in the horizontal direction.
#' @param ranker The algorithm for assigning a level to each node: "network-simplex" (the network
#'   simplex algorithm), "tight-tree" (the compact tree algorithm), or "longest-path"
#'   (the longest path algorithm).
#' @param nodeSize G6 custom attribute to specify the node size uniformly or for each node.
#'   Can be a single number (same width/height), an array \code{[width, height]},
#'   or a function that returns either.
#' @param controlPoints Whether to keep the control points of the edge.
#' @param ... Additional parameters to pass to the layout. See
#' \url{https://g6.antv.antgroup.com/en/manual/layout/dagre-layout}.
#'
#' @return A list containing the configuration for G6 AntV Dagre layout.
#' @export
#'
#' @examples
#' # Basic dagre layout
#' dagre_config <- dagre_layout()
#'
#' # Custom dagre layout with horizontal flow
#' dagre_config <- dagre_layout(
#'   rankdir = "LR",
#'   nodesep = 80,
#'   ranksep = 150,
#'   ranker = "tight-tree"
#' )
#'
#' # Custom dagre layout with specific node size
#' dagre_config <- dagre_layout(
#'   nodeSize = 40,
#'   controlPoints = TRUE
#' )
dagre_layout <- function(
  rankdir = c("TB", "BT", "LR", "RL"),
  align = c("UL", "UR", "DL", "DR"),
  nodesep = 50,
  ranksep = 100,
  ranker = c("network-simplex", "tight-tree", "longest-path"),
  nodeSize = NULL,
  controlPoints = FALSE,
  ...
) {
  # Validate rankdir
  rankdir <- match.arg(rankdir)
  # Validate align
  align <- match.arg(align)

  # Validate nodesep
  if (!is.numeric(nodesep) || nodesep < 0) {
    stop("nodesep must be a non-negative number")
  }

  # Validate ranksep
  if (!is.numeric(ranksep) || ranksep < 0) {
    stop("ranksep must be a non-negative number")
  }

  # Validate ranker
  ranker <- match.arg(ranker)

  # Validate nodeSize
  if (!is.null(nodeSize)) {
    if (!is.numeric(nodeSize) && !is_js(nodeSize)) {
      if (is.numeric(nodeSize) && is.vector(nodeSize)) {
        if (length(nodeSize) != 2) {
          stop(
            "nodeSize as a vector must have exactly 2 elements [width, height]"
          )
        }
      } else {
        stop(
          "nodeSize must be a number, a 2-element vector [width, height], or a JS function"
        )
      }
    }
  }

  # Validate controlPoints
  if (!is.logical(controlPoints)) {
    stop("controlPoints must be a logical value (TRUE or FALSE)")
  }
  build_layout("dagre", ...)
}


#' Generate G6 Force Atlas2 layout configuration
#'
#' This function creates a configuration list for G6 Force Atlas2 layout
#' with all available options as parameters.
#'
#' @param barnesHut Logical. Whether to enable quadtree acceleration.
#' When enabled, improves performance for large graphs but may affect layout quality.
#' By default, enabled if node count > 250.
#' @param dissuadeHubs Logical. Whether to enable hub mode.
#' If TRUE, nodes with higher in-degree are more likely to be placed at the center
#' than those with high out-degree. Defaults to FALSE.
#' @param height Numeric. Layout height. Defaults to container height.
#' @param kg Numeric. Gravity coefficient. The larger the value,
#' the more concentrated the layout is at the center. Defaults to 1.
#' @param kr Numeric. Repulsion coefficient. Adjusts the compactness of the layout.
#' The larger the value, the looser the layout. Defaults to 5.
#' @param ks Numeric. Controls the speed of node movement during iteration. Defaults to 0.1.
#' @param ksmax Numeric. Maximum node movement speed during iteration. Defaults to 10.
#' @param mode Character. Clustering mode.
#' Options are "normal" or "linlog". In "linlog" mode, clusters are more compact.
#' Defaults to "normal".
#' @param nodeSize Numeric or function. Node size (diameter).
#' Used for repulsion calculation when preventOverlap is enabled.
#'
#' If not set, uses data.size in node data.
#' @param preventOverlap Logical. Whether to prevent node overlap.
#' When enabled, layout considers node size to avoid overlap.
#' Node size is specified by \code{nodeSize} or \code{data.size} in node data. Defaults to FALSE.
#' @param prune Logical. Whether to enable auto-pruning.
#' By default, enabled if node count > 100. Pruning speeds up
#' convergence but may reduce layout quality. Set to FALSE to disable auto-activation.
#' @param tao Numeric. Tolerance for stopping oscillation when layout is near convergence. Defaults to 0.1.
#' @param width Numeric. Layout width. Defaults to container width.
#' @param center Numeric vector of length 2. Layout center in the form \code{[x, y]}.
#' Each node is attracted to this point, with gravity controlled by \code{kg}.
#' If not set, uses canvas center.
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/force-atlas2-layout}.
#'
#' @return A list containing the configuration for G6 force atlas2 layout.
#' @export
#'
#' @examples
#' if (interactive()) {
#'   g6(lesmis$nodes, lesmis$edges) |>
#'     g6_layout(force_atlas2_layout(
#'       kr = 20,
#'       preventOverlap = TRUE,
#'       center = c(250, 250))) |>
#'     g6_options(autoResize = TRUE) |>
#'     g6_behaviors(
#'       "zoom-canvas",
#'       drag_element()
#'     )
#' }
force_atlas2_layout <- function(
  barnesHut = NULL,
  dissuadeHubs = FALSE,
  height = NULL,
  kg = 1,
  kr = 5,
  ks = 0.1,
  ksmax = 10,
  mode = "normal",
  nodeSize = NULL,
  preventOverlap = FALSE,
  prune = NULL,
  tao = 0.1,
  width = NULL,
  center = NULL,
  ...
) {
  if (!is.null(barnesHut) && !is.logical(barnesHut)) {
    stop("'barnesHut' must be NULL or logical (TRUE/FALSE)")
  }

  if (!is.logical(dissuadeHubs)) {
    stop("'dissuadeHubs' must be logical (TRUE/FALSE)")
  }

  if (
    !is.null(height) &&
      (!is.numeric(height) || length(height) != 1 || height < 0)
  ) {
    stop("'height' must be a single non-negative number or NULL")
  }

  if (!is.numeric(kg) || length(kg) != 1 || kg < 0) {
    stop("'kg' must be a single non-negative number")
  }

  if (!is.numeric(kr) || length(kr) != 1 || kr < 0) {
    stop("'kr' must be a single non-negative number")
  }

  if (!is.numeric(ks) || length(ks) != 1 || ks < 0) {
    stop("'ks' must be a single non-negative number")
  }

  if (!is.numeric(ksmax) || length(ksmax) != 1 || ksmax < 0) {
    stop("'ksmax' must be a single non-negative number")
  }

  if (!is.character(mode) || !(mode %in% c("normal", "linlog"))) {
    stop("'mode' must be one of 'normal' or 'linlog'")
  }

  if (!is.null(nodeSize) && !is.numeric(nodeSize) && !is_js(nodeSize)) {
    stop("'nodeSize' must be a numeric value, a function, or NULL")
  }

  if (!is.logical(preventOverlap)) {
    stop("'preventOverlap' must be logical (TRUE/FALSE)")
  }

  if (!is.null(prune) && !is.logical(prune)) {
    stop("'prune' must be logical (TRUE/FALSE) or NULL")
  }

  if (!is.numeric(tao) || length(tao) != 1 || tao < 0) {
    stop("'tao' must be a single non-negative number")
  }

  if (
    !is.null(width) && (!is.numeric(width) || length(width) != 1 || width < 0)
  ) {
    stop("'width' must be a single non-negative number or NULL")
  }

  if (!is.null(center)) {
    if (!is.numeric(center) || length(center) != 2) {
      stop("'center' must be a numeric vector of length 2 or NULL")
    }
  }
  build_layout("force-atlas2", ...)
}

#' Generate G6 Fruchterman layout configuration
#'
#' This function creates a configuration list for G6 Fruchterman layout
#' with all available options as parameters.
#' @param height Numeric. Layout height. Defaults to container height.
#' @param width Numeric. Layout width. Defaults to container width.
#' @param gravity Numeric. Central force attracting nodes to the center.
#' Larger values make the layout more compact. Defaults to 10.
#' @param speed Numeric. Node movement speed per iteration. Higher values may cause oscillation. Defaults to 5.
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/fruchterman-layout}.
#'
#' @return A list containing the configuration for G6 fruchterman layout.
#' @export
#' @examples
#' if (interactive()) {
#'   g6(lesmis$nodes, lesmis$edges) |>
#'    g6_layout(fruchterman_layout(
#'      gravity = 5,
#'      speed = 5
#'    )) |>
#'    g6_behaviors(
#'      "zoom-canvas",
#'      drag_element()
#'    )
#' }
#'
fruchterman_layout <- function(
  height = NULL,
  width = NULL,
  gravity = 10,
  speed = 5,
  ...
) {
  if (
    !is.null(height) &&
      (!is.numeric(height) || length(height) != 1 || height < 0)
  ) {
    stop("'height' must be a single non-negative number or NULL")
  }

  if (
    !is.null(width) && (!is.numeric(width) || length(width) != 1 || width < 0)
  ) {
    stop("'width' must be a single non-negative number or NULL")
  }

  if (!is.numeric(gravity) || length(gravity) != 1 || gravity < 0) {
    stop("'gravity' must be a single non-negative number")
  }

  if (!is.numeric(speed) || length(speed) != 1 || speed < 0) {
    stop("'speed' must be a single non-negative number")
  }
  build_layout("fruchterman", ...)
}

#' Generate G6 Radial layout configuration
#'
#' This function creates a configuration list for G6 Radial layout
#' with all available options as parameters.
#'
#' @param center Numeric vector of length 2. Center coordinates.
#' @param focusNode Character, list (Node), or NULL. Radiating center node. Defaults to NULL.
#' @param height Numeric. Canvas height.
#' @param width Numeric. Canvas width.
#' @param nodeSize Numeric. Node size (diameter).
#' @param nodeSpacing Numeric or function. Minimum node spacing (effective when preventing overlap). Defaults to 10.
#' @param linkDistance Numeric. Edge length. Defaults to 50.
#' @param unitRadius Numeric or NULL. Radius per circle. Defaults to 100.
#' @param maxIteration Numeric. Maximum number of iterations. Defaults to 1000.
#' @param maxPreventOverlapIteration Numeric. Max iterations for overlap prevention. Defaults to 200.
#' @param preventOverlap Logical. Whether to prevent node overlap. Defaults to FALSE.
#' @param sortBy Character. Field for sorting nodes in the same layer.
#' @param sortStrength Numeric. Sorting strength for nodes in the same layer. Defaults to 10.
#' @param strictRadial Logical. Strictly place nodes in the same layer on the same ring. Defaults to TRUE.
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/radial-layout}.
#'
#' @return A list containing the configuration for G6 radial layout.
#' @export
#' @examples
#' if (interactive()) {
#'   g6(jsonUrl = "https://assets.antv.antgroup.com/g6/radial.json") |>
#'     g6_layout(
#'       radial_layout(
#'         unitRadius = 100,
#'         linkDistance = 200
#'       )
#'     ) |>
#'     g6_behaviors(
#'       "zoom-canvas",
#'       drag_element()
#'     )
#' }
#'
radial_layout <- function(
  center = NULL,
  focusNode = NULL,
  height = NULL,
  width = NULL,
  nodeSize = NULL,
  nodeSpacing = 10,
  linkDistance = 50,
  unitRadius = 100,
  maxIteration = 1000,
  maxPreventOverlapIteration = 200,
  preventOverlap = FALSE,
  sortBy = NULL,
  sortStrength = 10,
  strictRadial = TRUE,
  ...
) {
  if (!is.null(center)) {
    if (!is.numeric(center) || length(center) != 2) {
      stop("'center' must be a numeric vector of length 2 or NULL")
    }
  }

  if (!is.null(focusNode) && !(is.character(focusNode) || is.list(focusNode))) {
    stop("'focusNode' must be a character string, a list (Node), or NULL")
  }

  if (
    !is.null(height) &&
      (!is.numeric(height) || length(height) != 1 || height < 0)
  ) {
    stop("'height' must be a single non-negative number or NULL")
  }

  if (
    !is.null(width) && (!is.numeric(width) || length(width) != 1 || width < 0)
  ) {
    stop("'width' must be a single non-negative number or NULL")
  }

  if (
    !is.null(nodeSize) &&
      (!is.numeric(nodeSize) || length(nodeSize) != 1 || nodeSize < 0)
  ) {
    stop("'nodeSize' must be a single non-negative number or NULL")
  }

  if (!is.numeric(nodeSpacing) && !is.function(nodeSpacing)) {
    stop("'nodeSpacing' must be a number or a function")
  }

  if (
    !is.numeric(linkDistance) || length(linkDistance) != 1 || linkDistance < 0
  ) {
    stop("'linkDistance' must be a single non-negative number")
  }

  if (
    !is.null(unitRadius) &&
      (!is.numeric(unitRadius) || length(unitRadius) != 1 || unitRadius < 0)
  ) {
    stop("'unitRadius' must be a single non-negative number or NULL")
  }

  if (
    !is.numeric(maxIteration) || length(maxIteration) != 1 || maxIteration < 1
  ) {
    stop("'maxIteration' must be a positive number")
  }

  if (
    !is.numeric(maxPreventOverlapIteration) ||
      length(maxPreventOverlapIteration) != 1 ||
      maxPreventOverlapIteration < 1
  ) {
    stop("'maxPreventOverlapIteration' must be a positive number")
  }

  if (!is.logical(preventOverlap) || length(preventOverlap) != 1) {
    stop("'preventOverlap' must be a logical value")
  }

  if (!is.null(sortBy) && (!is.character(sortBy) || length(sortBy) != 1)) {
    stop("'sortBy' must be a character string or NULL")
  }

  if (
    !is.numeric(sortStrength) || length(sortStrength) != 1 || sortStrength < 0
  ) {
    stop("'sortStrength' must be a single non-negative number")
  }

  if (!is.logical(strictRadial) || length(strictRadial) != 1) {
    stop("'strictRadial' must be a logical value")
  }

  build_layout("radial", ...)
}

#' Generate G6 Dendrogram layout configuration
#'
#' This function creates a configuration list for G6 Dendrogram layout
#' with all available options as parameters.
#'
#' @param direction Character. Layout direction. Options: "LR", "RL", "TB", "BT", "H", "V". Defaults to "LR".
#' @param nodeSep Numeric. Node spacing, distance between nodes on the same level. Defaults to 20.
#' @param rankSep Numeric. Rank spacing, distance between different levels. Defaults to 200.
#' @param radial Logical. Whether to enable radial layout. Defaults to FALSE.
#' @param ... Additional parameters to pass to the layout.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/dendrogram-layout}.
#'
#' @return A list containing the configuration for G6 dendrogram layout.
#' @export
dendrogram_layout <- function(
  direction = c("LR", "RL", "TB", "BT", "H", "V"),
  nodeSep = 20,
  rankSep = 200,
  radial = FALSE,
  ...
) {
  directions <- match.arg(direction)

  if (!is.numeric(nodeSep) || length(nodeSep) != 1 || nodeSep < 0) {
    stop("'nodeSep' must be a single non-negative number")
  }

  if (!is.numeric(rankSep) || length(rankSep) != 1 || rankSep < 0) {
    stop("'rankSep' must be a single non-negative number")
  }

  if (!is.logical(radial) || length(radial) != 1) {
    stop("'radial' must be a logical value")
  }
  build_layout("dendrogram", ...)
}

#' Create an AntV Combo Combined Layout
#'
#' Creates a combo combined layout configuration for G6 graphs. This layout
#' algorithm combines different layout strategies for elements inside combos
#' and the outermost layer, providing hierarchical organization of graph elements.
#'
#' @param center Layout center coordinates. A numeric vector of length 2 \code{[x, y]}.
#'   If NULL, uses the graph center. Default is NULL.
#' @param comboPadding Padding value inside the combo, used only for force
#'   calculation, not for rendering. It is recommended to set the same value
#'   as the visual padding. Can be a number, numeric vector, function, or
#'   JS function. Default is 10.
#' @param innerLayout Layout algorithm for elements inside the combo. Should be
#'   a Layout object or function. If NULL, uses ConcentricLayout as default.
#' @param nodeSize Node size (diameter), used for collision detection. If not
#'   specified, it is calculated from the node's size property. Can be a number,
#'   numeric vector, function, or JS function. Default is 10.
#' @param outerLayout Layout algorithm for the outermost layer. Should be a
#'   Layout object or function. If NULL, uses ForceLayout as default.
#' @param spacing Minimum spacing between node/combo edges when preventNodeOverlap
#'   or preventOverlap is true. Can be a number, function, or JS function for
#'   different nodes. Default is NULL.
#' @param treeKey Tree key identifier as a character string. Default is NULL.
#' @param ... Additional parameters passed to the layout configuration.
#' See \url{https://g6.antv.antgroup.com/en/manual/layout/combo-combined-layout}.
#'
#' @return A layout configuration object for use with G6 graphs.
#'
#' @details
#' The combo combined layout is particularly useful for graphs with hierarchical
#' structures where you want different layout algorithms for different levels
#' of the hierarchy. The inner layout handles elements within combos, while
#' the outer layout manages the overall arrangement.
#'
#' @examples
#' # Basic combo combined layout
#' layout <- combo_combined_layout()
#'
#' # Custom configuration with specific center and padding
#' layout <- combo_combined_layout(
#'   comboPadding = 20,
#'   nodeSize = 15,
#'   spacing = 10
#' )
#'
#' @seealso \code{\link{antv_dagre_layout}} for dagre layout configuration
#'
#' @export
combo_combined_layout <- function(
  center = NULL,
  comboPadding = 10,
  innerLayout = NULL,
  nodeSize = 10,
  outerLayout = NULL,
  spacing = NULL,
  treeKey = NULL,
  ...
) {
  # Validate center
  if (!is.null(center)) {
    if (!is.numeric(center) || length(center) != 2) {
      stop("center must be a numeric vector with 2 elements [x, y]")
    }
  }

  # Validate comboPadding
  if (!is.null(comboPadding)) {
    if (is.numeric(comboPadding)) {
      if (is.vector(comboPadding) && length(comboPadding) > 1) {
        # Allow numeric vector
      } else if (length(comboPadding) == 1 && comboPadding < 0) {
        stop("comboPadding must be a non-negative number")
      }
    } else if (!is_js(comboPadding)) {
      stop(
        "comboPadding must be a number, numeric vector, JS function, or JS function"
      )
    }
  }

  # Validate innerLayout
  if (!is.null(innerLayout)) {
    if (!is.list(innerLayout) && !is_js(innerLayout)) {
      stop("innerLayout must be a Layout object or JS function")
    }
  }

  # Validate nodeSize
  if (!is.null(nodeSize)) {
    if (is.numeric(nodeSize)) {
      if (is.vector(nodeSize) && length(nodeSize) > 1) {
        # Allow numeric vector
      } else if (length(nodeSize) == 1 && nodeSize <= 0) {
        stop("nodeSize must be a positive number")
      }
    } else if (!is_js(nodeSize)) {
      stop(
        "nodeSize must be a number, numeric vector, function, or JS function"
      )
    }
  }

  # Validate outerLayout
  if (!is.null(outerLayout)) {
    if (!is.list(outerLayout) && !is_js(outerLayout)) {
      stop("outerLayout must be a Layout object or JS function")
    }
  }

  # Validate spacing
  if (!is.null(spacing)) {
    if (is.numeric(spacing)) {
      if (spacing < 0) {
        stop("spacing must be a non-negative number")
      }
    } else if (is.function(spacing)) {
      # Allow function
    } else if (!is_js(spacing)) {
      stop("spacing must be a number, function, or JS function")
    }
  }

  # Validate treeKey
  if (!is.null(treeKey)) {
    if (!is.character(treeKey) || length(treeKey) != 1) {
      stop("treeKey must be a single character string")
    }
  }

  build_layout("combo-combined", ...)
}

#' @keywords internal
valid_layouts <- c(
  # N = 20
  "antv-dagre" = antv_dagre_layout,
  "circular" = circular_layout,
  "combo-combined" = combo_combined_layout,
  "concentric" = concentric_layout,
  "d3-force" = d3_force_layout,
  #"d3-force-3d" = d3_force_3d_layout,
  #"dagre" = dagre_layout,
  #"fishbone" = fishbone_layout,
  #"force",
  "force-atlas2" = force_atlas2_layout,
  "fruchterman" = fruchterman_layout,
  #"grid" = grid_layout,
  #"mds" = mds_layout,
  "radial" = radial_layout,
  #"random",
  #"snake",
  "compact-box" = compact_box_layout,
  "dendrogram" = dendrogram_layout #,
  #"mindmap",
  #"indented"
)

test_that("validate layout works", {
  expect_snapshot(error = TRUE, {
    g6() |>
      g6_layout(NULL)
    g6() |>
      g6_layout(
        "blabla"
      )

    g6() |> g6_layout(list(test = 1))
  })
})

test_that("individual layout functions work correctly", {
  lapply(names(valid_layouts), function(layout) {
    expect_snapshot(
      {
        str(
          g6() |>
            g6_layout(layout)
        )
      }
    )
  })
})

test_that("layout validation works correctly", {
  # Test antv_dagre layout validation
  expect_error(antv_dagre_layout(rankdir = "invalid"))
  expect_error(antv_dagre_layout(align = "invalid"))
  expect_error(antv_dagre_layout(nodesep = -10))
  expect_error(antv_dagre_layout(ranksep = -20))
  expect_error(antv_dagre_layout(ranker = "invalid"))
  expect_error(antv_dagre_layout(controlPoints = "true"))
  expect_error(antv_dagre_layout(sortByCombo = "false"))
  expect_error(antv_dagre_layout(edgeLabelSpace = "true"))
  expect_error(antv_dagre_layout(radial = "false"))
  ad1 <- antv_dagre_layout()
  expect_type(ad1, "list")

  # Test circular layout validation
  expect_error(circular_layout(angleRatio = -1))
  expect_error(circular_layout(center = c(100)))
  expect_error(circular_layout(center = "center"))
  expect_error(circular_layout(clockwise = "true"))
  expect_error(circular_layout(divisions = -1))
  expect_error(circular_layout(ordering = "invalid"))
  expect_error(circular_layout(width = -100))
  expect_error(circular_layout(height = -50))
  c1 <- circular_layout()
  expect_type(c1, "list")

  # Test compact_box layout validation
  expect_error(compact_box_layout(direction = "invalid"))
  expect_error(compact_box_layout(radial = "false"))
  cb1 <- compact_box_layout()
  expect_type(cb1, "list")

  # Test concentric layout validation
  expect_error(concentric_layout(center = c(100)))
  expect_error(concentric_layout(clockwise = "true"))
  expect_error(concentric_layout(equidistant = "false"))
  expect_error(concentric_layout(width = -100))
  expect_error(concentric_layout(height = -50))
  expect_error(concentric_layout(sortBy = 123))
  expect_error(concentric_layout(maxLevelDiff = -5))
  expect_error(concentric_layout(preventOverlap = "true"))
  con1 <- concentric_layout()
  expect_type(con1, "list")

  # Test d3_force layout validation
  expect_error(d3_force_layout(link = "invalid"))
  expect_error(d3_force_layout(collide = "invalid"))
  expect_error(d3_force_layout(link = list(distance = -50)))
  expect_error(d3_force_layout(link = list(strength = -1)))
  expect_error(d3_force_layout(link = list(iterations = -1)))
  expect_error(d3_force_layout(collide = list(radius = -10)))
  expect_error(d3_force_layout(collide = list(strength = 1.5)))
  expect_error(d3_force_layout(collide = list(iterations = -1)))
  d3f1 <- d3_force_layout()
  expect_type(d3f1, "list")

  # Test force_atlas2 layout validation
  expect_error(force_atlas2_layout(barnesHut = "true"))
  expect_error(force_atlas2_layout(dissuadeHubs = "false"))
  expect_error(force_atlas2_layout(height = -100))
  expect_error(force_atlas2_layout(kg = -1))
  expect_error(force_atlas2_layout(kr = -5))
  expect_error(force_atlas2_layout(ks = -0.1))
  expect_error(force_atlas2_layout(ksmax = -10))
  expect_error(force_atlas2_layout(mode = "invalid"))
  expect_error(force_atlas2_layout(preventOverlap = "true"))
  expect_error(force_atlas2_layout(prune = "false"))
  expect_error(force_atlas2_layout(tao = -0.1))
  expect_error(force_atlas2_layout(width = -200))
  expect_error(force_atlas2_layout(center = c(100)))
  fa2_1 <- force_atlas2_layout()
  expect_type(fa2_1, "list")

  # Test fruchterman layout validation
  expect_error(fruchterman_layout(height = -100))
  expect_error(fruchterman_layout(width = -200))
  expect_error(fruchterman_layout(gravity = -10))
  expect_error(fruchterman_layout(speed = -1))
  fr1 <- fruchterman_layout()
  expect_type(fr1, "list")

  # Test radial layout validation
  expect_error(radial_layout(center = c(100)))
  expect_error(radial_layout(focusNode = 123))
  expect_error(radial_layout(height = -100))
  expect_error(radial_layout(width = -200))
  expect_error(radial_layout(nodeSize = -10))
  expect_error(radial_layout(linkDistance = -50))
  expect_error(radial_layout(unitRadius = -40))
  expect_error(radial_layout(maxIteration = -1))
  expect_error(radial_layout(maxPreventOverlapIteration = -1))
  expect_error(radial_layout(preventOverlap = "true"))
  expect_error(radial_layout(sortBy = 123))
  expect_error(radial_layout(sortStrength = -10))
  expect_error(radial_layout(strictRadial = "false"))
  r1 <- radial_layout()
  expect_type(r1, "list")

  # Test dendrogram layout validation
  expect_error(dendrogram_layout(direction = "invalid"))
  expect_error(dendrogram_layout(nodeSep = -10))
  expect_error(dendrogram_layout(rankSep = -20))
  expect_error(dendrogram_layout(radial = "false"))
  d1 <- dendrogram_layout()
  expect_type(d1, "list")

  # Test combo_combined layout validation
  expect_error(combo_combined_layout(center = c(100)))
  expect_error(combo_combined_layout(comboPadding = -10))
  cc1 <- combo_combined_layout()
  expect_type(cc1, "list")
})

test_that("layout match.arg validation works", {
  # Test match.arg cases for various layouts
  antv_dagre1 <- antv_dagre_layout(
    rankdir = "LR",
    align = "UR",
    ranker = "tight-tree"
  )
  expect_type(antv_dagre1, "list")

  circular1 <- circular_layout(ordering = "topology")
  expect_type(circular1, "list")

  compact_box1 <- compact_box_layout(direction = "TB")
  expect_type(compact_box1, "list")

  concentric1 <- concentric_layout(sortBy = "degree")
  expect_type(concentric1, "list")

  force_atlas2_1 <- force_atlas2_layout(mode = "linlog")
  expect_type(force_atlas2_1, "list")

  dendrogram1 <- dendrogram_layout(direction = "BT")
  expect_type(dendrogram1, "list")
})

test_that("layout complex configurations work", {
  # Test antv_dagre with all options
  antv_dagre_complex <- antv_dagre_layout(
    rankdir = "TB",
    align = "DL",
    nodesep = 30,
    ranksep = 50,
    ranker = "longest-path",
    controlPoints = TRUE,
    sortByCombo = TRUE,
    edgeLabelSpace = FALSE,
    radial = TRUE,
    focusNode = "node1"
  )
  expect_type(antv_dagre_complex, "list")

  # Test circular with complex configuration
  circular_complex <- circular_layout(
    angleRatio = 1.5,
    center = c(200, 150),
    clockwise = FALSE,
    divisions = 8,
    ordering = "degree",
    radius = 200,
    startAngle = pi / 4,
    endAngle = 3 * pi / 2,
    startRadius = 50,
    endRadius = 300
  )
  expect_type(circular_complex, "list")

  # Test concentric with complex configuration
  concentric_complex <- concentric_layout(
    center = c(300, 250),
    clockwise = TRUE,
    equidistant = TRUE,
    width = 600,
    height = 500,
    sortBy = "value",
    maxLevelDiff = 10,
    nodeSize = 25,
    nodeSpacing = 15,
    preventOverlap = TRUE,
    startAngle = 0,
    sweep = pi
  )
  expect_type(concentric_complex, "list")

  # Test d3_force with complex configuration
  d3_force_complex <- d3_force_layout(
    link = list(
      distance = 120,
      strength = 1.5,
      iterations = 3
    ),
    collide = list(
      radius = 25,
      strength = 0.9,
      iterations = 2
    )
  )
  expect_type(d3_force_complex, "list")

  # Test force_atlas2 with all options
  force_atlas2_complex <- force_atlas2_layout(
    barnesHut = TRUE,
    dissuadeHubs = TRUE,
    height = 400,
    kg = 2,
    kr = 10,
    ks = 0.2,
    ksmax = 15,
    mode = "linlog",
    preventOverlap = TRUE,
    prune = FALSE,
    tao = 0.05,
    width = 600,
    center = c(300, 200)
  )
  expect_type(force_atlas2_complex, "list")

  # Test fruchterman with all options
  fruchterman_complex <- fruchterman_layout(
    height = 500,
    width = 700,
    gravity = 15,
    speed = 3
  )
  expect_type(fruchterman_complex, "list")

  # Test radial with complex configuration
  radial_complex <- radial_layout(
    center = c(250, 250),
    focusNode = "central",
    height = 500,
    width = 500,
    nodeSize = 20,
    nodeSpacing = 12,
    linkDistance = 80,
    unitRadius = 60,
    maxIteration = 800,
    maxPreventOverlapIteration = 150,
    preventOverlap = TRUE,
    sortBy = "degree",
    sortStrength = 15,
    strictRadial = FALSE
  )
  expect_type(radial_complex, "list")

  # Test dendrogram with all options
  dendrogram_complex <- dendrogram_layout(
    direction = "RL",
    nodeSep = 40,
    rankSep = 300,
    radial = TRUE
  )
  expect_type(dendrogram_complex, "list")

  # Test combo_combined with complex configuration
  combo_combined_complex <- combo_combined_layout(
    center = c(200, 200),
    comboPadding = 25,
    nodeSize = 18,
    spacing = 8,
    treeKey = "parentId"
  )
  expect_type(combo_combined_complex, "list")
})

test_that("layout center parameter validation works", {
  # Test that center parameters accept numeric vectors of length 2 or 3
  expect_error(circular_layout(center = c(100, 200, 300, 400)))
  expect_error(concentric_layout(center = "center"))
  expect_error(radial_layout(center = c(100)))

  # Valid center configurations
  circular_center <- circular_layout(center = c(100, 200))
  expect_type(circular_center, "list")

  concentric_center <- concentric_layout(center = c(150, 150, 50))
  expect_type(concentric_center, "list")

  radial_center <- radial_layout(center = c(200, 100))
  expect_type(radial_center, "list")
})

test_that("layout nodeSize parameter validation works", {
  # Test nodeSize parameter variations
  concentric_nodeSize1 <- concentric_layout(nodeSize = 20)
  expect_type(concentric_nodeSize1, "list")

  concentric_nodeSize2 <- concentric_layout(nodeSize = c(30, 40))
  expect_type(concentric_nodeSize2, "list")

  # Remove the error expectations that don't actually error
  # concentric_layout apparently accepts these values
  concentric_nodeSize3 <- concentric_layout(nodeSize = c(30, 40, 50))
  expect_type(concentric_nodeSize3, "list")

  concentric_nodeSize4 <- concentric_layout(nodeSize = -10)
  expect_type(concentric_nodeSize4, "list")
})

test_that("layout numeric parameter bounds work", {
  # Test various numeric parameter bounds
  expect_error(circular_layout(angleRatio = 0))
  expect_error(circular_layout(divisions = 0))
  expect_error(force_atlas2_layout(kg = -1))
  expect_error(fruchterman_layout(gravity = -5))
  expect_error(radial_layout(maxIteration = 0))
  expect_error(dendrogram_layout(nodeSep = -1))

  # Valid configurations
  circular_valid <- circular_layout(angleRatio = 0.5, divisions = 4)
  expect_type(circular_valid, "list")

  force_atlas2_valid <- force_atlas2_layout(kg = 0.5)
  expect_type(force_atlas2_valid, "list")
})

test_that("dagre layout validation works correctly", {
  # Test dagre layout validation
  expect_error(dagre_layout(rankdir = "invalid"))
  expect_error(dagre_layout(align = "invalid"))
  expect_error(dagre_layout(nodesep = -10))
  expect_error(dagre_layout(ranksep = -20))
  expect_error(dagre_layout(ranker = "invalid"))
  d1 <- dagre_layout()
  expect_type(d1, "list")

  # Test match.arg cases for dagre layout
  dagre1 <- dagre_layout(
    rankdir = "LR",
    align = "UR",
    ranker = "tight-tree"
  )
  expect_type(dagre1, "list")

  # Test dagre with all options
  dagre_complex <- dagre_layout(
    rankdir = "TB",
    align = "DL",
    nodesep = 30,
    ranksep = 50,
    ranker = "longest-path"
  )
  expect_type(dagre_complex, "list")
})

#' @keywords internal
dropNulls <- function(x, except = character()) {
  keep <- !vapply(x, is.null, logical(1)) | names(x) %in% except
  x[keep]
}

#' Marks as string to be processed as a JS function
#'
#' Useful for htmlwidgets
#'
#' @param ... Any valid JS element.
#'
#' @export
#' @return A character vector with class "JS_EVAL" that can be used in htmlwidgets
#' to mark is as a JavaScript function.
JS <- function(...) {
  x <- c(...)
  if (is.null(x)) {
    return()
  }
  if (!is.character(x)) {
    stop("The arguments for JS() must be a character vector")
  }
  x <- paste(x, collapse = "\n")
  structure(x, class = unique(c("JS_EVAL", oldClass(x))))
}

#' @keywords internal
is_js <- function(x) {
  inherits(x, "JS_EVAL")
}

#' @keywords internal
validate_component <- function(x, mode) {
  valid_choices <- switch(
    mode,
    "layout" = valid_layouts,
    "behavior" = valid_behaviors,
    "plugin" = valid_plugins
  )

  # Allow to pass behavior as text
  if (!is.list(x)) {
    # Call with default values
    if (x %in% names(valid_choices)) {
      x <- valid_choices[[x]]()
    } else {
      stop(sprintf(
        "'%s' is not a valid %s. Valid choices are: %s.",
        x,
        mode,
        paste(names(valid_choices), collapse = ", ")
      ))
    }
  } else {
    if (is.null(x[["type"]]) || !(x[["type"]] %in% names(valid_choices))) {
      stop(sprintf(
        "'%s' is not a valid %s. Valid choices are: %s.",
        if (is.null(x[["type"]])) "unknown" else x[["type"]],
        mode,
        paste(names(valid_choices), collapse = ", ")
      ))
    }
  }
  x
}

lgl_ply <- function(x, fun, ..., length = 1L, use_names = FALSE) {
  vapply(x, fun, logical(length), ..., USE.NAMES = use_names)
}

# Used in zzz.R to register input handlers
register_selection_handler <- function(type) {
  shiny::registerInputHandler(paste0("g6R.", type), function(data, ...) {
    if (!length(data)) return(NULL)
    data <- unlist(data)
    attr(data, "eventType") <- type
    data
  }, force = TRUE)
}

test_that("dropNulls removes NULL elements", {
  input_list <- list(a = 1, b = NULL, c = 3, d = NULL)
  result <- dropNulls(input_list)
  expected <- list(a = 1, c = 3)
  expect_equal(result, expected)

  # Test with no NULLs
  no_null <- list(a = 1, b = 2)
  result_same <- dropNulls(no_null)
  expect_equal(result_same, no_null)
})

test_that("JS function creates JS_EVAL objects", {
  js_code <- "function() { return true; }"
  result <- JS(js_code)
  expect_s3_class(result, "JS_EVAL")
  expect_equal(as.character(result), js_code)

  # Test with multiple arguments
  js_multi <- JS("var a = 1;", "var b = 2;")
  expect_s3_class(js_multi, "JS_EVAL")
  expect_equal(as.character(js_multi), "var a = 1;\nvar b = 2;")
})

test_that("JS function handles edge cases", {
  # Test with NULL
  expect_null(JS(NULL))

  # Test with empty string
  result_empty <- JS("")
  expect_s3_class(result_empty, "JS_EVAL")
  expect_equal(as.character(result_empty), "")

  # Test error with non-character input
  expect_error(JS(123), "must be a character vector")
  expect_error(JS(list(a = 1)), "must be a character vector")
})

test_that("is_js identifies JS_EVAL objects", {
  js_obj <- JS("test")
  regular_string <- "test"

  expect_true(is_js(js_obj))
  expect_false(is_js(regular_string))
  expect_false(is_js(123))
  expect_false(is_js(NULL))
})

test_that("register_selection_handler works", {
  res <- register_selection_handler("test")
  expect_type(res, "closure")
  expect_null(res(list()))
  processed <- res(list(a = "res"))
  expect_named(processed, "a")
  expect_type(processed, "character")
  expect_identical(attr(processed, "eventType"), "test")
  expect_identical(processed[["a"]], "res")
  shiny::removeInputHandler("g6R.test")
})

#' Character network from "Les miserables" novel
#'
#' A dataset containing Les Misérables characters network, encoding interactions
#' between characters of Victor Hugo's novel. Two characters are connected
#' whenever they appear in the same chapter. This dataset was first created
#' by Donald Knuth as part of the Stanford Graph Base.
#' (https://people.sc.fsu.edu/~jburkardt/datasets/sgb/sgb.html).
#' It contains 77 nodes corresponding to characters of the novel, and 254 edges.
#'
#' @docType data
#'
#' @usage data(lesmis)
#'
#' @format A list with 2 data frames:
#' \describe{
#'   \item{nodes}{data frame with 77 rows for the nodes. Contains node labels and x/y coordinates.}
#'   \item{edges}{data frame with 254 rows for the edges. Contains souyrce/target and the number of times the interaction happened.}
#' }
#' @source \url{https://networks.skewed.de/net/lesmis}
"lesmis"


#' Example Network for radial layouts
#'
#' @docType data
#'
#' @usage data(radial)
#'
#' @format A list with 2 data frames:
#' \describe{
#'   \item{nodes}{data frame with 34 rows for the nodes.}
#'   \item{edges}{data frame with 58 rows for the edges.}
#' }
#' @source \url{https://assets.antv.antgroup.com/g6/radial.json}
"radial"


#' Example tree graph
#'
#' The graph contains a classification of algorithm categories
#'
#' @docType data
#'
#' @usage data(tree)
#'
#' @format A list with 2 data frames:
#' \describe{
#'   \item{nodes}{data frame with 31 rows for the nodes.}
#'   \item{edges}{data frame with 30 rows for the edges.}
#' }
#' @source \url{https://gw.alipayobjects.com/os/antvdemo/assets/data/algorithm-category.json}
"tree"


#' Example DAG graph
#'
#' The graph is a directed acyclic graph
#'
#' @docType data
#'
#' @usage data(dag)
#'
#' @format A list with 2 data frames:
#' \describe{
#'   \item{nodes}{data frame with 12 rows for the nodes.}
#'   \item{edges}{data frame with 12 rows for the edges.}
#'   \item{combo}{data frame with 3 rows for the combos.}
#' }
#' @source \url{https://gw.alipayobjects.com/os/antvdemo/assets/data/algorithm-category.json}
"dag"

#' Example pokemon data
#'
#' Pokemon evolution network to showcase combo features.
#'
#' @docType data
#'
#' @usage data(poke)
#'
#' @format A list with 3 nested lists:
#' \describe{
#'   \item{nodes}{list with 120 pokemon.}
#'   \item{edges}{list with 69 connections.}
#'   \item{combo}{list with 51 pokemon families.
#'    A family contains all evolutions of a pokemon.}
#' }
#' @source \url{https://pokeapi.co/docs/v2}
"poke"

#' Create G6 Graph Behaviors Configuration
#'
#' Configures interaction behaviors for a G6 graph visualization.
#' This function collects and combines multiple behavior configurations
#' into a list that can be passed to graph initialization functions.
#'
#' @param graph A g6 graph instance.
#' @param ... Behavior configuration objects created by behavior-specific functions.
#'   These can include any of the following behaviors:
#'
#'   \strong{Navigation behaviors:}
#'   \itemize{
#'     \item \code{drag_canvas()} - Drag the entire canvas view
#'     \item \code{zoom_canvas()} - Zoom the canvas view
#'     \item \code{scroll_canvas()} - Scroll the canvas using the wheel
#'     \item \code{optimize_viewport_transform()} - Optimize view transform performance
#'   }
#'
#'   \strong{Selection behaviors:}
#'   \itemize{
#'     \item \code{click_select()} - Click to select graph elements
#'     \item \code{brush_select()} - Select elements by dragging a rectangular area
#'     \item \code{lasso_select()} - Freely draw an area to select elements
#'   }
#'
#'   \strong{Editing behaviors:}
#'   \itemize{
#'     \item \code{create_edge()} - Interactively create new edges
#'     \item \code{drag_element()} - Drag nodes or combos
#'     \item \code{drag_element_force()} - Drag nodes in force-directed layout
#'   }
#'
#'   \strong{Data Exploration behaviors:}
#'   \itemize{
#'     \item \code{collapse_expand()} - Expand or collapse subtree nodes
#'     \item \code{focus_element()} - Focus on specific elements and automatically adjust the view
#'     \item \code{hover_activate()} - Highlight elements when hovering
#'   }
#'
#'   \strong{Visual Optimization behaviors:}
#'   \itemize{
#'     \item \code{fix_element_size()} - Fix the element size to a specified value
#'     \item \code{auto_adapt_label()} - Automatically adjust label position
#'   }
#'
#' @return A list of behavior configuration objects that can be passed to G6 graph initialization
#'
#' @note You can create custom behaviors from JavaScript and use them on the R side. See more
#' at \url{https://g6.antv.antgroup.com/en/manual/behavior/custom-behavior}.
#' @examples
#' # Create a basic set of behaviors
#' behaviors <- g6_behaviors(
#'   g6(),
#'   drag_canvas(),
#'   zoom_canvas(),
#'   click_select()
#' )
#'
#' # Create a more customized set of behaviors
#' behaviors <- g6_behaviors(
#'   g6(),
#'   drag_canvas(),
#'   zoom_canvas(sensitivity = 1.5),
#'   hover_activate(state = "highlight"),
#'   fix_element_size(
#'     node = list(
#'       list(shape = "circle", fields = c("r", "lineWidth"))
#'     )
#'   )
#' )
#'
#' @export
g6_behaviors <- function(graph, ...) {
  behaviors <- list(...)
  if (length(behaviors)) {
    graph$x$behaviors <- lapply(behaviors, validate_behavior)
  } else {
    stop("You must provide at least one behavior configuration.")
  }
  graph
}

#' @keywords internal
validate_behavior <- function(x) {
  validate_component(x, "behavior")
}

#' @keywords internal
valid_states <- c("selected", "active", "inactive", "disabled", "highlight")

#' Configure Auto Adapt Label Behavior
#'
#' Creates a configuration object for the auto-adapt-label behavior in G6.
#' This behavior automatically adjusts label positions to reduce overlapping and
#' improve readability in the graph visualization.
#'
#' @param key Unique identifier for the behavior (string, default: "auto-adapt-label").
#' @param enable Whether to enable this behavior (JS function, default: returns TRUE for all events).
#' @param throttle Throttle time in milliseconds to optimize performance (numeric, default: 100).
#' @param padding Padding space around labels in pixels (numeric, default: 0).
#' @param sort Global sorting rule for all element types (list or JS function, default: NULL).
#' @param sortNode Sorting rule specifically for node labels (list, default: list(type = "degree")).
#' @param sortEdge Sorting rule specifically for edge labels (list, default: NULL).
#' @param sortCombo Sorting rule specifically for combo labels (list, default: NULL).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/auto-adapt-label}.
#'
#' Sorting parameters determine which labels take priority when space is limited:
#' \itemize{
#'   \item When \code{sort} is provided, it applies to all element types and overrides type-specific settings
#'   \item Type-specific sorting (\code{sortNode}, \code{sortEdge}, \code{sortCombo}) only applies when \code{sort} is NULL
#'   \item The default sorting for nodes is by degree (higher degree nodes' labels are shown first)
#' }
#'
#' @return A list with the configuration settings for the auto-adapt-label behavior.
#' @export
#'
#' @examples
#' # Basic configuration with defaults
#' config <- auto_adapt_label()
#'
#' # Custom configuration with more padding and custom throttle
#' config <- auto_adapt_label(
#'   key = "my-label-adapter",
#'   throttle = 200,
#'   padding = 5
#' )
#'
#'
#' # Using a custom enable function
#' config <- auto_adapt_label(
#'   enable = JS("(e) => e.targetType === 'node'")
#' )
auto_adapt_label <- function(
  key = "auto-adapt-label",
  enable = TRUE,
  throttle = 100,
  padding = 0,
  sort = NULL,
  sortNode = list(type = "degree"),
  sortEdge = NULL,
  sortCombo = NULL,
  ...
) {
  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' must be a logical value or a JavaScript function wrapped by JS()"
    )
  }

  # 'throttle' should be a positive number
  if (!is.numeric(throttle) || length(throttle) != 1 || throttle < 0) {
    stop("'throttle' must be a single non-negative numeric value")
  }

  # 'padding' can be a single number or an array of numbers
  if (!is.numeric(padding)) {
    stop("'padding' must be a numeric value or a numeric vector")
  }

  # 'sort' is a custom sorting function or NULL
  if (!is.null(sort) && !is_js(sort)) {
    stop(
      "'sort' must be NULL or a JavaScript function wrapped by JS() that compares two elements"
    )
  }

  # 'sortNode' can be a list with configuration options or a function
  if (!is.null(sortNode) && !is.list(sortNode) && !is_js(sortNode)) {
    stop(
      "'sortNode' must be NULL, a list with configuration options, or a comparison JavaScript function wrapped by JS()"
    )
  }

  # If sortNode is a list and has a type field, validate it's one of the expected values
  if (is.list(sortNode) && !is.null(sortNode$type)) {
    valid_types <- c(
      "degree",
      "betweenness",
      "closeness",
      "eigenvector",
      "pagerank"
    )
    if (!(sortNode$type %in% valid_types)) {
      stop(
        "'sortNode$type' must be one of: ",
        paste(valid_types, collapse = ", ")
      )
    }
  }

  # 'sortEdge' can be a function or NULL
  if (!is.null(sortEdge) && !is_js(sortEdge)) {
    stop(
      "'sortEdge' must be NULL or a JavaScript function wrapped by JS() that compares two edges"
    )
  }

  # 'sortCombo' can be a function or NULL
  if (!is.null(sortCombo) && !is_js(sortCombo)) {
    stop(
      "'sortCombo' must be NULL or a JavaScript function wrapped by JS() that compares two combos"
    )
  }

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "auto-adapt-label"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Brush Selection Interaction
#'
#' Creates a configuration object for brush selection interaction in graph visualizations.
#' This function configures how elements are selected when using a brush selection tool.
#'
#' @param key Behavior unique identifier. Useful to modify this behavior from JS side.
#' @param animation Whether to enable animation (boolean, default: FALSE).
#' @param enable Whether to enable brush select functionality (boolean or function, default: TRUE).
#' @param enableElements Types of elements that can be selected (character vector, default: "node").
#' Can be \code{c("node", "edge", "combo")}.
#' @param immediately Whether to select immediately in default mode (boolean, default: FALSE).
#' @param mode Selection mode: "union", "intersect", "diff", or "default" (string, default: "default").
#' @param onSelect Callback for selected element state (JS function).
#' @param state State to switch to when selected (string, default: "selected").
#' @param style Style specification for the selection box (list).
#' See \url{https://g6.antv.antgroup.com/en/manual/behavior/brush-select#style}.
#' @param trigger Shortcut keys for selection (character vector).
#' @param outputId Manually pass the Shiny output ID. This is useful when the graph
#' is initialised outside the shiny render function and the ID cannot be automatically
#' inferred. This allows to set input values from the callback function
#' with the right namespace and graph ID. You must typically pass `session$ns("graphid")`
#' to ensure this also works in modules.
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/brush-select}.
#'
#' @return A list with the configuration settings for the brush select behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- brush_select()
#'
#' # Custom configuration
#' config <- brush_select(
#'   animation = TRUE,
#'   enableElements = c("node", "edge"),
#'   mode = "union",
#'   state = "highlight",
#'   style = list(
#'     fill = "rgba(0, 0, 255, 0.1)",
#'     stroke = "blue",
#'     lineWidth = 2
#'   ),
#'   trigger = c("Shift")
#' )
brush_select <- function(
  key = "brush-select",
  animation = FALSE,
  enable = JS(
    "(e) => {
      return true;
    }"
  ),
  enableElements = "node",
  immediately = FALSE,
  mode = c("default", "union", "intersect", "diff"),
  onSelect = NULL,
  state = c("selected", "active", "inactive", "disabled", "highlight"),
  style = NULL,
  trigger = "shift",
  outputId = NULL,
  ...
) {
  # Validate inputs
  if (!is.logical(animation)) {
    stop("'animation' should be a boolean value")
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop("'enable' should be a boolean or JS() function")
  }

  valid_elements <- c("node", "edge", "combo")
  if (!all(enableElements %in% valid_elements)) {
    stop(
      "'enableElements' should only contain: ",
      paste(valid_elements, collapse = ", ")
    )
  }

  if (!is.logical(immediately)) {
    stop("'immediately' should be a boolean value")
  }

  mode <- match.arg(mode)

  if (!is.null(onSelect) && !is_js(onSelect)) {
    stop("'onSelect' should be a JS function")
  }

  state <- match.arg(state)

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "brush-select"
  if (length(enableElements) == 1) {
    config$enableElements <- list(config$enableElements)
  }

  # We can't access the graph instance from within this JS
  # code as it is create oustide the widget factory
  if (is.null(config$outputId)) {
    config$outputId <- shiny::getCurrentOutputInfo()[["name"]]
  }

  # Provide default for onSelect for Shiny context
  if (is.null(config$onSelect) && !is.null(config$outputId)) {
    config$onSelect <- JS(
      sprintf(
        "(states) => {
          const selected = Object.getOwnPropertyNames(states);
          const nodes = [];
          const edges = [];
          const combos = [];
          const widget = HTMLWidgets.find('#%s').getWidget();
          if (selected.length === 0) return states;
          selected.forEach(id => {
            const type = widget.getElementType(id);
            if (type === 'node') nodes.push(id);
            else if (type === 'edge') edges.push(id);
            else if (type === 'combo') combos.push(id);
          });
          Shiny.setInputValue('%s-selected_node:g6R.brush_select', nodes, {priority: 'event'});
          Shiny.setInputValue('%s-selected_edge:g6R.brush_select', edges, {priority: 'event'});
          Shiny.setInputValue('%s-selected_combo:g6R.brush_select', combos, {priority: 'event'});
          return states;
        }",
        config$outputId,
        config$outputId,
        config$outputId,
        config$outputId
      )
    )
  }

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Click Select Behavior
#'
#' Creates a configuration object for the click-select behavior in G6.
#' This allows users to select graph elements by clicking.
#'
#' @param key Behavior unique identifier. Useful to modify this behavior from JS side.
#' @param animation Whether to enable animation effects when switching element states (boolean, default: TRUE).
#' @param degree Controls the highlight spread range (number or function, default: 0).
#' @param enable Whether to enable the click element function (boolean or function, default: TRUE).
#' @param multiple Whether to allow multiple selections (boolean, default: FALSE).
#' @param state The state applied when an element is selected (string, default: "selected").
#' @param neighborState The state applied to elements with n-degree relationships (string, default: "selected").
#' @param unselectedState The state applied to all other elements (string, default: NULL).
#' @param onClick Callback when an element is clicked (function, default: NULL).
#' @param trigger Keys for multi-selection (character vector, default: c("shift")).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/click-select}.
#'
#' @return A list with the configuration settings for the click select behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- click_select()
#'
#' # Custom configuration
#' config <- click_select(
#'   animation = FALSE,
#'   degree = 1,
#'   multiple = TRUE,
#'   state = "active",
#'   neighborState = "highlight",
#'   unselectedState = "inactive",
#'   trigger = c("Control")
#' )
#'
#' # Example leveraging the input[["<GRAPH_ID>-selected_<ELEMENT_TYPE>"]]
#' if (interactive()) {
#'   library(shiny)
#'   library(g6R)
#'   library(bslib)
#'
#'   nodes <- data.frame(id = c("node1", "node2"))
#'   edges <- data.frame(source = "node1", target = "node2")
#'   combos <- data.frame(id = "combo1", type = "rect")
#'
#'   ui <- page_fluid(
#'     g6_output("graph"),
#'     verbatimTextOutput("selected_elements")
#'   )
#'
#'   server <- function(input, output, session) {
#'     output$graph <- render_g6({
#'       g6(
#'         nodes = nodes,
#'         edges = edges,
#'         combos = combos
#'       ) |>
#'         g6_layout() |>
#'         g6_behaviors(
#'           click_select(multiple = TRUE),
#'           brush_select(
#'             enableElements = c("node", "edge", "combo"),
#'             immediately = TRUE
#'           )
#'         )
#'     })
#'
#'     output$selected_elements <- renderPrint({
#'       list(
#'         selected_nodes = input[["graph-selected_node"]],
#'         selected_edges = input[["graph-selected_edge"]],
#'         selected_combos = input[["graph-selected_combo"]]
#'       )
#'     })
#'   }
#'
#'   shinyApp(ui, server)
#' }
click_select <- function(
  key = "click-select",
  animation = TRUE,
  degree = 0,
  enable = TRUE,
  multiple = FALSE,
  state = c("selected", "active", "inactive", "disabled", "highlight"),
  neighborState = c("selected", "active", "inactive", "disabled", "highlight"),
  unselectedState = NULL,
  onClick = NULL,
  trigger = "shift",
  ...
) {
  # Validate inputs
  if (!is.logical(animation)) {
    stop("'animation' should be a boolean value")
  }

  if (!is.numeric(degree) && !is_js(degree)) {
    stop("'degree' should be a number or a JS function")
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop("'enable' should be a boolean or a JS function")
  }

  if (!is.logical(multiple)) {
    stop("'multiple' should be a boolean value")
  }

  state <- match.arg(state)
  neighborState <- match.arg(neighborState)

  if (
    !is.null(unselectedState) &&
      (!is.character(unselectedState) ||
        (length(unselectedState) == 1 && !unselectedState %in% valid_states))
  ) {
    stop(
      "'unselectedState' should be one of 'selected', 'active', 'inactive', 
      'disabled', 'highlight', or a custom string"
    )
  }

  if (!is.null(onClick) && !is_js(onClick)) {
    stop("'onClick' should be a JS function")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$trigger <- list(config$trigger)
  config$type <- "click-select"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Collapse Expand Behavior
#'
#' Creates a configuration object for the collapse-expand behavior in G6.
#' This allows users to collapse or expand nodes/combos with child elements.
#'
#' @param key Behavior unique identifier. Useful to modify this behavior from JS side.
#' @param animation Enable expand/collapse animation effects (boolean, default: TRUE).
#' @param enable Enable expand/collapse functionality (boolean or function, default: TRUE).
#' @param trigger Trigger method: "click" or "dblclick" (string, default: "dblclick").
#' @param onCollapse Callback function when collapse is completed (function, default: NULL).
#' @param onExpand Callback function when expand is completed (function, default: NULL).
#' @param align Align with the target element to avoid view offset (boolean, default: TRUE).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/collapse-expand}.
#'
#' @return A list with the configuration settings for the collapse-expand behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- collapse_expand()
collapse_expand <- function(
  key = "collapse-expand",
  animation = TRUE,
  enable = TRUE,
  trigger = "dblclick",
  onCollapse = NULL,
  onExpand = NULL,
  align = TRUE,
  ...
) {
  # Validate inputs
  if (!is.logical(animation)) {
    stop("'animation' should be a boolean value")
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop("'enable' should be a boolean or a JS function")
  }

  if (!trigger %in% c("click", "dblclick")) {
    stop("'trigger' should be one of 'click' or 'dblclick'")
  }

  if (!is.null(onCollapse) && !is_js(onCollapse)) {
    stop("'onCollapse' should be a JS function")
  }

  if (!is.null(onExpand) && !is_js(onExpand)) {
    stop("'onExpand' should be a JS function")
  }

  if (!is.logical(align)) {
    stop("'align' should be a boolean value")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "collapse-expand"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Create Edge Behavior
#'
#' Creates a configuration object for the create-edge behavior in G6.
#' This allows users to create edges between nodes by clicking or dragging.
#'
#' @param key Behavior unique identifier. Useful to modify this behavior from JS side.
#' @param trigger The way to trigger edge creation: "click" or "drag" (string, default: "drag").
#' @param enable Whether to enable this behavior (boolean or function, default: FALSE).
#' Our default implementation works in parallel with the \link{context_menu} plugin which is
#' responsible for activating the edge behavior when edge creation is selected.
#' @param onCreate Callback function for creating an edge, returns edge data (function, default: NULL).
#' @param onFinish Callback function for successfully creating an edge (function).
#' By default, we provide an internal implementation that disables the edge mode when the edge
#' creation is succesful so that it does not conflict with other drag behaviors.
#' @param style Style of the newly created edge (list, default: NULL).
#' @param notify Whether to show a feedback message in the ui.
#' @param outputId Manually pass the Shiny output ID. This is useful when the graph
#' is initialised outside the shiny render function and the ID cannot be automatically
#' inferred. This allows to set input values from the callback function
#' with the right namespace and graph ID. You must typically pass `session$ns("graphid")`
#' to ensure this also works in modules.
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/create-edge}.
#'
#' @return A list with the configuration settings for the create-edge behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- create_edge()
#' if (interactive()) {
#'   library(shiny)
#'   library(bslib)
#'   library(g6R)
#'
#'   nodes <- list(
#'     list(
#'       id = "node1"
#'     ),
#'     list(
#'       id = "node2"
#'     )
#'   )
#'
#'   modUI <- function(id) {
#'     ns <- NS(id)
#'     tagList(
#'       g6Output(ns("graph"))
#'     )
#'   }
#'
#'   modServer <- function(id) {
#'     moduleServer(id, function(input, output, session) {
#'       output$graph <- renderG6({
#'         g6(nodes) |>
#'           g6_options(
#'             animation = FALSE,
#'             edge = edge_options(
#'               style = list(
#'                 endArrow = TRUE
#'               )
#'             )
#'           ) |>
#'           g6_layout(d3_force_layout()) |>
#'           g6_behaviors(
#'             # avoid conflict with internal function
#'             g6R::create_edge(
#'               target = c("node", "combo", "canvas"),
#'               enable = JS(
#'                 "(e) => {
#'                   return e.shiftKey;
#'                 }"
#'               ),
#'               onFinish = JS(
#'                 sprintf(
#'                   "(edge) => {
#'                     const graph = HTMLWidgets.find('#%s').getWidget();
#'                     // Avoid to create edges in combos. If so, we remove it
#'                     if (edge.targetType === 'combo') {
#'                       graph.removeEdgeData([edge.id]);
#'                       return;
#'                     }
#'                     Shiny.setInputValue('%s', edge);
#'                   }",
#'                   session$ns("graph"),
#'                   session$ns("added_edge")
#'                 )
#'               )
#'             )
#'           )
#'       })
#'
#'       observeEvent(input[["added_edge"]], {
#'         showNotification(
#'           sprintf("Edge dropped on: %s", input[["added_edge"]]$targetType),
#'           type = "message"
#'         )
#'       })
#'     })
#'   }
#'
#'   ui <- page_fluid(
#'     modUI("test")
#'   )
#'
#'   server <- function(input, output, session) {
#'     modServer("test")
#'   }
#'
#'   shinyApp(ui, server)
#' }
create_edge <- function(
  key = "create-edge",
  trigger = "drag",
  enable = FALSE,
  onCreate = NULL,
  onFinish = NULL,
  style = NULL,
  notify = FALSE,
  outputId = NULL,
  ...
) {
  # Validate inputs
  if (!trigger %in% c("click", "drag")) {
    stop("'trigger' should be one of 'click' or 'drag'")
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop("'enable' should be a boolean or a JS function")
  }

  if (!is.null(onCreate) && !is_js(onCreate)) {
    stop("'onCreate' should be a JS function")
  }

  if (!is.null(onFinish) && !is_js(onFinish)) {
    stop("'onFinish' should be a JS function")
  }

  if (!is.null(style) && !is.list(style)) {
    stop("'style' should be a list")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "create-edge"

  # We can't access the graph instance from within this JS
  # code as it is create oustide the widget factory
  if (is.null(config$outputId)) {
    config$outputId <- shiny::getCurrentOutputInfo()[["name"]]
  }

  # Provide default in Shiny context only
  if (is.null(config$onFinish) && !is.null(config$outputId)) {
    config$onFinish <- JS(
      sprintf(
        "(edge) => {
          // Canvas drops don't have a real target node, skip processing
          const notify = %s;
          const graph = HTMLWidgets.find('#%s').getWidget();
          // Avoid to create edges in combos. If so, we remove it
          if (edge.targetType === 'combo') {
            graph.removeEdgeData([edge.id]);
            if (notify) {
              Shiny.notifications.show(
                { 
                  html: 'Edge can only created bewteen 2 nodes',
                  type: 'error' 
                }
              )
            }
          } else {
            // Re-enable drag element behaviors is done from the custom behavior class
            // in the customDrop method.
            if (notify) {
              Shiny.notifications.show(
                { 
                  html: 'Edge successfuly created',
                  type: 'success' 
                }
              )
            }
          }
        }",
        as.numeric(notify),
        config$outputId
      )
    )
  }

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Drag Canvas Behavior
#'
#' Creates a configuration object for the drag-canvas behavior in G6.
#' This allows users to drag the canvas to pan the view.
#'
#' @param key Behavior unique identifier. Useful to modify this behavior from JS side.
#' @param enable Whether to enable this behavior (boolean or function, default: function that enables dragging only on canvas).
#' @param animation Drag animation configuration for keyboard movement (list, default: NULL).
#' @param direction Allowed drag direction: "x", "y", or "both" (string, default: "both").
#' @param range Draggable viewport range in viewport size units (number or numeric vector, default: Inf).
#' @param sensitivity Distance to trigger a single keyboard movement (number, default: 10).
#' @param trigger Keyboard keys to trigger dragging (list, default: NULL).
#' @param onFinish Callback function when dragging is completed (function, default: NULL).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/drag-canvas}.
#'
#' @return A list with the configuration settings for the drag-canvas behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- drag_canvas()
#'
#' # Custom configuration
#' config <- drag_canvas(
#'   enable = TRUE,
#'   direction = "x",
#'   range = c(-100, 100),
#'   sensitivity = 5,
#'   trigger = list(
#'    up = "ArrowUp",
#'    down = "ArrowDown",
#'    left = "ArrowLeft",
#'    right = "ArrowRight"
#'   )
#' )
drag_canvas <- function(
  key = "drag-canvas",
  enable = NULL,
  animation = NULL,
  direction = c("both", "x", "y"),
  range = NULL,
  sensitivity = 10,
  trigger = NULL,
  onFinish = NULL,
  ...
) {
  # Validate inputs
  if (is.null(enable)) {
    enable <- JS(
      "(e) => {
        return e.targetType === 'canvas';
      }"
    )
  }
  if (!is.logical(enable) && !is_js(enable)) {
    stop("'enable' should be a boolean or a function")
  }

  if (!is.null(animation) && !is.list(animation)) {
    stop("'animation' should be a list of animation configuration options")
  }

  direction <- match.arg(direction)

  if (!is.null(range) && !is.numeric(range)) {
    stop("'range' should be a number, numeric vector, or Inf")
  }

  if (!is.numeric(sensitivity) || sensitivity <= 0) {
    stop("'sensitivity' should be a positive number")
  }

  if (!is.null(trigger) && !is.list(trigger)) {
    stop("'trigger' should be a list of keyboard trigger configuration")
  }

  if (!is.null(onFinish) && !is_js(onFinish)) {
    stop("'onFinish' should be a JS function")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "drag-canvas"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Drag Element Behavior
#'
#' Creates a configuration object for the drag-element behavior in G6.
#' This allows users to drag nodes and combos in the graph.
#'
#' @param key Unique identifier for the behavior, used for subsequent operations (string, default: NULL)
#' @param enable Whether to enable the drag function (boolean or function, default: function that enables dragging for nodes and combos).
#' @param animation Whether to enable drag animation (boolean, default: TRUE).
#' @param state Identifier for the selected state of nodes (string, default: "selected").
#' @param dropEffect Defines the operation effect after dragging ends: "link", "move", or "none" (string, default: "move").
#' @param hideEdge Controls the display state of edges during dragging: "none", "out", "in", "both", or "all" (string, default: "none").
#' @param shadow Whether to enable ghost nodes (boolean, default: FALSE).
#' @param cursor Customize the mouse style during dragging (list, default: NULL).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/drag-element}.
#'
#' @return A list with the configuration settings for the drag-element behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- drag_element()
#'
#' # Custom configuration
#' config <- drag_element(
#'   key = "my-drag-behavior",
#'   animation = FALSE,
#'   dropEffect = "link",
#'   hideEdge = "both",
#'   shadow = TRUE,
#'   cursor = list(
#'     default = "default",
#'     grab = "grab",
#'     grabbing = "grabbing"
#'   ),
#'   enable = JS(
#'    "(e) => {
#'      return e.targetType === 'node' || e.targetType === 'combo';
#'    }"
#'   )
#' )
drag_element <- function(
  key = "drag-element",
  enable = NULL,
  animation = TRUE,
  state = "selected",
  dropEffect = c("move", "link", "none"),
  hideEdge = c("none", "out", "in", "both", "all"),
  shadow = FALSE,
  cursor = NULL,
  ...
) {
  # Validate inputs
  if (is.null(enable)) {
    enable <- JS(
      "(e) => {
        const target = e.nativeEvent?.target;
        const graph = HTMLWidgets.find(`#${target?.closest?.('.g6')?.id}`)?.getWidget();
        try {
          if (graph?.getNodeData?.('g6-create-edge-assist-node-id')) return false;
        } catch (err) {}
        return true;
      }"
    )
  }
  if (!is.logical(enable) && !is_js(enable)) {
    stop("'enable' should be a boolean or a JS function")
  }

  if (!is.logical(animation)) {
    stop("'animation' should be a boolean value")
  }

  if (!is.character(state)) {
    stop("'state' should be a string")
  }

  dropEffect <- match.arg(dropEffect)
  hideEdge <- match.arg(hideEdge)

  if (!is.logical(shadow)) {
    stop("'shadow' should be a boolean value")
  }

  if (!is.null(cursor) && !is.list(cursor)) {
    stop("'cursor' should be a list of cursor style configurations")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "drag-element"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Drag Element Force Behavior
#'
#' Creates a configuration object for the drag-element-force behavior in G6.
#' This allows users to drag nodes and combos with force-directed layout interactions.
#'
#' @param key Unique identifier for the behavior, used for subsequent operations (string, default: "drag-element-force").
#' @param fixed Whether to keep the node position fixed after dragging ends (boolean, default: FALSE).
#' @param enable Whether to enable the drag function (boolean or JS function, default: JS function that enables dragging for nodes and combos).
#' @param state Identifier for the selected state of nodes (string, default: "selected").
#' @param hideEdge Controls the display state of edges during dragging: "none", "out", "in", "both", or "all" (string, default: "none").
#' @param cursor Customize the mouse style during dragging (list, default: NULL).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/drag-element-force}.
#'
#' @return A list with the configuration settings for the drag-element-force behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- drag_element_force()
#'
#' # Custom configuration with JavaScript arrow function and custom key
#' config <- drag_element_force(
#'   key = "my-custom-drag-force",
#'   fixed = TRUE,
#'   enable = JS("(event) => { return event.targetType === 'node'; }"),
#'   hideEdge = "both",
#'   cursor = list(
#'     default = "default",
#'     grab = "grab",
#'     grabbing = "grabbing"
#'   )
#' )
drag_element_force <- function(
  key = "drag-element-force",
  fixed = FALSE,
  enable = NULL,
  state = "selected",
  hideEdge = c("none", "out", "in", "both", "all"),
  cursor = NULL,
  ...
) {
  # Validate inputs
  if (!is.logical(fixed)) {
    stop("'fixed' should be a boolean value")
  }

  # Enable default
  if (is.null(enable)) {
    enable <- JS(
      "(e) => {
        const target = e.nativeEvent?.target;
        const graph = HTMLWidgets.find(`#${target?.closest?.('.g6')?.id}`)?.getWidget();
        try {
          if (graph?.getNodeData?.('g6-create-edge-assist-node-id')) return false;
        } catch (err) {}
        return ['node', 'combo'].includes(e.targetType);
      }"
    )
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  if (!is.character(state)) {
    stop("'state' should be a string")
  }

  hideEdge <- match.arg(hideEdge)

  if (!is.null(cursor) && !is.list(cursor)) {
    stop("'cursor' should be a list of cursor style configurations")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "drag-element-force"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Fix Element Size Behavior
#'
#' Creates a configuration object for the fix-element-size behavior in G6.
#' This allows maintaining fixed visual sizes for elements during zoom operations.
#'
#' @param key Unique identifier for the behavior, used for subsequent operations
#' (string, default: "fix-element-size").
#' @param enable Whether to enable this interaction (boolean or JS function, default: TRUE).
#' @param reset Whether to restore style when elements are redrawn (boolean, default: FALSE).
#' @param state Specify the state of elements to fix size (string, default: "")
#' @param node Node configuration item(s) to define which attributes maintain fixed
#' size (list or array of lists, default: NULL).
#' @param nodeFilter Node filter to determine which nodes maintain fixed size
#' (JS function, default: returns TRUE for all nodes).
#' @param edge Edge configuration item(s) to define which attributes maintain fixed size
#' (list or array of lists, default: predefined list).
#' @param edgeFilter Edge filter to determine which edges maintain fixed size
#' (JS function, default: returns TRUE for all edges).
#' @param combo Combo configuration item(s) to define which attributes maintain fixed size
#' (list or array of lists, default: NULL).
#' @param comboFilter Combo filter to determine which combos maintain fixed size
#' (JS function, default: returns TRUE for all combos).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/fix-element-size}.
#'
#' @return A list with the configuration settings for the fix-element-size behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- fix_element_size()
#'
#' # Custom configuration with filters and specific shape configurations
#' config <- fix_element_size(
#'   key = "my-fix-size-behavior",
#'   reset = TRUE,
#'   state = "active",
#'   node = list(
#'     list(shape = "circle", fields = c("r", "lineWidth")),
#'     list(shape = "label", fields = c("fontSize"))
#'   ),
#'   nodeFilter = JS("(node) => node.type === 'important'"),
#'   edge = list(shape = "line", fields = c("lineWidth", "lineDash")),
#'   edgeFilter = JS("(edge) => edge.weight > 5")
#' )
fix_element_size <- function(
  key = "fix-element-size",
  enable = TRUE,
  reset = FALSE,
  state = "",
  node = NULL,
  nodeFilter = JS("() => true"),
  edge = list(
    list(shape = "key", fields = c("lineWidth")),
    list(shape = "halo", fields = c("lineWidth")),
    list(shape = "label", fields = c("fontSize"))
  ),
  edgeFilter = JS("() => true"),
  combo = NULL,
  comboFilter = JS("() => true"),
  ...
) {
  # Validate inputs
  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  if (!is.logical(reset)) {
    stop("'reset' should be a boolean value")
  }

  if (!is.character(state)) {
    stop("'state' should be a string")
  }

  # Validate node configuration
  if (!is.null(node) && !is.list(node)) {
    stop("'node' should be a list or array of configuration objects")
  }

  # Validate node filter
  if (!is_js(nodeFilter)) {
    stop(
      "'nodeFilter' should be a JavaScript function wrapped with JS()"
    )
  }

  # Validate edge configuration
  if (!is.null(edge) && !is.list(edge)) {
    stop("'edge' should be a list or array of configuration objects")
  }

  # Validate edge filter
  if (!is_js(edgeFilter)) {
    stop(
      "'edgeFilter' should be a JavaScript function wrapped with JS()"
    )
  }

  # Validate combo configuration
  if (!is.null(combo) && !is.list(combo)) {
    stop("'combo' should be a list or array of configuration objects")
  }

  # Validate combo filter
  if (!is_js(comboFilter)) {
    stop(
      "'comboFilter' should be a JavaScript function wrapped with JS()"
    )
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "fix-element-size"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Focus Element Behavior
#'
#' Creates a configuration object for the focus-element behavior in G6.
#' This behavior allows focusing on specific elements by automatically adjusting the viewport.
#'
#' @param key Unique identifier for the behavior, used for subsequent operations
#' (string, default: "focus-element").
#' @param animation Focus animation settings (list, default: list with duration 500ms and easing "ease-in").
#' @param enable Whether to enable the focus feature (boolean or JS function, default: TRUE).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/focus-element}.
#'
#' @return A list with the configuration settings for the focus-element behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- focus_element()
#'
#' # Custom configuration
#' config <- focus_element(
#'   key = "my-focus-behavior",
#'   animation = list(duration = 1000, easing = "ease-out"),
#'   enable = JS("(event) => event.targetType === 'node'")
#' )
focus_element <- function(
  key = "focus-element",
  animation = list(duration = 500, easing = "ease-in"),
  enable = TRUE,
  ...
) {
  # Validate inputs
  if (!is.list(animation)) {
    stop("'animation' should be a list with animation settings")
  }

  # Define valid easing values
  valid_easing_values <- c("ease-in-out", "ease-in", "ease-out", "linear")

  if (is.list(animation)) {
    if (!is.null(animation$duration) && !is.numeric(animation$duration)) {
      stop("'animation$duration' should be a number")
    }

    if (!is.null(animation$easing)) {
      if (!is.character(animation$easing)) {
        stop("'animation$easing' should be a string")
      }

      if (!animation$easing %in% valid_easing_values) {
        stop(
          "'animation$easing' should be one of: ",
          paste(valid_easing_values, collapse = ", ")
        )
      }
    }
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "focus-element"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Hover Activate Behavior
#'
#' Creates a configuration object for the hover-activate behavior in G6.
#' This behavior activates elements when the mouse hovers over them.
#'
#' @param key Unique identifier for the behavior (string, default: "hover-activate").
#' @param animation Whether to enable animation (boolean, default: TRUE).
#' @param enable Whether to enable hover feature (boolean or JS function, default: TRUE).
#' @param degree Degree of relationship to activate elements (number or JS function, default: 0).
#' @param direction Specify edge direction: "both", "in", or "out" (string, default: "both").
#' @param state State of activated elements (string, default: "active").
#' @param inactiveState State of inactive elements (string, default: NULL).
#' @param onHover Callback when element is hovered (JS function, default: NULL).
#' @param onHoverEnd Callback when hover ends (JS function, default: NULL).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/hover-activate}.
#'
#' @return A list with the configuration settings for the hover-activate behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- hover_activate()
#'
#' # Custom configuration
#' config <- hover_activate(
#'   key = "my-hover-behavior",
#'   animation = FALSE,
#'   degree = 1,
#'   direction = "out",
#'   state = "highlight",
#'   inactiveState = "inactive",
#'   onHover = JS("(event) => { console.log('Hover on:', event.target.id); }")
#' )
hover_activate <- function(
  key = "hover-activate",
  animation = TRUE,
  enable = TRUE,
  degree = 0,
  direction = c("both", "in", "out"),
  state = "active",
  inactiveState = NULL,
  onHover = NULL,
  onHoverEnd = NULL,
  ...
) {
  # Validate inputs
  if (!is.logical(animation)) {
    stop("'animation' should be a boolean value")
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  if (!is.numeric(degree) && !is_js(degree)) {
    stop(
      "'degree' should be a number or a JavaScript function wrapped with JS()"
    )
  }

  direction <- match.arg(direction)

  if (!is.character(state)) {
    stop("'state' should be a string")
  }

  if (!is.null(inactiveState) && !is.character(inactiveState)) {
    stop("'inactiveState' should be a string")
  }

  if (!is.null(onHover) && !is_js(onHover)) {
    stop(
      "'onHover' should be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(onHoverEnd) && !is_js(onHoverEnd)) {
    stop(
      "'onHoverEnd' should be a JavaScript function wrapped with JS()"
    )
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "hover-activate"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Lasso Select Behavior
#'
#' Creates a configuration object for the lasso-select behavior in G6.
#' This behavior allows selecting elements by drawing a lasso around them.
#'
#' @param key Unique identifier for the behavior (string, default: "lasso-select").
#' @param animation Whether to enable animation (boolean, default: FALSE).
#' @param enable Whether to enable lasso selection (boolean or JS function, default: TRUE).
#' @param enableElements Types of elements that can be selected (character vector, default: c("node", "combo", "edge")).
#' @param immediately Whether to select immediately, only effective when selection mode is default (boolean, default: FALSE).
#' @param mode Selection mode: "union", "intersect", "diff", or "default" (string, default: "default").
#' @param onSelect Callback for selected element state (JS function, default: NULL).
#' @param state State to switch to when selected (string, default: "selected").
#' @param style Style of the lasso during selection (list, default: NULL).
#' @param trigger Press this shortcut key along with mouse click to select (character vector, default: c("shift")).
#' @param outputId Manually pass the Shiny output ID. This is useful when the graph
#' is initialised outside the shiny render function and the ID cannot be automatically
#' inferred. This allows to set input values from the callback function
#' with the right namespace and graph ID. You must typically pass `session$ns("graphid")`
#' to ensure this also works in modules.
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/lasso-select}.
#'
#' @return A list with the configuration settings for the lasso-select behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- lasso_select()
#'
#' # Custom configuration
#' config <- lasso_select(
#'   key = "my-lasso-select",
#'   animation = TRUE,
#'   enableElements = c("node", "combo"),
#'   mode = "union",
#'   state = "highlight",
#'   trigger = c("control"),
#'   style = list(
#'     stroke = "#1890FF",
#'     lineWidth = 2,
#'     fillOpacity = 0.1
#'   )
#' )
lasso_select <- function(
  key = "lasso-select",
  animation = FALSE,
  enable = TRUE,
  enableElements = "node",
  immediately = FALSE,
  mode = c("default", "union", "intersect", "diff"),
  onSelect = NULL,
  state = "selected",
  style = NULL,
  trigger = c("shift"),
  outputId = NULL,
  ...
) {
  # Validate inputs
  if (!is.character(key)) {
    stop("'key' should be a string")
  }

  if (!is.logical(animation)) {
    stop("'animation' should be a boolean value")
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  valid_elements <- c("node", "edge", "combo")
  if (!all(enableElements %in% valid_elements)) {
    stop(
      "'enableElements' should only contain: ",
      paste(valid_elements, collapse = ", ")
    )
  }

  if (!is.logical(immediately)) {
    stop("'immediately' should be a boolean value")
  }

  mode <- match.arg(mode)

  if (!is.null(onSelect) && !is_js(onSelect)) {
    stop(
      "'onSelect' should be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.character(state)) {
    stop("'state' should be a string")
  }

  if (!is.null(style) && !is.list(style)) {
    stop("'style' should be a list of style properties")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "lasso-select"
  if (length(enableElements) == 1) {
    config$enableElements <- list(config$enableElements)
  }

  # Recover output id in Shiny context
  if (is.null(config$outputId)) {
    config$outputId <- shiny::getCurrentOutputInfo()[["name"]]
  }

  if (is.null(config$onSelect) && !is.null(config$outputId)) {
    config$onSelect <- JS(
      sprintf(
        "(states) => {
          const selected = Object.getOwnPropertyNames(states);
          const nodes = [];
          const edges = [];
          const combos = [];
          const widget = HTMLWidgets.find('#%s').getWidget();
          if (selected.length === 0) return states;
          selected.forEach(id => {
            const type = widget.getElementType(id);
            if (type === 'node') nodes.push(id);
            else if (type === 'edge') edges.push(id);
            else if (type === 'combo') combos.push(id);
          });
          Shiny.setInputValue('%s-selected_node:g6R.lasso_select', nodes, {priority: 'event'});
          Shiny.setInputValue('%s-selected_edge:g6R.lasso_select', edges, {priority: 'event'});
          Shiny.setInputValue('%s-selected_combo:g6R.lasso_select', combos, {priority: 'event'});
          return states;
        }",
        config$outputId,
        config$outputId,
        config$outputId,
        config$outputId
      )
    )
  }

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Optimize Viewport Transform Behavior
#'
#' Creates a configuration object for the optimize-viewport-transform behavior in G6.
#' This behavior improves performance during viewport transformations by temporarily
#' hiding certain elements.
#'
#' @param key Unique identifier for the behavior (string, default: "optimize-viewport-transform").
#' @param enable Whether to enable this behavior (boolean or JS function, default: TRUE).
#' @param debounce How long after the operation ends to restore the visibility of all elements
#' in milliseconds (number, default: 200).
#' @param shapes Function to specify which graphical elements should remain visible during canvas
#' operations (JS function, default: returns TRUE for nodes).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/optimize-viewport-transform}.
#'
#' @return A list with the configuration settings for the optimize-viewport-transform behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- optimize_viewport_transform()
#'
#' # Custom configuration
#' config <- optimize_viewport_transform(
#'   key = "my-optimize-transform",
#'   debounce = 500,
#'   shapes = JS("(type) => type === 'node' || type === 'edge'")
#' )
#'
#' # With conditional enabling
#' config <- optimize_viewport_transform(
#'   enable = JS("(event) => event.getCurrentTransform().zoom < 0.5")
#' )
optimize_viewport_transform <- function(
  key = "optimize-viewport-transform",
  enable = TRUE,
  debounce = 200,
  shapes = JS("(type) => type === 'node'"),
  ...
) {
  # Validate inputs
  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  if (!is.numeric(debounce) || debounce < 0) {
    stop("'debounce' should be a non-negative number")
  }

  if (!is_js(shapes)) {
    stop(
      "'shapes' should be a JavaScript function wrapped with JS()"
    )
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "optimize-viewport-transform"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Scroll Canvas Behavior
#'
#' Creates a configuration object for the scroll-canvas behavior in G6.
#' This behavior allows scrolling the canvas with mouse wheel or keyboard.
#'
#' @param key Unique identifier for the behavior (string, default: "scroll-canvas").
#' @param enable Whether to enable this behavior (boolean or JS function, default: TRUE).
#' @param direction Allowed scrolling direction: "x", "y", or NULL for no limit
#' (string or NULL, default: NULL).
#' @param range Scrollable viewport range in viewport size units (numeric or numeric vector, default: 1).
#' @param sensitivity Scrolling sensitivity, the larger the value, the faster the scrolling
#' (numeric, default: 1).
#' @param trigger Keyboard shortcuts to trigger scrolling (list, default: NULL).
#' @param onFinish Callback function when scrolling is finished (JS function, default: NULL).
#' @param preventDefault Whether to prevent the browser's default event (boolean, default: TRUE).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/scroll-canvas}.
#'
#' @return A list with the configuration settings for the scroll-canvas behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- scroll_canvas()
#'
#' # Custom configuration
#' config <- scroll_canvas(
#'   key = "my-scroll-behavior",
#'   direction = "x",
#'   range = c(-2, 2),
#'   sensitivity = 1.5,
#'   preventDefault = FALSE
#' )
#'
#' # With keyboard triggers and callback
#' config <- scroll_canvas(
#'   enable = JS("(event) => !event.altKey"),
#'   trigger = list(
#'     up = "w",
#'     down = "s",
#'     left = "a",
#'     right = "d"
#'   ),
#'   onFinish = JS("() => { console.log('Scrolling finished'); }")
#' )
scroll_canvas <- function(
  key = "scroll-canvas",
  enable = TRUE,
  direction = NULL,
  range = 1,
  sensitivity = 1,
  trigger = NULL,
  onFinish = NULL,
  preventDefault = TRUE,
  ...
) {
  # Validate inputs
  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  valid_directions <- c("x", "y", NULL)
  if (!is.null(direction) && !direction %in% c("x", "y")) {
    stop("'direction' should be one of: 'x', 'y', or NULL")
  }

  if (!is.numeric(range) && !is.numeric(as.vector(range))) {
    stop("'range' should be a number or a numeric vector")
  }

  if (!is.numeric(sensitivity) || sensitivity <= 0) {
    stop("'sensitivity' should be a positive number")
  }

  if (!is.null(trigger) && !is.list(trigger)) {
    stop("'trigger' should be a list of keyboard shortcuts")
  }

  if (!is.null(onFinish) && !is_js(onFinish)) {
    stop(
      "'onFinish' should be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.logical(preventDefault)) {
    stop("'preventDefault' should be a boolean value")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "scroll-canvas"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' Configure Zoom Canvas Behavior
#'
#' Creates a configuration object for the zoom-canvas behavior in G6.
#' This behavior allows zooming the canvas with mouse wheel or keyboard shortcuts.
#'
#' @param key Unique identifier for the behavior (string, default: "zoom-canvas").
#' @param animation Zoom animation effect settings (list, default: list with duration 200ms).
#' @param enable Whether to enable this behavior (boolean or JS function, default: TRUE).
#' @param origin Zoom center point in viewport coordinates (list with x, y values, default: NULL).
#' @param onFinish Callback function when zooming is finished (JS function, default: NULL).
#' @param preventDefault Whether to prevent the browser's default event (boolean, default: TRUE).
#' @param sensitivity Zoom sensitivity, the larger the value, the faster the zoom (numeric, default: 1).
#' @param trigger How to trigger zooming, supports mouse wheel and keyboard shortcuts (list, default: NULL).
#' @param ... Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/zoom-canvas}.
#'
#' @return A list with the configuration settings for the zoom-canvas behavior.
#' @export
#'
#' @examples
#' # Basic configuration
#' config <- zoom_canvas()
#'
#' # Custom configuration
#' config <- zoom_canvas(
#'   key = "my-zoom-behavior",
#'   animation = list(duration = 300, easing = "ease-in-out"),
#'   origin = list(x = 0, y = 0),
#'   sensitivity = 1.5,
#'   preventDefault = FALSE
#' )
#'
#' # With keyboard triggers and callback
#' config <- zoom_canvas(
#'   enable = JS("(event) => !event.altKey"),
#'   trigger = list(
#'     zoomIn = "+",
#'     zoomOut = "-",
#'     reset = "0"
#'   ),
#'   onFinish = JS("() => { console.log('Zooming finished'); }")
#' )
zoom_canvas <- function(
  key = "zoom-canvas",
  animation = list(duration = 200),
  enable = TRUE,
  origin = NULL,
  onFinish = NULL,
  preventDefault = TRUE,
  sensitivity = 1,
  trigger = NULL,
  ...
) {
  # Validate inputs
  if (!is.character(key)) {
    stop("'key' should be a string")
  }

  if (!is.list(animation)) {
    stop("'animation' should be a list with animation settings")
  }

  if (is.list(animation) && !is.null(animation$duration)) {
    if (!is.numeric(animation$duration) || animation$duration < 0) {
      stop("'animation$duration' should be a non-negative number")
    }
  }

  if (!is.logical(enable) && !is_js(enable)) {
    stop(
      "'enable' should be a boolean or a JavaScript function wrapped with JS()"
    )
  }

  if (!is.null(origin)) {
    if (!is.list(origin)) {
      stop("'origin' should be a list")
    }
  }

  if (!is.null(onFinish) && !is_js(onFinish)) {
    stop(
      "'onFinish' should be a JavaScript function wrapped with JS()"
    )
  }

  if (!is.logical(preventDefault)) {
    stop("'preventDefault' should be a boolean value")
  }

  if (!is.numeric(sensitivity) || sensitivity <= 0) {
    stop("'sensitivity' should be a positive number")
  }

  if (!is.null(trigger) && !is.list(trigger)) {
    stop("'trigger' should be a list of trigger configurations")
  }

  # Create the configuration list with internal type parameter
  arg_names <- names(formals())
  arg_names <- arg_names[arg_names != "..."]
  # Get values of only the named parameters
  config <- mget(arg_names)
  config$type <- "zoom-canvas"

  # Drop NULL elements
  dropNulls(c(config, list(...)))
}

#' @keywords internal
valid_behaviors <- c(
  "auto-adapt-label" = auto_adapt_label,
  "brush-select" = brush_select,
  "click-select" = click_select,
  "collapse-expand" = collapse_expand,
  "create-edge" = create_edge,
  "drag-canvas" = drag_canvas,
  "drag-element" = drag_element,
  "drag-element-force" = drag_element_force,
  "fix-element-size" = fix_element_size,
  "focus-element" = focus_element,
  "hover-activate" = hover_activate,
  "lasso-select" = lasso_select,
  "optimize-viewport-transform" = optimize_viewport_transform,
  "scroll-canvas" = scroll_canvas,
  "zoom-canvas" = zoom_canvas
)

test_that("g6_behaviors creates proper behavior configuration", {
  expect_snapshot(error = TRUE, {
    g6() |> g6_behaviors()
  })

  g <- g6() |>
    g6_behaviors(
      "drag-canvas",
      zoom_canvas(),
      click_select()
    )

  expect_type(g$x$behaviors, "list")
  expect_length(g$x$behaviors, 3)
  keys <- vapply(g$x$behaviors, function(x) x$key, character(1))
  expect_true("drag-canvas" %in% keys)
  expect_true("zoom-canvas" %in% keys)
  expect_true("click-select" %in% keys)
})

test_that("validate behavior works", {
  expect_snapshot(error = TRUE, {
    g6() |>
      g6_behaviors(
        "blabla"
      )

    g6() |> g6_behaviors(list(test = 1))
  })
})

test_that("individual behavior functions work correctly", {
  lapply(names(valid_behaviors), function(behavior) {
    expect_snapshot(
      {
        str(
          g6() |>
            g6_behaviors(behavior)
        )
      }
    )
  })
})


test_that("behavior functions work with valid parameters", {
  # Test auto_adapt_label with various parameters
  expect_no_error(auto_adapt_label())
  expect_no_error(auto_adapt_label(throttle = 200, padding = 5))
  expect_no_error(auto_adapt_label(sortNode = list(type = "betweenness")))
  expect_no_error(auto_adapt_label(sortNode = list(type = "closeness")))
  expect_no_error(auto_adapt_label(sortNode = list(type = "eigenvector")))
  expect_no_error(auto_adapt_label(sortNode = list(type = "pagerank")))
  expect_no_error(auto_adapt_label(enable = FALSE))
  expect_no_error(auto_adapt_label(enable = JS("() => true")))
  expect_no_error(auto_adapt_label(sort = JS("(a, b) => a.id - b.id")))
  expect_no_error(auto_adapt_label(
    sortEdge = JS("(a, b) => a.weight - b.weight")
  ))
  expect_no_error(auto_adapt_label(sortCombo = JS("(a, b) => a.size - b.size")))

  # Test brush_select with various parameters
  expect_no_error(brush_select())
  expect_no_error(brush_select(animation = TRUE))
  expect_no_error(brush_select(enableElements = c("node", "edge")))
  expect_no_error(brush_select(enableElements = c("node", "edge", "combo")))
  expect_no_error(brush_select(mode = "union"))
  expect_no_error(brush_select(mode = "intersect"))
  expect_no_error(brush_select(mode = "diff"))
  expect_no_error(brush_select(state = "active"))
  expect_no_error(brush_select(state = "highlight"))
  expect_no_error(brush_select(immediately = TRUE))
  expect_no_error(brush_select(enable = FALSE))
  expect_no_error(brush_select(enable = JS("() => true")))
  expect_no_error(brush_select(onSelect = JS("(states) => states")))

  # Test click_select with various parameters
  expect_no_error(click_select())
  expect_no_error(click_select(animation = FALSE))
  expect_no_error(click_select(degree = 1))
  expect_no_error(click_select(degree = JS("() => 2")))
  expect_no_error(click_select(multiple = TRUE))
  expect_no_error(click_select(state = "active"))
  expect_no_error(click_select(neighborState = "highlight"))
  expect_no_error(click_select(unselectedState = "inactive"))
  expect_no_error(click_select(enable = FALSE))
  expect_no_error(click_select(enable = JS("() => true")))
  expect_no_error(click_select(onClick = JS("(event) => {}")))

  # Test collapse_expand with various parameters
  expect_no_error(collapse_expand())
  expect_no_error(collapse_expand(animation = FALSE))
  expect_no_error(collapse_expand(trigger = "click"))
  expect_no_error(collapse_expand(align = FALSE))
  expect_no_error(collapse_expand(enable = FALSE))
  expect_no_error(collapse_expand(enable = JS("() => true")))
  expect_no_error(collapse_expand(onCollapse = JS("() => {}")))
  expect_no_error(collapse_expand(onExpand = JS("() => {}")))

  # Test create_edge with various parameters
  expect_no_error(create_edge())
  expect_no_error(create_edge(trigger = "click"))
  expect_no_error(create_edge(enable = TRUE))
  expect_no_error(create_edge(enable = JS("() => true")))
  expect_no_error(create_edge(notify = TRUE))
  expect_no_error(create_edge(onCreate = JS("() => {}")))
  expect_no_error(create_edge(onFinish = JS("() => {}")))
  expect_no_error(create_edge(style = list(stroke = "red")))

  # Test drag_canvas with various parameters
  expect_no_error(drag_canvas())
  expect_no_error(drag_canvas(direction = "x"))
  expect_no_error(drag_canvas(direction = "y"))
  expect_no_error(drag_canvas(range = c(-100, 100)))
  expect_no_error(drag_canvas(range = 50))
  expect_no_error(drag_canvas(sensitivity = 5))
  expect_no_error(drag_canvas(enable = FALSE))
  expect_no_error(drag_canvas(enable = JS("() => true")))
  expect_no_error(drag_canvas(animation = list(duration = 300)))
  expect_no_error(drag_canvas(trigger = list(up = "w", down = "s")))
  expect_no_error(drag_canvas(onFinish = JS("() => {}")))

  # Test drag_element with various parameters
  expect_no_error(drag_element())
  expect_no_error(drag_element(animation = FALSE))
  expect_no_error(drag_element(dropEffect = "link"))
  expect_no_error(drag_element(dropEffect = "none"))
  expect_no_error(drag_element(hideEdge = "out"))
  expect_no_error(drag_element(hideEdge = "in"))
  expect_no_error(drag_element(hideEdge = "both"))
  expect_no_error(drag_element(hideEdge = "all"))
  expect_no_error(drag_element(shadow = TRUE))
  expect_no_error(drag_element(enable = FALSE))
  expect_no_error(drag_element(enable = JS("() => true")))
  expect_no_error(drag_element(
    cursor = list(default = "default", grab = "grab")
  ))

  # Test drag_element_force with various parameters
  expect_no_error(drag_element_force())
  expect_no_error(drag_element_force(fixed = TRUE))
  expect_no_error(drag_element_force(hideEdge = "out"))
  expect_no_error(drag_element_force(hideEdge = "in"))
  expect_no_error(drag_element_force(hideEdge = "both"))
  expect_no_error(drag_element_force(hideEdge = "all"))
  expect_no_error(drag_element_force(enable = FALSE))
  expect_no_error(drag_element_force(enable = JS("() => true")))
  expect_no_error(drag_element_force(cursor = list(default = "default")))

  # Test fix_element_size with various parameters
  expect_no_error(fix_element_size())
  expect_no_error(fix_element_size(reset = TRUE))
  expect_no_error(fix_element_size(state = "active"))
  expect_no_error(fix_element_size(enable = FALSE))
  expect_no_error(fix_element_size(enable = JS("() => true")))
  expect_no_error(fix_element_size(
    node = list(list(shape = "circle", fields = c("r")))
  ))
  expect_no_error(fix_element_size(nodeFilter = JS("() => false")))
  expect_no_error(fix_element_size(
    edge = list(list(shape = "line", fields = c("lineWidth")))
  ))
  expect_no_error(fix_element_size(edgeFilter = JS("() => false")))
  expect_no_error(fix_element_size(
    combo = list(list(shape = "rect", fields = c("width")))
  ))
  expect_no_error(fix_element_size(comboFilter = JS("() => false")))

  # Test focus_element with various parameters
  expect_no_error(focus_element())
  expect_no_error(focus_element(animation = list(duration = 1000)))
  expect_no_error(focus_element(animation = list(easing = "ease-out")))
  expect_no_error(focus_element(animation = list(easing = "ease-in-out")))
  expect_no_error(focus_element(animation = list(easing = "linear")))
  expect_no_error(focus_element(enable = FALSE))
  expect_no_error(focus_element(enable = JS("() => true")))

  # Test hover_activate with various parameters
  expect_no_error(hover_activate())
  expect_no_error(hover_activate(animation = FALSE))
  expect_no_error(hover_activate(degree = 1))
  expect_no_error(hover_activate(degree = JS("() => 2")))
  expect_no_error(hover_activate(direction = "in"))
  expect_no_error(hover_activate(direction = "out"))
  expect_no_error(hover_activate(state = "highlight"))
  expect_no_error(hover_activate(inactiveState = "inactive"))
  expect_no_error(hover_activate(enable = FALSE))
  expect_no_error(hover_activate(enable = JS("() => true")))
  expect_no_error(hover_activate(onHover = JS("() => {}")))
  expect_no_error(hover_activate(onHoverEnd = JS("() => {}")))

  # Test lasso_select with various parameters
  expect_no_error(lasso_select())
  expect_no_error(lasso_select(animation = TRUE))
  expect_no_error(lasso_select(enableElements = c("node", "combo")))
  expect_no_error(lasso_select(enableElements = c("node", "edge", "combo")))
  expect_no_error(lasso_select(immediately = TRUE))
  expect_no_error(lasso_select(mode = "union"))
  expect_no_error(lasso_select(mode = "intersect"))
  expect_no_error(lasso_select(mode = "diff"))
  expect_no_error(lasso_select(state = "highlight"))
  expect_no_error(lasso_select(enable = FALSE))
  expect_no_error(lasso_select(enable = JS("() => true")))
  expect_no_error(lasso_select(onSelect = JS("() => {}")))
  expect_no_error(lasso_select(style = list(stroke = "blue")))
  expect_no_error(lasso_select(trigger = c("control")))

  # Test optimize_viewport_transform with various parameters
  expect_no_error(optimize_viewport_transform())
  expect_no_error(optimize_viewport_transform(debounce = 500))
  expect_no_error(optimize_viewport_transform(enable = FALSE))
  expect_no_error(optimize_viewport_transform(enable = JS("() => true")))
  expect_no_error(optimize_viewport_transform(shapes = JS("() => false")))

  # Test scroll_canvas with various parameters
  expect_no_error(scroll_canvas())
  expect_no_error(scroll_canvas(direction = "x"))
  expect_no_error(scroll_canvas(direction = "y"))
  expect_no_error(scroll_canvas(range = c(-2, 2)))
  expect_no_error(scroll_canvas(range = 2))
  expect_no_error(scroll_canvas(sensitivity = 1.5))
  expect_no_error(scroll_canvas(preventDefault = FALSE))
  expect_no_error(scroll_canvas(enable = FALSE))
  expect_no_error(scroll_canvas(enable = JS("() => true")))
  expect_no_error(scroll_canvas(trigger = list(up = "w")))
  expect_no_error(scroll_canvas(onFinish = JS("() => {}")))

  # Test zoom_canvas with various parameters
  expect_no_error(zoom_canvas())
  expect_no_error(zoom_canvas(animation = list(duration = 300)))
  expect_no_error(zoom_canvas(origin = list(x = 0, y = 0)))
  expect_no_error(zoom_canvas(sensitivity = 1.5))
  expect_no_error(zoom_canvas(preventDefault = FALSE))
  expect_no_error(zoom_canvas(enable = FALSE))
  expect_no_error(zoom_canvas(enable = JS("() => true")))
  expect_no_error(zoom_canvas(trigger = list(zoomIn = "+")))
  expect_no_error(zoom_canvas(onFinish = JS("() => {}")))
})

test_that("behavior functions handle invalid parameters correctly", {
  # Test auto_adapt_label error conditions
  expect_error(
    auto_adapt_label(enable = "invalid"),
    "must be a logical value or a JavaScript function"
  )
  expect_error(
    auto_adapt_label(throttle = -1),
    "must be a single non-negative numeric value"
  )
  expect_error(
    auto_adapt_label(throttle = c(1, 2)),
    "must be a single non-negative numeric value"
  )
  expect_error(auto_adapt_label(padding = "invalid"), "must be a numeric value")
  expect_error(
    auto_adapt_label(sort = "invalid"),
    "must be NULL or a JavaScript function"
  )
  expect_error(auto_adapt_label(sortNode = "invalid"), "must be NULL, a list")
  expect_error(
    auto_adapt_label(sortNode = list(type = "invalid")),
    "must be one of:"
  )
  expect_error(
    auto_adapt_label(sortEdge = "invalid"),
    "must be NULL or a JavaScript function"
  )
  expect_error(
    auto_adapt_label(sortCombo = "invalid"),
    "must be NULL or a JavaScript function"
  )

  # Test brush_select error conditions
  expect_error(brush_select(animation = "invalid"), "should be a boolean value")
  expect_error(brush_select(enable = "invalid"), "should be a boolean or JS")
  expect_error(brush_select(enableElements = "invalid"), "should only contain:")
  expect_error(
    brush_select(immediately = "invalid"),
    "should be a boolean value"
  )
  expect_error(brush_select(onSelect = "invalid"), "should be a JS function")

  # Test click_select error conditions
  expect_error(click_select(animation = "invalid"), "should be a boolean value")
  expect_error(
    click_select(degree = "invalid"),
    "should be a number or a JS function"
  )
  expect_error(
    click_select(enable = "invalid"),
    "should be a boolean or a JS function"
  )
  expect_error(click_select(multiple = "invalid"), "should be a boolean value")
  expect_error(click_select(unselectedState = 123), "should be one of")
  expect_error(click_select(onClick = "invalid"), "should be a JS function")

  # Test collapse_expand error conditions
  expect_error(
    collapse_expand(animation = "invalid"),
    "should be a boolean value"
  )
  expect_error(
    collapse_expand(enable = "invalid"),
    "should be a boolean or a JS function"
  )
  expect_error(collapse_expand(trigger = "invalid"), "should be one of")
  expect_error(collapse_expand(align = "invalid"), "should be a boolean value")
  expect_error(
    collapse_expand(onCollapse = "invalid"),
    "should be a JS function"
  )
  expect_error(collapse_expand(onExpand = "invalid"), "should be a JS function")

  # Test create_edge error conditions
  expect_error(create_edge(trigger = "invalid"), "should be one of")
  expect_error(
    create_edge(enable = "invalid"),
    "should be a boolean or a JS function"
  )
  expect_error(create_edge(onCreate = "invalid"), "should be a JS function")
  expect_error(create_edge(onFinish = "invalid"), "should be a JS function")
  expect_error(create_edge(style = "invalid"), "should be a list")

  # Test drag_canvas error conditions
  expect_error(
    drag_canvas(enable = "invalid"),
    "should be a boolean or a function"
  )
  expect_error(drag_canvas(animation = "invalid"), "should be a list")
  expect_error(drag_canvas(range = "invalid"), "should be a number")
  expect_error(drag_canvas(sensitivity = -1), "should be a positive number")
  expect_error(drag_canvas(sensitivity = 0), "should be a positive number")
  expect_error(drag_canvas(trigger = "invalid"), "should be a list")
  expect_error(drag_canvas(onFinish = "invalid"), "should be a JS function")

  # Test drag_element error conditions
  expect_error(
    drag_element(enable = "invalid"),
    "should be a boolean or a JS function"
  )
  expect_error(drag_element(animation = "invalid"), "should be a boolean value")
  expect_error(drag_element(state = 123), "should be a string")
  expect_error(drag_element(shadow = "invalid"), "should be a boolean value")
  expect_error(drag_element(cursor = "invalid"), "should be a list")

  # Test drag_element_force error conditions
  expect_error(
    drag_element_force(fixed = "invalid"),
    "should be a boolean value"
  )
  expect_error(
    drag_element_force(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )
  expect_error(drag_element_force(state = 123), "should be a string")
  expect_error(drag_element_force(cursor = "invalid"), "should be a list")

  # Test fix_element_size error conditions
  expect_error(
    fix_element_size(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )
  expect_error(fix_element_size(reset = "invalid"), "should be a boolean value")
  expect_error(fix_element_size(state = 123), "should be a string")
  expect_error(fix_element_size(node = "invalid"), "should be a list")
  expect_error(
    fix_element_size(nodeFilter = "invalid"),
    "should be a JavaScript function"
  )
  expect_error(fix_element_size(edge = "invalid"), "should be a list")
  expect_error(
    fix_element_size(edgeFilter = "invalid"),
    "should be a JavaScript function"
  )
  expect_error(fix_element_size(combo = "invalid"), "should be a list")
  expect_error(
    fix_element_size(comboFilter = "invalid"),
    "should be a JavaScript function"
  )

  # Test focus_element error conditions
  expect_error(focus_element(animation = "invalid"), "should be a list")
  expect_error(
    focus_element(animation = list(duration = "invalid")),
    "should be a number"
  )
  expect_error(
    focus_element(animation = list(easing = 123)),
    "should be a string"
  )
  expect_error(
    focus_element(animation = list(easing = "invalid")),
    "should be one of:"
  )
  expect_error(
    focus_element(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )

  # Test hover_activate error conditions
  expect_error(
    hover_activate(animation = "invalid"),
    "should be a boolean value"
  )
  expect_error(
    hover_activate(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )
  expect_error(
    hover_activate(degree = "invalid"),
    "should be a number or a JavaScript function"
  )
  expect_error(hover_activate(state = 123), "should be a string")
  expect_error(hover_activate(inactiveState = 123), "should be a string")
  expect_error(
    hover_activate(onHover = "invalid"),
    "should be a JavaScript function"
  )
  expect_error(
    hover_activate(onHoverEnd = "invalid"),
    "should be a JavaScript function"
  )

  # Test lasso_select error conditions
  expect_error(lasso_select(key = 123), "should be a string")
  expect_error(lasso_select(animation = "invalid"), "should be a boolean value")
  expect_error(
    lasso_select(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )
  expect_error(lasso_select(enableElements = "invalid"), "should only contain:")
  expect_error(
    lasso_select(immediately = "invalid"),
    "should be a boolean value"
  )
  expect_error(
    lasso_select(onSelect = "invalid"),
    "should be a JavaScript function"
  )
  expect_error(lasso_select(state = 123), "should be a string")
  expect_error(lasso_select(style = "invalid"), "should be a list")

  # Test optimize_viewport_transform error conditions
  expect_error(
    optimize_viewport_transform(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )
  expect_error(
    optimize_viewport_transform(debounce = -1),
    "should be a non-negative number"
  )
  expect_error(
    optimize_viewport_transform(shapes = "invalid"),
    "should be a JavaScript function"
  )

  # Test scroll_canvas error conditions
  expect_error(
    scroll_canvas(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )
  expect_error(scroll_canvas(direction = "invalid"), "should be one of:")
  expect_error(scroll_canvas(range = "invalid"), "should be a number")
  expect_error(scroll_canvas(sensitivity = -1), "should be a positive number")
  expect_error(scroll_canvas(sensitivity = 0), "should be a positive number")
  expect_error(scroll_canvas(trigger = "invalid"), "should be a list")
  expect_error(
    scroll_canvas(onFinish = "invalid"),
    "should be a JavaScript function"
  )
  expect_error(
    scroll_canvas(preventDefault = "invalid"),
    "should be a boolean value"
  )

  # Test zoom_canvas error conditions
  expect_error(zoom_canvas(key = 123), "should be a string")
  expect_error(zoom_canvas(animation = "invalid"), "should be a list")
  expect_error(
    zoom_canvas(animation = list(duration = -1)),
    "should be a non-negative number"
  )
  expect_error(
    zoom_canvas(enable = "invalid"),
    "should be a boolean or a JavaScript function"
  )
  expect_error(zoom_canvas(origin = "invalid"), "should be a list")
  expect_error(
    zoom_canvas(onFinish = "invalid"),
    "should be a JavaScript function"
  )
  expect_error(
    zoom_canvas(preventDefault = "invalid"),
    "should be a boolean value"
  )
  expect_error(zoom_canvas(sensitivity = -1), "should be a positive number")
  expect_error(zoom_canvas(sensitivity = 0), "should be a positive number")
  expect_error(zoom_canvas(trigger = "invalid"), "should be a list")
})

#' Configure collapse button for nodes
#'
#' @param collapsed Logical. Whether the node should be collapsed initially. Default is FALSE.
#' @param visibility Character. Visibility mode of the collapse button.
#'   Either `"visible"` (always shown when children exist) or `"hover"`
#'   (shown only on mouse hover). Default is `"visible"`.
#' @param placement Character or numeric vector. Position of the collapse button.
#'   Can be one of: "top", "right", "bottom", "left", "right-top", "right-bottom",
#'   "left-top", "left-bottom", or a numeric vector of length 2 for custom coordinates.
#' @param r Numeric. Radius of the button. Default is 8.
#' @param fill Character. Fill color of the button background. Default is "#fff".
#' @param stroke Character. Stroke color of the button border. Default is "#CED4D9".
#' @param lineWidth Numeric. Width of the button border. Default is 1.
#' @param iconStroke Character. Stroke color of the +/- icon. Default is "#000".
#' @param iconLineWidth Numeric. Width of the +/- icon lines. Default is 1.4.
#' @param cursor Character. CSS cursor style. Default is "pointer".
#' @param zIndex Numeric. Z-index for layering. Default is 999.
#'
#' @return A list of collapse button configuration options.
#' @export
#'
#' @examples
#' # Default collapse button
#' g6_collapse_options()
#'
#' # Custom styled collapse button
#' g6_collapse_options(
#'   placement = "right-top",
#'   fill = "#f0f0f0",
#'   stroke = "#333",
#'   r = 10
#' )
#'
#' # Collapse button with custom coordinates
#' g6_collapse_options(placement = c(1, 0.2))
#' @rdname g6_collapse_options
g6_collapse_options <- function(
  collapsed = FALSE,
  visibility = c("visible", "hover"),
  placement = "right-top",
  r = 6,
  fill = "#fff",
  stroke = "#9cabd4",
  lineWidth = 1,
  iconStroke = "#9cabd4",
  iconLineWidth = 1.4,
  cursor = "pointer",
  zIndex = 999
) {
  visibility <- match.arg(visibility)

  # Validate placement
  if (is.character(placement)) {
    placement <- match.arg(
      placement,
      choices = c(
        "top",
        "right",
        "bottom",
        "left",
        "right-top",
        "right-bottom",
        "left-top",
        "left-bottom"
      )
    )
  } else if (is.numeric(placement)) {
    if (length(placement) != 2) {
      stop("Numeric placement must be a vector of length 2.")
    }
    # At least one coordinate must be 0 or 1 (on the edge)
    if (!any(placement %in% c(0, 1))) {
      stop(
        "Invalid collapse button placement: at least one coordinate must be 0 or 1 (i.e., on the node edge). ",
        "You supplied: c(",
        paste(placement, collapse = ", "),
        ")"
      )
    }
  } else {
    stop("placement must be a character string or numeric vector of length 2.")
  }

  structure(
    list(
      collapsed = collapsed,
      visibility = visibility,
      placement = placement,
      r = r,
      fill = fill,
      stroke = stroke,
      lineWidth = lineWidth,
      iconStroke = iconStroke,
      iconLineWidth = iconLineWidth,
      cursor = cursor,
      zIndex = zIndex
    ),
    class = "g6_collapse_options"
  )
}

#' Check if object is a g6_collapse_options configuration
#'
#' @param x An object to check.
#' @return Logical indicating if x is of class g6_collapse_options.
#' @export
#' @rdname g6_collapse_options
is_g6_collapse_options <- function(x) {
  inherits(x, "g6_collapse_options")
}

#' G6 Graph Elements
#'
#' Constructors and validators for G6 node, edge, and combo elements.
#'
#' @param id Character. Unique identifier for the node or combo (required).
#' For edges, this is optional (id is constructed as source-target if not provided).
#' @param type Character. Element type (optional).
#' @param data List. Custom data for the element (optional).
#' @param style List. Element style (optional).
#' @param states Character vector. Initial states for the element (optional).
#' @param combo Character or NULL. Combo ID or parent combo ID (optional).
#' @param children Character vector. Child node IDs (optional, nodes only).
#' @param ports List. Ports definition (optional, nodes only).
#' See \link{g6_ports} and \link{g6_port} for details.
#' @param collapse List. Collapse button configuration (optional, nodes and combos).
#' See \link{g6_collapse_options} for details. For nodes, only used when node has children.
#' For combos, when provided and `type` is NULL, auto-sets type to `"rect-combo-with-extra-button"`.
#' @param source Character. Source node ID (required, edges only).
#' @param target Character. Target node ID (required, edges only).
#' @param ... Additional arguments (unused).
#' the checks on source and target.
#'
#' @return An S3 object of class \code{g6_node}, \code{g6_edge}, or \code{g6_combo} (and \code{g6_element}).
#' @examples
#' # Create a node
#' node <- g6_node(
#'   id = "n1",
#'   type = "rect",
#'   data = list(label = "Node 1"),
#'   style = list(color = "red"),
#'   states = list("selected"),
#'   combo = NULL,
#'   children = c("n2", "n3")
#' )
#'
#' # Create an edge
#' edge <- g6_edge(
#'   source = "n1",
#'   target = "n2",
#'   type = "line",
#'   style = list(width = 2)
#' )
#'
#' # Create a combo
#' combo <- g6_combo(
#'   id = "combo1",
#'   type = "rect",
#'   data = list(label = "Combo 1"),
#'   style = list(border = "dashed"),
#'   states = list("active"),
#'   combo = NULL
#' )
#'
#' # Validate a node explicitly
#' validate_element(node)
#' @rdname g6_element
#' @export
g6_node <- function(
  id,
  type = NULL,
  data = NULL,
  style = NULL,
  states = NULL,
  combo = NULL,
  children = NULL,
  ports = NULL,
  collapse = NULL
) {
  node <- dropNulls(list(
    id = id,
    type = type,
    data = data,
    style = style,
    states = states,
    combo = combo,
    children = children,
    ports = ports,
    collapse = collapse
  ))
  node <- structure(node, class = c("g6_node", "g6_element"))
  validate_element(node)
}

#' @rdname g6_element
#' @export
g6_edge <- function(
  source,
  target,
  id = paste(source, target, sep = "-"),
  type = NULL,
  data = NULL,
  style = NULL,
  states = NULL
) {
  edge <- dropNulls(list(
    source = source,
    target = target,
    id = id,
    type = type,
    data = data,
    style = style,
    states = states
  ))
  edge <- structure(edge, class = c("g6_edge", "g6_element"))
  validate_element(edge)
}

#' @keywords internal
#' @note To manage updating an edge
#' since source and target are not mandatory in that case.
g6_edge_update <- function(id, ...) {
  edge <- dropNulls(list(id = id, ...))
  edge <- structure(edge, class = c("g6_edge", "g6_element"), update = TRUE)
  validate_element(edge) # update validation
}

#' @rdname g6_element
#' @export
g6_combo <- function(
  id,
  type = NULL,
  data = NULL,
  style = NULL,
  states = NULL,
  combo = NULL,
  collapse = NULL
) {
  combo <- dropNulls(list(
    id = id,
    type = type,
    data = data,
    style = style,
    states = states,
    combo = combo,
    collapse = collapse
  ))
  combo <- structure(combo, class = c("g6_combo", "g6_element"))
  validate_element(combo)
}

#' Check if an object is a G6 element
#'
#' @rdname is_g6_element
#' @param x An object of class \code{g6_node}, \code{g6_edge}, or \code{g6_combo}.
#' @return Logical indicating if \code{x} is of the specified class.
#' @export
is_g6_node <- function(x) {
  inherits(x, "g6_node")
}

#' @rdname is_g6_element
#' @export
is_g6_edge <- function(x) {
  inherits(x, "g6_edge")
}

#' @rdname is_g6_element
#' @export
is_g6_combo <- function(x) {
  inherits(x, "g6_combo")
}

#' @rdname g6_element
#' @param x An object of class \code{g6_element},
#' \code{g6_node}, \code{g6_edge}, or \code{g6_combo}.
#' @export
validate_element <- function(x, ...) {
  UseMethod("validate_element")
}

#' @rdname g6_element
#' @export
validate_element.g6_element <- function(x, ...) {
  # id: character, length 1, non-empty
  if (
    !is.null(x$id) &&
      (length(x$id) != 1 || is.na(x$id) || x$id == "")
  ) {
    stop("'id' must be a non-empty character string if provided.")
  }
  # type: character, length 1
  if (!is.null(x$type) && (!is.character(x$type) || length(x$type) != 1)) {
    stop("'type' must be a character string if provided.")
  }
  # data: list
  if (!is.null(x$data) && !is.list(x$data)) {
    stop("'data' must be a list if provided.")
  }
  # style: list
  if (!is.null(x$style) && !is.list(x$style)) {
    stop("'style' must be a list if provided.")
  }

  is_unnamed_list <- function(x) {
    is.list(x) && (is.null(names(x)) || all(names(x) == ""))
  }

  # states: unnamed list
  if (!is.null(x$states) && !is_unnamed_list(x$states)) {
    stop("'states' must be an unnamed list if provided.")
  }
  x
}

#' @keywords internal
check_custom_node_type <- function(x) {
  if (length(x[["type"]]) && !grepl("custom", x[["type"]])) {
    stop(
      "Node 'ports' are typically used with custom node types. ",
      "Ensure the node type supports ports like 'custom-<TYPE>-node.'"
    )
  }
}

#' @rdname g6_element
#' @export
validate_element.g6_node <- function(x, ...) {
  # Node-specific validation
  if (is.null(x$id)) {
    stop("Node 'id' is required.")
  }
  if (
    !is.null(x$combo) &&
      !(length(x$combo) == 1 || is.null(x$combo))
  ) {
    stop("Node 'combo' must be a character string or NULL.")
  }
  if (!is.null(x$children)) {
    if (!is.vector(x$children) && !is.list(x$children)) {
      stop("Node 'children' must be a character vector or list if provided.")
    }

    # Convert to character vector for validation and processing
    children_vec <- as.character(unlist(x$children))

    # Check if node includes itself as a child
    if (as.character(x$id) %in% children_vec) {
      stop("Node 'children' cannot include the node itself (id: ", x$id, ").")
    }

    # Auto-enable directed graph mode when children are present
    if (
      length(children_vec) > 0 && !("g6R.directed_graph" %in% names(options()))
    ) {
      set_g6_directed_graph(TRUE)
    }

    # Convert to list for JavaScript
    x$children <- as.list(children_vec)
  }
  if (length(x$ports)) {
    check_custom_node_type(x)
    if (!is_g6_ports(x$ports)) {
      stop("Node 'ports' must be of class 'g6_ports'.")
    }
    x$style$ports <- x$ports
    x$ports <- NULL
  }
  if (length(x$collapse)) {
    check_custom_node_type(x)
    if (!is_g6_collapse_options(x$collapse)) {
      stop("Node 'collapse' must be of class 'g6_collapse_options'.")
    }
    x$style$collapse <- x$collapse
    x$collapse <- NULL
  }
  NextMethod()
}

#' @rdname g6_element
#' @export
validate_element.g6_edge <- function(x, ...) {
  # Skip source/target validation if updating
  if (isTRUE(attr(x, "update"))) {
    attributes(x)$update <- NULL
    return(NextMethod())
  }

  if (
    is.null(x$source) ||
      length(x$source) != 1 ||
      is.na(x$source) ||
      x$source == ""
  ) {
    stop("Edge 'source' is required and must be a non-empty character string.")
  }
  if (
    is.null(x$target) ||
      length(x$target) != 1 ||
      is.na(x$target) ||
      x$target == ""
  ) {
    stop("Edge 'target' is required and must be a non-empty character string.")
  }
  if (!is.null(x$id) && (!is.character(x$id) || length(x$id) != 1)) {
    stop("Edge 'id' must be a character string if provided.")
  }
  NextMethod()
}

#' @rdname g6_element
#' @export
validate_element.g6_combo <- function(x, ...) {
  # Combo-specific validation
  if (is.null(x$id)) {
    stop("Combo 'id' is required.")
  }
  if (
    !is.null(x$combo) &&
      !(length(x$combo) == 1 || is.null(x$combo))
  ) {
    stop("Combo 'combo' must be a character string or NULL.")
  }
  if (length(x$collapse)) {
    if (!is_g6_collapse_options(x$collapse)) {
      stop("Combo 'collapse' must be of class 'g6_collapse_options'.")
    }
    if (is.null(x$type)) {
      x$type <- "rect-combo-with-extra-button"
    } else {
      if (!grepl("-combo-with-extra-button", x$type)) {
        x$type <- paste0(x$type, "-combo-with-extra-button")
      }
    }
    x$style$collapse <- x$collapse
    x$collapse <- NULL
  }
  NextMethod()
}

#' Create and validate lists of G6 elements
#'
#' Constructors for lists of G6 node, edge, and combo elements.
#' Each function accepts multiple validated elements and returns a
#' list with the appropriate class.
#' All elements are validated on construction.
#'
#' @param ... One or more \code{g6_node}, \code{g6_edge}, or \code{g6_combo} objects.
#' @return An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.
#' @examples
#' nodes <- g6_nodes(
#'   g6_node(id = "n1"),
#'   g6_node(id = "n2")
#' )
#' edges <- g6_edges(
#'   g6_edge(source = "n1", target = "n2")
#' )
#' combos <- g6_combos(
#'   g6_combo(id = "c1")
#' )
#' @rdname g6_elements
#' @export
g6_nodes <- function(...) {
  nodes <- structure(list(...), class = "g6_nodes")
  validate_elements(nodes)
}

#' @rdname g6_elements
#' @export
g6_edges <- function(...) {
  edges <- structure(list(...), class = "g6_edges")
  validate_elements(edges)
}

#' @rdname g6_elements
#' @export
g6_combos <- function(...) {
  combos <- structure(list(...), class = "g6_combos")
  validate_elements(combos)
}

#' Check if an object is a list of G6 elements
#'
#' @param x An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.
#' @rdname is_g6_elements
#' @export
is_g6_nodes <- function(x) {
  inherits(x, "g6_nodes")
}

#' @rdname is_g6_elements
#' @export
is_g6_edges <- function(x) {
  inherits(x, "g6_edges")
}

#' @rdname is_g6_elements
#' @export
is_g6_combos <- function(x) {
  inherits(x, "g6_combos")
}

#' Validate a list of G6 elements
#'
#' S3 generic for validating lists of G6 elements.
#'
#' @param x An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.
#' @param ... Additional arguments (unused).
#' @return Invisibly returns the validated object.
#' @rdname g6_elements
#' @export
validate_elements <- function(x, ...) {
  stopifnot(length(x) > 0)
  UseMethod("validate_elements")
}

#' @rdname g6_elements
#' @export
validate_elements.g6_nodes <- function(x, ...) {
  if (!all(lgl_ply(x, is_g6_node))) {
    stop("All elements must be of class 'g6_node'.")
  }
  x
}

#' @rdname g6_elements
#' @export
validate_elements.g6_edges <- function(x, ...) {
  if (!all(lgl_ply(x, is_g6_edge))) {
    stop("All elements must be of class 'g6_edge'.")
  }
  x
}

#' @rdname g6_elements
#' @export
validate_elements.g6_combos <- function(x, ...) {
  if (!all(lgl_ply(x, is_g6_combo))) {
    stop("All elements must be of class 'g6_combo'.")
  }
  x
}

#' Coerce to a g6 element object
#' @export
#' @param x An object to be coerced. List or g6 element are supported.
#' @param ... Additional arguments (unused).
#' @rdname as_g6_element
#' @return An element of class \code{g6_node}, \code{g6_edge}, or \code{g6_combo}.
as_g6_node <- function(x, ...) {
  UseMethod("as_g6_node")
}

#' @export
#' @rdname as_g6_element
as_g6_node.g6_node <- function(x, ...) {
  x
}

#' @export
#' @rdname as_g6_element
as_g6_node.list <- function(x, ...) {
  # Coerce ports if present
  if (length(x$style$ports)) {
    x$ports <- as_g6_ports(x$style$ports)
  }

  if ("combo" %in% names(x)) {
    # Drop nulls except for combo field
    node <- dropNulls(x, except = "combo")
    node <- structure(node, class = c("g6_node", "g6_element"))
    node <- validate_element(node)
    return(node)
  }

  # Default behavior
  do.call(g6_node, x)
}

#' Coerce to a list of g6_elements objects
#' @export
#' @param x An object to be coerced. Data frame, list
#' or g6 element are supported.
#' @param ... Additional arguments (unused).
#' @rdname as_g6_elements
#' @return An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.
as_g6_nodes <- function(x, ...) {
  UseMethod("as_g6_nodes")
}

#' @export
#' @rdname g6_elements
as_g6_nodes.g6_nodes <- function(x, ...) {
  x
}

#' @export
#' @rdname g6_elements
as_g6_nodes.data.frame <- function(x, ...) {
  lst <- lapply(seq_len(nrow(x)), function(i) {
    lapply(x[i, , drop = FALSE], `[[`, 1)
  })
  as_g6_nodes(lst)
}

#' @export
#' @rdname g6_elements
as_g6_nodes.list <- function(x, ...) {
  nodes <- lapply(x, as_g6_node)
  structure(nodes, class = "g6_nodes")
}

#' @export
#' @rdname as_g6_element
as_g6_edge <- function(x, ...) {
  UseMethod("as_g6_edge")
}

#' @export
#' @rdname as_g6_element
as_g6_edge.g6_edge <- function(x, ...) {
  x
}

#' @export
#' @rdname as_g6_element
as_g6_edge.list <- function(x, ...) {
  if (!is.null(x$id) && is.null(x$source) && is.null(x$target)) {
    do.call(g6_edge_update, x)
  } else {
    do.call(g6_edge, x)
  }
}

#' @export
#' @rdname as_g6_elements
as_g6_edges <- function(x, ...) {
  UseMethod("as_g6_edges")
}

#' @export
#' @rdname as_g6_elements
as_g6_edges.g6_edges <- function(x, ...) {
  x
}

#' @export
#' @rdname as_g6_elements
as_g6_edges.data.frame <- function(x, ...) {
  lst <- lapply(seq_len(nrow(x)), function(i) {
    lapply(x[i, , drop = FALSE], `[[`, 1)
  })
  as_g6_edges(lst)
}

#' @export
#' @rdname as_g6_elements
as_g6_edges.list <- function(x, ...) {
  edges <- lapply(x, as_g6_edge)
  structure(edges, class = "g6_edges")
}

#' @export
#' @rdname as_g6_element
as_g6_combo <- function(x, ...) {
  UseMethod("as_g6_combo")
}

#' @export
#' @rdname as_g6_element
as_g6_combo.g6_combo <- function(x, ...) {
  x
}

#' @export
#' @rdname as_g6_element
as_g6_combo.list <- function(x, ...) {
  if ("combo" %in% names(x)) {
    # Drop nulls except for combo field
    combo <- dropNulls(x, except = "combo")
    combo <- structure(combo, class = c("g6_combo", "g6_element"))
    validate_element(combo)
    return(combo)
  }
  do.call(g6_combo, x)
}

#' @export
#' @rdname as_g6_elements
as_g6_combos <- function(x, ...) {
  UseMethod("as_g6_combos")
}

#' @export
#' @rdname as_g6_elements
as_g6_combos.g6_combos <- function(x, ...) {
  x
}

#' @export
#' @rdname as_g6_elements
as_g6_combos.data.frame <- function(x, ...) {
  lst <- lapply(seq_len(nrow(x)), function(i) {
    lapply(x[i, , drop = FALSE], `[[`, 1)
  })
  as_g6_combos(lst)
}

#' @export
#' @rdname as_g6_elements
as_g6_combos.list <- function(x, ...) {
  combos <- lapply(x, as_g6_combo)
  structure(combos, class = "g6_combos")
}

#' Create a g6_data object
#'
#' Create compatible data structure for [g6_add_data()], [g6_set_data()] or
#' simply [g6()].
#'
#' @param nodes Nodes data.
#' @param edges Edges data.
#' @param combos Combo data.
#' @param x A list with elements \code{nodes}, \code{edges}, and/or \code{combos}.
#' @param ... Additional arguments (unused).
#'
#' @export
#' @rdname g6_data
#' @return An object of class \code{g6_data}.
g6_data <- function(nodes = NULL, edges = NULL, combos = NULL) {
  dat <- list()

  if (!is.null(nodes)) {
    dat[["nodes"]] <- as_g6_nodes(nodes)
  }
  if (!is.null(edges)) {
    dat[["edges"]] <- as_g6_edges(edges)
  }
  if (!is.null(combos)) {
    dat[["combos"]] <- as_g6_combos(combos)
  }

  structure(dat, class = "g6_data")
}

#' Check if an object is a g6_data object
#'
#' @param x An object to check.
#' @return Logical indicating if \code{x} is of class \code{g6_data}.
#' @export
#' @rdname is_g6_data
is_g6_data <- function(x) {
  inherits(x, "g6_data")
}

#' Coerce to a g6_data object
#'
#' @export
#' @rdname g6_data
as_g6_data <- function(x, ...) {
  UseMethod("as_g6_data")
}

#' @export
#' @rdname g6_data
as_g6_data.g6_data <- function(x, ...) {
  x
}

#' @export
#' @rdname g6_data
as_g6_data.list <- function(x, ...) {
  g6_data(
    nodes = x$nodes,
    edges = x$edges,
    combos = x$combos
  )
}

#' @keywords internal
flatten_g6_elements <- function(el, max_depth = 2) {
  depth <- 1
  while (
    length(el) == 1 &&
      is.list(el[[1]]) &&
      is.null(names(el[[1]]))
  ) {
    el <- el[[1]]
    depth <- depth + 1
    if (depth > max_depth) {
      stop(
        "Input is too deeply nested. Please provide a 
        list of elements, not a list of lists of lists."
      )
    }
  }
  el
}

#' @keywords internal
as_g6_elements <- function(..., coerc_func) {
  el <- list(...)
  # If user passed a single data frame, use it directly
  if (length(el) == 1 && inherits(el[[1]], "data.frame")) {
    return(coerc_func(el[[1]]))
  }
  el <- flatten_g6_elements(el)
  coerc_func(el)
}

library(testthat)
library(g6R)

test_that("g6_collapse_options constructs with defaults", {
  collapse <- g6_collapse_options()
  expect_s3_class(collapse, "g6_collapse_options")
  expect_true(is_g6_collapse_options(collapse))
  expect_equal(collapse$collapsed, FALSE)
  expect_equal(collapse$visibility, "visible")
  expect_equal(collapse$placement, "right-top")
  expect_equal(collapse$r, 6)
  expect_equal(collapse$fill, "#fff")
  expect_equal(collapse$stroke, "#9cabd4")
  expect_equal(collapse$lineWidth, 1)
  expect_equal(collapse$iconStroke, "#9cabd4")
  expect_equal(collapse$iconLineWidth, 1.4)
  expect_equal(collapse$cursor, "pointer")
  expect_equal(collapse$zIndex, 999)
})

test_that("g6_collapse_options accepts visibility = 'hover'", {
  collapse <- g6_collapse_options(visibility = "hover")
  expect_equal(collapse$visibility, "hover")
})

test_that("g6_collapse_options rejects invalid visibility", {
  expect_snapshot(error = TRUE, {
    g6_collapse_options(visibility = "invalid")
  })
})

test_that("g6_collapse_options validates placement with match.arg", {
  expect_snapshot(error = TRUE, {
    g6_collapse_options(placement = "invalid-placement")
  })
})

test_that("g6_collapse_options accepts numeric placement", {
  collapse <- g6_collapse_options(placement = c(0, 0.5))
  expect_equal(collapse$placement, c(0, 0.5))

  collapse2 <- g6_collapse_options(placement = c(0.3, 1))
  expect_equal(collapse2$placement, c(0.3, 1))
})

test_that("g6_collapse_options validates numeric placement length", {
  expect_snapshot(error = TRUE, {
    g6_collapse_options(placement = c(0))
  })

  expect_snapshot(error = TRUE, {
    g6_collapse_options(placement = c(0, 0.5, 1))
  })
})

test_that("g6_collapse_options validates numeric placement has at least one edge coordinate", {
  expect_snapshot(error = TRUE, {
    g6_collapse_options(placement = c(0.5, 0.5))
  })

  expect_snapshot(error = TRUE, {
    g6_collapse_options(placement = c(0.3, 0.7))
  })
})

test_that("g6_collapse_options rejects invalid placement type", {
  expect_snapshot(error = TRUE, {
    g6_collapse_options(placement = list(0, 0.5))
  })

  expect_snapshot(error = TRUE, {
    g6_collapse_options(placement = NULL)
  })
})

test_that("is_g6_collapse_options works correctly", {
  collapse <- g6_collapse_options()
  expect_true(is_g6_collapse_options(collapse))
  expect_false(is_g6_collapse_options(list()))
  expect_false(is_g6_collapse_options("not a collapse"))
  expect_false(is_g6_collapse_options(NULL))
})

test_that("g6_node constructs and validates a node", {
  node <- g6_node(
    id = "n1",
    type = "rect",
    data = list(label = "Node 1"),
    style = list(color = "red"),
    states = list("selected"),
    combo = NULL,
    children = c("n2", "n3")
  )
  expect_s3_class(node, "g6_node")
  expect_s3_class(node, "g6_element")
  expect_equal(node$id, "n1")
  expect_equal(node$type, "rect")
  expect_equal(node$data$label, "Node 1")
})

test_that("g6_node with collapse options moves collapse to style", {
  node <- g6_node(
    id = "n1",
    type = "custom-rect-node",
    children = c("n2"),
    collapse = g6_collapse_options(collapsed = TRUE, placement = "top")
  )
  expect_s3_class(node, "g6_node")
  expect_null(node$collapse)
  expect_s3_class(node$style$collapse, "g6_collapse_options")
  expect_equal(node$style$collapse$collapsed, TRUE)
  expect_equal(node$style$collapse$placement, "top")
})

test_that("g6_node validation fails for invalid collapse", {
  expect_snapshot(error = TRUE, {
    g6_node(
      id = "n1",
      children = c("n2"),
      collapse = list(collapsed = TRUE, placement = "top")
    )
  })

  expect_snapshot(error = TRUE, {
    g6_node(
      id = "n1",
      children = c("n2"),
      collapse = "invalid"
    )
  })
})

test_that("g6_edge constructs and validates an edge", {
  edge <- g6_edge(
    source = "n1",
    target = "n2",
    type = "line",
    style = list(width = 2)
  )
  expect_s3_class(edge, "g6_edge")
  expect_s3_class(edge, "g6_element")
  expect_equal(edge$source, "n1")
  expect_equal(edge$target, "n2")
  expect_equal(edge$type, "line")
})

test_that("g6_combo constructs and validates a combo", {
  combo <- g6_combo(
    id = "combo1",
    type = "rect",
    data = list(label = "Combo 1"),
    style = list(border = "dashed"),
    states = list("active"),
    combo = NULL
  )
  expect_s3_class(combo, "g6_combo")
  expect_s3_class(combo, "g6_element")
  expect_equal(combo$id, "combo1")
  expect_equal(combo$type, "rect")
  expect_equal(combo$data$label, "Combo 1")
})

test_that("g6_node validation fails for missing id", {
  expect_snapshot(error = TRUE, {
    g6_node(type = "rect")
  })
})

test_that("g6_edge validation fails for missing source/target", {
  expect_snapshot(error = TRUE, {
    g6_edge(source = NULL, target = "n2")
  })
  expect_snapshot(error = TRUE, {
    g6_edge(source = "n1", target = NULL)
  })
})

test_that("g6_combo validation fails for missing id", {
  expect_snapshot(error = TRUE, {
    g6_combo(type = "rect")
  })
})

test_that("g6_nodes constructs and validates a list of nodes", {
  nodes <- g6_nodes(
    g6_node(id = "n1"),
    g6_node(id = "n2")
  )
  expect_s3_class(nodes, "g6_nodes")
  expect_length(nodes, 2)
  expect_s3_class(nodes[[1]], "g6_node")
  expect_equal(nodes[[1]]$id, "n1")
})

test_that("g6_edges constructs and validates a list of edges", {
  edges <- g6_edges(
    g6_edge(source = "n1", target = "n2"),
    g6_edge(source = "n2", target = "n3")
  )
  expect_s3_class(edges, "g6_edges")
  expect_length(edges, 2)
  expect_s3_class(edges[[2]], "g6_edge")
  expect_equal(edges[[2]]$source, "n2")
  expect_equal(edges[[2]]$target, "n3")
})

test_that("g6_combos constructs and validates a list of combos", {
  combos <- g6_combos(
    g6_combo(id = "c1"),
    g6_combo(id = "c2")
  )
  expect_s3_class(combos, "g6_combos")
  expect_length(combos, 2)
  expect_s3_class(combos[[1]], "g6_combo")
  expect_equal(combos[[2]]$id, "c2")
})

test_that("g6_nodes fails if any element is not a g6_node", {
  expect_snapshot(error = TRUE, {
    g6_nodes(g6_node(id = "n1"), g6_edge(source = "n1", target = "n2"))
  })
})

test_that("g6_edges fails if any element is not a g6_edge", {
  expect_snapshot(error = TRUE, {
    g6_edges(g6_edge(source = "n1", target = "n2"), g6_node(id = "n2"))
  })
})

test_that("g6_combos fails if any element is not a g6_combo", {
  expect_snapshot(error = TRUE, {
    g6_combos(g6_combo(id = "c1"), g6_node(id = "n1"))
  })
})

test_that("g6_nodes fails if no nodes provided", {
  expect_snapshot(error = TRUE, {
    g6_nodes()
  })
})

test_that("g6_edges fails if no edges provided", {
  expect_snapshot(error = TRUE, {
    g6_edges()
  })
})

test_that("g6_combos fails if no combos provided", {
  expect_snapshot(error = TRUE, {
    g6_combos()
  })
})

test_that("node, edge, combo coercion and predicates", {
  n <- as_g6_node(list(id = "n1"))
  expect_true(is_g6_node(n))
  nodes <- as_g6_nodes(list(list(id = "n1"), list(id = "n2")))
  expect_true(is_g6_nodes(nodes))
  expect_true(all(vapply(nodes, is_g6_node, logical(1))))

  e <- as_g6_edge(list(source = "n1", target = "n2"))
  expect_true(is_g6_edge(e))
  edges <- as_g6_edges(list(
    list(source = "n1", target = "n2"),
    list(source = "n2", target = "n1")
  ))
  expect_true(is_g6_edges(edges))
  expect_true(all(vapply(edges, is_g6_edge, logical(1))))

  c <- as_g6_combo(list(id = "combo1"))
  expect_true(is_g6_combo(c))
  combos <- as_g6_combos(list(list(id = "combo1"), list(id = "combo2")))
  expect_true(is_g6_combos(combos))
  expect_true(all(vapply(combos, is_g6_combo, logical(1))))
})

test_that("g6_data and related functions", {
  nodes <- list(list(id = "n1"), list(id = "n2"))
  edges <- list(list(source = "n1", target = "n2"))
  combos <- list(list(id = "combo1"))
  dat <- g6_data(nodes = nodes, edges = edges, combos = combos)
  expect_true(is_g6_data(dat))
  expect_s3_class(dat, "g6_data")
  dat2 <- as_g6_data(list(nodes = nodes, edges = edges, combos = combos))
  expect_true(is_g6_data(dat2))
  expect_equal(dat, dat2)
})

test_that("package data objects are accessible and properly structured", {
  # Test dag data
  expect_true(exists("dag"))
  data("dag", package = "g6R")
  expect_type(dag, "list")
  expect_true("nodes" %in% names(dag))
  expect_true("edges" %in% names(dag))

  # Test lesmis data
  expect_true(exists("lesmis"))
  data("lesmis", package = "g6R")
  expect_type(lesmis, "list")
  expect_true("nodes" %in% names(lesmis))
  expect_true("edges" %in% names(lesmis))

  # Test tree data
  expect_true(exists("tree"))
  data("tree", package = "g6R")
  expect_type(tree, "list")
  expect_true("nodes" %in% names(tree))

  # Test radial data
  expect_true(exists("radial"))
  data("radial", package = "g6R")
  expect_type(radial, "list")
  expect_true("nodes" %in% names(radial))

  # Test poke data
  expect_true(exists("poke"))
  data("poke", package = "g6R")
  expect_type(poke, "list")
})

test_that("data objects have required node structure", {
  data("dag", package = "g6R")

  # Check nodes structure
  if (length(dag$nodes) > 0) {
    first_node <- dag$nodes[1, ]
    expect_true("id" %in% names(first_node))
    expect_type(first_node$id, "character")
  }

  # Check edges structure if present
  if ("edges" %in% names(dag) && length(dag$edges) > 0) {
    first_edge <- dag$edges[1, ]
    expect_true("source" %in% names(first_edge))
    expect_true("target" %in% names(first_edge))
  }
})

test_that("data objects work with g6 function", {
  data("dag", package = "g6R")

  # Should be able to create g6 visualizations with the data
  expect_no_error({
    g <- g6(nodes = dag$nodes, edges = dag$edges)
  })

  g <- g6(nodes = dag$nodes, edges = dag$edges)
  expect_s3_class(g, "htmlwidget")
  expect_s3_class(g, "g6")
})

---
title: "g6R"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{g6R}
  %\VignetteEngine{quarto::html}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, message=FALSE}
library(g6R)
library(shiny)
```

## Graph data

### Data formats

To setup a `{g6R}` graph, you first have to define __node__, __edges__ and __combos__ (collection of nodes), like so:

```{r}
nodes <- data.frame(id = 1:10)
```

You could also pass these elements as __list__, which will be faster than the dataframe approach since under the hood, `{g6R}` has to send list to JavaScript:

```r
nodes <- lapply(seq_len(10), function(i) {
  list(id = i)
})
```

We then define some random edges:

```{r}
# Set a seed for reproducibility
set.seed(123)

# Define the number of edges to create (e.g., 200 random connections)
num_edges <- 5

# Generate random edges
edges <- data.frame(
  source = sample(nodes$id, num_edges, replace = TRUE),
  target = sample(nodes$id, num_edges, replace = TRUE)
)

edges$id <- paste0(edges$source, edges$target)
duplicated_id <- which(duplicated(edges$id) == TRUE)
if (length(duplicated_id)) {
  edges <- edges[-duplicated_id, ]
}
```

In `{g6R}`, the recommended way to assemble graph data is by using the helper functions `g6_data()` and `as_g6_data()`.  
These functions provide a consistent and user-friendly interface for combining nodes, edges, and combos into a single graph data object.  
`g6_data()` allows you to explicitly construct a graph data object from validated nodes, edges, and combos, while `as_g6_data()` 
can coerce various input formats (lists, data frames) into a graph data object.

You can list available methods for `as_g6_data` with:

```{r}
as.list(methods("as_g6_data"))
```

Using these helpers ensures compatibility with all `{g6R}` features and is the recommended method for graph data creation.

For example:

```{r}
# Create nodes and edges
my_nodes <- data.frame(id = c("A", "B"))
my_edges <- data.frame(source = "A", target = "B")

# Assemble graph data using g6_data()
graph <- g6_data(
  nodes = my_nodes,
  edges = my_edges
)
graph

# Or use as_g6_data() directly with a list
lst <- list(
  nodes = list(
    list(id = "A"),
    list(id = "B")
  ),
  edges = list(
    list(source = "A", target = "B")
  )
)
graph2 <- as_g6_data(lst)

all.equal(graph, graph2)
```

### Performance

If you are dealing with large graphs, you might want to use the `jsonUrl` parameter to __fetch__ the graph data from a hosted JSON file instead. If so, then leave `nodes`, `edges` and `combos` `NULL`. This allows to bypass the serialization of the graph data to JavaScript, which can be slow for large graphs. That's done in the `shinyAppDir(system.file("examples", "json", package = "g6R"))` example.

```{r g6R-json, eval=FALSE, echo = TRUE}
shinyAppDir(system.file("examples", "json", package = "g6R"))
```

```{r shinylive_url_2, echo = FALSE, results = 'asis'}
# extract the code from knitr code chunks by ID
code <- paste0(
  c(
    "webr::install(\"g6R\", repos = \"https://cynkra.github.io/blockr.webR/\")",
    knitr::knit_code$get("g6R-json")
  ),
  collapse = "\n"
)

url <- roxy.shinylive::create_shinylive_url(code, header = FALSE)
```

```{r shinylive_iframe_2, echo = FALSE, eval = TRUE}
tags$iframe(
  class = "border border-5 rounded shadow-lg",
  src = url,
  style = "zoom: 0.75;",
  width = "100%",
  height = "1100px"
)
```

## Initialise the graph

We leverage `g6` to create an __instance__ of our network:

```{r}
g6(nodes, edges, width = 200, height = 200)
```

As you can see, the nodes don't render well yet. Let's add it some layout.

## Layout

`g6_layout()` allows to pass any supported [layout](https://g6.antv.antgroup.com/en/manual/layout/overview). For this example, we select the `d3_force_layout()`:

```{r}
g <- g6(nodes, edges) |>
  g6_layout(d3_force_layout())
g
```

That's better! We could go further by displaying nodes __label__.

## Tweak options

`g6_options()` is your to go tool when it comes to change the style of the graph element such as nodes. Properties are selected from the [documentation](https://g6.antv.antgroup.com/en/manual/element/node/build-in/base-node#main-graphic-style):

```{r}
g <- g |>
  g6_options(
    node = list(
      style = list(
        labelBackground = TRUE,
        labelBackgroundFill = '#FFB6C1',
        labelBackgroundRadius = 4,
        labelFontFamily = 'Arial',
        labelPadding = c(0, 4),
        labelText = JS(
          "(d) => {
              return d.id
            }"
        )
      )
    )
  )
g
```

## Plugins

Plugins allowing users to improve the user experience by adding graphical components to the canvas like __minimaps__ or __tooltips__.
We can pass them inside `g6_plugins` either as a character string with a reference to the plugin name or using the correponding function to pass more configuration options:

```r
# Use defaults
g6_plugins("minimap")

# Custom options
g6_plugins(
  minimap(size = c(100, 100))
)
```

```{r}
g <- g |>
  g6_plugins(
    minimap(size = c(100, 100))
  )
g
```

## Behaviors

Behaviors correspond to interactions between users and the graph elements, such as dragging the nodes and selecting nodes. Be mindful that not all behaviors are compatible: dragging canvas and canvas elements would require to specify different triggers for each behavior.
With `{g6R}` behaviors can be added with `g6_behaviors`, like plugins:

```{r}
g <- g |>
  g6_behaviors(
    "zoom-canvas",
    drag_element_force(fixed = TRUE),
    click_select(
      multiple = TRUE,
      onClick = JS(
        "(e) => {
            console.log(e);
          }"
      )
    ),
    brush_select()
  )
g
```

Notice that we can pass __callback__ functions from R to JavaScript. This is useful in combination with Shiny to set custom inputs, for instance.

# g6_port validation fails for missing or invalid key/type/multiple

    Code
      g6_port(key = "", type = "input")
    Condition
      Error in `validate_port.g6_port()`:
      ! 'key' must be a non-empty character string.

---

    Code
      g6_port(key = "x", type = "foo")
    Condition
      Error in `match.arg()`:
      ! 'arg' should be one of "input", "output"

---

    Code
      g6_port(key = "x", type = "input", arity = -1)
    Condition
      Error in `validate_port.g6_port()`:
      ! 'arity' must be a single non-negative number (0, Inf, or positive integer).

# g6_ports fails if keys are not unique or elements are not g6_port

    Code
      g6_ports(g6_port("a", type = "input"), g6_port("a", type = "output"))
    Condition
      Error in `validate_ports.list()`:
      ! All port 'key' values must be unique. Duplicates: a

---

    Code
      g6_ports(g6_port("a", type = "input"), list(key = "b", type = "output"))
    Condition
      Error in `validate_ports.list()`:
      ! All elements must be of class 'g6_port'.

# g6 function creates proper htmlwidget

    Code
      g6(iconsUrl = NULL)
    Condition
      Error in `g6()`:
      ! !is.null(iconsUrl) is not TRUE
    Code
      g6(nodes = data.frame(), jsonUrl = "https://assets.antv.antgroup.com/g6/cluster.json")
    Condition
      Error in `g6()`:
      ! Can't use jsonUrl` argument at the same time as `nodes`, `edges`, and `combos`.

# g6_plugins creates proper plugin configuration

    Code
      g6_plugins(g6())
    Condition
      Error in `g6_plugins()`:
      ! You must provide at least one plugin configuration.

# validate plugins works

    Code
      g6_plugins(g6(), "blabla")
    Condition
      Error in `validate_component()`:
      ! 'blabla' is not a valid plugin. Valid choices are: background, bubble-sets, contextmenu, edge-bundling, edge-filter-lens, fisheye, fullscreen, grid-line, history, hull, legend, minimap, snapline, timebar, toolbar, tooltip, watermark.
    Code
      g6_plugins(g6(), list(test = 1))
    Condition
      Error in `validate_component()`:
      ! 'unknown' is not a valid plugin. Valid choices are: background, bubble-sets, contextmenu, edge-bundling, edge-filter-lens, fisheye, fullscreen, grid-line, history, hull, legend, minimap, snapline, timebar, toolbar, tooltip, watermark.

# individual plugin functions work correctly

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 6
        .. .. ..$ width         : chr "100%"
        .. .. ..$ height        : chr "100%"
        .. .. ..$ backgroundSize: chr "cover"
        .. .. ..$ transition    : chr "background 0.5s"
        .. .. ..$ zIndex        : chr "-1"
        .. .. ..$ type          : chr "background"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Condition
      Error in `valid_choices[[x]]()`:
      ! argument "members" is missing, with no default

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 8
        .. .. ..$ key      : chr "contextmenu"
        .. .. ..$ className: chr "g6-contextmenu"
        .. .. ..$ trigger  : chr "contextmenu"
        .. .. ..$ offset   : num [1:2] 4 4
        .. .. ..$ onClick  : 'JS_EVAL' chr "(value, target, current) => {\n        const graph = HTMLWidgets\n          .find(`#${target.closest('.g6').id}"| __truncated__
        .. .. ..$ getItems : 'JS_EVAL' chr "() => {\n        return [\n          { name: 'Create edge', value: 'create_edge' },\n          { name: 'Remove "| __truncated__
        .. .. ..$ enable   : 'JS_EVAL' chr "(e) => e.targetType === 'node'"
        .. .. ..$ type     : chr "contextmenu"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 10
        .. .. ..$ key            : chr "edge-bundling"
        .. .. ..$ bundleThreshold: num 0.6
        .. .. ..$ cycles         : num 6
        .. .. ..$ divisions      : num 1
        .. .. ..$ divRate        : num 2
        .. .. ..$ iterations     : num 90
        .. .. ..$ iterRate       : num 0.667
        .. .. ..$ K              : num 0.1
        .. .. ..$ lambda         : num 0.1
        .. .. ..$ type           : chr "edge-bundling"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 10
        .. .. ..$ key           : chr "edge-filter-lens"
        .. .. ..$ trigger       : chr "pointermove"
        .. .. ..$ r             : num 60
        .. .. ..$ minR          : num 0
        .. .. ..$ scaleRBy      : chr "wheel"
        .. .. ..$ nodeType      : chr "both"
        .. .. ..$ nodeStyle     :List of 1
        .. .. .. ..$ label: logi FALSE
        .. .. ..$ edgeStyle     :List of 1
        .. .. .. ..$ label: logi TRUE
        .. .. ..$ preventDefault: logi TRUE
        .. .. ..$ type          : chr "edge-filter-lens"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 11
        .. .. ..$ key           : chr "fish-eye"
        .. .. ..$ trigger       : chr "pointermove"
        .. .. ..$ r             : num 120
        .. .. ..$ minR          : num 0
        .. .. ..$ d             : num 1.5
        .. .. ..$ maxD          : num 5
        .. .. ..$ minD          : num 0
        .. .. ..$ showDPercent  : logi TRUE
        .. .. ..$ nodeStyle     :List of 1
        .. .. .. ..$ label: logi TRUE
        .. .. ..$ preventDefault: logi TRUE
        .. .. ..$ type          : chr "fisheye"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 3
        .. .. ..$ key    : chr "fullscreen"
        .. .. ..$ autoFit: logi TRUE
        .. .. ..$ type   : chr "fullscreen"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 10
        .. .. ..$ key            : chr "grid-line"
        .. .. ..$ border         : logi TRUE
        .. .. ..$ borderLineWidth: num 1
        .. .. ..$ borderStroke   : chr "#eee"
        .. .. ..$ borderStyle    : chr "solid"
        .. .. ..$ follow         : logi FALSE
        .. .. ..$ lineWidth      : num 1
        .. .. ..$ size           : num 20
        .. .. ..$ stroke         : chr "#eee"
        .. .. ..$ type           : chr "grid-line"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 3
        .. .. ..$ key      : chr "history"
        .. .. ..$ stackSize: num 0
        .. .. ..$ type     : chr "history"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Condition
      Error in `valid_choices[[x]]()`:
      ! argument "members" is missing, with no default

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 15
        .. .. ..$ key              : chr "legend"
        .. .. ..$ trigger          : chr "hover"
        .. .. ..$ position         : chr "bottom"
        .. .. ..$ orientation      : chr "horizontal"
        .. .. ..$ layout           : chr "flex"
        .. .. ..$ showTitle        : logi FALSE
        .. .. ..$ titleText        : chr ""
        .. .. ..$ width            : num 240
        .. .. ..$ height           : num 160
        .. .. ..$ itemSpacing      : num 4
        .. .. ..$ rowPadding       : num 10
        .. .. ..$ colPadding       : num 10
        .. .. ..$ itemMarkerSize   : num 16
        .. .. ..$ itemLabelFontSize: num 16
        .. .. ..$ type             : chr "legend"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 7
        .. .. ..$ key     : chr "minimap"
        .. .. ..$ delay   : num 128
        .. .. ..$ padding : num 10
        .. .. ..$ position: chr "right-bottom"
        .. .. ..$ shape   : chr "key"
        .. .. ..$ size    : num [1:2] 240 160
        .. .. ..$ type    : chr "minimap"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 8
        .. .. ..$ key                : chr "snapline"
        .. .. ..$ tolerance          : num 5
        .. .. ..$ offset             : num 20
        .. .. ..$ autoSnap           : logi TRUE
        .. .. ..$ shape              : chr "key"
        .. .. ..$ verticalLineStyle  :List of 1
        .. .. .. ..$ stroke: chr "#1783FF"
        .. .. ..$ horizontalLineStyle:List of 1
        .. .. .. ..$ stroke: chr "#1783FF"
        .. .. ..$ type               : chr "snapline"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Condition
      Error in `valid_choices[[x]]()`:
      ! argument "data" is missing, with no default

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 5
        .. .. ..$ getItems: 'JS_EVAL' chr "( ) => [\n        { id : 'zoom-in' , value : 'zoom-in' },\n        { id : 'zoom-out' , value : 'zoom-out' },\n "| __truncated__
        .. .. ..$ key     : chr "toolbar"
        .. .. ..$ position: chr "top-left"
        .. .. ..$ onClick : 'JS_EVAL' chr "( value, target, current ) => {   \n        // Handle button click events\n        const graph = HTMLWidgets\n "| __truncated__
        .. .. ..$ type    : chr "toolbar"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 7
        .. .. ..$ key      : chr "tooltip"
        .. .. ..$ position : chr "top-right"
        .. .. ..$ enable   : logi TRUE
        .. .. ..$ trigger  : chr "hover"
        .. .. ..$ offset   : num [1:2] 10 10
        .. .. ..$ enterable: logi FALSE
        .. .. ..$ type     : chr "tooltip"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_plugins(g6(), plugin))
    Condition
      Warning in `valid_choices[[x]]()`:
      Neither 'imageURL' nor 'text' is provided; watermark may not be visible
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ plugins         :List of 1
        .. ..$ :List of 11
        .. .. ..$ key             : chr "watermark"
        .. .. ..$ width           : num 200
        .. .. ..$ height          : num 100
        .. .. ..$ opacity         : num 0.2
        .. .. ..$ rotate          : num 0.262
        .. .. ..$ textFill        : chr "#000"
        .. .. ..$ textFontSize    : num 16
        .. .. ..$ textAlign       : chr "center"
        .. .. ..$ textBaseline    : chr "alphabetic"
        .. .. ..$ backgroundRepeat: chr "repeat"
        .. .. ..$ type            : chr "watermark"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

# validate layout works

    Code
      g6_layout(g6(), NULL)
    Condition
      Error in `g6_layout()`:
      ! layout must be specified.
    Code
      g6_layout(g6(), "blabla")
    Condition
      Error in `validate_component()`:
      ! 'blabla' is not a valid layout. Valid choices are: antv-dagre, circular, combo-combined, concentric, d3-force, force-atlas2, fruchterman, radial, compact-box, dendrogram.
    Code
      g6_layout(g6(), list(test = 1))
    Condition
      Error in `validate_component()`:
      ! 'unknown' is not a valid layout. Valid choices are: antv-dagre, circular, combo-combined, concentric, d3-force, force-atlas2, fruchterman, radial, compact-box, dendrogram.

# individual layout functions work correctly

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 10
        .. ..$ type          : chr "antv-dagre"
        .. ..$ rankdir       : chr "TB"
        .. ..$ align         : chr "UL"
        .. ..$ nodesep       : num 50
        .. ..$ ranksep       : num 100
        .. ..$ ranker        : chr "network-simplex"
        .. ..$ controlPoints : logi FALSE
        .. ..$ sortByCombo   : logi FALSE
        .. ..$ edgeLabelSpace: logi TRUE
        .. ..$ radial        : logi FALSE
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 8
        .. ..$ type       : chr "circular"
        .. ..$ angleRatio : num 1
        .. ..$ clockwise  : logi TRUE
        .. ..$ divisions  : num 1
        .. ..$ nodeSize   : num 10
        .. ..$ nodeSpacing: num 10
        .. ..$ startAngle : num 0
        .. ..$ endAngle   : num 6.28
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 3
        .. ..$ type        : chr "combo-combined"
        .. ..$ comboPadding: num 10
        .. ..$ nodeSize    : num 10
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 8
        .. ..$ type          : chr "concentric"
        .. ..$ clockwise     : logi FALSE
        .. ..$ equidistant   : logi FALSE
        .. ..$ sortBy        : chr "degree"
        .. ..$ nodeSize      : num 30
        .. ..$ nodeSpacing   : num 10
        .. ..$ preventOverlap: logi FALSE
        .. ..$ startAngle    : num 4.71
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 3
        .. ..$ type   : chr "d3-force"
        .. ..$ link   :List of 2
        .. .. ..$ distance: num 100
        .. .. ..$ strength: num 2
        .. ..$ collide:List of 1
        .. .. ..$ radius: num 40
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 9
        .. ..$ type          : chr "force-atlas2"
        .. ..$ dissuadeHubs  : logi FALSE
        .. ..$ kg            : num 1
        .. ..$ kr            : num 5
        .. ..$ ks            : num 0.1
        .. ..$ ksmax         : num 10
        .. ..$ mode          : chr "normal"
        .. ..$ preventOverlap: logi FALSE
        .. ..$ tao           : num 0.1
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 3
        .. ..$ type   : chr "fruchterman"
        .. ..$ gravity: num 10
        .. ..$ speed  : num 5
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 9
        .. ..$ type                      : chr "radial"
        .. ..$ nodeSpacing               : num 10
        .. ..$ linkDistance              : num 50
        .. ..$ unitRadius                : num 100
        .. ..$ maxIteration              : num 1000
        .. ..$ maxPreventOverlapIteration: num 200
        .. ..$ preventOverlap            : logi FALSE
        .. ..$ sortStrength              : num 10
        .. ..$ strictRadial              : logi TRUE
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 3
        .. ..$ type     : chr "compact-box"
        .. ..$ direction: chr "LR"
        .. ..$ radial   : logi FALSE
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_layout(g6(), layout))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi TRUE
        ..$ maxCollapseDepth: num Inf
        ..$ layout          :List of 5
        .. ..$ type     : chr "dendrogram"
        .. ..$ direction: chr [1:6] "LR" "RL" "TB" "BT" ...
        .. ..$ nodeSep  : num 20
        .. ..$ rankSep  : num 200
        .. ..$ radial   : logi FALSE
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

# g6_data_proxy handles inputs correctly

    Code
      g6_data_proxy(list(), test_df, "add", "Node")
    Condition
      Error in `g6_data_proxy()`:
      ! Can't use g6_add_* with g6 object. Only within shiny and using g6_proxy.

# g6_fit_center works correctly

    Code
      g6_fit_center(proxy, "invalid")
    Condition
      Error in `g6_fit_center()`:
      ! is.list(animation) is not TRUE

# g6_igraph errors for non-igraph objects

    Code
      g6_igraph("not an igraph")
    Condition
      Error in `ensure_igraph()`:
      ! Must provide a graph object (provided wrong object type).
    Code
      g6_igraph(list())
    Condition
      Error in `ensure_igraph()`:
      ! Must provide a graph object (provided wrong object type).
    Code
      g6_igraph(123)
    Condition
      Error in `ensure_igraph()`:
      ! Must provide a graph object (provided wrong object type).

# g6_behaviors creates proper behavior configuration

    Code
      g6_behaviors(g6())
    Condition
      Error in `g6_behaviors()`:
      ! You must provide at least one behavior configuration.

# validate behavior works

    Code
      g6_behaviors(g6(), "blabla")
    Condition
      Error in `validate_component()`:
      ! 'blabla' is not a valid behavior. Valid choices are: auto-adapt-label, brush-select, click-select, collapse-expand, create-edge, drag-canvas, drag-element, drag-element-force, fix-element-size, focus-element, hover-activate, lasso-select, optimize-viewport-transform, scroll-canvas, zoom-canvas.
    Code
      g6_behaviors(g6(), list(test = 1))
    Condition
      Error in `validate_component()`:
      ! 'unknown' is not a valid behavior. Valid choices are: auto-adapt-label, brush-select, click-select, collapse-expand, create-edge, drag-canvas, drag-element, drag-element-force, fix-element-size, focus-element, hover-activate, lasso-select, optimize-viewport-transform, scroll-canvas, zoom-canvas.

# individual behavior functions work correctly

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 6
        .. .. ..$ key     : chr "auto-adapt-label"
        .. .. ..$ enable  : logi TRUE
        .. .. ..$ throttle: num 100
        .. .. ..$ padding : num 0
        .. .. ..$ sortNode:List of 1
        .. .. .. ..$ type: chr "degree"
        .. .. ..$ type    : chr "auto-adapt-label"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 9
        .. .. ..$ key           : chr "brush-select"
        .. .. ..$ animation     : logi FALSE
        .. .. ..$ enable        : 'JS_EVAL' chr "(e) => {\n      return true;\n    }"
        .. .. ..$ enableElements:List of 1
        .. .. .. ..$ : chr "node"
        .. .. ..$ immediately   : logi FALSE
        .. .. ..$ mode          : chr "default"
        .. .. ..$ state         : chr "selected"
        .. .. ..$ trigger       : chr "shift"
        .. .. ..$ type          : chr "brush-select"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 9
        .. .. ..$ key          : chr "click-select"
        .. .. ..$ animation    : logi TRUE
        .. .. ..$ degree       : num 0
        .. .. ..$ enable       : logi TRUE
        .. .. ..$ multiple     : logi FALSE
        .. .. ..$ state        : chr "selected"
        .. .. ..$ neighborState: chr "selected"
        .. .. ..$ trigger      :List of 1
        .. .. .. ..$ : chr "shift"
        .. .. ..$ type         : chr "click-select"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 6
        .. .. ..$ key      : chr "collapse-expand"
        .. .. ..$ animation: logi TRUE
        .. .. ..$ enable   : logi TRUE
        .. .. ..$ trigger  : chr "dblclick"
        .. .. ..$ align    : logi TRUE
        .. .. ..$ type     : chr "collapse-expand"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 5
        .. .. ..$ key    : chr "create-edge"
        .. .. ..$ trigger: chr "drag"
        .. .. ..$ enable : logi FALSE
        .. .. ..$ notify : logi FALSE
        .. .. ..$ type   : chr "create-edge"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 5
        .. .. ..$ key        : chr "drag-canvas"
        .. .. ..$ enable     : 'JS_EVAL' chr "(e) => {\n        return e.targetType === 'canvas';\n      }"
        .. .. ..$ direction  : chr "both"
        .. .. ..$ sensitivity: num 10
        .. .. ..$ type       : chr "drag-canvas"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 8
        .. .. ..$ key       : chr "drag-element"
        .. .. ..$ enable    : 'JS_EVAL' chr "(e) => {\n        const target = e.nativeEvent?.target;\n        const graph = HTMLWidgets.find(`#${target?.clo"| __truncated__
        .. .. ..$ animation : logi TRUE
        .. .. ..$ state     : chr "selected"
        .. .. ..$ dropEffect: chr "move"
        .. .. ..$ hideEdge  : chr "none"
        .. .. ..$ shadow    : logi FALSE
        .. .. ..$ type      : chr "drag-element"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 6
        .. .. ..$ key     : chr "drag-element-force"
        .. .. ..$ fixed   : logi FALSE
        .. .. ..$ enable  : 'JS_EVAL' chr "(e) => {\n        const target = e.nativeEvent?.target;\n        const graph = HTMLWidgets.find(`#${target?.clo"| __truncated__
        .. .. ..$ state   : chr "selected"
        .. .. ..$ hideEdge: chr "none"
        .. .. ..$ type    : chr "drag-element-force"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 9
        .. .. ..$ key        : chr "fix-element-size"
        .. .. ..$ enable     : logi TRUE
        .. .. ..$ reset      : logi FALSE
        .. .. ..$ state      : chr ""
        .. .. ..$ nodeFilter : 'JS_EVAL' chr "() => true"
        .. .. ..$ edge       :List of 3
        .. .. .. ..$ :List of 2
        .. .. .. .. ..$ shape : chr "key"
        .. .. .. .. ..$ fields: chr "lineWidth"
        .. .. .. ..$ :List of 2
        .. .. .. .. ..$ shape : chr "halo"
        .. .. .. .. ..$ fields: chr "lineWidth"
        .. .. .. ..$ :List of 2
        .. .. .. .. ..$ shape : chr "label"
        .. .. .. .. ..$ fields: chr "fontSize"
        .. .. ..$ edgeFilter : 'JS_EVAL' chr "() => true"
        .. .. ..$ comboFilter: 'JS_EVAL' chr "() => true"
        .. .. ..$ type       : chr "fix-element-size"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 4
        .. .. ..$ key      : chr "focus-element"
        .. .. ..$ animation:List of 2
        .. .. .. ..$ duration: num 500
        .. .. .. ..$ easing  : chr "ease-in"
        .. .. ..$ enable   : logi TRUE
        .. .. ..$ type     : chr "focus-element"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 7
        .. .. ..$ key      : chr "hover-activate"
        .. .. ..$ animation: logi TRUE
        .. .. ..$ enable   : logi TRUE
        .. .. ..$ degree   : num 0
        .. .. ..$ direction: chr "both"
        .. .. ..$ state    : chr "active"
        .. .. ..$ type     : chr "hover-activate"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 9
        .. .. ..$ key           : chr "lasso-select"
        .. .. ..$ animation     : logi FALSE
        .. .. ..$ enable        : logi TRUE
        .. .. ..$ enableElements:List of 1
        .. .. .. ..$ : chr "node"
        .. .. ..$ immediately   : logi FALSE
        .. .. ..$ mode          : chr "default"
        .. .. ..$ state         : chr "selected"
        .. .. ..$ trigger       : chr "shift"
        .. .. ..$ type          : chr "lasso-select"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 5
        .. .. ..$ key     : chr "optimize-viewport-transform"
        .. .. ..$ enable  : logi TRUE
        .. .. ..$ debounce: num 200
        .. .. ..$ shapes  : 'JS_EVAL' chr "(type) => type === 'node'"
        .. .. ..$ type    : chr "optimize-viewport-transform"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 6
        .. .. ..$ key           : chr "scroll-canvas"
        .. .. ..$ enable        : logi TRUE
        .. .. ..$ range         : num 1
        .. .. ..$ sensitivity   : num 1
        .. .. ..$ preventDefault: logi TRUE
        .. .. ..$ type          : chr "scroll-canvas"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

---

    Code
      str(g6_behaviors(g6(), behavior))
    Output
      List of 8
       $ x            :List of 8
        ..$ data            : list()
        .. ..- attr(*, "class")= chr "g6_data"
        ..$ jsonUrl         : NULL
        ..$ iconsUrl        : chr "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js"
        ..$ mode            : chr "prod"
        ..$ preservePosition: logi FALSE
        ..$ directed        : logi FALSE
        ..$ maxCollapseDepth: num Inf
        ..$ behaviors       :List of 1
        .. ..$ :List of 6
        .. .. ..$ key           : chr "zoom-canvas"
        .. .. ..$ animation     :List of 1
        .. .. .. ..$ duration: num 200
        .. .. ..$ enable        : logi TRUE
        .. .. ..$ preventDefault: logi TRUE
        .. .. ..$ sensitivity   : num 1
        .. .. ..$ type          : chr "zoom-canvas"
       $ width        : chr "100%"
       $ height       : NULL
       $ sizingPolicy :List of 7
        ..$ defaultWidth : NULL
        ..$ defaultHeight: NULL
        ..$ padding      : NULL
        ..$ fill         : NULL
        ..$ viewer       :List of 6
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi TRUE
        .. ..$ suppress     : logi FALSE
        .. ..$ paneHeight   : NULL
        ..$ browser      :List of 5
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ padding      : NULL
        .. ..$ fill         : logi FALSE
        .. ..$ external     : logi FALSE
        ..$ knitr        :List of 3
        .. ..$ defaultWidth : NULL
        .. ..$ defaultHeight: NULL
        .. ..$ figure       : logi TRUE
       $ dependencies : NULL
       $ elementId    : NULL
       $ preRenderHook: NULL
       $ jsHooks      : list()
       - attr(*, "class")= chr [1:2] "g6" "htmlwidget"
       - attr(*, "package")= chr "g6R"

# g6_collapse_options rejects invalid visibility

    Code
      g6_collapse_options(visibility = "invalid")
    Condition
      Error in `match.arg()`:
      ! 'arg' should be one of "visible", "hover"

# g6_collapse_options validates placement with match.arg

    Code
      g6_collapse_options(placement = "invalid-placement")
    Condition
      Error in `match.arg()`:
      ! 'arg' should be one of "top", "right", "bottom", "left", "right-top", "right-bottom", "left-top", "left-bottom"

# g6_collapse_options validates numeric placement length

    Code
      g6_collapse_options(placement = c(0))
    Condition
      Error in `g6_collapse_options()`:
      ! Numeric placement must be a vector of length 2.

---

    Code
      g6_collapse_options(placement = c(0, 0.5, 1))
    Condition
      Error in `g6_collapse_options()`:
      ! Numeric placement must be a vector of length 2.

# g6_collapse_options validates numeric placement has at least one edge coordinate

    Code
      g6_collapse_options(placement = c(0.5, 0.5))
    Condition
      Error in `g6_collapse_options()`:
      ! Invalid collapse button placement: at least one coordinate must be 0 or 1 (i.e., on the node edge). You supplied: c(0.5, 0.5)

---

    Code
      g6_collapse_options(placement = c(0.3, 0.7))
    Condition
      Error in `g6_collapse_options()`:
      ! Invalid collapse button placement: at least one coordinate must be 0 or 1 (i.e., on the node edge). You supplied: c(0.3, 0.7)

# g6_collapse_options rejects invalid placement type

    Code
      g6_collapse_options(placement = list(0, 0.5))
    Condition
      Error in `g6_collapse_options()`:
      ! placement must be a character string or numeric vector of length 2.

---

    Code
      g6_collapse_options(placement = NULL)
    Condition
      Error in `g6_collapse_options()`:
      ! placement must be a character string or numeric vector of length 2.

# g6_node validation fails for invalid collapse

    Code
      g6_node(id = "n1", children = c("n2"), collapse = list(collapsed = TRUE,
        placement = "top"))
    Condition
      Error in `validate_element.g6_node()`:
      ! Node 'collapse' must be of class 'g6_collapse_options'.

---

    Code
      g6_node(id = "n1", children = c("n2"), collapse = "invalid")
    Condition
      Error in `validate_element.g6_node()`:
      ! Node 'collapse' must be of class 'g6_collapse_options'.

# g6_node validation fails for missing id

    Code
      g6_node(type = "rect")
    Condition
      Error in `g6_node()`:
      ! argument "id" is missing, with no default

# g6_edge validation fails for missing source/target

    Code
      g6_edge(source = NULL, target = "n2")
    Condition
      Error in `validate_element.g6_edge()`:
      ! Edge 'source' is required and must be a non-empty character string.

---

    Code
      g6_edge(source = "n1", target = NULL)
    Condition
      Error in `validate_element.g6_edge()`:
      ! Edge 'target' is required and must be a non-empty character string.

# g6_combo validation fails for missing id

    Code
      g6_combo(type = "rect")
    Condition
      Error in `g6_combo()`:
      ! argument "id" is missing, with no default

# g6_nodes fails if any element is not a g6_node

    Code
      g6_nodes(g6_node(id = "n1"), g6_edge(source = "n1", target = "n2"))
    Condition
      Error in `validate_elements.g6_nodes()`:
      ! All elements must be of class 'g6_node'.

# g6_edges fails if any element is not a g6_edge

    Code
      g6_edges(g6_edge(source = "n1", target = "n2"), g6_node(id = "n2"))
    Condition
      Error in `validate_elements.g6_edges()`:
      ! All elements must be of class 'g6_edge'.

# g6_combos fails if any element is not a g6_combo

    Code
      g6_combos(g6_combo(id = "c1"), g6_node(id = "n1"))
    Condition
      Error in `validate_elements.g6_combos()`:
      ! All elements must be of class 'g6_combo'.

# g6_nodes fails if no nodes provided

    Code
      g6_nodes()
    Condition
      Error in `validate_elements()`:
      ! length(x) > 0 is not TRUE

# g6_edges fails if no edges provided

    Code
      g6_edges()
    Condition
      Error in `validate_elements()`:
      ! length(x) > 0 is not TRUE

# g6_combos fails if no combos provided

    Code
      g6_combos()
    Condition
      Error in `validate_elements()`:
      ! length(x) > 0 is not TRUE

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6.R
\name{g6-shiny}
\alias{g6-shiny}
\alias{g6Output}
\alias{g6_output}
\alias{renderG6}
\alias{render_g6}
\title{Shiny bindings for g6}
\usage{
g6Output(outputId, width = "100\%", height = "400px")

g6_output(outputId, width = "100\%", height = "400px")

renderG6(expr, env = parent.frame(), quoted = FALSE)

render_g6(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{outputId}{output variable to read from}

\item{width, height}{Must be a valid CSS unit (like \code{'100\%'},
\code{'400px'}, \code{'auto'}) or a number, which will be coerced to a
string and have \code{'px'} appended.}

\item{expr}{An expression that generates a g6}

\item{env}{The environment in which to evaluate \code{expr}.}

\item{quoted}{Is \code{expr} a quoted expression (with \code{quote()})? This
is useful if you want to save an expression in a variable.}
}
\value{
\code{g6Output} and \code{g6_output}
return a Shiny output function that can be used in the UI part of a Shiny app.
\code{renderG6} and \code{render_g6} return a
Shiny render function that can be used in the server part of a Shiny app to
render a \code{g6} element.
}
\description{
Output and render functions for using g6 within Shiny
applications and interactive Rmd documents.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg.R
\docType{package}
\name{g6R-package}
\alias{g6R}
\alias{g6R-package}
\title{g6R: Graph Visualisation Engine Widget for R and 'shiny' Apps}
\description{
Create stunning network experiences powered by the 'G6' graph visualisation engine 'JavaScript' library \url{https://g6.antv.antgroup.com/en}. In 'shiny' mode, modify your graph directly from the server function to dynamically interact with nodes and edges. Select your favorite layout among 20 choices. 15 behaviors are available such as interactive edge creation, collapse-expand and brush select. 17 plugins designed to improve the user experience such as a mini-map, toolbars and grid lines. Customise the look and feel of your graph with comprehensive options for nodes, edges and more.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/cynkra/g6R}
  \item \url{https://cynkra.github.io/g6R/}
  \item Report bugs at \url{https://github.com/cynkra/g6R/issues}
}

}
\author{
\strong{Maintainer}: David Granjon \email{dgranjon@ymail.com}

Authors:
\itemize{
  \item David Schoch \email{david@schochastics.net}
}

Other contributors:
\itemize{
  \item  cynkra GmbH [funder]
  \item  Bristol Myers Squibb [funder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{edge_bundling}
\alias{edge_bundling}
\title{Configure Edge Bundling Plugin}
\usage{
edge_bundling(
  key = "edge-bundling",
  bundleThreshold = 0.6,
  cycles = 6,
  divisions = 1,
  divRate = 2,
  iterations = 90,
  iterRate = 2/3,
  K = 0.1,
  lambda = 0.1,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{bundleThreshold}{Edge compatibility threshold, determines which edges
should be bundled together (number, default: 0.6).}

\item{cycles}{Number of simulation cycles (number, default: 6).}

\item{divisions}{Initial number of cut points (number, default: 1).}

\item{divRate}{Growth rate of cut points (number, default: 2).}

\item{iterations}{Number of iterations executed in the first cycle (number, default: 90).}

\item{iterRate}{Iteration decrement rate (number, default: 2/3).}

\item{K}{Edge strength, affects attraction and repulsion between edges (number, default: 0.1).}

\item{lambda}{Initial step size (number, default: 0.1).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/edge-bundling}.}
}
\value{
A list with the configuration settings for the edge-bundling plugin.
}
\description{
Creates a configuration object for the edge-bundling plugin in G6.
This plugin automatically bundles similar edges together to reduce visual clutter.
}
\examples{
# Basic configuration
config <- edge_bundling()

# Custom configuration
config <- edge_bundling(
  key = "my-edge-bundling",
  bundleThreshold = 0.8,
  cycles = 8,
  K = 0.2
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{fix_element_size}
\alias{fix_element_size}
\title{Configure Fix Element Size Behavior}
\usage{
fix_element_size(
  key = "fix-element-size",
  enable = TRUE,
  reset = FALSE,
  state = "",
  node = NULL,
  nodeFilter = JS("() => true"),
  edge = list(list(shape = "key", fields = c("lineWidth")), list(shape = "halo", fields =
    c("lineWidth")), list(shape = "label", fields = c("fontSize"))),
  edgeFilter = JS("() => true"),
  combo = NULL,
  comboFilter = JS("() => true"),
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior, used for subsequent operations
(string, default: "fix-element-size").}

\item{enable}{Whether to enable this interaction (boolean or JS function, default: TRUE).}

\item{reset}{Whether to restore style when elements are redrawn (boolean, default: FALSE).}

\item{state}{Specify the state of elements to fix size (string, default: "")}

\item{node}{Node configuration item(s) to define which attributes maintain fixed
size (list or array of lists, default: NULL).}

\item{nodeFilter}{Node filter to determine which nodes maintain fixed size
(JS function, default: returns TRUE for all nodes).}

\item{edge}{Edge configuration item(s) to define which attributes maintain fixed size
(list or array of lists, default: predefined list).}

\item{edgeFilter}{Edge filter to determine which edges maintain fixed size
(JS function, default: returns TRUE for all edges).}

\item{combo}{Combo configuration item(s) to define which attributes maintain fixed size
(list or array of lists, default: NULL).}

\item{comboFilter}{Combo filter to determine which combos maintain fixed size
(JS function, default: returns TRUE for all combos).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/fix-element-size}.}
}
\value{
A list with the configuration settings for the fix-element-size behavior.
}
\description{
Creates a configuration object for the fix-element-size behavior in G6.
This allows maintaining fixed visual sizes for elements during zoom operations.
}
\examples{
# Basic configuration
config <- fix_element_size()

# Custom configuration with filters and specific shape configurations
config <- fix_element_size(
  key = "my-fix-size-behavior",
  reset = TRUE,
  state = "active",
  node = list(
    list(shape = "circle", fields = c("r", "lineWidth")),
    list(shape = "label", fields = c("fontSize"))
  ),
  nodeFilter = JS("(node) => node.type === 'important'"),
  edge = list(shape = "line", fields = c("lineWidth", "lineDash")),
  edgeFilter = JS("(edge) => edge.weight > 5")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_focus_elements}
\alias{g6_focus_elements}
\alias{g6_focus_nodes}
\alias{g6_focus_edges}
\alias{g6_focus_combos}
\title{Focus on specific elements in a g6 graph via proxy}
\usage{
g6_focus_elements(graph, ids, animation = NULL)

g6_focus_nodes(graph, ids, animation = NULL)

g6_focus_edges(graph, ids, animation = NULL)

g6_focus_combos(graph, ids, animation = NULL)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{ids}{Character vector containing the IDs of the elements (nodes/edges/combos) to focus on.}

\item{animation}{Optional list containing animation configuration parameters for the focus action.
Common parameters include:
\itemize{
\item \code{duration}: Duration of the animation in milliseconds.
\item \code{easing}: Animation easing function name (e.g., "ease-in", "ease-out").
}
If NULL, no animation will be applied.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function focuses on one or more elements (nodes/edges) in an existing g6 graph instance
using a proxy object. It highlights the specified elements and can optionally
animate the view to focus on them.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

See \url{https://g6.antv.antgroup.com/en/api/element#graphfocuselementid-animation} for more details.
}
\seealso{
\code{\link{g6_proxy}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{combo_options}
\alias{combo_options}
\title{Create Combo Options Configuration for G6 Graphs}
\usage{
combo_options(
  type = "circle",
  style = NULL,
  state = NULL,
  palette = NULL,
  animation = NULL
)
}
\arguments{
\item{type}{Combo type. Can be a built-in combo type name or a custom combo name.
Built-in types include "circle", "rect", "polygon", etc. Default: "circle".}

\item{style}{Combo style configuration. Controls the appearance of combos including color,
size, border, etc.
Default: NULL.}

\item{state}{Defines the style of the combo in different states, such as hover, selected,
disabled, etc. Should be a list mapping state names to style configurations.
Default: NULL.}

\item{palette}{Defines the color palette of the combo, used to map colors based on different data.
Default: NULL.}

\item{animation}{Defines the animation effect of the combo. Can be created with
\code{animation_config()}.
Default: NULL.}
}
\value{
A list containing combo options configuration that can be passed to \code{\link[=g6_options]{g6_options()}}.
}
\description{
Configures the general options for combos in a G6 graph. These settings control
the type, style, state, palette, and animation of combos.
}
\details{
Combo options allow defining how combos (node groupings) appear and behave in a G6 graph.
This includes selecting combo types, setting styles, configuring state-based appearances,
defining color palettes, and specifying animation effects.
}
\examples{
# Basic combo options with default circle type
options <- combo_options()

# Rectangle combo with custom style
options <- combo_options(
  type = "rect",
  style = list(
    fill = "#F6F6F6",
    stroke = "#CCCCCC",
    lineWidth = 1
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6.R
\name{g6}
\alias{g6}
\title{Create a G6 Graph Visualization}
\usage{
g6(
  nodes = NULL,
  edges = NULL,
  combos = NULL,
  jsonUrl = NULL,
  iconsUrl = "//at.alicdn.com/t/font_2678727_za4qjydwkkh.js",
  width = "100\%",
  height = NULL,
  elementId = NULL
)
}
\arguments{
\item{nodes}{A data frame or list of nodes in the graph. Each node should have at least
an "id" field. See 'Data Structure' section for more details.
Default: NULL.}

\item{edges}{A data frame or list of edges in the graph. Each edge should have "source"
and "target" fields identifying the connected nodes. See 'Data Structure' section
for more details.
Default: NULL.}

\item{combos}{A data frame or list of combo groups in the graph. Each combo should have
at least an "id" field. Nodes can be assigned to combos using their "combo" field.
See 'Data Structure' section for more details.
Default: NULL.}

\item{jsonUrl}{An url pointing to a valid JSON containing the graph data in G6 format.
See \url{https://assets.antv.antgroup.com/g6/20000.json} for
an example. Can't be used at the same time as \code{nodes}, \code{edges}, and \code{combos}.}

\item{iconsUrl}{A URL pointing to a JavaScript file containing custom icons. Default provides
reasonable set of icons from
\url{https://at.alicdn.com/t/project/2678727/caef142c-804a-4a2f-a914-ae82666a31ee.html?spm=a313x.7781069.1998910419.35}.}

\item{width}{Width of the graph container in pixels or as a valid CSS unit.
Default: NULL (automatic sizing).}

\item{height}{Height of the graph container in pixels or as a valid CSS unit.
Default: NULL (automatic sizing).}

\item{elementId}{A unique ID for the graph HTML element.
Default: NULL (automatically generated).}
}
\value{
An htmlwidget object that can be printed, included in R Markdown documents,
or used in Shiny applications. This widget contains the graph data and configuration
necessary to render the G6 graph visualization.
}
\description{
Creates an interactive graph visualization using the G6 graph visualization library.
This function is the main entry point for creating G6 graph visualizations in R.
}
\details{
The \code{g6} function creates a G6 graph as an htmlwidget that can be used in R Markdown,
Shiny applications, or rendered to HTML. It takes graph data in the form of nodes, edges,
and optional combo groupings, along with various configuration options for customizing
the appearance and behavior of the graph.

\subsection{Nodes}{
The \code{nodes} parameter should be a data frame or list of nodes with at least an \code{id} field
for each node. Additional fields can include:
\itemize{
\item \code{id} (required): Unique identifier for the node. Must be a character.
\item \code{type}: Node type (e.g., "circle", "rect", "diamond").
\item \code{data}: Custom data associated with the node.
\item \code{style}: List of style attributes (color, size, etc.).
\item \code{states}: String. Initial states for the node, such as selected, active, hover, etc.
\item \code{combo}: ID of the combo this node belongs to.
\item \code{ports}: Node ports. Can also be passed inside the style field.
}}

\subsection{Edges}{
The \code{edges} parameter should be a data frame or list of edges with at least \code{source} and
\code{target} fields identifying the connected nodes. Additional fields can include:
\itemize{
\item \code{source} (required): ID of the source node. Must be a character.
\item \code{target} (required): ID of the target node. Must be a character.
\item \code{id}: Unique identifier for the edge.
\item \code{type}: Edge type (e.g., "line", "cubic", "arc").
\item \code{data}: Custom data associated with the edge.
\item \code{style}: List of style attributes (color, width, etc.).
\item \code{states}: String. Initial states for the edge.
}}

\subsection{Combos}{
The \code{combos} parameter is used for grouping nodes and can be a data frame or list with
combo definitions. Fields include:
\itemize{
\item \code{id} (required): Unique identifier for the combo. Must be a character.
\item \code{type}: String: Combo type. It can be the type of built-in Combo, or the custom Combo.
\item \code{data}: Custom data associated with the combo.
\item \code{style}: List of style attributes.
\item \code{states}: String. Initial states for the combo.
\item \code{combo}: String. Parent combo ID. If there is no parent combo, it is null.
}}

Nodes are assigned to combos by setting their \code{combo} field to the ID of the combo.
}
\examples{
# Create a simple graph with two nodes and one edge
nodes <- data.frame(
  id = c("node1", "node2")
)

edges <- data.frame(
  source = "node1",
  target = "node2"
)

g6(nodes = nodes, edges = edges)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{collapse_expand}
\alias{collapse_expand}
\title{Configure Collapse Expand Behavior}
\usage{
collapse_expand(
  key = "collapse-expand",
  animation = TRUE,
  enable = TRUE,
  trigger = "dblclick",
  onCollapse = NULL,
  onExpand = NULL,
  align = TRUE,
  ...
)
}
\arguments{
\item{key}{Behavior unique identifier. Useful to modify this behavior from JS side.}

\item{animation}{Enable expand/collapse animation effects (boolean, default: TRUE).}

\item{enable}{Enable expand/collapse functionality (boolean or function, default: TRUE).}

\item{trigger}{Trigger method: "click" or "dblclick" (string, default: "dblclick").}

\item{onCollapse}{Callback function when collapse is completed (function, default: NULL).}

\item{onExpand}{Callback function when expand is completed (function, default: NULL).}

\item{align}{Align with the target element to avoid view offset (boolean, default: TRUE).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/collapse-expand}.}
}
\value{
A list with the configuration settings for the collapse-expand behavior.
}
\description{
Creates a configuration object for the collapse-expand behavior in G6.
This allows users to collapse or expand nodes/combos with child elements.
}
\examples{
# Basic configuration
config <- collapse_expand()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_update_plugin}
\alias{g6_update_plugin}
\title{Update a plugin in a g6 graph via proxy}
\usage{
g6_update_plugin(graph, key, ...)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{key}{Character string identifying the plugin to update.}

\item{...}{Named arguments representing the plugin configuration options to update
and their new values.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function allows updating the configuration of an existing plugin in a g6 graph
instance using a proxy object within a Shiny application.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

The function allows dynamically updating the configuration of an existing plugin
without having to reinitialize it. This is useful for changing plugin behavior
or appearance in response to user interactions.
}
\examples{
if (interactive()) {
  library(shiny)
  library(g6R)
  library(bslib)

  color_to_hex <- function(col) {
    rgb <- col2rgb(col)
    sprintf("#\%02X\%02X\%02X", rgb[1], rgb[2], rgb[3])
  }

  nodes <- data.frame(id = c("node1", "node2", "node3"))
  edges <- data.frame(source = "node1", target = "node2")

  ui <- page_fluid(
    g6_output("graph"),
    actionButton("add_hull", "Add Hull Plugin"),
    selectInput(
      "hull_color",
      "Hull Color",
      choices = c("red", "blue", "green", "orange", "purple"),
      selected = "red"
    )
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(nodes = nodes, edges = edges) |>
        g6_layout() |>
        g6_options(animation = FALSE)
    })

    observeEvent(input$add_hull, {
      g6_add_plugin(
        g6_proxy("graph"),
        hull(
          members = c("node1", "node2", "node3"),
          fill = color_to_hex(input$hull_color),
          fillOpacity = 0.2
        )
      )
    })

    observeEvent(input$hull_color, {
      # Only update if hull plugin exists
      g6_update_plugin(
        g6_proxy("graph"),
        key = "hull",
        fill = color_to_hex(input$hull_color),
        stroke = color_to_hex(input$hull_color)
      )
    })
  }

  shinyApp(ui, server)
}
}
\seealso{
\code{\link{g6_proxy}}, \code{\link{g6_add_plugin}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{tooltips}
\alias{tooltips}
\title{Configure Tooltip Plugin}
\usage{
tooltips(
  key = "tooltip",
  position = c("top-right", "top", "bottom", "left", "right", "top-left", "bottom-left",
    "bottom-right"),
  enable = TRUE,
  getContent = NULL,
  onOpenChange = NULL,
  trigger = c("hover", "click"),
  container = NULL,
  offset = c(10, 10),
  enterable = FALSE,
  title = NULL,
  style = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{position}{Tooltip position relative to the element (string, default: "top-right").}

\item{enable}{Whether the plugin is enabled (boolean or function, default: TRUE).}

\item{getContent}{Function to generate custom tooltip content (JS function, default: NULL).}

\item{onOpenChange}{Callback for tooltip show/hide events (JS function, default: NULL).}

\item{trigger}{Trigger behavior: "hover" or "click" (string, default: "hover").}

\item{container}{Custom rendering container for tooltip (string or HTML element, default: NULL).}

\item{offset}{Offset distance as a vector of two numbers \code{[x, y]} (numeric vector, default: c(10, 10)).}

\item{enterable}{Whether the pointer can enter the tooltip (boolean, default: FALSE).}

\item{title}{Title for the tooltip (string, default: NULL).}

\item{style}{Custom style for the tooltip (list or JS object, default: list(".tooltip" = list(visibility = "hidden"))).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/tooltip}.}
}
\value{
A list with the configuration settings for the tooltip plugin.
}
\description{
Creates a configuration object for the tooltip plugin in G6.
This plugin displays tooltips when interacting with graph elements.
}
\examples{
# Basic tooltip
config <- tooltips()

# Tooltip with custom position and content
config <- tooltips(
  position = "bottom",
  getContent = JS("(event, items) => {
    let result = `<h4>Custom Content</h4>`;
    items.forEach((item) => {
      result += `<p>Type: ${item.data.description}</p>`;
    });
    return result;
  }")
)

# Click-triggered tooltip with custom style
config <- tooltips(
  trigger = "click",
  position = "bottom-left",
  offset = c(15, 20),
  style = list(
    ".tooltip" = list(
      backgroundColor = "#fff",
      border = "1px solid #ccc",
      borderRadius = "4px",
      boxShadow = "0 2px 6px rgba(0,0,0,0.1)",
      padding = "10px",
      maxWidth = "300px"
    )
  )
)

# Conditional tooltip based on node type
config <- tooltips(
  enable = JS("(event, items) => {
    // Only show tooltip for nodes with type 'important'
    const item = items[0];
    return item.type === 'important';
  }"),
  enterable = TRUE,
  onOpenChange = JS("(open) => {
    console.log('Tooltip visibility changed:', open);
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{g6_ports}
\alias{g6_ports}
\title{Create a List of G6 Ports}
\usage{
g6_ports(...)
}
\arguments{
\item{...}{One or more g6_port objects.}
}
\value{
An S3 object of class 'g6_ports'.
}
\description{
Create a List of G6 Ports
}
\examples{
g6_ports(
  g6_port("input-1", type = "input", placement = "left"),
  g6_port("output-1", type = "output", placement = "right")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{d3_force_layout}
\alias{d3_force_layout}
\title{Generate G6 D3 Force layout configuration}
\usage{
d3_force_layout(
  link = list(distance = 100, strength = 2),
  collide = list(radius = 40),
  ...
)
}
\arguments{
\item{link}{A list specifying force parameters for links (edges), with components:
\describe{
\item{id}{Edge id generation function, format: \code{function(edge, index, edges) { return string }}. Default is \code{function(e) e.id}}
\item{distance}{Ideal edge length that edges will tend toward. Can be a number or a function
\code{function(edge, index, edges) { return number }}. Default is 30}
\item{strength}{The strength of the force. Higher values make edge lengths closer to the ideal length.
Can be a number or a function \code{function(edge, index, edges) { return number }}. Default is 1}
\item{iterations}{Number of iterations of link force. Default is 1}
}}

\item{collide}{A list specifying collision force parameters for nodes, with components:
\describe{
\item{radius}{Collision radius. Nodes closer than this distance will experience a repulsive force.
Can be a number or a function \code{function(node, index, nodes) { return number }}. Default is 10}
\item{strength}{The strength of the repulsive force. Higher values produce more obvious repulsion. Default is 1}
\item{iterations}{The number of iterations for collision detection. Default is 1}
}}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/d3-force}.}
}
\value{
A list containing the configuration for G6 AntV D3 Force layout.
}
\description{
This function creates a configuration list for G6 D3 Force layout
with all available options as parameters.
}
\examples{
# Basic D3 force layout
d3_force_config <- d3_force_layout()

# Custom link distance and collision radius
d3_force_config <- d3_force_layout(
  link = list(
    distance = 150,
    strength = 0.5,
    iterations = 3
  ),
  collide = list(
    radius = 30,
    strength = 0.8
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_hide_elements}
\alias{g6_hide_elements}
\alias{g6_hide_nodes}
\alias{g6_hide_edges}
\alias{g6_hide_combos}
\alias{g6_show_elements}
\alias{g6_show_nodes}
\alias{g6_show_edges}
\alias{g6_show_combos}
\title{Hide/show elements in a g6 graph}
\usage{
g6_hide_elements(graph, ids, animation = NULL)

g6_hide_nodes(graph, ids, animation = NULL)

g6_hide_edges(graph, ids, animation = NULL)

g6_hide_combos(graph, ids, animation = NULL)

g6_show_elements(graph, ids, animation = NULL)

g6_show_nodes(graph, ids, animation = NULL)

g6_show_edges(graph, ids, animation = NULL)

g6_show_combos(graph, ids, animation = NULL)
}
\arguments{
\item{graph}{A g6 graph object or a g6_proxy object for Shiny applications.}

\item{ids}{Character vector specifying the IDs of elements to hide/show.
Can include node IDs, edge IDs, or combo IDs.}

\item{animation}{Boolean to toggle animation.}
}
\value{
The modified g6 graph or g6_proxy object, allowing for method chaining.
}
\description{
This function hides/shows specified elements (nodes, edges, or combos) in a g6 graph.
Hidden elements are removed from view but remain in the graph data structure.
}
\details{
When elements are hidden, they are removed from the visual display but still exist
in the underlying data structure. This means they can be shown again later using
\code{\link{g6_show_elements}} without having to recreate them.

Hidden elements will not participate in layout calculations, which may cause other
elements to reposition. When elements are shown again, the graph
may recalculate layout positions, which can cause other elements to reposition.
}
\seealso{
\code{\link{g6_show_elements}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{g6_node}
\alias{g6_node}
\alias{g6_edge}
\alias{g6_combo}
\alias{validate_element}
\alias{validate_element.g6_element}
\alias{validate_element.g6_node}
\alias{validate_element.g6_edge}
\alias{validate_element.g6_combo}
\title{G6 Graph Elements}
\usage{
g6_node(
  id,
  type = NULL,
  data = NULL,
  style = NULL,
  states = NULL,
  combo = NULL,
  children = NULL,
  ports = NULL,
  collapse = NULL
)

g6_edge(
  source,
  target,
  id = paste(source, target, sep = "-"),
  type = NULL,
  data = NULL,
  style = NULL,
  states = NULL
)

g6_combo(
  id,
  type = NULL,
  data = NULL,
  style = NULL,
  states = NULL,
  combo = NULL,
  collapse = NULL
)

validate_element(x, ...)

\method{validate_element}{g6_element}(x, ...)

\method{validate_element}{g6_node}(x, ...)

\method{validate_element}{g6_edge}(x, ...)

\method{validate_element}{g6_combo}(x, ...)
}
\arguments{
\item{id}{Character. Unique identifier for the node or combo (required).
For edges, this is optional (id is constructed as source-target if not provided).}

\item{type}{Character. Element type (optional).}

\item{data}{List. Custom data for the element (optional).}

\item{style}{List. Element style (optional).}

\item{states}{Character vector. Initial states for the element (optional).}

\item{combo}{Character or NULL. Combo ID or parent combo ID (optional).}

\item{children}{Character vector. Child node IDs (optional, nodes only).}

\item{ports}{List. Ports definition (optional, nodes only).
See \link{g6_ports} and \link{g6_port} for details.}

\item{collapse}{List. Collapse button configuration (optional, nodes and combos).
See \link{g6_collapse_options} for details. For nodes, only used when node has children.
For combos, when provided and \code{type} is NULL, auto-sets type to \code{"rect-combo-with-extra-button"}.}

\item{source}{Character. Source node ID (required, edges only).}

\item{target}{Character. Target node ID (required, edges only).}

\item{x}{An object of class \code{g6_element},
\code{g6_node}, \code{g6_edge}, or \code{g6_combo}.}

\item{...}{Additional arguments (unused).
the checks on source and target.}
}
\value{
An S3 object of class \code{g6_node}, \code{g6_edge}, or \code{g6_combo} (and \code{g6_element}).
}
\description{
Constructors and validators for G6 node, edge, and combo elements.
}
\examples{
# Create a node
node <- g6_node(
  id = "n1",
  type = "rect",
  data = list(label = "Node 1"),
  style = list(color = "red"),
  states = list("selected"),
  combo = NULL,
  children = c("n2", "n3")
)

# Create an edge
edge <- g6_edge(
  source = "n1",
  target = "n2",
  type = "line",
  style = list(width = 2)
)

# Create a combo
combo <- g6_combo(
  id = "combo1",
  type = "rect",
  data = list(label = "Combo 1"),
  style = list(border = "dashed"),
  states = list("active"),
  combo = NULL
)

# Validate a node explicitly
validate_element(node)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{g6_data}
\alias{g6_data}
\alias{as_g6_data}
\alias{as_g6_data.g6_data}
\alias{as_g6_data.list}
\title{Create a g6_data object}
\usage{
g6_data(nodes = NULL, edges = NULL, combos = NULL)

as_g6_data(x, ...)

\method{as_g6_data}{g6_data}(x, ...)

\method{as_g6_data}{list}(x, ...)
}
\arguments{
\item{nodes}{Nodes data.}

\item{edges}{Edges data.}

\item{combos}{Combo data.}

\item{x}{A list with elements \code{nodes}, \code{edges}, and/or \code{combos}.}

\item{...}{Additional arguments (unused).}
}
\value{
An object of class \code{g6_data}.
}
\description{
Create compatible data structure for \code{\link[=g6_add_data]{g6_add_data()}}, \code{\link[=g6_set_data]{g6_set_data()}} or
simply \code{\link[=g6]{g6()}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{compact_box_layout}
\alias{compact_box_layout}
\title{Generate G6 AntV Compact Box layout configuration}
\usage{
compact_box_layout(
  direction = c("LR", "RL", "TB", "BT", "H", "V"),
  getSide = NULL,
  getId = NULL,
  getWidth = NULL,
  getHeight = NULL,
  getHGap = NULL,
  getVGap = NULL,
  radial = FALSE,
  ...
)
}
\arguments{
\item{direction}{Layout direction: "LR" (left to right), "RL" (right to left),
"TB" (top to bottom), "BT" (bottom to top), "H" (horizontal), or "V" (vertical)}

\item{getSide}{Function to set the nodes to be arranged on the left/right side
of the root node. If not set, the algorithm automatically assigns the nodes
to the left/right side. Note: This parameter is only effective when the layout
direction is "H". Function format: \code{function(node) { return "left" or "right" }}}

\item{getId}{Callback function for generating node IDs.
Function format: \code{function(node) { return string }}}

\item{getWidth}{Function to calculate the width of each node.
Function format: \code{function(node) { return number }}}

\item{getHeight}{Function to calculate the height of each node.
Function format: \code{function(node) { return number }}}

\item{getHGap}{Function to calculate the horizontal gap for each node.
Function format: \code{function(node) { return number }}}

\item{getVGap}{Function to calculate the vertical gap for each node.
Function format: \code{function(node) { return number }}}

\item{radial}{Whether to enable radial layout}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/compact-box-layout}.}
}
\value{
A list containing the configuration for G6 AntV Compact Box layout.
}
\description{
This function creates a configuration list for G6 AntV Compact Box layout
with all available options as parameters. The Compact Box layout is designed
for efficiently laying out trees and hierarchical structures.
}
\examples{
# Basic compact box layout
box_config <- compact_box_layout()

# Vertical compact box layout
box_config <- compact_box_layout(
  direction = "TB"
)

# Radial layout
box_config <- compact_box_layout(
  radial = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igraph.R
\name{g6_igraph}
\alias{g6_igraph}
\title{Create a G6 Graph from an igraph Object}
\usage{
g6_igraph(
  graph,
  combos = NULL,
  width = "100\%",
  height = NULL,
  elementId = NULL
)
}
\arguments{
\item{graph}{An object of class \code{\link[igraph:aaa-igraph-package]{igraph::igraph}} containing the graph to visualize.}

\item{combos}{A data frame or list of combo groups in the graph. Each combo should have
at least an "id" field. Nodes can be assigned to combos using their "combo" field.
See 'Data Structure' section for more details.
Default: NULL.}

\item{width}{Width of the graph container in pixels or as a valid CSS unit.
Default: NULL (automatic sizing).}

\item{height}{Height of the graph container in pixels or as a valid CSS unit.
Default: NULL (automatic sizing).}

\item{elementId}{A unique ID for the graph HTML element.
Default: NULL (automatically generated).}
}
\value{
An htmlwidget object that can be rendered in R Markdown, Shiny apps, or the R console.
}
\description{
Converts an \code{igraph} graph object into the format required by the \code{g6()} function
and creates an interactive G6 graph visualization. This is a convenience wrapper around
\code{\link[=g6]{g6()}} that allows you to work directly with \code{igraph} objects.
}
\details{
This function extracts the node and edge data from an \code{igraph} object, converts them
into the appropriate format for G6, and passes them to \code{\link[=g6]{g6()}}. Node styling is derived
from vertex attributes, and edge styling from edge attributes.

If the graph is directed, edges will automatically be rendered with arrows.
}
\examples{
if (require(igraph)) {
  g <- igraph::make_ring(5)
  g6_igraph(g)
}

}
\seealso{
\code{\link[=g6]{g6()}} for more information about node, edge, and combo structure.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{bubble_sets}
\alias{bubble_sets}
\title{Configure Bubble Sets Plugin for G6}
\usage{
bubble_sets(
  members,
  key = "bubble-sets",
  avoidMembers = NULL,
  label = TRUE,
  labelPlacement = c("bottom", "left", "right", "top", "center"),
  labelBackground = FALSE,
  labelPadding = 0,
  labelCloseToPath = TRUE,
  labelAutoRotate = TRUE,
  labelOffsetX = 0,
  labelOffsetY = 0,
  labelMaxWidth = NULL,
  maxRoutingIterations = 100,
  maxMarchingIterations = 20,
  pixelGroup = 4,
  edgeR0 = NULL,
  edgeR1 = NULL,
  nodeR0 = NULL,
  nodeR1 = NULL,
  morphBuffer = NULL,
  threshold = NULL,
  memberInfluenceFactor = NULL,
  edgeInfluenceFactor = NULL,
  nonMemberInfluenceFactor = NULL,
  virtualEdges = NULL,
  ...
)
}
\arguments{
\item{members}{Member elements, including nodes and edges (character/numeric vector, required).}

\item{key}{Unique identifier for updates (string, default: NULL).}

\item{avoidMembers}{Elements to avoid when drawing contours (character/numeric vector, default: NULL).}

\item{label}{Whether to display labels (boolean, default: TRUE).}

\item{labelPlacement}{Label position (string, default: "bottom").}

\item{labelBackground}{Whether to display background (boolean, default: FALSE).}

\item{labelPadding}{Label padding (numeric or numeric vector, default: 0).}

\item{labelCloseToPath}{Whether the label is close to the contour (boolean, default: TRUE).}

\item{labelAutoRotate}{Whether the label rotates with the contour (boolean, default: TRUE).}

\item{labelOffsetX}{Label x-axis offset (numeric, default: 0).}

\item{labelOffsetY}{Label y-axis offset (numeric, default: 0).}

\item{labelMaxWidth}{Maximum width of the text (numeric, default: NULL).}

\item{maxRoutingIterations}{Maximum iterations for path calculation (numeric, default: 100).}

\item{maxMarchingIterations}{Maximum iterations for contour calculation (numeric, default: 20).}

\item{pixelGroup}{Number of pixels per potential area group (numeric, default: 4).}

\item{edgeR0}{Edge radius parameter R0 (numeric, default: NULL).}

\item{edgeR1}{Edge radius parameter R1 (numeric, default: NULL).}

\item{nodeR0}{Node radius parameter R0 (numeric, default: NULL).}

\item{nodeR1}{Node radius parameter R1 (numeric, default: NULL).}

\item{morphBuffer}{Morph buffer size (numeric, default: NULL).}

\item{threshold}{Threshold (numeric, default: NULL).}

\item{memberInfluenceFactor}{Member influence factor (numeric, default: NULL).}

\item{edgeInfluenceFactor}{Edge influence factor (numeric, default: NULL).}

\item{nonMemberInfluenceFactor}{Non-member influence factor (numeric, default: NULL).}

\item{virtualEdges}{Whether to use virtual edges (boolean, default: NULL).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/bubble-sets}.}
}
\value{
A list with the configuration settings for the bubble-sets plugin.
}
\description{
Creates a configuration object for the bubble-sets plugin in G6.
This plugin creates bubble-like contours around groups of specified elements.
}
\examples{
# Basic bubble set around specific nodes
bubble <- bubble_sets(
  members = c("node1", "node2", "node3"),
  label = TRUE
)

# More customized bubble set
bubble <- bubble_sets(
  key = "team-a",
  members = c("node1", "node2", "node3", "edge1", "edge2"),
  avoidMembers = c("node4", "node5"),
  labelPlacement = "top",
  labelBackground = TRUE,
  labelPadding = c(4, 2),
  maxRoutingIterations = 150
)

# Bubble set with advanced parameters
bubble <- bubble_sets(
  members = c("node1", "node2", "node3"),
  pixelGroup = 6,
  edgeR0 = 10,
  nodeR0 = 5,
  memberInfluenceFactor = 0.8,
  edgeInfluenceFactor = 0.5,
  nonMemberInfluenceFactor = 0.3,
  virtualEdges = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{node_style_options}
\alias{node_style_options}
\title{Create Node Style Options for G6 Graphs}
\usage{
node_style_options(
  collapsed = FALSE,
  cursor = "default",
  fill = "#1783FF",
  fillOpacity = 1,
  increasedLineWidthForHitTesting = 0,
  lineCap = c("butt", "round", "square"),
  lineDash = NULL,
  lineDashOffset = NULL,
  lineJoin = c("miter", "round", "bevel"),
  lineWidth = 1,
  opacity = 1,
  shadowBlur = NULL,
  shadowColor = NULL,
  shadowOffsetX = NULL,
  shadowOffsetY = NULL,
  shadowType = c("outer", "inner"),
  size = 32,
  stroke = "#000",
  strokeOpacity = 1,
  transform = NULL,
  transformOrigin = NULL,
  visibility = c("visible", "hidden"),
  x = NULL,
  y = NULL,
  z = NULL,
  zIndex = NULL,
  ...
)
}
\arguments{
\item{collapsed}{Whether the current node/group is collapsed. Default: FALSE.}

\item{cursor}{Node mouse hover cursor style. Common values include "default",
"pointer", "move", etc. Default: "default".}

\item{fill}{Node fill color. Can be any valid CSS color value. Default: "#1783FF".}

\item{fillOpacity}{Node fill color opacity. Value between 0 and 1. Default: 1.}

\item{increasedLineWidthForHitTesting}{When lineWidth is small, this value increases
the interactive area to make "thin lines" easier to interact with. Default: 0.}

\item{lineCap}{Node stroke end style. Options: "round", "square", "butt". Default: "butt".}

\item{lineDash}{Node stroke dash style. Vector of numbers specifying dash pattern.}

\item{lineDashOffset}{Node stroke dash offset. Default: NULL.}

\item{lineJoin}{Node stroke join style. Options: "round", "bevel", "miter". Default: "miter".}

\item{lineWidth}{Node stroke width. Default: 1.}

\item{opacity}{Node overall opacity. Value between 0 and 1. Default: 1.}

\item{shadowBlur}{Node shadow blur amount. Default: NULL.}

\item{shadowColor}{Node shadow color. Default: NULL.}

\item{shadowOffsetX}{Node shadow offset in the x-axis direction. Default: NULL.}

\item{shadowOffsetY}{Node shadow offset in the y-axis direction. Default: NULL.}

\item{shadowType}{Node shadow type. Options: "inner", "outer". Default: "outer".}

\item{size}{Node size. Can be a single number for equal width/height or a vector
of two numbers \code{[width, height]}. Default: 32.}

\item{stroke}{Node stroke (border) color. Default: "#000".}

\item{strokeOpacity}{Node stroke color opacity. Value between 0 and 1. Default: 1.}

\item{transform}{CSS transform attribute to rotate, scale, skew, or translate the node.
Default: NULL.}

\item{transformOrigin}{Rotation and scaling center point. Default: NULL.}

\item{visibility}{Whether the node is visible. Options: "visible", "hidden". Default: "visible".}

\item{x}{Node x coordinate. Default: 0.}

\item{y}{Node y coordinate. Default: 0.}

\item{z}{Node z coordinate (for 3D). Default: 0.}

\item{zIndex}{Node rendering level (for layering). Default: 0.}

\item{...}{Other parameters.}
}
\value{
A list containing node style options that can be passed to \code{\link[=node_options]{node_options()}}.
}
\description{
Configures the styling options for nodes in a G6 graph. These settings control
the appearance and interaction behavior of nodes. Used in \link{node_options}.
}
\details{
Node style options allow fine-grained control over how nodes are rendered and
behave in a G6 graph. This includes colors, sizes, borders, shadows, visibility,
positioning, and interaction states.
}
\examples{
# Basic node style with blue fill and red border
styles <- node_style_options(
  fill = "#1783FF",
  stroke = "#FF0000",
  lineWidth = 2
)

# Create a node with shadow effects
styles <- node_style_options(
  fill = "#FFFFFF",
  stroke = "#333333",
  lineWidth = 1,
  shadowBlur = 10,
  shadowColor = "rgba(0,0,0,0.3)",
  shadowOffsetX = 5,
  shadowOffsetY = 5
)

# Custom sized node with dashed border
styles <- node_style_options(
  size = c(100, 50),
  fill = "#E8F7FF",
  stroke = "#1783FF",
  lineDash = c(5, 5),
  opacity = 0.8
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{click_select}
\alias{click_select}
\title{Configure Click Select Behavior}
\usage{
click_select(
  key = "click-select",
  animation = TRUE,
  degree = 0,
  enable = TRUE,
  multiple = FALSE,
  state = c("selected", "active", "inactive", "disabled", "highlight"),
  neighborState = c("selected", "active", "inactive", "disabled", "highlight"),
  unselectedState = NULL,
  onClick = NULL,
  trigger = "shift",
  ...
)
}
\arguments{
\item{key}{Behavior unique identifier. Useful to modify this behavior from JS side.}

\item{animation}{Whether to enable animation effects when switching element states (boolean, default: TRUE).}

\item{degree}{Controls the highlight spread range (number or function, default: 0).}

\item{enable}{Whether to enable the click element function (boolean or function, default: TRUE).}

\item{multiple}{Whether to allow multiple selections (boolean, default: FALSE).}

\item{state}{The state applied when an element is selected (string, default: "selected").}

\item{neighborState}{The state applied to elements with n-degree relationships (string, default: "selected").}

\item{unselectedState}{The state applied to all other elements (string, default: NULL).}

\item{onClick}{Callback when an element is clicked (function, default: NULL).}

\item{trigger}{Keys for multi-selection (character vector, default: c("shift")).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/click-select}.}
}
\value{
A list with the configuration settings for the click select behavior.
}
\description{
Creates a configuration object for the click-select behavior in G6.
This allows users to select graph elements by clicking.
}
\examples{
# Basic configuration
config <- click_select()

# Custom configuration
config <- click_select(
  animation = FALSE,
  degree = 1,
  multiple = TRUE,
  state = "active",
  neighborState = "highlight",
  unselectedState = "inactive",
  trigger = c("Control")
)

# Example leveraging the input[["<GRAPH_ID>-selected_<ELEMENT_TYPE>"]]
if (interactive()) {
  library(shiny)
  library(g6R)
  library(bslib)

  nodes <- data.frame(id = c("node1", "node2"))
  edges <- data.frame(source = "node1", target = "node2")
  combos <- data.frame(id = "combo1", type = "rect")

  ui <- page_fluid(
    g6_output("graph"),
    verbatimTextOutput("selected_elements")
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(
        nodes = nodes,
        edges = edges,
        combos = combos
      ) |>
        g6_layout() |>
        g6_behaviors(
          click_select(multiple = TRUE),
          brush_select(
            enableElements = c("node", "edge", "combo"),
            immediately = TRUE
          )
        )
    })

    output$selected_elements <- renderPrint({
      list(
        selected_nodes = input[["graph-selected_node"]],
        selected_edges = input[["graph-selected_edge"]],
        selected_combos = input[["graph-selected_combo"]]
      )
    })
  }

  shinyApp(ui, server)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/g6.R
\name{set_g6_max_collapse_depth}
\alias{set_g6_max_collapse_depth}
\title{Set max collapse depth}
\usage{
set_g6_max_collapse_depth(val)
}
\arguments{
\item{val}{A single number >= -1. Use \code{Inf} for no limit,
\code{-1} to disable all collapsing.}
}
\value{
Invisibly returns \code{val}.
}
\description{
Controls which nodes display a collapse button based on their depth in
the graph. Only nodes at depth \code{<= maxCollapseDepth} will show
collapse buttons. Set to \code{Inf} (the default) to allow all nodes
with children to be collapsible. Set to \code{0} to only allow root
nodes to collapse. Set to \code{-1} to disable collapsing entirely.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{lasso_select}
\alias{lasso_select}
\title{Configure Lasso Select Behavior}
\usage{
lasso_select(
  key = "lasso-select",
  animation = FALSE,
  enable = TRUE,
  enableElements = "node",
  immediately = FALSE,
  mode = c("default", "union", "intersect", "diff"),
  onSelect = NULL,
  state = "selected",
  style = NULL,
  trigger = c("shift"),
  outputId = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior (string, default: "lasso-select").}

\item{animation}{Whether to enable animation (boolean, default: FALSE).}

\item{enable}{Whether to enable lasso selection (boolean or JS function, default: TRUE).}

\item{enableElements}{Types of elements that can be selected (character vector, default: c("node", "combo", "edge")).}

\item{immediately}{Whether to select immediately, only effective when selection mode is default (boolean, default: FALSE).}

\item{mode}{Selection mode: "union", "intersect", "diff", or "default" (string, default: "default").}

\item{onSelect}{Callback for selected element state (JS function, default: NULL).}

\item{state}{State to switch to when selected (string, default: "selected").}

\item{style}{Style of the lasso during selection (list, default: NULL).}

\item{trigger}{Press this shortcut key along with mouse click to select (character vector, default: c("shift")).}

\item{outputId}{Manually pass the Shiny output ID. This is useful when the graph
is initialised outside the shiny render function and the ID cannot be automatically
inferred. This allows to set input values from the callback function
with the right namespace and graph ID. You must typically pass \code{session$ns("graphid")}
to ensure this also works in modules.}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/lasso-select}.}
}
\value{
A list with the configuration settings for the lasso-select behavior.
}
\description{
Creates a configuration object for the lasso-select behavior in G6.
This behavior allows selecting elements by drawing a lasso around them.
}
\examples{
# Basic configuration
config <- lasso_select()

# Custom configuration
config <- lasso_select(
  key = "my-lasso-select",
  animation = TRUE,
  enableElements = c("node", "combo"),
  mode = "union",
  state = "highlight",
  trigger = c("control"),
  style = list(
    stroke = "#1890FF",
    lineWidth = 2,
    fillOpacity = 0.1
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{optimize_viewport_transform}
\alias{optimize_viewport_transform}
\title{Configure Optimize Viewport Transform Behavior}
\usage{
optimize_viewport_transform(
  key = "optimize-viewport-transform",
  enable = TRUE,
  debounce = 200,
  shapes = JS("(type) => type === 'node'"),
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior (string, default: "optimize-viewport-transform").}

\item{enable}{Whether to enable this behavior (boolean or JS function, default: TRUE).}

\item{debounce}{How long after the operation ends to restore the visibility of all elements
in milliseconds (number, default: 200).}

\item{shapes}{Function to specify which graphical elements should remain visible during canvas
operations (JS function, default: returns TRUE for nodes).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/optimize-viewport-transform}.}
}
\value{
A list with the configuration settings for the optimize-viewport-transform behavior.
}
\description{
Creates a configuration object for the optimize-viewport-transform behavior in G6.
This behavior improves performance during viewport transformations by temporarily
hiding certain elements.
}
\examples{
# Basic configuration
config <- optimize_viewport_transform()

# Custom configuration
config <- optimize_viewport_transform(
  key = "my-optimize-transform",
  debounce = 500,
  shapes = JS("(type) => type === 'node' || type === 'edge'")
)

# With conditional enabling
config <- optimize_viewport_transform(
  enable = JS("(event) => event.getCurrentTransform().zoom < 0.5")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{drag_element}
\alias{drag_element}
\title{Configure Drag Element Behavior}
\usage{
drag_element(
  key = "drag-element",
  enable = NULL,
  animation = TRUE,
  state = "selected",
  dropEffect = c("move", "link", "none"),
  hideEdge = c("none", "out", "in", "both", "all"),
  shadow = FALSE,
  cursor = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior, used for subsequent operations (string, default: NULL)}

\item{enable}{Whether to enable the drag function (boolean or function, default: function that enables dragging for nodes and combos).}

\item{animation}{Whether to enable drag animation (boolean, default: TRUE).}

\item{state}{Identifier for the selected state of nodes (string, default: "selected").}

\item{dropEffect}{Defines the operation effect after dragging ends: "link", "move", or "none" (string, default: "move").}

\item{hideEdge}{Controls the display state of edges during dragging: "none", "out", "in", "both", or "all" (string, default: "none").}

\item{shadow}{Whether to enable ghost nodes (boolean, default: FALSE).}

\item{cursor}{Customize the mouse style during dragging (list, default: NULL).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/drag-element}.}
}
\value{
A list with the configuration settings for the drag-element behavior.
}
\description{
Creates a configuration object for the drag-element behavior in G6.
This allows users to drag nodes and combos in the graph.
}
\examples{
# Basic configuration
config <- drag_element()

# Custom configuration
config <- drag_element(
  key = "my-drag-behavior",
  animation = FALSE,
  dropEffect = "link",
  hideEdge = "both",
  shadow = TRUE,
  cursor = list(
    default = "default",
    grab = "grab",
    grabbing = "grabbing"
  ),
  enable = JS(
   "(e) => {
     return e.targetType === 'node' || e.targetType === 'combo';
   }"
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{scroll_canvas}
\alias{scroll_canvas}
\title{Configure Scroll Canvas Behavior}
\usage{
scroll_canvas(
  key = "scroll-canvas",
  enable = TRUE,
  direction = NULL,
  range = 1,
  sensitivity = 1,
  trigger = NULL,
  onFinish = NULL,
  preventDefault = TRUE,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior (string, default: "scroll-canvas").}

\item{enable}{Whether to enable this behavior (boolean or JS function, default: TRUE).}

\item{direction}{Allowed scrolling direction: "x", "y", or NULL for no limit
(string or NULL, default: NULL).}

\item{range}{Scrollable viewport range in viewport size units (numeric or numeric vector, default: 1).}

\item{sensitivity}{Scrolling sensitivity, the larger the value, the faster the scrolling
(numeric, default: 1).}

\item{trigger}{Keyboard shortcuts to trigger scrolling (list, default: NULL).}

\item{onFinish}{Callback function when scrolling is finished (JS function, default: NULL).}

\item{preventDefault}{Whether to prevent the browser's default event (boolean, default: TRUE).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/scroll-canvas}.}
}
\value{
A list with the configuration settings for the scroll-canvas behavior.
}
\description{
Creates a configuration object for the scroll-canvas behavior in G6.
This behavior allows scrolling the canvas with mouse wheel or keyboard.
}
\examples{
# Basic configuration
config <- scroll_canvas()

# Custom configuration
config <- scroll_canvas(
  key = "my-scroll-behavior",
  direction = "x",
  range = c(-2, 2),
  sensitivity = 1.5,
  preventDefault = FALSE
)

# With keyboard triggers and callback
config <- scroll_canvas(
  enable = JS("(event) => !event.altKey"),
  trigger = list(
    up = "w",
    down = "s",
    left = "a",
    right = "d"
  ),
  onFinish = JS("() => { console.log('Scrolling finished'); }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_set_theme}
\alias{g6_set_theme}
\title{Set the theme for a g6 graph via proxy}
\usage{
g6_set_theme(graph, theme)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{theme}{A character string representing the theme to apply to the graph.
There are 2 internal predefined themes: \code{light} and \code{dark}.
Alternatively, you can pass a custom theme object that conforms to the G6 theme specifications,
according to the documentation at \url{https://g6.antv.antgroup.com/en/manual/theme/custom-theme}.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function sets the theme for an existing g6 graph instance
}
\examples{
if (interactive()) {
  library(shiny)
  library(bslib)
  library(g6R)

  themes <- list(
    "light" = list(
      theme = 'light',
      node = list(
        style = list(
          size = 4
        ),
        palette = list(
          type = 'group',
          field = 'cluster'
        )
      ),
      plugins = list(
        list(
          type = 'background',
          background = '#fff'
        )
      )
    ),
    "dark" = list(
      theme = 'dark',
      node = list(
        style = list(
          size = 4
        ),
        palette = list(
          type = 'group',
          field = 'cluster'
        )
      ),
      plugins = list(
        list(
          type = 'background',
          background = '#000'
        )
      )
    ),
    "blue" = list(
      theme = 'light',
      node = list(
        style = list(
          size = 4
        ),
        palette = list(
          type = 'group',
          field = 'cluster',
          color = "blues",
          invert = TRUE
        )
      ),
      plugins = list(
        list(
          type = 'background',
          background = '#f3faff'
        )
      )
    ),
    "yellow" = list(
      theme = 'light',
      node = list(
        style = list(
          size = 4
        ),
        palette = list(
          type = 'group',
          field = 'cluster',
          color = c(
            '#ffe7ba',
            '#ffd591',
            '#ffc069',
            '#ffa940',
            '#fa8c16',
            '#d46b08',
            '#ad4e00',
            '#873800',
            '#612500'
          )
        )
      ),
      plugins = list(
        list(
          type = 'background',
          background = '#fcf9f1'
        )
      )
    )
  )

  ui <- page_fluid(
    selectInput(
      "theme",
      "Theme",
      choices = c("light", "dark", "blue", "yellow"),
      selected = "default"
    ),
    g6Output("graph", height = "100vh")
  )

  server <- function(input, output, session) {
    output$graph <- renderG6({
      g6(
        jsonUrl = "https://assets.antv.antgroup.com/g6/20000.json"
      ) |>
        g6_options(
          animation = FALSE,
          autoFit = "view",
          padding = 20,
          node = list(
            style = list(size = 4),
            palette = list(
              type = "group",
              field = "cluster"
            )
          )
        ) |>
        g6_behaviors(
          "zoom-canvas",
          "drag-canvas",
          "optimize-viewport-transform"
        )
    })

    observeEvent(input$theme, {
      g6_proxy("graph") |>
        g6_set_theme(themes[[input$theme]])
    })

    observeEvent(input[["graph-initialized"]], {
      showNotification(
        "Graph initialized",
        type = "message"
      )
    })

    observeEvent(input[["graph-state"]], {
      showNotification(
        "Graph state updated",
        type = "message"
      )
    })
  }

  shinyApp(ui, server)

}
}
\seealso{
\code{\link[=g6_proxy]{g6_proxy()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{toolbar}
\alias{toolbar}
\title{Configure Toolbar Plugin}
\usage{
toolbar(
  getItems = NULL,
  key = "toolbar",
  className = NULL,
  position = c("top-left", "top", "top-right", "right", "bottom-right", "bottom",
    "bottom-left", "left"),
  style = NULL,
  onClick = NULL,
  ...
)
}
\arguments{
\item{getItems}{Function that returns the list of toolbar items (JS function, required).}

\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{className}{Additional CSS class name for the toolbar DOM element (string, default: NULL).}

\item{position}{Toolbar position relative to the canvas (string, default: "top-left").}

\item{style}{Custom style for the toolbar DOM element (list or JS object, default: NULL).}

\item{onClick}{Callback function after a toolbar item is clicked (JS function, default: NULL).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/toolbar}.}
}
\value{
A list with the configuration settings for the toolbar plugin.
}
\description{
Creates a configuration object for the toolbar plugin in G6.
This plugin adds a customizable toolbar with items for graph operations.
}
\examples{
# Basic toolbar with zoom controls
config <- g6R::toolbar(
  position = "top-right",
  getItems = JS("() => [
    { id: 'zoom-in', value: 'zoom-in' },
    { id: 'zoom-out', value: 'zoom-out' },
    { id: 'undo', value: 'undo' },
    { id: 'redo', value: 'redo' },
    { id: 'auto-fit', value: 'fit' }
  ]"),
  onClick = JS("(value) => {
    // redo, undo need to be used with the history plugin
    const history = graph.getPluginInstance('history');
    switch (value) {
      case 'zoom-in':
        graph.zoomTo(1.1);
        break;
      case 'zoom-out':
        graph.zoomTo(0.9);
        break;
      case 'undo':
        history?.undo();
        break;
      case 'redo':
        history?.redo();
        break;
      case 'fit':
        graph.fitView();
        break;
      default:
        break;
    }
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_update_layout}
\alias{g6_update_layout}
\title{Execute layout for a g6 graph via proxy}
\usage{
g6_update_layout(graph, ...)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{...}{Any option to pass to the layout. If so, the layout is updated
before running it.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function order the execution of the layout of the current
graph. It can also update layout options before running it.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.
}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{edge_filter_lens}
\alias{edge_filter_lens}
\title{Configure Edge Filter Lens Plugin}
\usage{
edge_filter_lens(
  key = "edge-filter-lens",
  trigger = c("pointermove", "click", "drag"),
  r = 60,
  maxR = NULL,
  minR = 0,
  scaleRBy = "wheel",
  nodeType = c("both", "source", "target", "either"),
  filter = NULL,
  style = NULL,
  nodeStyle = list(label = FALSE),
  edgeStyle = list(label = TRUE),
  preventDefault = TRUE,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{trigger}{Method to move the lens: "pointermove", "click", or "drag" (string, default: "pointermove").}

\item{r}{Radius of the lens (number, default: 60).}

\item{maxR}{Maximum radius of the lens (number, default: NULL - half of the smaller canvas dimension).}

\item{minR}{Minimum radius of the lens (number, default: 0).}

\item{scaleRBy}{Method to scale the lens radius (string, default: "wheel").}

\item{nodeType}{Edge display condition: "both", "source", "target", or "either" (string, default: "both").}

\item{filter}{Filter out elements that are never displayed in the lens (JS function, default: NULL).}

\item{style}{Style of the lens (list, default: NULL).}

\item{nodeStyle}{Style of nodes in the lens (list or JS function, default: list(label = FALSE)).}

\item{edgeStyle}{Style of edges in the lens (list or JS function, default: list(label = TRUE)).}

\item{preventDefault}{Whether to prevent default events (boolean, default: TRUE).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/edge-filter-lens}.}
}
\value{
A list with the configuration settings for the edge-filter-lens plugin.
}
\description{
Creates a configuration object for the edge-filter-lens plugin in G6.
This plugin creates a lens that filters and displays edges within a specific area.
}
\examples{
# Basic configuration
config <- edge_filter_lens()

# Custom configuration
config <- edge_filter_lens(
  key = "my-edge-lens",
  trigger = "drag",
  r = 100,
  nodeType = "either",
  style = list(
    fill = "rgba(200, 200, 200, 0.3)",
    stroke = "#999",
    lineWidth = 2
  ),
  filter = JS("(id, type) => {
    // Only display edges connected to specific nodes
    if (type === 'edge') {
      const edge = graph.getEdgeData(id);
      return edge.source === 'node1' || edge.target === 'node1';
    }
    return true;
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{force_atlas2_layout}
\alias{force_atlas2_layout}
\title{Generate G6 Force Atlas2 layout configuration}
\usage{
force_atlas2_layout(
  barnesHut = NULL,
  dissuadeHubs = FALSE,
  height = NULL,
  kg = 1,
  kr = 5,
  ks = 0.1,
  ksmax = 10,
  mode = "normal",
  nodeSize = NULL,
  preventOverlap = FALSE,
  prune = NULL,
  tao = 0.1,
  width = NULL,
  center = NULL,
  ...
)
}
\arguments{
\item{barnesHut}{Logical. Whether to enable quadtree acceleration.
When enabled, improves performance for large graphs but may affect layout quality.
By default, enabled if node count > 250.}

\item{dissuadeHubs}{Logical. Whether to enable hub mode.
If TRUE, nodes with higher in-degree are more likely to be placed at the center
than those with high out-degree. Defaults to FALSE.}

\item{height}{Numeric. Layout height. Defaults to container height.}

\item{kg}{Numeric. Gravity coefficient. The larger the value,
the more concentrated the layout is at the center. Defaults to 1.}

\item{kr}{Numeric. Repulsion coefficient. Adjusts the compactness of the layout.
The larger the value, the looser the layout. Defaults to 5.}

\item{ks}{Numeric. Controls the speed of node movement during iteration. Defaults to 0.1.}

\item{ksmax}{Numeric. Maximum node movement speed during iteration. Defaults to 10.}

\item{mode}{Character. Clustering mode.
Options are "normal" or "linlog". In "linlog" mode, clusters are more compact.
Defaults to "normal".}

\item{nodeSize}{Numeric or function. Node size (diameter).
Used for repulsion calculation when preventOverlap is enabled.

If not set, uses data.size in node data.}

\item{preventOverlap}{Logical. Whether to prevent node overlap.
When enabled, layout considers node size to avoid overlap.
Node size is specified by \code{nodeSize} or \code{data.size} in node data. Defaults to FALSE.}

\item{prune}{Logical. Whether to enable auto-pruning.
By default, enabled if node count > 100. Pruning speeds up
convergence but may reduce layout quality. Set to FALSE to disable auto-activation.}

\item{tao}{Numeric. Tolerance for stopping oscillation when layout is near convergence. Defaults to 0.1.}

\item{width}{Numeric. Layout width. Defaults to container width.}

\item{center}{Numeric vector of length 2. Layout center in the form \code{[x, y]}.
Each node is attracted to this point, with gravity controlled by \code{kg}.
If not set, uses canvas center.}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/force-atlas2-layout}.}
}
\value{
A list containing the configuration for G6 force atlas2 layout.
}
\description{
This function creates a configuration list for G6 Force Atlas2 layout
with all available options as parameters.
}
\examples{
if (interactive()) {
  g6(lesmis$nodes, lesmis$edges) |>
    g6_layout(force_atlas2_layout(
      kr = 20,
      preventOverlap = TRUE,
      center = c(250, 250))) |>
    g6_options(autoResize = TRUE) |>
    g6_behaviors(
      "zoom-canvas",
      drag_element()
    )
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_update_ports}
\alias{g6_update_ports}
\title{Update ports for one or more nodes in a g6 graph via proxy}
\usage{
g6_update_ports(graph, ops)
}
\arguments{
\item{graph}{A g6_proxy object.}

\item{ops}{A named list of operations, one entry per node (named by node ID).
Each entry can contain:
\itemize{
\item \code{add}: a list of port objects to add.
\item \code{remove}: a character vector of port keys to remove
\item \code{update}: a list of port objects (with key) to update
}}
}
\value{
The g6_proxy object.
}
\description{
This function sends a message to the client
to update (add, remove, or modify) ports
for one or more nodes. The actual update is handled
on the JS side.
}
\details{
Removing a port that is currently used by an edge
removes the edge as well. Conversely, removing an edge
does not remove the ports it was using.
}
\examples{
if (interactive()) {
  library(shiny)
  library(g6R)

  ui <- fluidPage(
    g6_output("graph", height = "500px"),
    actionButton("update_ports", "Update Ports")
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(
        nodes = g6_nodes(
          g6_node(
            id = "A",
            type = "custom-rect-node",
            ports = g6_ports(
              g6_input_port(key = "in1", label = "in1", placement = "left"),
              g6_output_port(key = "out1", label = "out1", placement = "right"),
              g6_output_port(key = "out2", label = "out2", placement = c(1, 1))
            ),
            style = list(x = 100, y = 200, labelText = "Node A")
          ),
          g6_node(
            id = "B",
            type = "custom-rect-node",
            ports = g6_ports(
              g6_input_port(key = "in2", label = "in2", placement = "left"),
              g6_output_port(key = "out3", label = "out3", placement = "right"),
              g6_output_port(key = "out4", label = "out4", placement = c(1, 0))
            ),
            style = list(x = 300, y = 200, labelText = "Node B")
          )
        ),
        edges = g6_edges(
          g6_edge(source = "A", target = "B", style = list(sourcePort = "out1", targetPort = "in2"))
        )
      ) |> g6_behaviors(click_select(), drag_element(), drag_canvas())
    })

    observeEvent(input$update_ports, {
      g6_update_ports(
        g6_proxy("graph"),
        list(
          A = list(remove = c("out1", "out2")),
          B = list(
            add = list(g6_port(key = "new", label = "new", placement = "top")),
            update = list(g6_port(key = "in2", label = "Updated label"))
          )
        )
      )
    })
  }

  shinyApp(ui, server)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{edge_style_options}
\alias{edge_style_options}
\title{Create Edge Style Options for G6 Graphs}
\usage{
edge_style_options(
  class = NULL,
  cursor = valid_cursors,
  fill = NULL,
  fillRule = c("nonzero", "evenodd"),
  filter = NULL,
  increasedLineWidthForHitTesting = NULL,
  isBillboard = TRUE,
  lineDash = 0,
  lineDashOffset = 0,
  lineWidth = 1,
  opacity = 1,
  pointerEvents = NULL,
  shadowBlur = NULL,
  shadowColor = NULL,
  shadowOffsetX = NULL,
  shadowOffsetY = NULL,
  shadowType = NULL,
  sourcePort = NULL,
  stroke = "#000",
  strokeOpacity = 1,
  targetPort = NULL,
  transform = NULL,
  transformOrigin = NULL,
  visibility = c("visible", "hidden"),
  zIndex = -10000,
  ...
)
}
\arguments{
\item{class}{Edge class name for custom styling with CSS. Default: NULL.}

\item{cursor}{Edge mouse hover cursor style. Common values include "default",
"pointer", "move", etc. Default: "default".}

\item{fill}{Edge area fill color (for edges with area, like loops). Default: NULL.}

\item{fillRule}{Edge internal fill rule. Options: "nonzero", "evenodd". Default: NULL.}

\item{filter}{Edge shadow filter effect. Default: NULL.}

\item{increasedLineWidthForHitTesting}{When the edge width is too small, this value increases
the interaction area to make edges easier to interact with. Default: NULL.}

\item{isBillboard}{Effective in 3D scenes, always facing the screen so the line width
is not affected by perspective projection. Default: TRUE.}

\item{lineDash}{Edge dashed line style. Numeric vector specifying dash pattern. Default: 0.}

\item{lineDashOffset}{Edge dashed line offset. Default: 0.}

\item{lineWidth}{Edge width in pixels. Default: 1.}

\item{opacity}{Overall opacity of the edge. Value between 0 and 1. Default: 1.}

\item{pointerEvents}{Whether the edge responds to pointer events. Default: NULL.}

\item{shadowBlur}{Edge shadow blur effect amount. Default: NULL.}

\item{shadowColor}{Edge shadow color. Default: NULL.}

\item{shadowOffsetX}{Edge shadow X-axis offset. Default: NULL.}

\item{shadowOffsetY}{Edge shadow Y-axis offset. Default: NULL.}

\item{shadowType}{Edge shadow type. Options: "inner", "outer", "both". Default: NULL.}

\item{sourcePort}{Source port of the edge connection. Default: NULL.}

\item{stroke}{Edge color. Default: "#000".}

\item{strokeOpacity}{Edge color opacity. Value between 0 and 1. Default: 1.}

\item{targetPort}{Target port of the edge connection. Default: NULL.}

\item{transform}{CSS transform attribute to rotate, scale, skew, or translate the edge.
Default: NULL.}

\item{transformOrigin}{Rotation and scaling center point. Default: NULL.}

\item{visibility}{Whether the edge is visible. Options: "visible", "hidden". Default: "visible".}

\item{zIndex}{Edge rendering level (for layering). Default: 1.}

\item{...}{Extra parameters.}
}
\value{
A list containing edge style options that can be passed to \code{\link[=edge_options]{edge_options()}}.
}
\description{
Configures the styling options for edges in a G6 graph. These settings control
the appearance and interaction behavior of edges.
}
\details{
Edge style options allow fine-grained control over how edges are rendered and
behave in a G6 graph. This includes colors, widths, line styles, shadows, visibility,
and interaction properties.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_fit_center}
\alias{g6_fit_center}
\title{Center graph}
\usage{
g6_fit_center(graph, animation = NULL)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{animation}{Optional list containing animation configuration parameters for the focus action.
Common parameters include:
\itemize{
\item \code{duration}: Duration of the animation in milliseconds.
\item \code{easing}: Animation easing function name (e.g., "ease-in", "ease-out").
}
If NULL, no animation will be applied.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function pans the graph to the center of the viewport
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

See \url{https://g6.antv.antgroup.com/en/api/viewport#graphfitcenteranimation} for more details.
}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{background}
\alias{background}
\title{Configure Background Plugin for G6}
\usage{
background(
  key = NULL,
  width = "100\%",
  height = "100\%",
  backgroundColor = NULL,
  backgroundImage = NULL,
  backgroundSize = "cover",
  backgroundPosition = NULL,
  backgroundRepeat = NULL,
  opacity = NULL,
  transition = "background 0.5s",
  zIndex = "-1",
  ...
)
}
\arguments{
\item{key}{Unique identifier for updates (string, default: NULL).}

\item{width}{Background width (string, default: "100\%").}

\item{height}{Background height (string, default: "100\%").}

\item{backgroundColor}{Background color (string, default: NULL).}

\item{backgroundImage}{Background image URL (string, default: NULL).}

\item{backgroundSize}{Background size (string, default: "cover").}

\item{backgroundPosition}{Background position (string, default: NULL).}

\item{backgroundRepeat}{Background repeat (string, default: NULL).}

\item{opacity}{Background opacity (string, default: NULL).}

\item{transition}{Transition animation (string, default: "background 0.5s").}

\item{zIndex}{Stacking order (string, default: "-1").}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/background}.}
}
\value{
A list with the configuration settings for the background plugin.
}
\description{
Creates a configuration object for the background plugin in G6.
This plugin adds a customizable background to the graph canvas.
}
\examples{
# Basic background color
bg <- background(backgroundColor = "#f0f0f0")

# Background with image
bg <- background(
  backgroundImage = "https://example.com/background.jpg",
  backgroundSize = "contain",
  backgroundRepeat = "no-repeat",
  backgroundPosition = "center"
)

# Semi-transparent background with transition
bg <- background(
  backgroundColor = "#000000",
  opacity = "0.3",
  transition = "all 1s ease-in-out"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_get_nodes}
\alias{g6_get_nodes}
\alias{g6_get_edges}
\alias{g6_get_combos}
\title{Get the state of nodes/edges/combos in a g6 graph via proxy}
\usage{
g6_get_nodes(graph, nodes)

g6_get_edges(graph, edges)

g6_get_combos(graph, combos)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{nodes, edges, combos}{A string or character vector with the IDs of the nodes/edges/combos.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function gets the state of one or more nodes/edges/combos to an existing g6 graph instance
using a proxy object.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

If a node with the same ID already exists, it will not be added again.
See \url{https://g6.antv.antgroup.com/en/api/data#graphgetnodedata} for more details.
}
\examples{
 if (interactive()) {
   library(shiny)
   library(bslib)

    ui <- page_fluid(
      verbatimTextOutput("res"),
      g6Output("graph")
    )

    server <- function(input, output, session) {
      output$graph <- renderG6({
        g6(
          nodes = data.frame(id = c("node1", "node2"))
        ) |>
          g6_options(animation = FALSE) |>
          g6_layout() |>
          g6_behaviors(click_select())
      })

      # Send query to JS
      observeEvent(req(input[["graph-initialized"]]), {
        g6_proxy("graph") |> g6_get_nodes(c("node1", "node2"))
      })

      # Recover query result inside input[["<GRAPH_ID>-<ELEMENT_ID>-state"]]
      output$res <- renderPrint({
        list(
          node1_state = input[["graph-node1-state"]],
          node2_state = input[["graph-node2-state"]]
        )
      })
    }
    shinyApp(ui, server)
 }
}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_remove_nodes}
\alias{g6_remove_nodes}
\alias{g6_remove_edges}
\alias{g6_remove_combos}
\title{Remove nodes/edge/combos from a g6 graph via proxy}
\usage{
g6_remove_nodes(graph, ids)

g6_remove_edges(graph, ids)

g6_remove_combos(graph, ids)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{ids}{Character vector or list containing the IDs of the nodes/edges/combos to be removed.
If a single ID is provided, it will be converted to a list internally. You can't mix
nodes, edges and combos ids, elements have to be of the same type.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function removes one or more nodes/edges/combos from an existing g6 graph instance
using a proxy object. This allows updating the graph without completely
re-rendering it.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

See \url{https://g6.antv.antgroup.com/en/api/data#graphremovenodedata},
\url{https://g6.antv.antgroup.com/en/api/data#graphremoveedgedata} and
\url{https://g6.antv.antgroup.com/en/api/data#graphremovecombodata} for more details.
}
\note{
When a node is removed, its connected edges are also removed.
}
\examples{
if (interactive()) {
  library(shiny)
  library(g6R)
  library(bslib)

  # Static data defined globally
  nodes <- data.frame(id = 1:3)
  edges <- data.frame(source = c(1, 2), target = c(2, 3))

  ui <- page_fluid(
    title = "Remove Nodes Dynamically",
    g6_output("graph"),
    actionButton("remove_node", "Remove Last Node")
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(nodes = nodes, edges = edges) |> g6_layout()
    })

    # Track the next node id and current node ids
    current_ids <- reactiveVal(nodes$id)

    observeEvent(input$remove_node, {
      ids <- current_ids()
      if (length(ids) > 0) {
        remove_id <- tail(ids, 1)
        g6_remove_nodes(g6_proxy("graph"), remove_id)
        current_ids(ids[-length(ids)])
      }
    })
  }

  shinyApp(ui, server)

}
}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_proxy}
\alias{g6_proxy}
\title{Create a proxy object to modify an existing g6 graph instance}
\usage{
g6_proxy(id, session = shiny::getDefaultReactiveDomain())
}
\arguments{
\item{id}{Character string matching the ID of the g6 graph instance to be modified.}

\item{session}{The Shiny session object within which the graph exists.
By default, this uses the current reactive domain.}
}
\value{
A proxy object of class "g6_proxy" that can be used with g6 proxy methods
such as \code{g6_add_nodes()}, \code{g6_remove_nodes()}, etc.
}
\description{
This function creates a proxy object that can be used to update an existing g6 graph
instance after it has been rendered in the UI. The proxy allows for server-side
modifications of the graph without completely re-rendering it.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{antv_dagre_layout}
\alias{antv_dagre_layout}
\title{Generate G6 AntV Dagre layout configuration}
\usage{
antv_dagre_layout(
  rankdir = c("TB", "BT", "LR", "RL"),
  align = c("UL", "UR", "DL", "DR"),
  nodesep = 50,
  nodesepFunc = NULL,
  ranksep = 100,
  ranksepFunc = NULL,
  ranker = c("network-simplex", "tight-tree", "longest-path"),
  nodeSize = NULL,
  controlPoints = FALSE,
  begin = NULL,
  sortByCombo = FALSE,
  edgeLabelSpace = TRUE,
  nodeOrder = NULL,
  radial = FALSE,
  focusNode = NULL,
  preset = NULL,
  ...
)
}
\arguments{
\item{rankdir}{Layout direction: "TB" (top to bottom), "BT" (bottom to top),
"LR" (left to right), or "RL" (right to left).}

\item{align}{Node alignment: "UL" (upper left), "UR" (upper right),
"DL" (down left), or "DR" (down right).}

\item{nodesep}{Node spacing (px). When rankdir is "TB" or "BT", it's the horizontal.
spacing of nodes; when rankdir is "LR" or "RL", it's the vertical spacing of nodes.}

\item{nodesepFunc}{Function to customize node spacing for different nodes,
in the form of function(node) that returns a number. Has higher priority than nodesep.}

\item{ranksep}{Layer spacing (px). When rankdir is "TB" or "BT", it's the vertical spacing
between adjacent layers; when rankdir is "LR" or "RL", it's the horizontal spacing.}

\item{ranksepFunc}{Function to customize layer spacing, in the form of function(node)
that returns a number. Has higher priority than ranksep.}

\item{ranker}{Algorithm for assigning ranks to nodes: "network-simplex", "tight-tree",
or "longest-path".}

\item{nodeSize}{Node size for collision detection. Can be a single number (same width/height),
an array \code{[width, height]}, or a function that returns either.}

\item{controlPoints}{Whether to retain edge control points.}

\item{begin}{Alignment position of the upper left corner of the layout.
Can be \code{[x, y]} or \code{[x, y, z]}.}

\item{sortByCombo}{Whether to sort nodes on the same layer by parentId to prevent combo overlap.}

\item{edgeLabelSpace}{Whether to leave space for edge labels.}

\item{nodeOrder}{Reference array of node order on the same layer, containing node id values.}

\item{radial}{Whether to perform a radial layout based on dagre.}

\item{focusNode}{Focused node (only used when radial=TRUE). Can be a node ID or node object.}

\item{preset}{Node positions to reference during layout calculation.}

\item{...}{Additional parameters to pass to the layout.}
}
\value{
A list containing the configuration for G6 AntV Dagre layout.
}
\description{
This function creates a configuration list for G6 AntV Dagre layout
with all available options as parameters.
}
\examples{
# Basic dagre layout
dagre_config <- antv_dagre_layout()

# Horizontal layout with custom spacing
dagre_config <- antv_dagre_layout(
  rankdir = "LR",
  align = "UL",
  nodesep = 80,
  ranksep = 150
)

# Radial layout with focus node
dagre_config <- antv_dagre_layout(
  radial = TRUE,
  focusNode = "node1",
  ranker = "tight-tree"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{is_g6_node}
\alias{is_g6_node}
\alias{is_g6_edge}
\alias{is_g6_combo}
\title{Check if an object is a G6 element}
\usage{
is_g6_node(x)

is_g6_edge(x)

is_g6_combo(x)
}
\arguments{
\item{x}{An object of class \code{g6_node}, \code{g6_edge}, or \code{g6_combo}.}
}
\value{
Logical indicating if \code{x} is of the specified class.
}
\description{
Check if an object is a G6 element
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{lesmis}
\alias{lesmis}
\title{Character network from "Les miserables" novel}
\format{
A list with 2 data frames:
\describe{
\item{nodes}{data frame with 77 rows for the nodes. Contains node labels and x/y coordinates.}
\item{edges}{data frame with 254 rows for the edges. Contains souyrce/target and the number of times the interaction happened.}
}
}
\source{
\url{https://networks.skewed.de/net/lesmis}
}
\usage{
data(lesmis)
}
\description{
A dataset containing Les Misérables characters network, encoding interactions
between characters of Victor Hugo's novel. Two characters are connected
whenever they appear in the same chapter. This dataset was first created
by Donald Knuth as part of the Stanford Graph Base.
(https://people.sc.fsu.edu/~jburkardt/datasets/sgb/sgb.html).
It contains 77 nodes corresponding to characters of the novel, and 254 edges.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{watermark}
\alias{watermark}
\title{Configure Watermark Plugin}
\usage{
watermark(
  key = "watermark",
  width = 200,
  height = 100,
  opacity = 0.2,
  rotate = pi/12,
  imageURL = NULL,
  text = NULL,
  textFill = "#000",
  textFontSize = 16,
  textFontFamily = NULL,
  textFontWeight = NULL,
  textFontVariant = NULL,
  textAlign = c("center", "end", "left", "right", "start"),
  textBaseline = c("alphabetic", "bottom", "hanging", "ideographic", "middle", "top"),
  backgroundRepeat = "repeat",
  backgroundAttachment = NULL,
  backgroundBlendMode = NULL,
  backgroundClip = NULL,
  backgroundColor = NULL,
  backgroundImage = NULL,
  backgroundOrigin = NULL,
  backgroundPosition = NULL,
  backgroundPositionX = NULL,
  backgroundPositionY = NULL,
  backgroundSize = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{width}{Width of a single watermark (number, default: 200).}

\item{height}{Height of a single watermark (number, default: 100).}

\item{opacity}{Opacity of the watermark (number, default: 0.2).}

\item{rotate}{Rotation angle of the watermark in radians (number, default: pi/12).}

\item{imageURL}{Image watermark URL, higher priority than text watermark (string, default: NULL).}

\item{text}{Watermark text content (string, default: NULL).}

\item{textFill}{Color of the text watermark (string, default: "#000").}

\item{textFontSize}{Font size of the text watermark (number, default: 16).}

\item{textFontFamily}{Font of the text watermark (string, default: NULL).}

\item{textFontWeight}{Font weight of the text watermark (string, default: NULL).}

\item{textFontVariant}{Font variant of the text watermark (string, default: NULL).}

\item{textAlign}{Text alignment of the watermark (string, default: "center").}

\item{textBaseline}{Baseline alignment of the text watermark (string, default: "middle").}

\item{backgroundRepeat}{Repeat mode of the watermark (string, default: "repeat").}

\item{backgroundAttachment}{Background attachment behavior of the watermark (string, default: NULL).}

\item{backgroundBlendMode}{Background blend mode of the watermark (string, default: NULL).}

\item{backgroundClip}{Background clip of the watermark (string, default: NULL).}

\item{backgroundColor}{Background color of the watermark (string, default: NULL).}

\item{backgroundImage}{Background image of the watermark (string, default: NULL).}

\item{backgroundOrigin}{Background origin of the watermark (string, default: NULL).}

\item{backgroundPosition}{Background position of the watermark (string, default: NULL).}

\item{backgroundPositionX}{Horizontal position of the watermark background (string, default: NULL).}

\item{backgroundPositionY}{Vertical position of the watermark background (string, default: NULL).}

\item{backgroundSize}{Background size of the watermark (string, default: NULL).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/watermark}.}
}
\value{
A list with the configuration settings for the watermark plugin.
}
\description{
Creates a configuration object for the watermark plugin in G6.
This plugin adds a watermark to the graph canvas.
}
\examples{
# Basic text watermark
config <- watermark(
  text = "G6 Graph",
  opacity = 0.1
)

# Image watermark
config <- watermark(
  imageURL = "https://gw.alipayobjects.com/os/s/prod/antv/assets/image/logo-with-text-73b8a.svg",
  width = 150,
  height = 75,
  opacity = 0.15,
  rotate = 0
)

# Customized text watermark
config <- watermark(
  text = "CONFIDENTIAL",
  textFill = "#ff0000",
  textFontSize = 24,
  textFontWeight = "bold",
  opacity = 0.08,
  rotate = pi/6,
  backgroundRepeat = "repeat-x"
)

# Watermark with background styling
config <- watermark(
  text = "Draft Document",
  textFill = "#333",
  backgroundColor = "#f9f9f9",
  backgroundClip = "content-box",
  backgroundSize = "cover"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{circular_layout}
\alias{circular_layout}
\title{Generate G6 AntV circular layout configuration}
\usage{
circular_layout(
  angleRatio = 1,
  center = NULL,
  clockwise = TRUE,
  divisions = 1,
  nodeSize = 10,
  nodeSpacing = 10,
  ordering = NULL,
  radius = NULL,
  startAngle = 0,
  endAngle = 2 * pi,
  startRadius = NULL,
  endRadius = NULL,
  width = NULL,
  height = NULL,
  ...
)
}
\arguments{
\item{angleRatio}{How many \code{2*PI} are there between the first node and the last node.}

\item{center}{Center of layout as vector \code{c(x, y)} or \code{c(x, y, z)}.}

\item{clockwise}{Is it arranged clockwise?}

\item{divisions}{Number of segments that nodes are placed on the ring.}

\item{nodeSize}{Node size (diameter) for collision detection.}

\item{nodeSpacing}{Minimum distance between rings.}

\item{ordering}{Basis for sorting nodes ("topology", "topology-directed", or "degree").}

\item{radius}{Radius of the circle (overrides startRadius and endRadius).}

\item{startAngle}{Starting angle of the layout.}

\item{endAngle}{End angle of the layout.}

\item{startRadius}{Starting radius of the spiral layout.}

\item{endRadius}{End radius of the spiral layout.}

\item{width}{Width of layout.}

\item{height}{Height of layout.}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/circular-layout}.}
}
\value{
A list containing the configuration for G6 AntV circular layout.
}
\description{
This function creates a configuration list for G6 AntV circular layout
with all available options as parameters.
}
\examples{
circular_config <- circular_layout(
  radius = 200,
  startAngle = 0,
  endAngle = pi,
  clockwise = FALSE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{as_g6_ports}
\alias{as_g6_ports}
\alias{as_g6_ports.g6_ports}
\alias{as_g6_ports.list}
\title{Coerce to a list of g6_port objects (g6_ports)}
\usage{
as_g6_ports(x, ...)

\method{as_g6_ports}{g6_ports}(x, ...)

\method{as_g6_ports}{list}(x, ...)
}
\arguments{
\item{x}{An object to coerce.}

\item{...}{Additional arguments (unused).}
}
\value{
An object of class \code{g6_ports}.
}
\description{
Coerce to a list of g6_port objects (g6_ports)
}
\examples{
as_g6_ports(list(
  list(key = "input-1", type = "input", placement = "left"),
  list(key = "output-1", type = "output", placement = "right")
))
as_g6_ports(g6_ports(
  g6_port("input-1", type = "input"),
  g6_port("output-1", type = "output")
))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_set_options}
\alias{g6_set_options}
\title{Set options for a g6 graph via proxy}
\usage{
g6_set_options(graph, ...)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{...}{Named arguments representing the options to update and their new values.
These can include any valid g6 graph options such as fitView, animate, modes, etc.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function allows updating various configuration options of an existing g6 graph
instance using a proxy object within a Shiny application.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

The function allows updating various graph options dynamically without having to
re-render the entire graph. This is useful for changing behavior, appearance,
or interaction modes in response to user input.
}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{radial_layout}
\alias{radial_layout}
\title{Generate G6 Radial layout configuration}
\usage{
radial_layout(
  center = NULL,
  focusNode = NULL,
  height = NULL,
  width = NULL,
  nodeSize = NULL,
  nodeSpacing = 10,
  linkDistance = 50,
  unitRadius = 100,
  maxIteration = 1000,
  maxPreventOverlapIteration = 200,
  preventOverlap = FALSE,
  sortBy = NULL,
  sortStrength = 10,
  strictRadial = TRUE,
  ...
)
}
\arguments{
\item{center}{Numeric vector of length 2. Center coordinates.}

\item{focusNode}{Character, list (Node), or NULL. Radiating center node. Defaults to NULL.}

\item{height}{Numeric. Canvas height.}

\item{width}{Numeric. Canvas width.}

\item{nodeSize}{Numeric. Node size (diameter).}

\item{nodeSpacing}{Numeric or function. Minimum node spacing (effective when preventing overlap). Defaults to 10.}

\item{linkDistance}{Numeric. Edge length. Defaults to 50.}

\item{unitRadius}{Numeric or NULL. Radius per circle. Defaults to 100.}

\item{maxIteration}{Numeric. Maximum number of iterations. Defaults to 1000.}

\item{maxPreventOverlapIteration}{Numeric. Max iterations for overlap prevention. Defaults to 200.}

\item{preventOverlap}{Logical. Whether to prevent node overlap. Defaults to FALSE.}

\item{sortBy}{Character. Field for sorting nodes in the same layer.}

\item{sortStrength}{Numeric. Sorting strength for nodes in the same layer. Defaults to 10.}

\item{strictRadial}{Logical. Strictly place nodes in the same layer on the same ring. Defaults to TRUE.}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/radial-layout}.}
}
\value{
A list containing the configuration for G6 radial layout.
}
\description{
This function creates a configuration list for G6 Radial layout
with all available options as parameters.
}
\examples{
if (interactive()) {
  g6(jsonUrl = "https://assets.antv.antgroup.com/g6/radial.json") |>
    g6_layout(
      radial_layout(
        unitRadius = 100,
        linkDistance = 200
      )
    ) |>
    g6_behaviors(
      "zoom-canvas",
      drag_element()
    )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_update_nodes}
\alias{g6_update_nodes}
\alias{g6_update_edges}
\alias{g6_update_combos}
\title{Update nodes/edges/combos to a g6 graph via proxy}
\usage{
g6_update_nodes(graph, ...)

g6_update_edges(graph, ...)

g6_update_combos(graph, ...)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{...}{Nodes or edges or combos. You can pass a list of nodes/edges/combos, a dataframe
or leverage the g6_nodes(), g6_edges() or g6_combos() helpers or pass individual elements like
g6_node(), g6_edge() or g6_combo(). Elements structure must be compliant with specifications listed at
\url{https://g6.antv.antgroup.com/manual/element/overview}.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function updates one or more nodes/edges/combos to an existing g6 graph instance
using a proxy object. This allows updating the graph without completely
re-rendering it.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

See \url{https://g6.antv.antgroup.com/en/api/data#graphupdatenodedata},
\url{https://g6.antv.antgroup.com/en/api/data#graphupdateedgedata} and
\url{https://g6.antv.antgroup.com/en/api/data#graphupdatecombodata} for more details.
}
\examples{
if (interactive()) {
  library(shiny)
  library(g6R)
  library(bslib)

  # Static data defined globally
  nodes <- data.frame(id = 1:3)
  edges <- data.frame(source = c(1, 2), target = c(2, 3))

  ui <- page_fluid(
    title = "Update Nodes Dynamically",
    g6_output("graph"),
    actionButton("update_node", "Update Node 1 Label")
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(nodes = nodes, edges = edges) |> g6_layout()
    })

    observeEvent(input$update_node, {
      # Update label for node 1
      g6_update_nodes(
        g6_proxy("graph"),
        g6_node(id = 1, style = list(labelText = "Node label updated"))
      )
    })
  }

  shinyApp(ui, server)
}

}
\seealso{
\code{\link{g6_proxy}}, \code{\link{g6_remove_nodes}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{validate_edges_ports}
\alias{validate_edges_ports}
\title{Validate that all edges connect input/output ports}
\usage{
validate_edges_ports(edges, nodes)
}
\arguments{
\item{edges}{List of g6_edge objects.}

\item{nodes}{List of g6_node objects.}
}
\value{
Invisibly TRUE if all edges are valid, otherwise stops with an error.
}
\description{
Validate that all edges connect input/output ports
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{hull}
\alias{hull}
\title{Configure Hull Plugin}
\usage{
hull(
  members,
  key = "hull",
  concavity = Inf,
  corner = c("rounded", "smooth", "sharp"),
  padding = 10,
  label = TRUE,
  labelText = NULL,
  labelPlacement = c("bottom", "left", "right", "top", "center"),
  labelBackground = FALSE,
  labelPadding = 0,
  labelCloseToPath = TRUE,
  labelAutoRotate = TRUE,
  labelOffsetX = 0,
  labelOffsetY = 0,
  labelMaxWidth = NULL,
  ...
)
}
\arguments{
\item{members}{Elements within the hull, including nodes and edges (character/numeric vector, required).}

\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{concavity}{Concavity parameter, larger values create less concave hulls (number, default: Infinity).}

\item{corner}{Corner type: "rounded", "smooth", or "sharp" (string, default: "rounded").}

\item{padding}{Padding around the elements (number, default: 10).}

\item{label}{Whether to display the label (boolean, default: TRUE).}

\item{labelText}{Label text content. Default to NULL.}

\item{labelPlacement}{Label position: "left", "right", "top", "bottom", or "center" (string, default: "bottom").}

\item{labelBackground}{Whether to display the background (boolean, default: FALSE).}

\item{labelPadding}{Label padding (number or numeric vector, default: 0).}

\item{labelCloseToPath}{Whether the label is close to the hull (boolean, default: TRUE).}

\item{labelAutoRotate}{Whether the label rotates with the hull, effective only when
closeToPath is true (boolean, default: TRUE).}

\item{labelOffsetX}{X-axis offset (number, default: 0).}

\item{labelOffsetY}{Y-axis offset (number, default: 0).}

\item{labelMaxWidth}{Maximum width of the text, exceeding will be ellipsized (number or NULL, default: NULL).}

\item{...}{Other options.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/hull}.}
}
\value{
A list with the configuration settings for the hull plugin.
}
\description{
Creates a configuration object for the hull plugin in G6.
This plugin creates a hull (convex or concave) that surrounds specified graph elements.
}
\examples{
# Basic configuration
config <- hull(members = c("node1", "node2", "node3"))

# Custom configuration for a cluster
config <- hull(
  key = "cluster-hull",
  members = c("node1", "node2", "node3", "node4"),
  concavity = 0.8,
  corner = "smooth",
  padding = 15,
  label = TRUE,
  labelPlacement = "top",
  labelBackground = TRUE,
  labelPadding = c(4, 8),
  labelMaxWidth = 100
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{canvas_config}
\alias{canvas_config}
\title{Create Canvas Configuration for G6 Graphs}
\usage{
canvas_config(
  container = NULL,
  devicePixelRatio = NULL,
  width = NULL,
  height = NULL,
  cursor = NULL,
  background = NULL,
  renderer = NULL,
  enableMultiLayer = NULL
)
}
\arguments{
\item{container}{The container element for the canvas. Can be a CSS selector string
or an HTML element reference.}

\item{devicePixelRatio}{The device pixel ratio to use for rendering. Higher values
provide sharper rendering on high-DPI displays but may impact performance.
If not specified, the device's pixel ratio will be used.}

\item{width}{The width of the canvas in pixels.}

\item{height}{The height of the canvas in pixels.}

\item{cursor}{The CSS cursor style to use when hovering over the canvas.
Common values include "default", "pointer", "move", etc.}

\item{background}{The background color of the canvas. Can be any valid CSS color
value (hex, rgb, rgba, named colors).}

\item{renderer}{A function that returns a renderer for different layers.
The function takes a layer parameter which can be 'background', 'main',
'label', or 'transient'.}

\item{enableMultiLayer}{Whether to enable multi-layer rendering. This is a
non-dynamic parameter and is only effective during initialization.
Multi-layer rendering can improve performance for complex graphs by
separating elements into different rendering layers.}
}
\value{
A list containing the canvas configuration that can be passed to \code{\link[=g6_options]{g6_options()}}.
}
\description{
Configures the canvas settings for a G6 graph. The canvas is the rendering
surface where the graph is drawn.
}
\details{
Canvas configuration controls how the graph is rendered, including its size,
scaling, background, and rendering layer settings. This function provides a
structured way to configure all canvas-related options.

Note that many of these settings (container, width, height, devicePixelRatio,
background, cursor) can also be set directly in the main graph configuration,
which will be automatically converted to canvas configuration items.
}
\examples{
# Basic canvas configuration
config <- canvas_config(
  container = "#graph-container",
  width = 800,
  height = 600
)

# Canvas with multi-layer rendering enabled
config <- canvas_config(
  container = "#graph-container",
  width = 1000,
  height = 700,
  enableMultiLayer = TRUE,
  cursor = "grab"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{as_g6_node}
\alias{as_g6_node}
\alias{as_g6_node.g6_node}
\alias{as_g6_node.list}
\alias{as_g6_edge}
\alias{as_g6_edge.g6_edge}
\alias{as_g6_edge.list}
\alias{as_g6_combo}
\alias{as_g6_combo.g6_combo}
\alias{as_g6_combo.list}
\title{Coerce to a g6 element object}
\usage{
as_g6_node(x, ...)

\method{as_g6_node}{g6_node}(x, ...)

\method{as_g6_node}{list}(x, ...)

as_g6_edge(x, ...)

\method{as_g6_edge}{g6_edge}(x, ...)

\method{as_g6_edge}{list}(x, ...)

as_g6_combo(x, ...)

\method{as_g6_combo}{g6_combo}(x, ...)

\method{as_g6_combo}{list}(x, ...)
}
\arguments{
\item{x}{An object to be coerced. List or g6 element are supported.}

\item{...}{Additional arguments (unused).}
}
\value{
An element of class \code{g6_node}, \code{g6_edge}, or \code{g6_combo}.
}
\description{
Coerce to a g6 element object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{dag}
\alias{dag}
\title{Example DAG graph}
\format{
A list with 2 data frames:
\describe{
\item{nodes}{data frame with 12 rows for the nodes.}
\item{edges}{data frame with 12 rows for the edges.}
\item{combo}{data frame with 3 rows for the combos.}
}
}
\source{
\url{https://gw.alipayobjects.com/os/antvdemo/assets/data/algorithm-category.json}
}
\usage{
data(dag)
}
\description{
The graph is a directed acyclic graph
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{edge_options}
\alias{edge_options}
\title{Create Edge Options Configuration for G6 Graphs}
\usage{
edge_options(
  type = c("line", "polyline", "arc", "quadratic", "cubic", "cubic-vertical",
    "cubic-horizontal", "loop"),
  style = edge_style_options(),
  state = NULL,
  palette = NULL,
  animation = NULL
)
}
\arguments{
\item{type}{Edge type. Can be a built-in edge type name or a custom edge name.
Built-in types include "line", "polyline", "arc", "quadratic", "cubic", "cubic-vertical",
"cubic-horizontal", "loop", etc. Default: "line".}

\item{style}{Edge style configuration. Controls the appearance of edges including color,
width, dash patterns, etc. Can be created with \code{edge_style_options()}.
Default: NULL.}

\item{state}{Defines the style of the edge in different states, such as hover, selected,
disabled, etc. Should be a list mapping state names to style configurations.
Default: NULL.}

\item{palette}{Defines the color palette of the edge, used to map colors based on different data.
Default: NULL.}

\item{animation}{Defines the animation effect of the edge. Can be created with
\code{animation_config()}.
Default: NULL.}
}
\value{
A list containing edge options configuration that can be passed to \code{\link[=g6_options]{g6_options()}}.
}
\description{
Configures the general options for edges in a G6 graph. These settings control
the type, style, state, palette, and animation of edges.
}
\details{
Edge options allow defining how edges appear and behave in a G6 graph. This includes
selecting edge types, setting styles, configuring state-based appearances, defining
color palettes, and specifying animation effects.
}
\examples{
# Basic edge options with default line type
options <- edge_options()

# Curved edge with custom style
options <- edge_options(
  type = "cubic",
  style = edge_style_options(
    stroke = "#1783FF",
    lineWidth = 2,
    endArrow = TRUE
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{grid_line}
\alias{grid_line}
\title{Configure Grid Line Plugin}
\usage{
grid_line(
  key = "grid-line",
  border = TRUE,
  borderLineWidth = 1,
  borderStroke = "#eee",
  borderStyle = "solid",
  follow = FALSE,
  lineWidth = 1,
  size = 20,
  stroke = "#eee",
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{border}{Whether to display the border (boolean, default: TRUE).}

\item{borderLineWidth}{Border line width (number, default: 1).}

\item{borderStroke}{Border color (string, default: "#eee").}

\item{borderStyle}{Border style (string, default: "solid").}

\item{follow}{Whether the grid follows canvas movements (boolean or list, default: FALSE).}

\item{lineWidth}{Grid line width (number or string, default: 1).}

\item{size}{Grid unit size in pixels (number, default: 20).}

\item{stroke}{Grid line color (string, default: "#eee").}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/grid-line}.}
}
\value{
A list with the configuration settings for the grid-line plugin.
}
\description{
Creates a configuration object for the grid-line plugin in G6.
This plugin adds a background grid to the graph canvas.
}
\examples{
# Basic configuration
config <- grid_line()

# Custom configuration
config <- grid_line(
  key = "my-grid",
  border = TRUE,
  borderLineWidth = 2,
  borderStroke = "#ccc",
  borderStyle = "dashed",
  follow = list(
    translate = TRUE,
    zoom = FALSE
  ),
  lineWidth = 0.5,
  size = 30,
  stroke = "#e0e0e0"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{combo_combined_layout}
\alias{combo_combined_layout}
\title{Create an AntV Combo Combined Layout}
\usage{
combo_combined_layout(
  center = NULL,
  comboPadding = 10,
  innerLayout = NULL,
  nodeSize = 10,
  outerLayout = NULL,
  spacing = NULL,
  treeKey = NULL,
  ...
)
}
\arguments{
\item{center}{Layout center coordinates. A numeric vector of length 2 \code{[x, y]}.
If NULL, uses the graph center. Default is NULL.}

\item{comboPadding}{Padding value inside the combo, used only for force
calculation, not for rendering. It is recommended to set the same value
as the visual padding. Can be a number, numeric vector, function, or
JS function. Default is 10.}

\item{innerLayout}{Layout algorithm for elements inside the combo. Should be
a Layout object or function. If NULL, uses ConcentricLayout as default.}

\item{nodeSize}{Node size (diameter), used for collision detection. If not
specified, it is calculated from the node's size property. Can be a number,
numeric vector, function, or JS function. Default is 10.}

\item{outerLayout}{Layout algorithm for the outermost layer. Should be a
Layout object or function. If NULL, uses ForceLayout as default.}

\item{spacing}{Minimum spacing between node/combo edges when preventNodeOverlap
or preventOverlap is true. Can be a number, function, or JS function for
different nodes. Default is NULL.}

\item{treeKey}{Tree key identifier as a character string. Default is NULL.}

\item{...}{Additional parameters passed to the layout configuration.
See \url{https://g6.antv.antgroup.com/en/manual/layout/combo-combined-layout}.}
}
\value{
A layout configuration object for use with G6 graphs.
}
\description{
Creates a combo combined layout configuration for G6 graphs. This layout
algorithm combines different layout strategies for elements inside combos
and the outermost layer, providing hierarchical organization of graph elements.
}
\details{
The combo combined layout is particularly useful for graphs with hierarchical
structures where you want different layout algorithms for different levels
of the hierarchy. The inner layout handles elements within combos, while
the outer layout manages the overall arrangement.
}
\examples{
# Basic combo combined layout
layout <- combo_combined_layout()

# Custom configuration with specific center and padding
layout <- combo_combined_layout(
  comboPadding = 20,
  nodeSize = 15,
  spacing = 10
)

}
\seealso{
\code{\link{antv_dagre_layout}} for dagre layout configuration
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_add_plugin}
\alias{g6_add_plugin}
\title{Add a plugin to a g6 graph via proxy}
\usage{
g6_add_plugin(graph, ...)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{...}{Named arguments where each name is a plugin type and each value is a
list of configuration options for that plugin.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function adds one or more plugins to an existing g6 graph instance
using a proxy object within a Shiny application.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

G6 plugins extend the functionality of the graph visualization with features like
minimaps, toolbar controls, contextual menus, and more. This function allows adding
these plugins dynamically after the graph has been initialized.
}
\seealso{
\code{\link{g6_proxy}}, \code{\link{g6_update_plugin}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{g6_options}
\alias{g6_options}
\title{Configure Global Options for G6 Graph}
\usage{
g6_options(
  graph,
  node = NULL,
  edge = NULL,
  combo = NULL,
  autoFit = NULL,
  canvas = NULL,
  animation = TRUE,
  autoResize = FALSE,
  background = NULL,
  cursor = valid_cursors,
  devicePixelRatio = NULL,
  renderer = NULL,
  padding = NULL,
  rotation = 0,
  x = NULL,
  y = NULL,
  zoom = 1,
  zoomRange = c(0.01, 10),
  theme = "light",
  ...
)
}
\arguments{
\item{graph}{g6 graph instance.}

\item{node}{Node configuration. Controls the default appearance and behavior of nodes.
Created with \code{node_options()}. Default: NULL.}

\item{edge}{Edge configuration. Controls the default appearance and behavior of edges.
Created with \code{edge_options()}. Default: NULL.}

\item{combo}{Combo configuration. Controls the default appearance and behavior of combo nodes.
Created with \code{combo_options()}. Default: NULL.}

\item{autoFit}{Automatically fit the graph content to the canvas.
Created with \code{auto_fit_config()}. Default: NULL.}

\item{canvas}{Canvas configuration for the graph rendering surface.
Created with \code{canvas_config()}. Default: NULL.}

\item{animation}{Global animation configuration for graph transitions.
Created with \code{animation_config()}. Default: TRUE.}

\item{autoResize}{Whether the graph should automatically resize when the window size changes.
Default: FALSE.}

\item{background}{Background color of the graph. If not specified, the background will be transparent.
Default: NULL.}

\item{cursor}{Default mouse cursor style when hovering over the graph.
Options include: "default", "pointer", "move", etc.
Default: "default".}

\item{devicePixelRatio}{Device pixel ratio for rendering on high-DPI displays. If NULL,
the browser's default device pixel ratio will be used.
Default: NULL.}

\item{renderer}{Rendering engine to use. A JS function. To render
as svg, you can pass \verb{() => new SVGRenderer()}.
Default: NULL (G6 will choose the appropriate renderer).}

\item{padding}{Padding around the graph content in pixels. Can be a single number for equal padding
on all sides or a vector of four numbers \code{[top, right, bottom, left]}.
Default: NULL.}

\item{rotation}{Rotation angle of the entire graph in degrees.
Default: 0.}

\item{x}{X-coordinate of the graph's center relative to the container.
Default: NULL (will use container center).}

\item{y}{Y-coordinate of the graph's center relative to the container.
Default: NULL (will use container center).}

\item{zoom}{Initial zoom level of the graph. 1 represents 100\% (original size).
Default: 1.}

\item{zoomRange}{Minimum and maximum allowed zoom levels, specified as a vector
with two elements: c(min_zoom, max_zoom).
Default: c(0.01, 10).}

\item{theme}{Color theme for the graph. Either \code{light} or \code{dark} or a list
representing a custom theme:
see \url{https://g6.antv.antgroup.com/en/manual/theme/custom-theme}.}

\item{...}{Other configuration parameters.}
}
\value{
A list containing all specified G6 graph configuration options.
}
\description{
Sets up the global configuration options for a G6 graph including node, edge and
combo styles, layout, canvas, animation, and interactive behavior settings.
}
\details{
The \code{g6_options} function provides a comprehensive configuration interface for G6 graphs.
It allows you to control all aspects of graph rendering and behavior, from styling of
individual elements to global visualization settings.
}
\examples{
# Basic usage with defaults
opts <- g6_options(g6())

# Customize node and edge styles
opts <- g6_options(
  g6(),
  node = node_options(
    type = "circle",
    style = node_style_options(
      fill = "#1783FF",
      stroke = "#0066CC"
    )
  ),
  edge = edge_options(
    type = "cubic",
    style = edge_style_options(
      stroke = "#999999",
      lineWidth = 1.5
    )
  )
)

# Configure graph with dark theme, auto-resize, and custom background
opts <- g6_options(
  g6(),
  theme = "dark",
  autoResize = TRUE,
  background = "#222222",
  padding = 20,
  zoom = 0.8,
  zoomRange = c(0.5, 2)
)

# Configure with custom animations
opts <- g6_options(
  g6(),
  animation = animation_config(
    duration = 500,
    easing = "easeCubic"
  ),
  autoFit = auto_fit_config(duration = 300, easing = "ease-out")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{JS}
\alias{JS}
\title{Marks as string to be processed as a JS function}
\usage{
JS(...)
}
\arguments{
\item{...}{Any valid JS element.}
}
\value{
A character vector with class "JS_EVAL" that can be used in htmlwidgets
to mark is as a JavaScript function.
}
\description{
Useful for htmlwidgets
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{dendrogram_layout}
\alias{dendrogram_layout}
\title{Generate G6 Dendrogram layout configuration}
\usage{
dendrogram_layout(
  direction = c("LR", "RL", "TB", "BT", "H", "V"),
  nodeSep = 20,
  rankSep = 200,
  radial = FALSE,
  ...
)
}
\arguments{
\item{direction}{Character. Layout direction. Options: "LR", "RL", "TB", "BT", "H", "V". Defaults to "LR".}

\item{nodeSep}{Numeric. Node spacing, distance between nodes on the same level. Defaults to 20.}

\item{rankSep}{Numeric. Rank spacing, distance between different levels. Defaults to 200.}

\item{radial}{Logical. Whether to enable radial layout. Defaults to FALSE.}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/dendrogram-layout}.}
}
\value{
A list containing the configuration for G6 dendrogram layout.
}
\description{
This function creates a configuration list for G6 Dendrogram layout
with all available options as parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{g6_port}
\alias{g6_port}
\alias{g6_input_port}
\alias{g6_output_port}
\title{Create a G6 Port}
\usage{
g6_port(
  key,
  label = key,
  type = c("input", "output"),
  arity = 1,
  visibility = c("visible", "hover", "hidden"),
  ...
)

g6_input_port(
  key,
  label = key,
  arity = 1,
  visibility = c("visible", "hover", "hidden"),
  fill = "#52C41A",
  ...
)

g6_output_port(
  key,
  label = key,
  arity = 1,
  visibility = c("visible", "hover", "hidden"),
  fill = "#FF4D4F",
  ...
)
}
\arguments{
\item{key}{Character. Unique identifier for the port (required).}

\item{label}{Character. Label text for the port (optional).}

\item{type}{Character. Either "input" or "output" (required).}

\item{arity}{Numeric. Maximum number of
connections this port can accept (default: 1).
Use 0, Inf, or any non-negative integer.}

\item{visibility}{Character. Controls port visibility behavior:
\itemize{
\item \code{"visible"}: Ports are always shown (default).
\item \code{"hover"}: Ports appear only when hovering over the node.
\item \code{"hidden"}: Ports are never visible.
}}

\item{...}{Additional port style parameters, e.g. \code{placement} and
\code{r} (radius, default 6). See
\url{https://g6.antv.antgroup.com/en/manual/element/node/base-node#portstyleprops}.
In addition to G6's native placements (\code{"left"}, \code{"right"},
\code{"top"}, \code{"bottom"}, or a \code{c(x, y)} pair), g6R adds
\code{placement = "label-bottom"}: the port snaps to the bottom-centre of the
node's label background when the node has a bottom label, falling back to a
normal bottom-of-node port otherwise. Pass \code{ripple = FALSE} to disable
the hover ripple for a port, or \code{haloFill} to override the colour of the
ring that makes the port look set into the node surface.}

\item{fill}{Character. Color of the port
(default: "#52C41A" for input ports,
"#FF4D4F" for output ports).}
}
\value{
An S3 object of class 'g6_port'.
}
\description{
Create a G6 Port

Create a single-connection input port

Create an output port (single connection by default)
}
\note{
To create an (input/output) port with multiple connections,
simply set the arity to \code{Inf} or any positive integer.
}
\examples{
g6_port("input-1", label = "port 1", type = "input", arity = 2, placement = "left")
g6_port("output-1", label = "port 2", type = "output", placement = "right")
g6_port("input-2", type = "input", visibility = "hover")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{g6_collapse_options}
\alias{g6_collapse_options}
\alias{is_g6_collapse_options}
\title{Configure collapse button for nodes}
\usage{
g6_collapse_options(
  collapsed = FALSE,
  visibility = c("visible", "hover"),
  placement = "right-top",
  r = 6,
  fill = "#fff",
  stroke = "#9cabd4",
  lineWidth = 1,
  iconStroke = "#9cabd4",
  iconLineWidth = 1.4,
  cursor = "pointer",
  zIndex = 999
)

is_g6_collapse_options(x)
}
\arguments{
\item{collapsed}{Logical. Whether the node should be collapsed initially. Default is FALSE.}

\item{visibility}{Character. Visibility mode of the collapse button.
Either \code{"visible"} (always shown when children exist) or \code{"hover"}
(shown only on mouse hover). Default is \code{"visible"}.}

\item{placement}{Character or numeric vector. Position of the collapse button.
Can be one of: "top", "right", "bottom", "left", "right-top", "right-bottom",
"left-top", "left-bottom", or a numeric vector of length 2 for custom coordinates.}

\item{r}{Numeric. Radius of the button. Default is 8.}

\item{fill}{Character. Fill color of the button background. Default is "#fff".}

\item{stroke}{Character. Stroke color of the button border. Default is "#CED4D9".}

\item{lineWidth}{Numeric. Width of the button border. Default is 1.}

\item{iconStroke}{Character. Stroke color of the +/- icon. Default is "#000".}

\item{iconLineWidth}{Numeric. Width of the +/- icon lines. Default is 1.4.}

\item{cursor}{Character. CSS cursor style. Default is "pointer".}

\item{zIndex}{Numeric. Z-index for layering. Default is 999.}

\item{x}{An object to check.}
}
\value{
A list of collapse button configuration options.

Logical indicating if x is of class g6_collapse_options.
}
\description{
Configure collapse button for nodes

Check if object is a g6_collapse_options configuration
}
\examples{
# Default collapse button
g6_collapse_options()

# Custom styled collapse button
g6_collapse_options(
  placement = "right-top",
  fill = "#f0f0f0",
  stroke = "#333",
  r = 10
)

# Collapse button with custom coordinates
g6_collapse_options(placement = c(1, 0.2))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_update_behavior}
\alias{g6_update_behavior}
\title{Update a behavior in a g6 graph via proxy}
\usage{
g6_update_behavior(graph, key, ...)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{key}{Character string identifying the behavior to update.}

\item{...}{Named arguments representing the behavior configuration options to update
and their new values.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function allows updating the configuration of an existing behavior in a g6 graph
instance using a proxy object within a Shiny application.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

Behaviors in G6 define how the graph responds to user interactions like dragging,
zooming, clicking, etc. This function allows dynamically updating the configuration
of these behaviors without having to reinitialize the graph.
}
\examples{
if (interactive()) {
  library(shiny)
  library(g6R)
  library(bslib)

  nodes <- data.frame(id = c("node1", "node2", "node3"))
  edges <- data.frame(source = "node1", target = "node2")

  ui <- page_fluid(
    g6_output("graph"),
    checkboxInput("enable_click_select", "Enable Click Select", value = TRUE)
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(nodes = nodes, edges = edges) |>
        g6_layout() |>
        g6_options(animation = FALSE) |>
        g6_behaviors(
          click_select()
        )
    })

    observeEvent(input$enable_click_select, {
      g6_update_behavior(
        g6_proxy("graph"),
        key = "click-select",
        enable = input$enable_click_select
      )
    })
  }

  shinyApp(ui, server)
}

}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{g6_nodes}
\alias{g6_nodes}
\alias{g6_edges}
\alias{g6_combos}
\alias{validate_elements}
\alias{validate_elements.g6_nodes}
\alias{validate_elements.g6_edges}
\alias{validate_elements.g6_combos}
\alias{as_g6_nodes.g6_nodes}
\alias{as_g6_nodes.data.frame}
\alias{as_g6_nodes.list}
\title{Create and validate lists of G6 elements}
\usage{
g6_nodes(...)

g6_edges(...)

g6_combos(...)

validate_elements(x, ...)

\method{validate_elements}{g6_nodes}(x, ...)

\method{validate_elements}{g6_edges}(x, ...)

\method{validate_elements}{g6_combos}(x, ...)

\method{as_g6_nodes}{g6_nodes}(x, ...)

\method{as_g6_nodes}{data.frame}(x, ...)

\method{as_g6_nodes}{list}(x, ...)
}
\arguments{
\item{...}{Additional arguments (unused).}

\item{x}{An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.}
}
\value{
An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.

Invisibly returns the validated object.
}
\description{
Constructors for lists of G6 node, edge, and combo elements.
Each function accepts multiple validated elements and returns a
list with the appropriate class.
All elements are validated on construction.

S3 generic for validating lists of G6 elements.
}
\examples{
nodes <- g6_nodes(
  g6_node(id = "n1"),
  g6_node(id = "n2")
)
edges <- g6_edges(
  g6_edge(source = "n1", target = "n2")
)
combos <- g6_combos(
  g6_combo(id = "c1")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{dagre_layout}
\alias{dagre_layout}
\title{Generate G6 AntV Dagre layout configuration}
\usage{
dagre_layout(
  rankdir = c("TB", "BT", "LR", "RL"),
  align = c("UL", "UR", "DL", "DR"),
  nodesep = 50,
  ranksep = 100,
  ranker = c("network-simplex", "tight-tree", "longest-path"),
  nodeSize = NULL,
  controlPoints = FALSE,
  ...
)
}
\arguments{
\item{rankdir}{Layout direction: "TB" (top to bottom), "BT" (bottom to top),
"LR" (left to right), or "RL" (right to left).}

\item{align}{Node alignment: "UL" (upper left), "UR" (upper right),
"DL" (down left), or "DR" (down right).}

\item{nodesep}{Node spacing (px). When rankdir is "TB" or "BT", it's the horizontal
spacing of nodes; when rankdir is "LR" or "RL", it's the vertical spacing of nodes.}

\item{ranksep}{Interlayer spacing (px). When rankdir is "TB" or "BT", it's the
spacing between adjacent layers in the vertical direction; when rankdir is "LR" or "RL",
it represents the spacing between adjacent layers in the horizontal direction.}

\item{ranker}{The algorithm for assigning a level to each node: "network-simplex" (the network
simplex algorithm), "tight-tree" (the compact tree algorithm), or "longest-path"
(the longest path algorithm).}

\item{nodeSize}{G6 custom attribute to specify the node size uniformly or for each node.
Can be a single number (same width/height), an array \code{[width, height]},
or a function that returns either.}

\item{controlPoints}{Whether to keep the control points of the edge.}

\item{...}{Additional parameters to pass to the layout. See
\url{https://g6.antv.antgroup.com/en/manual/layout/dagre-layout}.}
}
\value{
A list containing the configuration for G6 AntV Dagre layout.
}
\description{
This function creates a configuration list for G6 AntV Dagre layout
with all available options as parameters. The Dagre layout is designed
for directed graphs, creating hierarchical layouts with nodes arranged in layers.
}
\examples{
# Basic dagre layout
dagre_config <- dagre_layout()

# Custom dagre layout with horizontal flow
dagre_config <- dagre_layout(
  rankdir = "LR",
  nodesep = 80,
  ranksep = 150,
  ranker = "tight-tree"
)

# Custom dagre layout with specific node size
dagre_config <- dagre_layout(
  nodeSize = 40,
  controlPoints = TRUE
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{drag_canvas}
\alias{drag_canvas}
\title{Configure Drag Canvas Behavior}
\usage{
drag_canvas(
  key = "drag-canvas",
  enable = NULL,
  animation = NULL,
  direction = c("both", "x", "y"),
  range = NULL,
  sensitivity = 10,
  trigger = NULL,
  onFinish = NULL,
  ...
)
}
\arguments{
\item{key}{Behavior unique identifier. Useful to modify this behavior from JS side.}

\item{enable}{Whether to enable this behavior (boolean or function, default: function that enables dragging only on canvas).}

\item{animation}{Drag animation configuration for keyboard movement (list, default: NULL).}

\item{direction}{Allowed drag direction: "x", "y", or "both" (string, default: "both").}

\item{range}{Draggable viewport range in viewport size units (number or numeric vector, default: Inf).}

\item{sensitivity}{Distance to trigger a single keyboard movement (number, default: 10).}

\item{trigger}{Keyboard keys to trigger dragging (list, default: NULL).}

\item{onFinish}{Callback function when dragging is completed (function, default: NULL).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/drag-canvas}.}
}
\value{
A list with the configuration settings for the drag-canvas behavior.
}
\description{
Creates a configuration object for the drag-canvas behavior in G6.
This allows users to drag the canvas to pan the view.
}
\examples{
# Basic configuration
config <- drag_canvas()

# Custom configuration
config <- drag_canvas(
  enable = TRUE,
  direction = "x",
  range = c(-100, 100),
  sensitivity = 5,
  trigger = list(
   up = "ArrowUp",
   down = "ArrowDown",
   left = "ArrowLeft",
   right = "ArrowRight"
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{validate_ports}
\alias{validate_ports}
\alias{validate_ports.list}
\alias{validate_ports.g6_ports}
\title{Validate a list of G6 ports}
\usage{
validate_ports(x, ...)

\method{validate_ports}{list}(x, ...)

\method{validate_ports}{g6_ports}(x, ...)
}
\arguments{
\item{x}{A list of g6_port objects.}

\item{...}{Generic consistency.}
}
\value{
The validated list (invisibly).
}
\description{
Validate a list of G6 ports
}
\examples{
validate_ports(list(
  g6_port("input-1", type = "input"),
  g6_port("output-1", type = "output")
))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{node_options}
\alias{node_options}
\title{Create Node Options Configuration for G6 Graphs}
\usage{
node_options(
  type = c("circle", "rect", "ellipse", "diamond", "triangle", "star", "image",
    "modelRect"),
  style = node_style_options(),
  state = NULL,
  palette = NULL,
  animation = NULL
)
}
\arguments{
\item{type}{Node type. Can be a built-in node type name or a custom node name.
Built-in types include "circle", "rect", "ellipse", "diamond", "triangle", etc.
Default: "circle".}

\item{style}{Node style configuration. Controls the appearance of nodes including color,
size, border, etc. Can be created with \code{node_style_options()}.
Default: NULL.}

\item{state}{Defines the style of the node in different states, such as hover, selected,
disabled, etc. Should be a list mapping state names to style configurations.
Default: NULL.}

\item{palette}{Defines the color palette of the node, used to map colors based on different data.
Default: NULL.}

\item{animation}{Defines the animation effect of the node. Can be created with
\code{animation_config()}.
Default: NULL.}
}
\value{
A list containing node options configuration that can be passed to \code{\link[=g6_options]{g6_options()}}.
}
\description{
Configures the general options for nodes in a G6 graph. These settings control
the type, style, state, palette, and animation of nodes.
}
\details{
Node options allow defining how nodes appear and behave in a G6 graph. This includes
selecting node types, setting styles, configuring state-based appearances, defining
color palettes, and specifying animation effects.
}
\examples{
# Basic node options with default circle type
options <- node_options()

# Rectangle node with custom style
options <- node_options(
  type = "rect",
  style = node_style_options(
    fill = "#E8F7FF",
    stroke = "#1783FF",
    lineWidth = 2
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{timebar}
\alias{timebar}
\title{Configure Timebar Plugin}
\usage{
timebar(
  data,
  key = "timebar",
  className = "g6-timebar",
  x = NULL,
  y = NULL,
  width = 450,
  height = 60,
  position = c("bottom", "top"),
  padding = 10,
  timebarType = c("time", "chart"),
  elementTypes = c("node", "edge", "combo"),
  mode = c("modify", "visibility"),
  values = NULL,
  loop = FALSE,
  getTime = NULL,
  labelFormatter = NULL,
  onChange = NULL,
  onReset = NULL,
  onSpeedChange = NULL,
  onPlay = NULL,
  onPause = NULL,
  onBackward = NULL,
  onForward = NULL,
  ...
)
}
\arguments{
\item{data}{Time data, either a vector of timestamps or a list of objects with time and value (required).}

\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{className}{Additional class name for the timebar DOM (string, default: "g6-timebar").}

\item{x}{X position, will be ignored if position is set (number, default: NULL).}

\item{y}{Y position, will be ignored if position is set (number, default: NULL).}

\item{width}{Timebar width (number, default: 450).}

\item{height}{Timebar height (number, default: 60).}

\item{position}{Timebar position: "bottom" or "top" (string, default: "bottom").}

\item{padding}{Padding around the timebar (number or numeric vector, default: 10).}

\item{timebarType}{Display type: "time" or "chart" (string, default: "time").}

\item{elementTypes}{Filter element types: vector of "node", "edge", and/or "combo"
(character vector, default: c("node")).}

\item{mode}{Control element filtering method: "modify" or "visibility" (string, default: "modify").}

\item{values}{Current time value (number, vector of two numbers, Date, or vector of two Dates, default: NULL).}

\item{loop}{Whether to loop playback (boolean, default: FALSE).}

\item{getTime}{Method to get element time (JS function, default: NULL).}

\item{labelFormatter}{Custom time formatting in chart mode (JS function, default: NULL).}

\item{onChange}{Callback when time interval changes (JS function, default: NULL).}

\item{onReset}{Callback when reset (JS function, default: NULL).}

\item{onSpeedChange}{Callback when playback speed changes (JS function, default: NULL).}

\item{onPlay}{Callback when playback starts (JS function, default: NULL).}

\item{onPause}{Callback when paused (JS function, default: NULL).}

\item{onBackward}{Callback when moving backward (JS function, default: NULL).}

\item{onForward}{Callback when moving forward (JS function, default: NULL).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/timebar}.}
}
\value{
A list with the configuration settings for the timebar plugin.
}
\description{
Creates a configuration object for the timebar plugin in G6.
This plugin adds a timeline or chart-based control for time-related data visualization.
}
\examples{
# Basic timebar with array of timestamps
config <- timebar(
  data = c(1609459200000, 1609545600000, 1609632000000)  # Jan 1-3, 2021 in milliseconds
)

# Chart-type timebar with time-value pairs
config <- timebar(
  data = list(
    list(time = 1609459200000, value = 10),
    list(time = 1609545600000, value = 25),
    list(time = 1609632000000, value = 15)
  ),
  timebarType = "chart",
  width = 600,
  height = 100,
  position = "top"
)

# With custom callbacks
config <- timebar(
  data = c(1609459200000, 1609545600000, 1609632000000),
  onChange = JS("(values) => {
    console.log('Time changed:', values);
  }"),
  onPlay = JS("() => {
    console.log('Playback started');
  }")
)

# With custom time getter function for elements
config <- timebar(
  data = c(1609459200000, 1609545600000, 1609632000000),
  getTime = JS("(datum) => {
    return datum.created_at; // Get time from created_at property
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{animation_config}
\alias{animation_config}
\title{Create Animation Configuration for G6 Graphs}
\usage{
animation_config(
  delay = NULL,
  direction = c("forward", "alternate", "alternate-reverse", "normal", "reverse"),
  duration = NULL,
  easing = NULL,
  fill = c("none", "auto", "backwards", "both", "forwards"),
  iterations = NULL
)
}
\arguments{
\item{delay}{Animation delay time in milliseconds. The time to wait before the
animation begins. Must be a non-negative numeric value.}

\item{direction}{Animation playback direction. Options:
\itemize{
\item \code{"forward"}: Plays normally (default)
\item \code{"alternate"}: Plays forward, then in reverse
\item \code{"alternate-reverse"}: Plays in reverse, then forward
\item \code{"normal"}: Same as forward
\item \code{"reverse"}: Plays in reverse direction
}}

\item{duration}{Animation duration in milliseconds. The length of time the
animation will take to complete one cycle. Must be a non-negative numeric value.}

\item{easing}{Animation easing function. Controls the rate of change during
the animation. Common values include "linear", "ease", "ease-in", "ease-out",
"ease-in-out", or cubic-bezier values.}

\item{fill}{Fill mode after animation ends. Options:
\itemize{
\item \code{"none"}: Element returns to its initial state when animation ends (default)
\item \code{"auto"}: Follows the rules of the animation effect
\item \code{"backwards"}: Element retains first keyframe values during delay period
\item \code{"both"}: Combines forwards and backwards behavior
\item \code{"forwards"}: Element retains final keyframe values after animation ends
}}

\item{iterations}{Number of times the animation should repeat. A value of
\code{Inf} will cause the animation to repeat indefinitely. Must be a
non-negative numeric value.}
}
\value{
A list containing animation configuration that can be passed to \code{\link[=g6_options]{g6_options()}}.
}
\description{
Configures animation settings for G6 graph elements. These settings control
how graph elements animate when changes occur.
}
\details{
Animation configuration allows fine-tuning the timing and behavior of animations
in G6 graphs. This includes controlling the duration, delay, easing function,
direction, and other aspects of how graph elements animate.
}
\examples{
# Basic animation with duration
config <- animation_config(
  duration = 500
)

# Complex animation configuration
config <- animation_config(
  delay = 100,
  duration = 800,
  easing = "ease-in-out",
  direction = "alternate",
  fill = "forwards",
  iterations = 2
)

# Infinite animation
config <- animation_config(
  duration = 1000,
  easing = "linear",
  iterations = Inf
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{context_menu}
\alias{context_menu}
\title{Configure Context Menu Behavior}
\usage{
context_menu(
  key = "contextmenu",
  className = "g6-contextmenu",
  trigger = "contextmenu",
  offset = c(4, 4),
  onClick = NULL,
  getItems = NULL,
  getContent = NULL,
  loadingContent = NULL,
  enable = JS("(e) => e.targetType === 'node'"),
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior, used for
subsequent operations (string, default: "context-menu").}

\item{className}{Additional class name for the menu DOM (string, default: "g6-contextmenu").}

\item{trigger}{How to trigger the menu: "contextmenu" for right-click,
"click" for click (string, default: "contextmenu").}

\item{offset}{Offset of the menu display in X and Y directions (numeric vector, default: c(4, 4)).}

\item{onClick}{Callback method triggered after menu item is clicked (JS function). Our default allows
to create edge or either remove the current node.}

\item{getItems}{Returns the list of menu items, supports Promise (JS function, default: NULL).}

\item{getContent}{Returns the content of the menu, supports Promise (JS function, default: NULL).}

\item{loadingContent}{Menu content used when getContent returns a
Promise (string or HTML element, default: NULL).}

\item{enable}{Whether the context menu is available (boolean or JS function, default: TRUE).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/contextmenu}.}
}
\value{
A list with the configuration settings for the context menu plugin.
}
\description{
Creates a configuration object for the context-menu behavior in G6.
This allows users to display a context menu when right-clicking or clicking on graph elements.
}
\examples{
# Basic configuration
config <- context_menu()

# Custom configuration with JavaScript functions
config <- context_menu(
  key = "my-context-menu",
  className = "my-context-menu",
  trigger = "click",
  offset = c(10, 10),
  getItems = JS("(event) => {
    const type = event.itemType;
    const isNode = type === 'node';
    return [
      { key: 'delete', text: 'Delete' },
      { key: 'edit', text: 'Edit' },
      { key: 'details', text: 'View Details', disabled: !isNode }
    ];
  }"),
  onClick = JS("(value, target, current) => {
    if (value === 'delete') {
      // do stuff
  }")
)
if (interactive()) {
  library(shiny)
  library(g6R)
  library(bslib)

  nodes <- data.frame(id = c("node1", "node2"))
  edges <- data.frame(source = "node1", target = "node2")

  ui <- page_fluid(
    g6_output("graph"),
    verbatimTextOutput("contextmenu_info")
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(
        nodes = nodes,
        edges = edges
      ) |>
        g6_layout() |>
        g6_plugins(
          context_menu()
        )
    })

    output$contextmenu_info <- renderPrint({
      input[["graph-contextmenu"]]
    })
  }

  shinyApp(ui, server)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{fullscreen}
\alias{fullscreen}
\title{Configure Fullscreen Plugin}
\usage{
fullscreen(
  key = "fullscreen",
  autoFit = TRUE,
  trigger = NULL,
  onEnter = NULL,
  onExit = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{autoFit}{Whether to auto-fit the canvas size to the screen when in
fullscreen mode (boolean, default: TRUE).}

\item{trigger}{Methods to trigger fullscreen, e.g.,
list(request = "button", exit = "escape") (list, default: NULL).}

\item{onEnter}{Callback function after entering fullscreen mode (JS function, default: NULL).}

\item{onExit}{Callback function after exiting fullscreen mode (JS function, default: NULL).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/fullscreen}.}
}
\value{
A list with the configuration settings for the fullscreen plugin.
}
\description{
Creates a configuration object for the fullscreen plugin in G6.
This plugin enables fullscreen mode for the graph.
}
\examples{
# Basic configuration
config <- fullscreen()

# Custom configuration
config <- fullscreen(
  key = "my-fullscreen",
  autoFit = TRUE,
  trigger = list(
    request = "F",
    exit = "Esc"
  ),
  onEnter = JS("() => {
    console.log('Entered fullscreen mode');
  }"),
  onExit = JS("() => {
    console.log('Exited fullscreen mode');
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{is_g6_port}
\alias{is_g6_port}
\alias{is_g6_ports}
\title{Check if an object is a g6_port}
\usage{
is_g6_port(x)

is_g6_ports(x)
}
\arguments{
\item{x}{An object to check.}
}
\value{
Logical indicating if \code{x} is of class \code{g6_port}.
}
\description{
Check if an object is a g6_port
}
\examples{
p <- g6_port("input-1", type = "input")
is_g6_port(p)
is_g6_port(list(key = "input-1", type = "input"))
is_g6_port(as_g6_port(list(key = "input-1", type = "input")))
is_g6_ports(g6_ports(
  g6_port("input-1", type = "input"),
  g6_port("output-1", type = "output")
))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{legend}
\alias{legend}
\title{Configure Legend Plugin}
\usage{
legend(
  key = "legend",
  trigger = c("hover", "click"),
  position = c("bottom", "top", "left", "right", "top-left", "top-right", "bottom-left",
    "bottom-right"),
  container = NULL,
  className = NULL,
  containerStyle = NULL,
  nodeField = NULL,
  edgeField = NULL,
  comboField = NULL,
  orientation = c("horizontal", "vertical"),
  layout = c("flex", "grid"),
  showTitle = FALSE,
  titleText = "",
  x = NULL,
  y = NULL,
  width = 240,
  height = 160,
  itemSpacing = 4,
  rowPadding = 10,
  colPadding = 10,
  itemMarkerSize = 16,
  itemLabelFontSize = 16,
  gridCol = NULL,
  gridRow = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{trigger}{How legend items trigger highlighting: "hover" or "click" (string, default: "hover").}

\item{position}{Relative position of the legend on the canvas (string, default: "bottom").}

\item{container}{Container to which the legend is mounted (HTML element or string, default: NULL).}

\item{className}{Legend canvas class name (string, default: NULL).}

\item{containerStyle}{Style of the legend container (list or JS object, default: NULL).}

\item{nodeField}{Node classification identifier (string or JS function, default: NULL).}

\item{edgeField}{Edge classification identifier (string or JS function, default: NULL).}

\item{comboField}{Combo classification identifier (string or JS function, default: NULL).}

\item{orientation}{Layout direction: "horizontal" or "vertical" (string, default: "horizontal").}

\item{layout}{Layout method: "flex" or "grid" (string, default: "flex").}

\item{showTitle}{Whether to display the title (boolean, default: FALSE).}

\item{titleText}{Title content (string, default: "").}

\item{x}{Relative horizontal position (number, default: NULL).}

\item{y}{Relative vertical position (number, default: NULL).}

\item{width}{Width of the legend (number, default: 240).}

\item{height}{Height of the legend (number, default: 160).}

\item{itemSpacing}{Spacing between text and marker (number, default: 4).}

\item{rowPadding}{Spacing between rows (number, default: 10).}

\item{colPadding}{Spacing between columns (number, default: 10).}

\item{itemMarkerSize}{Size of the legend item marker (number, default: 16).}

\item{itemLabelFontSize}{Font size of the legend item text (number, default: 16).}

\item{gridCol}{Maximum number of columns for grid layout (number, default: NULL).}

\item{gridRow}{Maximum number of rows for grid layout (number, default: NULL).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/legend}.}
}
\value{
A list with the configuration settings for the legend plugin.
}
\description{
Creates a configuration object for the legend plugin in G6.
This plugin adds a legend to the graph, allowing users to identify and interact with
different categories of elements.
}
\examples{
# Basic configuration for node categories
config <- legend(
  nodeField = "category"
)

# Advanced configuration
config <- legend(
  key = "my-legend",
  position = "top-right",
  nodeField = "type",
  edgeField = "relation",
  orientation = "vertical",
  layout = "grid",
  showTitle = TRUE,
  titleText = "Graph Elements",
  width = 300,
  height = 200,
  gridCol = 2,
  containerStyle = list(
    background = "#f9f9f9",
    border = "1px solid #ddd",
    borderRadius = "4px",
    padding = "8px"
  )
)

# Using a function for classification
config <- legend(
  nodeField = JS("(item) => {
    return item.data.importance > 0.5 ? 'Important' : 'Regular';
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{validate_port}
\alias{validate_port}
\alias{validate_port.g6_port}
\title{Validate a single G6 port}
\usage{
validate_port(x, ...)

\method{validate_port}{g6_port}(x, ...)
}
\arguments{
\item{x}{An object of class 'g6_port'.}

\item{...}{Generic consistency.}
}
\value{
The validated port (invisibly).
}
\description{
Validate a single G6 port
}
\examples{
validate_port(g6_port("input-1", type = "input"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{poke}
\alias{poke}
\title{Example pokemon data}
\format{
A list with 3 nested lists:
\describe{
\item{nodes}{list with 120 pokemon.}
\item{edges}{list with 69 connections.}
\item{combo}{list with 51 pokemon families.
A family contains all evolutions of a pokemon.}
}
}
\source{
\url{https://pokeapi.co/docs/v2}
}
\usage{
data(poke)
}
\description{
Pokemon evolution network to showcase combo features.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{hover_activate}
\alias{hover_activate}
\title{Configure Hover Activate Behavior}
\usage{
hover_activate(
  key = "hover-activate",
  animation = TRUE,
  enable = TRUE,
  degree = 0,
  direction = c("both", "in", "out"),
  state = "active",
  inactiveState = NULL,
  onHover = NULL,
  onHoverEnd = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior (string, default: "hover-activate").}

\item{animation}{Whether to enable animation (boolean, default: TRUE).}

\item{enable}{Whether to enable hover feature (boolean or JS function, default: TRUE).}

\item{degree}{Degree of relationship to activate elements (number or JS function, default: 0).}

\item{direction}{Specify edge direction: "both", "in", or "out" (string, default: "both").}

\item{state}{State of activated elements (string, default: "active").}

\item{inactiveState}{State of inactive elements (string, default: NULL).}

\item{onHover}{Callback when element is hovered (JS function, default: NULL).}

\item{onHoverEnd}{Callback when hover ends (JS function, default: NULL).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/hover-activate}.}
}
\value{
A list with the configuration settings for the hover-activate behavior.
}
\description{
Creates a configuration object for the hover-activate behavior in G6.
This behavior activates elements when the mouse hovers over them.
}
\examples{
# Basic configuration
config <- hover_activate()

# Custom configuration
config <- hover_activate(
  key = "my-hover-behavior",
  animation = FALSE,
  degree = 1,
  direction = "out",
  state = "highlight",
  inactiveState = "inactive",
  onHover = JS("(event) => { console.log('Hover on:', event.target.id); }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{concentric_layout}
\alias{concentric_layout}
\title{Generate G6 AntV Concentric layout configuration}
\usage{
concentric_layout(
  center = NULL,
  clockwise = FALSE,
  equidistant = FALSE,
  width = NULL,
  height = NULL,
  sortBy = "degree",
  maxLevelDiff = NULL,
  nodeSize = 30,
  nodeSpacing = 10,
  preventOverlap = FALSE,
  startAngle = 3/2 * pi,
  sweep = NULL,
  ...
)
}
\arguments{
\item{center}{The center position of the circular layout \code{[x, y]} or \code{[x, y, z]}.
By default, it's the center position of the current container.}

\item{clockwise}{Whether nodes are arranged in clockwise order.}

\item{equidistant}{Whether the distances between rings are equal.}

\item{width}{The width of the layout. By default, the container width is used.}

\item{height}{The height of the layout. By default, the container height is used.}

\item{sortBy}{Specify the sorting basis (node attribute name). The higher the value,
the more central the node will be placed. If it is "degree", the degree of the node
will be calculated. The higher the degree, the more central the node will be placed.}

\item{maxLevelDiff}{If the maximum attribute difference of nodes in the same layer
is undefined, it will be set to maxValue / 4, where maxValue is the maximum
attribute value of the sorting basis. For example, if sortBy is 'degree', then
maxValue is the degree of the node with the largest degree among all nodes.}

\item{nodeSize}{Node size (diameter). Used to prevent collision detection when nodes overlap.
Can be a number, a 2-element vector \code{[width, height]}, or a function that returns a number.}

\item{nodeSpacing}{Minimum distance between rings, used to adjust the radius.
Can be a number, a vector of numbers, or a function that returns a number.}

\item{preventOverlap}{Whether to prevent node overlap. Must be coordinated with the nodeSize
attribute or the data.size attribute in the node data. Only when data.size is set in the
data or the nodeSize value that is the same as the current graph node size is configured
in the layout, can node overlap collision detection be performed.}

\item{startAngle}{The arc at which to start layout of nodes (in radians)}

\item{sweep}{If the radian difference between the first and last nodes in the same layer
is undefined, it will be set to \code{2 * Math.PI * (1 - 1 / |level.nodes|)}, where level.nodes
is the number of nodes in each layer calculated by the algorithm, and |level.nodes| is
the number of nodes in the layer.}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/concentric-layout}.}
}
\value{
A list containing the configuration for G6 AntV Concentric layout
}
\description{
This function creates a configuration list for G6 AntV Concentric layout
with all available options as parameters. The Concentric layout places nodes in
concentric circles based on a centrality measure.
}
\examples{
# Basic concentric layout
concentric_config <- concentric_layout()

# Custom concentric layout with degree-based sorting and overlap prevention
concentric_config <- concentric_layout(
  clockwise = TRUE,
  sortBy = "degree",
  preventOverlap = TRUE,
  nodeSize = 30,
  nodeSpacing = 20
)

# Custom concentric layout with specific center and dimensions
concentric_config <- concentric_layout(
  center = c(300, 300),
  width = 600,
  height = 600,
  equidistant = TRUE,
  startAngle = pi
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_get_ports}
\alias{g6_get_ports}
\alias{g6_get_input_ports}
\alias{g6_get_output_ports}
\title{Extract ports from a g6 graph state via proxy}
\usage{
g6_get_ports(graph)

g6_get_input_ports(graph)

g6_get_output_ports(graph)
}
\arguments{
\item{graph}{A g6_proxy object.}
}
\value{
A named list of ports for each node, optionally filtered by type.
}
\description{
These helpers extract ports from the graph state stored in Shiny input,
accessed via a g6_proxy object. Ports are grouped by node and can be filtered by type.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{brush_select}
\alias{brush_select}
\title{Configure Brush Selection Interaction}
\usage{
brush_select(
  key = "brush-select",
  animation = FALSE,
  enable = JS("(e) => {\\n      return true;\\n    }"),
  enableElements = "node",
  immediately = FALSE,
  mode = c("default", "union", "intersect", "diff"),
  onSelect = NULL,
  state = c("selected", "active", "inactive", "disabled", "highlight"),
  style = NULL,
  trigger = "shift",
  outputId = NULL,
  ...
)
}
\arguments{
\item{key}{Behavior unique identifier. Useful to modify this behavior from JS side.}

\item{animation}{Whether to enable animation (boolean, default: FALSE).}

\item{enable}{Whether to enable brush select functionality (boolean or function, default: TRUE).}

\item{enableElements}{Types of elements that can be selected (character vector, default: "node").
Can be \code{c("node", "edge", "combo")}.}

\item{immediately}{Whether to select immediately in default mode (boolean, default: FALSE).}

\item{mode}{Selection mode: "union", "intersect", "diff", or "default" (string, default: "default").}

\item{onSelect}{Callback for selected element state (JS function).}

\item{state}{State to switch to when selected (string, default: "selected").}

\item{style}{Style specification for the selection box (list).
See \url{https://g6.antv.antgroup.com/en/manual/behavior/brush-select#style}.}

\item{trigger}{Shortcut keys for selection (character vector).}

\item{outputId}{Manually pass the Shiny output ID. This is useful when the graph
is initialised outside the shiny render function and the ID cannot be automatically
inferred. This allows to set input values from the callback function
with the right namespace and graph ID. You must typically pass \code{session$ns("graphid")}
to ensure this also works in modules.}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/brush-select}.}
}
\value{
A list with the configuration settings for the brush select behavior.
}
\description{
Creates a configuration object for brush selection interaction in graph visualizations.
This function configures how elements are selected when using a brush selection tool.
}
\examples{
# Basic configuration
config <- brush_select()

# Custom configuration
config <- brush_select(
  animation = TRUE,
  enableElements = c("node", "edge"),
  mode = "union",
  state = "highlight",
  style = list(
    fill = "rgba(0, 0, 255, 0.1)",
    stroke = "blue",
    lineWidth = 2
  ),
  trigger = c("Shift")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{minimap}
\alias{minimap}
\title{Configure Minimap Plugin}
\usage{
minimap(
  key = "minimap",
  className = NULL,
  container = NULL,
  containerStyle = NULL,
  delay = 128,
  filter = NULL,
  maskStyle = NULL,
  padding = 10,
  position = "right-bottom",
  renderer = NULL,
  shape = "key",
  size = c(240, 160),
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{className}{Class name of the thumbnail canvas (string, default: NULL).}

\item{container}{Container to which the thumbnail is mounted (HTML element or string, default: NULL).}

\item{containerStyle}{Style of the thumbnail container (list or JS object, default: NULL).}

\item{delay}{Delay update time in milliseconds for performance optimization (number, default: 128).}

\item{filter}{Function to filter elements to display in minimap (JS function, default: NULL).}

\item{maskStyle}{Style of the mask (list or JS object, default: NULL).}

\item{padding}{Padding around the minimap (number or numeric vector, default: 10).}

\item{position}{Position of the thumbnail relative to the canvas
(string or numeric vector, default: "right-bottom").}

\item{renderer}{Custom renderer (JS object, default: NULL).}

\item{shape}{Method for generating element thumbnails (string or JS function, default: "key").}

\item{size}{Width and height of the minimap \code{[width, height]} (numeric vector, default: c(240, 160)).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/minimap}.}
}
\value{
A list with the configuration settings for the minimap plugin.
}
\description{
Creates a configuration object for the minimap plugin in G6.
This plugin adds a minimap/thumbnail view of the entire graph.
}
\examples{
# Basic configuration
config <- minimap()

# Custom configuration
config <- minimap(
  key = "my-minimap",
  position = "left-top",
  size = c(200, 150),
  padding = 15,
  containerStyle = list(
    border = "1px solid #ddd",
    borderRadius = "4px",
    boxShadow = "0 0 8px rgba(0,0,0,0.1)"
  ),
  maskStyle = list(
    stroke = "#1890ff",
    strokeWidth = 2,
    fill = "rgba(24, 144, 255, 0.1)"
  )
)

# With custom filtering function
config <- minimap(
  filter = JS("(id, elementType) => {
    // Only show nodes and important edges in the minimap
    if (elementType === 'node') return true;
    if (elementType === 'edge') {
      // Assuming edges have an 'important' attribute
      const edge = graph.findById(id);
      return edge.getModel().important === true;
    }
    return false;
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_collapse_combo}
\alias{g6_collapse_combo}
\alias{g6_expand_combo}
\title{Collapse or expand a combo element in a g6 graph}
\usage{
g6_collapse_combo(graph, id, options = NULL)

g6_expand_combo(graph, id, options = NULL)
}
\arguments{
\item{graph}{A g6 graph object or a g6_proxy object for Shiny applications.}

\item{id}{Character string specifying the ID of the combo to collapse/expand.}

\item{options}{List containing optional configuration parameters for the collapse/expand action:
\itemize{
\item \code{animate}: Logical value indicating whether to animate the collapsing process. Default is \code{TRUE}.
\item \code{align}: Logical value to ensure the position of expanded/collapsed nodes remains unchanged.
}}
}
\value{
The modified g6 graph or g6_proxy object, allowing for method chaining.
}
\description{
This function collapses/expands a specified combo (a group of nodes) in a g6 graph,
hiding its member nodes and edges while maintaining the combo itself visible.
This is useful for simplifying complex graphs with multiple hierarchical groups.
}
\details{
When a combo is collapsed, its member nodes and edges are hidden from view
while the combo itself remains visible, typically shown as a single node.
This helps to reduce visual complexity in large graphs with hierarchical groupings.
}
\references{
\url{https://g6.antv.antgroup.com/en/api/element#graphcollapseelementid-options},
\url{https://g6.antv.antgroup.com/en/api/element#graphexpandelementid-options}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{g6_layout}
\alias{g6_layout}
\title{Set the layout algorithm for a g6 graph}
\usage{
g6_layout(graph, layout = d3_force_layout())
}
\arguments{
\item{graph}{A g6 graph object created with \code{g6()}.}

\item{layout}{An existing layout function like \link{circular_layout} or a string like \code{circular-layout}.
At minimum, this can be a list that should contain a \code{type} element specifying the layout algorithm.
Additional parameters depend on the layout type chosen, for instance \code{list(type = "force")}.}
}
\value{
The modified g6 graph object with the specified layout, allowing for method chaining.
}
\description{
This function configures the layout algorithm used to position nodes in a g6 graph.
}
\details{
G6 provides several layout algorithms, each suitable for different graph structures:

\itemize{
\item \strong{force}: Force-directed layout using physical simulation of forces.
\item \strong{random}: Random layout placing nodes randomly.
\item \strong{circular}: Arranges nodes on a circle.
\item \strong{radial}: Radial layout with nodes arranged outward from a central node.
\item \strong{grid}: Arranges nodes in a grid pattern.
\item \strong{concentric}: Concentric circles with important nodes in the center.
\item \strong{dagre}: Hierarchical layout for directed acyclic graphs.
\item \strong{fruchterman}: Force-directed layout based on the Fruchterman-Reingold algorithm.
\item \strong{mds}: Multidimensional scaling layout.
\item \strong{comboForce}: Force-directed layout specially designed for combo graphs.
}

Each layout algorithm has specific configuration options. See the G6 documentation
for detailed information on each layout and its parameters:
\url{https://g6.antv.antgroup.com/en/manual/layout/overview}.
}
\seealso{
\code{\link[=g6]{g6()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{snapline}
\alias{snapline}
\title{Configure Snapline Plugin}
\usage{
snapline(
  key = "snapline",
  tolerance = 5,
  offset = 20,
  autoSnap = TRUE,
  shape = "key",
  verticalLineStyle = list(stroke = "#1783FF"),
  horizontalLineStyle = list(stroke = "#1783FF"),
  filter = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{tolerance}{The alignment accuracy in pixels (number, default: 5).}

\item{offset}{The extension distance of the snapline (number, default: 20).}

\item{autoSnap}{Whether to enable automatic snapping (boolean, default: TRUE).}

\item{shape}{Specifies which shape to use as reference: "key" or a function
(string or JS function, default: "key").}

\item{verticalLineStyle}{Vertical snapline style (list or JS object, default: list(stroke = "#1783FF")).}

\item{horizontalLineStyle}{Horizontal snapline style (list or JS object, default: list(stroke = "#1783FF")).}

\item{filter}{Function to filter nodes that don't participate in alignment (JS function, default: NULL).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/plugin/snapline}.}
}
\value{
A list with the configuration settings for the snapline plugin.
}
\description{
Creates a configuration object for the snapline plugin in G6.
This plugin provides alignment guidelines when moving nodes.
}
\examples{
# Basic configuration
config <- snapline()

# Custom configuration
config <- snapline(
  key = "my-snapline",
  tolerance = 8,
  offset = 30,
  verticalLineStyle = list(
    stroke = "#f00",
    strokeWidth = 1.5,
    lineDash = c(5, 2)
  ),
  horizontalLineStyle = list(
    stroke = "#00f",
    strokeWidth = 1.5,
    lineDash = c(5, 2)
  )
)

# With custom filter function
config <- snapline(
  filter = JS("(node) => {
    // Only allow regular nodes to participate in alignment
    // Exclude special nodes like 'start' or 'end'
    const model = node.getModel();
    return model.type !== 'start' && model.type !== 'end';
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{auto_adapt_label}
\alias{auto_adapt_label}
\title{Configure Auto Adapt Label Behavior}
\usage{
auto_adapt_label(
  key = "auto-adapt-label",
  enable = TRUE,
  throttle = 100,
  padding = 0,
  sort = NULL,
  sortNode = list(type = "degree"),
  sortEdge = NULL,
  sortCombo = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior (string, default: "auto-adapt-label").}

\item{enable}{Whether to enable this behavior (JS function, default: returns TRUE for all events).}

\item{throttle}{Throttle time in milliseconds to optimize performance (numeric, default: 100).}

\item{padding}{Padding space around labels in pixels (numeric, default: 0).}

\item{sort}{Global sorting rule for all element types (list or JS function, default: NULL).}

\item{sortNode}{Sorting rule specifically for node labels (list, default: list(type = "degree")).}

\item{sortEdge}{Sorting rule specifically for edge labels (list, default: NULL).}

\item{sortCombo}{Sorting rule specifically for combo labels (list, default: NULL).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/auto-adapt-label}.

Sorting parameters determine which labels take priority when space is limited:
\itemize{
\item When \code{sort} is provided, it applies to all element types and overrides type-specific settings
\item Type-specific sorting (\code{sortNode}, \code{sortEdge}, \code{sortCombo}) only applies when \code{sort} is NULL
\item The default sorting for nodes is by degree (higher degree nodes' labels are shown first)
}}
}
\value{
A list with the configuration settings for the auto-adapt-label behavior.
}
\description{
Creates a configuration object for the auto-adapt-label behavior in G6.
This behavior automatically adjusts label positions to reduce overlapping and
improve readability in the graph visualization.
}
\examples{
# Basic configuration with defaults
config <- auto_adapt_label()

# Custom configuration with more padding and custom throttle
config <- auto_adapt_label(
  key = "my-label-adapter",
  throttle = 200,
  padding = 5
)


# Using a custom enable function
config <- auto_adapt_label(
  enable = JS("(e) => e.targetType === 'node'")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{g6_behaviors}
\alias{g6_behaviors}
\title{Create G6 Graph Behaviors Configuration}
\usage{
g6_behaviors(graph, ...)
}
\arguments{
\item{graph}{A g6 graph instance.}

\item{...}{Behavior configuration objects created by behavior-specific functions.
These can include any of the following behaviors:

\strong{Navigation behaviors:}
\itemize{
\item \code{drag_canvas()} - Drag the entire canvas view
\item \code{zoom_canvas()} - Zoom the canvas view
\item \code{scroll_canvas()} - Scroll the canvas using the wheel
\item \code{optimize_viewport_transform()} - Optimize view transform performance
}

\strong{Selection behaviors:}
\itemize{
\item \code{click_select()} - Click to select graph elements
\item \code{brush_select()} - Select elements by dragging a rectangular area
\item \code{lasso_select()} - Freely draw an area to select elements
}

\strong{Editing behaviors:}
\itemize{
\item \code{create_edge()} - Interactively create new edges
\item \code{drag_element()} - Drag nodes or combos
\item \code{drag_element_force()} - Drag nodes in force-directed layout
}

\strong{Data Exploration behaviors:}
\itemize{
\item \code{collapse_expand()} - Expand or collapse subtree nodes
\item \code{focus_element()} - Focus on specific elements and automatically adjust the view
\item \code{hover_activate()} - Highlight elements when hovering
}

\strong{Visual Optimization behaviors:}
\itemize{
\item \code{fix_element_size()} - Fix the element size to a specified value
\item \code{auto_adapt_label()} - Automatically adjust label position
}}
}
\value{
A list of behavior configuration objects that can be passed to G6 graph initialization
}
\description{
Configures interaction behaviors for a G6 graph visualization.
This function collects and combines multiple behavior configurations
into a list that can be passed to graph initialization functions.
}
\note{
You can create custom behaviors from JavaScript and use them on the R side. See more
at \url{https://g6.antv.antgroup.com/en/manual/behavior/custom-behavior}.
}
\examples{
# Create a basic set of behaviors
behaviors <- g6_behaviors(
  g6(),
  drag_canvas(),
  zoom_canvas(),
  click_select()
)

# Create a more customized set of behaviors
behaviors <- g6_behaviors(
  g6(),
  drag_canvas(),
  zoom_canvas(sensitivity = 1.5),
  hover_activate(state = "highlight"),
  fix_element_size(
    node = list(
      list(shape = "circle", fields = c("r", "lineWidth"))
    )
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layouts.R
\name{fruchterman_layout}
\alias{fruchterman_layout}
\title{Generate G6 Fruchterman layout configuration}
\usage{
fruchterman_layout(height = NULL, width = NULL, gravity = 10, speed = 5, ...)
}
\arguments{
\item{height}{Numeric. Layout height. Defaults to container height.}

\item{width}{Numeric. Layout width. Defaults to container width.}

\item{gravity}{Numeric. Central force attracting nodes to the center.
Larger values make the layout more compact. Defaults to 10.}

\item{speed}{Numeric. Node movement speed per iteration. Higher values may cause oscillation. Defaults to 5.}

\item{...}{Additional parameters to pass to the layout.
See \url{https://g6.antv.antgroup.com/en/manual/layout/fruchterman-layout}.}
}
\value{
A list containing the configuration for G6 fruchterman layout.
}
\description{
This function creates a configuration list for G6 Fruchterman layout
with all available options as parameters.
}
\examples{
if (interactive()) {
  g6(lesmis$nodes, lesmis$edges) |>
   g6_layout(fruchterman_layout(
     gravity = 5,
     speed = 5
   )) |>
   g6_behaviors(
     "zoom-canvas",
     drag_element()
   )
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_canvas_resize}
\alias{g6_canvas_resize}
\title{Resize the canvas of a g6 graph via proxy}
\usage{
g6_canvas_resize(graph, width, height)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{width}{Numeric value specifying the new width of the canvas in pixels.}

\item{height}{Numeric value specifying the new height of the canvas in pixels.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function changes the size of the canvas of an existing g6 graph instance
using a proxy object. This allows updating the graph dimensions without completely
re-rendering it.
}
\details{
This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

See \url{https://g6.antv.antgroup.com/en/api/canvas#graphsetsizewidth-height} for more details.
}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{focus_element}
\alias{focus_element}
\title{Configure Focus Element Behavior}
\usage{
focus_element(
  key = "focus-element",
  animation = list(duration = 500, easing = "ease-in"),
  enable = TRUE,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior, used for subsequent operations
(string, default: "focus-element").}

\item{animation}{Focus animation settings (list, default: list with duration 500ms and easing "ease-in").}

\item{enable}{Whether to enable the focus feature (boolean or JS function, default: TRUE).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/focus-element}.}
}
\value{
A list with the configuration settings for the focus-element behavior.
}
\description{
Creates a configuration object for the focus-element behavior in G6.
This behavior allows focusing on specific elements by automatically adjusting the viewport.
}
\examples{
# Basic configuration
config <- focus_element()

# Custom configuration
config <- focus_element(
  key = "my-focus-behavior",
  animation = list(duration = 1000, easing = "ease-out"),
  enable = JS("(event) => event.targetType === 'node'")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{radial}
\alias{radial}
\title{Example Network for radial layouts}
\format{
A list with 2 data frames:
\describe{
\item{nodes}{data frame with 34 rows for the nodes.}
\item{edges}{data frame with 58 rows for the edges.}
}
}
\source{
\url{https://assets.antv.antgroup.com/g6/radial.json}
}
\usage{
data(radial)
}
\description{
Example Network for radial layouts
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{zoom_canvas}
\alias{zoom_canvas}
\title{Configure Zoom Canvas Behavior}
\usage{
zoom_canvas(
  key = "zoom-canvas",
  animation = list(duration = 200),
  enable = TRUE,
  origin = NULL,
  onFinish = NULL,
  preventDefault = TRUE,
  sensitivity = 1,
  trigger = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior (string, default: "zoom-canvas").}

\item{animation}{Zoom animation effect settings (list, default: list with duration 200ms).}

\item{enable}{Whether to enable this behavior (boolean or JS function, default: TRUE).}

\item{origin}{Zoom center point in viewport coordinates (list with x, y values, default: NULL).}

\item{onFinish}{Callback function when zooming is finished (JS function, default: NULL).}

\item{preventDefault}{Whether to prevent the browser's default event (boolean, default: TRUE).}

\item{sensitivity}{Zoom sensitivity, the larger the value, the faster the zoom (numeric, default: 1).}

\item{trigger}{How to trigger zooming, supports mouse wheel and keyboard shortcuts (list, default: NULL).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/zoom-canvas}.}
}
\value{
A list with the configuration settings for the zoom-canvas behavior.
}
\description{
Creates a configuration object for the zoom-canvas behavior in G6.
This behavior allows zooming the canvas with mouse wheel or keyboard shortcuts.
}
\examples{
# Basic configuration
config <- zoom_canvas()

# Custom configuration
config <- zoom_canvas(
  key = "my-zoom-behavior",
  animation = list(duration = 300, easing = "ease-in-out"),
  origin = list(x = 0, y = 0),
  sensitivity = 1.5,
  preventDefault = FALSE
)

# With keyboard triggers and callback
config <- zoom_canvas(
  enable = JS("(event) => !event.altKey"),
  trigger = list(
    zoomIn = "+",
    zoomOut = "-",
    reset = "0"
  ),
  onFinish = JS("() => { console.log('Zooming finished'); }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{g6_plugins}
\alias{g6_plugins}
\title{Create a List of G6 Plugins}
\usage{
g6_plugins(graph, ...)
}
\arguments{
\item{graph}{G6 graph instance.}

\item{...}{G6 plugin configuration objects created with plugin-specific functions}
}
\value{
A list of G6 plugin configurations that can be passed to a G6 graph.
}
\description{
Combines multiple G6 plugins into a list that can be passed to a G6 graph configuration.
G6 plugins extend the functionality of the base graph visualization with additional features.
}
\details{
G6 plugins provide extended functionality beyond the core graph visualization capabilities.
Plugins are divided into several categories:

\subsection{Visual Style Enhancement}{
\itemize{
\item \strong{Grid Line (grid-line):} Displays grid reference lines on the canvas
\item \strong{Background (background):} Adds background images or colors to the canvas
\item \strong{Watermark (watermark):} Adds a watermark to the canvas to protect copyright
\item \strong{Hull (hull):} Creates an outline for a specified set of nodes
\item \strong{Bubble Sets (bubble-sets):} Creates smooth bubble-like element outlines
\item \strong{Snapline (snapline):} Displays alignment reference lines when dragging elements
}
}

\subsection{Navigation and Overview}{
\itemize{
\item \strong{Minimap (minimap):} Displays a thumbnail preview of the graph, supporting navigation
\item \strong{Fullscreen (fullscreen):} Supports full-screen display and exit for charts
\item \strong{Timebar (timebar):} Provides filtering and playback control for temporal data
}
}

\subsection{Interactive Controls}{
\itemize{
\item \strong{Toolbar (toolbar):} Provides a collection of common operation buttons
\item \strong{Context Menu (contextmenu):} Displays a menu of selectable operations on right-click
\item \strong{Tooltip (tooltip):} Displays detailed information about elements on hover
\item \strong{Legend (legend):} Displays categories and corresponding style descriptions of chart data
}
}

\subsection{Data Exploration}{
\itemize{
\item \strong{Fisheye (fisheye):} Provides a focus + context exploration experience
\item \strong{Edge Filter Lens (edge-filter-lens):} Filters and displays edges within a specified area
\item \strong{Edge Bundling (edge-bundling):} Bundles edges with similar paths together to reduce visual clutter
}
}

\subsection{Advanced Features}{
\itemize{
\item \strong{History (history):} Supports undo/redo operations
\item \strong{Camera Setting (camera-setting):} Configures camera parameters in a 3D scene
}
}
}
\note{
You can also build your own plugins as described at
\url{https://g6.antv.antgroup.com/en/manual/plugin/custom-plugin}.
}
\examples{
# Create a configuration with multiple plugins
plugins <- g6_plugins(
  g6(),
  minimap(),
  grid_line(),
  tooltips(
    getContent = JS("(e, items) => {
      return `<div>${items[0].id}</div>`;
    }")
  )
)

# Add a context menu and toolbar
plugins <- g6_plugins(
  g6(),
  context_menu(
    key = "my-context-menu",
    className = "my-context-menu",
    trigger = "click",
    offset = c(10, 10),
    getItems = JS("(event) => {
      const type = event.itemType;
      const isNode = type === 'node';
      return [
        { key: 'delete', text: 'Delete' },
        { key: 'edit', text: 'Edit' },
        { key: 'details', text: 'View Details', disabled: !isNode }
      ];
    }"),
    onClick = JS("(value, target, current) => {
      if (value === 'delete') {
        // do stuff
    }")
  ),
  g6R::toolbar(
    position = "top-right",
    getItems = JS("() => [
      { id: 'zoom-in', value: 'zoom-in' },
      { id: 'zoom-out', value: 'zoom-out' },
      { id: 'fit', value: 'fit' }
    ]"),
    onClick = JS("(value) => {
      if (value === 'zoom-in') graph.zoomTo(1.1);
      else if (value === 'zoom-out') graph.zoomTo(0.9);
      else if (value === 'fit') graph.fitView();
    }")
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{is_g6_nodes}
\alias{is_g6_nodes}
\alias{is_g6_edges}
\alias{is_g6_combos}
\title{Check if an object is a list of G6 elements}
\usage{
is_g6_nodes(x)

is_g6_edges(x)

is_g6_combos(x)
}
\arguments{
\item{x}{An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.}
}
\description{
Check if an object is a list of G6 elements
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{drag_element_force}
\alias{drag_element_force}
\title{Configure Drag Element Force Behavior}
\usage{
drag_element_force(
  key = "drag-element-force",
  fixed = FALSE,
  enable = NULL,
  state = "selected",
  hideEdge = c("none", "out", "in", "both", "all"),
  cursor = NULL,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the behavior, used for subsequent operations (string, default: "drag-element-force").}

\item{fixed}{Whether to keep the node position fixed after dragging ends (boolean, default: FALSE).}

\item{enable}{Whether to enable the drag function (boolean or JS function, default: JS function that enables dragging for nodes and combos).}

\item{state}{Identifier for the selected state of nodes (string, default: "selected").}

\item{hideEdge}{Controls the display state of edges during dragging: "none", "out", "in", "both", or "all" (string, default: "none").}

\item{cursor}{Customize the mouse style during dragging (list, default: NULL).}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/drag-element-force}.}
}
\value{
A list with the configuration settings for the drag-element-force behavior.
}
\description{
Creates a configuration object for the drag-element-force behavior in G6.
This allows users to drag nodes and combos with force-directed layout interactions.
}
\examples{
# Basic configuration
config <- drag_element_force()

# Custom configuration with JavaScript arrow function and custom key
config <- drag_element_force(
  key = "my-custom-drag-force",
  fixed = TRUE,
  enable = JS("(event) => { return event.targetType === 'node'; }"),
  hideEdge = "both",
  cursor = list(
    default = "default",
    grab = "grab",
    grabbing = "grabbing"
  )
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/proxy.R
\name{g6_set_nodes}
\alias{g6_set_nodes}
\alias{g6_set_edges}
\alias{g6_set_combos}
\alias{g6_set_data}
\title{Set the state of nodes/edges/combos in a g6 graph via proxy}
\usage{
g6_set_nodes(graph, nodes)

g6_set_edges(graph, edges)

g6_set_combos(graph, combos)

g6_set_data(graph, data)
}
\arguments{
\item{graph}{A g6_proxy object created with \code{\link{g6_proxy}}.}

\item{nodes}{A key value pair list with the node id and its state.}

\item{edges}{A key value pair list with the edge id and its state.}

\item{combos}{A key value pair list with the combo id and its state.}

\item{data}{A nested list containing all nodes, edges and combo data.
Alternatively you can use \code{\link{g6_data}} to generate compatible data.}
}
\value{
The g6_proxy object (invisibly), allowing for method chaining.
}
\description{
This function sets the state of one or more nodes/edges/combos to an existing g6 graph instance
using a proxy object. This allows updating the graph without completely
re-rendering it. Valid states are "selected", "active", "inactive", "disabled", or "highlight".
}
\details{
\link{g6_set_data} allows to set all graph data at once (nodes, edges and combos).

This function can only be used with a g6_proxy object within a Shiny application.
It will not work with regular g6 objects outside of Shiny.

If a node with the same ID already exists, it will not be added again.
See \url{https://g6.antv.antgroup.com/en/api/element#graphsetelementstateid-state-options} for more details.
}
\examples{
if (interactive()) {
  # Example setting node states
  library(shiny)
  library(g6R)
  library(bslib)

  nodes <- data.frame(id = 1:2)
  edges <- data.frame(source = 1, target = 2)

  ui <- page_fluid(
    g6_output("graph"),
    actionButton("set_state", "Set Node States")
  )

  server <- function(input, output, session) {
    output$graph <- render_g6({
      g6(nodes = nodes, edges = edges) |> g6_layout()
    })

    observeEvent(input$set_state, {
      g6_set_nodes(
        g6_proxy("graph"),
        list(`1` = "selected", `2` = "disabled")
      )
    })
  }

  shinyApp(ui, server)

  # Replace data dynamically
  ui <- page_fluid(
    g6_output("graph"),
    actionButton("remove", "Remove All"),
    actionButton("reset", "Reset Graph"),
    verbatimTextOutput("state")
  )

  server <- function(input, output, session) {
    # Store initial state after first render
    initial_state <- reactiveVal(NULL)

    output$graph <- render_g6({
      g6(nodes = nodes, edges = edges) |>
        g6_layout() |>
        g6_options(animation = FALSE)
    })

    # Save initial state once available
    observe({
      state <- input[["graph-state"]]
      if (!is.null(state) && is.null(initial_state())) {
        initial_state(state)
      }
    })

    # Remove all nodes and edges
    observeEvent(input$remove, {
      g6_set_data(g6_proxy("graph"), list(nodes = list(), edges = list()))
    })

    # Reset graph to initial state
    observeEvent(input$reset, {
      state <- initial_state()
      if (!is.null(state)) {
        g6_set_data(g6_proxy("graph"), state)
      }
    })

    output$state <- renderPrint({
      input[["graph-state"]]
    })
  }

  shinyApp(ui, server)
}
}
\seealso{
\code{\link{g6_proxy}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{as_g6_nodes}
\alias{as_g6_nodes}
\alias{as_g6_edges}
\alias{as_g6_edges.g6_edges}
\alias{as_g6_edges.data.frame}
\alias{as_g6_edges.list}
\alias{as_g6_combos}
\alias{as_g6_combos.g6_combos}
\alias{as_g6_combos.data.frame}
\alias{as_g6_combos.list}
\title{Coerce to a list of g6_elements objects}
\usage{
as_g6_nodes(x, ...)

as_g6_edges(x, ...)

\method{as_g6_edges}{g6_edges}(x, ...)

\method{as_g6_edges}{data.frame}(x, ...)

\method{as_g6_edges}{list}(x, ...)

as_g6_combos(x, ...)

\method{as_g6_combos}{g6_combos}(x, ...)

\method{as_g6_combos}{data.frame}(x, ...)

\method{as_g6_combos}{list}(x, ...)
}
\arguments{
\item{x}{An object to be coerced. Data frame, list
or g6 element are supported.}

\item{...}{Additional arguments (unused).}
}
\value{
An object of class \code{g6_nodes}, \code{g6_edges}, or \code{g6_combos}.
}
\description{
Coerce to a list of g6_elements objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/behaviors.R
\name{create_edge}
\alias{create_edge}
\title{Configure Create Edge Behavior}
\usage{
create_edge(
  key = "create-edge",
  trigger = "drag",
  enable = FALSE,
  onCreate = NULL,
  onFinish = NULL,
  style = NULL,
  notify = FALSE,
  outputId = NULL,
  ...
)
}
\arguments{
\item{key}{Behavior unique identifier. Useful to modify this behavior from JS side.}

\item{trigger}{The way to trigger edge creation: "click" or "drag" (string, default: "drag").}

\item{enable}{Whether to enable this behavior (boolean or function, default: FALSE).
Our default implementation works in parallel with the \link{context_menu} plugin which is
responsible for activating the edge behavior when edge creation is selected.}

\item{onCreate}{Callback function for creating an edge, returns edge data (function, default: NULL).}

\item{onFinish}{Callback function for successfully creating an edge (function).
By default, we provide an internal implementation that disables the edge mode when the edge
creation is succesful so that it does not conflict with other drag behaviors.}

\item{style}{Style of the newly created edge (list, default: NULL).}

\item{notify}{Whether to show a feedback message in the ui.}

\item{outputId}{Manually pass the Shiny output ID. This is useful when the graph
is initialised outside the shiny render function and the ID cannot be automatically
inferred. This allows to set input values from the callback function
with the right namespace and graph ID. You must typically pass \code{session$ns("graphid")}
to ensure this also works in modules.}

\item{...}{Extra parameters. See \url{https://g6.antv.antgroup.com/en/manual/behavior/create-edge}.}
}
\value{
A list with the configuration settings for the create-edge behavior.
}
\description{
Creates a configuration object for the create-edge behavior in G6.
This allows users to create edges between nodes by clicking or dragging.
}
\examples{
# Basic configuration
config <- create_edge()
if (interactive()) {
  library(shiny)
  library(bslib)
  library(g6R)

  nodes <- list(
    list(
      id = "node1"
    ),
    list(
      id = "node2"
    )
  )

  modUI <- function(id) {
    ns <- NS(id)
    tagList(
      g6Output(ns("graph"))
    )
  }

  modServer <- function(id) {
    moduleServer(id, function(input, output, session) {
      output$graph <- renderG6({
        g6(nodes) |>
          g6_options(
            animation = FALSE,
            edge = edge_options(
              style = list(
                endArrow = TRUE
              )
            )
          ) |>
          g6_layout(d3_force_layout()) |>
          g6_behaviors(
            # avoid conflict with internal function
            g6R::create_edge(
              target = c("node", "combo", "canvas"),
              enable = JS(
                "(e) => {
                  return e.shiftKey;
                }"
              ),
              onFinish = JS(
                sprintf(
                  "(edge) => {
                    const graph = HTMLWidgets.find('#\%s').getWidget();
                    // Avoid to create edges in combos. If so, we remove it
                    if (edge.targetType === 'combo') {
                      graph.removeEdgeData([edge.id]);
                      return;
                    }
                    Shiny.setInputValue('\%s', edge);
                  }",
                  session$ns("graph"),
                  session$ns("added_edge")
                )
              )
            )
          )
      })

      observeEvent(input[["added_edge"]], {
        showNotification(
          sprintf("Edge dropped on: \%s", input[["added_edge"]]$targetType),
          type = "message"
        )
      })
    })
  }

  ui <- page_fluid(
    modUI("test")
  )

  server <- function(input, output, session) {
    modServer("test")
  }

  shinyApp(ui, server)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/options.R
\name{auto_fit_config}
\alias{auto_fit_config}
\title{Create Auto-Fit Configuration for G6 Graphs}
\usage{
auto_fit_config(
  type = c("view", "center"),
  when = c("overflow", "always"),
  direction = c("x", "y", "both"),
  duration = 1000,
  easing = c("ease-in-out", "ease", "ease-in", "ease-out", "linear", "cubic-bezier",
    "step-start", "step-end")
)
}
\arguments{
\item{type}{The auto-fit mode to use. Options:
\itemize{
\item \code{"view"}: Scale and translate the graph to fit all elements within the view (default)
\item \code{"center"}: Only translate the graph to center elements without scaling
}}

\item{when}{When the auto-fit should be triggered. Options:
\itemize{
\item \code{"overflow"}: Trigger auto-fit only when elements overflow the canvas (default)
\item \code{"always"}: Always perform auto-fit when the graph data changes
}}

\item{direction}{The direction for auto-fit adjustment. Options:
\itemize{
\item \code{"x"}: Adjust only along the x-axis
\item \code{"y"}: Adjust only along the y-axis
\item \code{"both"}: Adjust in both x and y directions (default)
}}

\item{duration}{The duration of the auto-fit animation in milliseconds (default: 1000)}

\item{easing}{The animation easing function to use. Options:
\itemize{
\item \code{"ease-in-out"}: Slow at the beginning and end of the animation (default)
\item \code{"ease"}: Standard easing
\item \code{"ease-in"}: Slow at the beginning
\item \code{"ease-out"}: Slow at the end
\item \code{"linear"}: Constant speed throughout
\item \code{"cubic-bezier"}: Custom cubic-bezier curve
\item \code{"step-start"}: Jump immediately to the end state
\item \code{"step-end"}: Jump at the end to the end state
}}
}
\value{
A list containing the auto-fit configuration that can be passed to \code{\link[=g6_options]{g6_options()}}.
}
\description{
Configures the auto-fit behavior for a G6 graph. Auto-fit automatically adjusts
the view to fit all elements or centers them within the canvas.
}
\details{
The auto-fit feature helps ensure that graph elements remain visible within the canvas.
It can be configured to either fit all elements to view or center them, and can be
triggered under different conditions.
}
\examples{
# Basic auto-fit configuration with default settings
config <- auto_fit_config()

# Auto-fit with only centering (no scaling)
config <- auto_fit_config(type = "center")

# Auto-fit that always triggers when graph data changes
config <- auto_fit_config(when = "always")

# Auto-fit only in the x direction
config <- auto_fit_config(direction = "x")

# Auto-fit with a fast animation
config <- auto_fit_config(duration = 300, easing = "ease-out")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/elements.R
\name{is_g6_data}
\alias{is_g6_data}
\title{Check if an object is a g6_data object}
\usage{
is_g6_data(x)
}
\arguments{
\item{x}{An object to check.}
}
\value{
Logical indicating if \code{x} is of class \code{g6_data}.
}
\description{
Check if an object is a g6_data object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plugins.R
\name{fish_eye}
\alias{fish_eye}
\title{Configure Fish Eye Plugin}
\usage{
fish_eye(
  key = "fish-eye",
  trigger = c("pointermove", "click", "drag"),
  r = 120,
  maxR = NULL,
  minR = 0,
  d = 1.5,
  maxD = 5,
  minD = 0,
  scaleRBy = NULL,
  scaleDBy = NULL,
  showDPercent = TRUE,
  style = NULL,
  nodeStyle = list(label = TRUE),
  preventDefault = TRUE,
  ...
)
}
\arguments{
\item{key}{Unique identifier for the plugin (string, default: NULL).}

\item{trigger}{Method to move the fisheye: "pointermove", "click", or "drag" (string, default: "pointermove").}

\item{r}{Radius of the fisheye (number, default: 120).}

\item{maxR}{Maximum adjustable radius of the fisheye
(number, default: NULL - half of the smaller canvas dimension).}

\item{minR}{Minimum adjustable radius of the fisheye (number, default: 0).}

\item{d}{Distortion factor (number, default: 1.5).}

\item{maxD}{Maximum adjustable distortion factor (number, default: 5).}

\item{minD}{Minimum adjustable distortion factor (number, default: 0).}

\item{scaleRBy}{Method to adjust the fisheye radius: "wheel" or "drag" (string, default: NULL).}

\item{scaleDBy}{Method to adjust the fisheye distortion factor: "wheel" or "drag" (string, default: NULL).}

\item{showDPercent}{Whether to show the distortion factor value in the fisheye (boolean, default: TRUE).}

\item{style}{Style of the fisheye (list, default: NULL).}

\item{nodeStyle}{Style of nodes in the fisheye (list or JS function, default: list(label = TRUE)).}

\item{preventDefault}{Whether to prevent default events (boolean, default: TRUE).}

\item{...}{Extra parameters.
See \url{https://g6.antv.antgroup.com/en/manual/fisheye}.}
}
\value{
A list with the configuration settings for the fisheye plugin.
}
\description{
Creates a configuration object for the fisheye plugin in G6.
This plugin creates a fisheye lens effect that magnifies elements within a specific area.
}
\examples{
# Basic configuration
config <- fish_eye()

# Custom configuration
config <- fish_eye(
  key = "my-fisheye",
  trigger = "drag",
  r = 200,
  d = 2.5,
  scaleRBy = "wheel",
  scaleDBy = "drag",
  style = list(
    stroke = "#1890ff",
    fill = "rgba(24, 144, 255, 0.1)",
    lineWidth = 2
  ),
  nodeStyle = JS("(datum) => {
    return {
      label: true,
      labelCfg: {
        style: {
          fill: '#003a8c',
          fontSize: 14
        }
      }
    };
  }")
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ports.R
\name{as_g6_port}
\alias{as_g6_port}
\alias{as_g6_port.g6_port}
\alias{as_g6_port.list}
\title{Coerce to a g6_port object}
\usage{
as_g6_port(x, ...)

\method{as_g6_port}{g6_port}(x, ...)

\method{as_g6_port}{list}(x, ...)
}
\arguments{
\item{x}{An object to coerce.}

\item{...}{Additional arguments (unused).}
}
\value{
An object of class \code{g6_port}.
}
\description{
Coerce to a g6_port object
}
\examples{
as_g6_port(list(key = "input-1", type = "input", placement = "left"))
as_g6_port(g6_port("input-1", type = "input"))
}
